﻿using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using Leap;
using UnityEngine.XR;
using System.Security.Policy;
using UnityEngine.UI.Extensions;

namespace JayJayWon
{
    public class UIAssist : MVRScript {
        const string pluginName = "UIAssist";
        const string pluginAuthor = "JayJayWon";
        const string pluginVersion = "1.10.0";
        const float pluginVersionFloat = 1.10f;
        const float _gazeFOVNarrow = 10f;
        const float _gazeFOVWide = 60f;

        public string _presetAppRef = "1";
        public string _presetSkinRef = "1";
        public string _lastLoadedThumbnailFolder = "";

        public static List<string> VAMSaveRestoreActions = new List<string>() { "SaveToStore1", "SaveToStore2", "SaveToStore3", "RestoreAllFromStore1", "RestoreAllFromStore2", "RestoreAllFromStore3", "RestorePhysicsFromStore1", "RestorePhysicsFromStore2", "RestorePhysicsFromStore3", "RestoreAppearanceFromStore1", "RestoreAppearanceFromStore2", "RestoreAppearanceFromStore3", "RestoreAllFromDefaults", "RestorePhysicalFromDefaults", "RestoreAppearanceFromDefaults" };
        public override void Init()
        {
            try
            {

                _isActivePluginJSON = new JSONStorableBool("activeUIAssist", false);
                RegisterBool(_isActivePluginJSON);
                _isActivePluginJSON.isStorable = false;
                _isActivePluginJSON.isRestorable = false;
                _isActivePlugin = false;

                if (XRSettings.enabled) _vrActive = true;

                if (containingAtom.name == "CoreControl" && containingAtom.type =="SessionPluginManager")
                {
                    if (GetActiveUIAssistCount() < 2)
                    {
                        _isActivePluginJSON.val = true;
                        _isActivePlugin = true;

                        List<string> handChoices = new List<string>();
                        handChoices.Add("Left Hand");
                        handChoices.Add("Right Hand");
                        handChoices.Add("Disabled");

                        _vrHandControlJSON = new JSONStorableStringChooser("handControl", handChoices, "Left Hand", "VR Watch Wrist", SwitchVRHand);
                        RegisterStringChooser(_vrHandControlJSON);

                        _vrPinModeJSON = new JSONStorableStringChooser("pinMode", null, "Static", "VR Pin Mode", SwitchVRPinMode);
                        _gazeMenuSelectModeJSON = new JSONStorableStringChooser("gazeMenuSelectMode", null, "Wrist UI Only", "Gaze UI Control");
                        _toolboxJSON = new JSONStorableStringChooser("toolBoxMode", null, "Select...", "Toolbox", SwitchToolBox);

                        _vrHUDHorizontalOffsetJSON = new JSONStorableFloat("vrHUDHorizontalOffset", 0.5f, 0.3f, 0.8f);
                        _vrHUDVerticalOffsetJSON = new JSONStorableFloat("vrHUDVerticalOffset", 0.5f, 0.2f, 0.8f);
                        _vrHUDDepthOffsetJSON = new JSONStorableFloat("vrHUDDepth", 0.6f, 0.1f, 2f);

                        _pluginVersionJSON = new JSONStorableString("PluginVersion", pluginVersion);
                        RegisterString(_pluginVersionJSON);
                        _pluginVersionJSON.isStorable = true;
                        _pluginVersionJSON.isRestorable = false;

                        _pluginVersionFloatJSON = new JSONStorableFloat("PluginVersionID", pluginVersionFloat, 1f, 10f);
                        RegisterFloat(_pluginVersionFloatJSON);
                        _pluginVersionFloatJSON.isStorable = true;
                        _pluginVersionFloatJSON.isRestorable = false;

                        _isPatreonOnlyVersionJSON = new JSONStorableBool("PatreonOnlyVersion", Constants._patreonContentEnabled);
                        RegisterBool(_isPatreonOnlyVersionJSON);
                        _isPatreonOnlyVersionJSON.isStorable = true;
                        _isPatreonOnlyVersionJSON.isRestorable = false;

                        _isPrimaryUIAJSON = new JSONStorableBool("PrimaryUIA", false);
                        RegisterBool(_isPrimaryUIAJSON);
                        _isPrimaryUIAJSON.isStorable = false;
                        _isPrimaryUIAJSON.isRestorable = false;

                        _uiMainCanvas = new WristUI(this);
                        _uidMainButtonCanvasDisplay = new UIDButtonCanvasDisplay(this);

                        _defaultButtonSkin = new ButtonSkin(2, this);
                        _defaultButtonSkinDisplay = new ButtonSkinDisplay(_defaultButtonSkin, this, true);

                        _mainButton = _uiMainCanvas.CreateUIButtonInCanvas(System.DateTime.Now.ToString("HH:mm"), 0f, 0f, 0f, 70f,40f);
                        _mainButton.button.onClick.AddListener(() => { ToggleUIActivateButton(); });
                        EventTrigger trigger = _mainButton.button.gameObject.AddComponent<EventTrigger>();
                        var pointerDown = new EventTrigger.Entry();
                        pointerDown.eventID = EventTriggerType.PointerDown;
                        pointerDown.callback.AddListener((e) => MainButtonClickDown());
                        trigger.triggers.Add(pointerDown);

                        var pointerUp = new EventTrigger.Entry();
                        pointerUp.eventID = EventTriggerType.PointerUp;
                        pointerUp.callback.AddListener((e) => MainButtonClickUp());
                        trigger.triggers.Add(pointerUp);

                        _leftButton = _uiMainCanvas.CreateUIButtonInCanvas("<<", -50f, 0f, 0f, 30f, 30f);
                        _leftButton.gameObject.SetActive(false);
                        _leftButton.button.onClick.AddListener(() => { ScrollLeftButton(); });
                        _rightButton = _uiMainCanvas.CreateUIButtonInCanvas(">>", 50f, 0f, 0f, 30f, 30f);
                        _rightButton.gameObject.SetActive(false);
                        _rightButton.button.onClick.AddListener(() => { ScrollRightButton(); });

                        _mainPinButton = _uiMainCanvas.CreateUIButtonInCanvas("X", -100f, 0f, 0f, 30f, 30f);
                        _mainPinButton.gameObject.SetActive(false);
                        _mainPinButton.button.onClick.AddListener(() => { TogglePin(); });

                        _quickLaunchButtonGrid = new ButtonGrid(6, 1, this, true);
                        _quickLaunchButtonGrid._buttonSize = "Small";
                                               
                        _activateQuickLaunchBarJSON = new JSONStorableBool("activateQuickLaunchBar", false, SwitchActivateQuickLaunchBar);
                        _buttonGridSelectorJSON = new JSONStorableStringChooser("gridSelection", GetGridChoices(), "1", "", SwitchDisplayGrid);
                        
                        _displayGazeSelectionsJSON = new JSONStorableBool("displayGazeSelections", true, SwitchDisplayGazeSelections);
                        _displayScreenLabelsJSON = new JSONStorableBool("displayScreenLabels", false, SwitchDisplayScreenLabels);
                        _gazeActivateVRScreenJSON = new JSONStorableBool("gazeActivateVRScreen", true);

                        string pluginDetails = "";
                        if (Constants._patreonContentEnabled) { pluginDetails = pluginName + " v" + pluginVersion + "\nby " + pluginAuthor + "\nPatron version"; }
                        else { pluginDetails = pluginName + " v" + pluginVersion + "\nby " + pluginAuthor + "\nFree version"; }

                        _pluginDetailsJSON = new JSONStorableString("PluginDetails", pluginDetails);

                        _toolbox = new ToolBox(this);

                        _atomRenameHandle = new SuperController.OnAtomUIDRename(this.AtomNameUpdate);
                        SuperController.singleton.onAtomUIDRenameHandlers += _atomRenameHandle;

                        _atomRemoveHandle = new SuperController.OnAtomRemoved (this.AtomRemovedUpdate);
                        SuperController.singleton.onAtomRemovedHandlers += _atomRemoveHandle;                        

                        List<string> test = new List<string>();
                        test.Add("Temp");
                        PatreonFeatures.GridLayoutFeature(test);
                        if (test != null && test.Count == 19) _secondValidation = true;

                        //					trans = SuperController.singleton.leapHandRight.Find("R_Arm_Cut").Find("R_Wrist").Find("R_Finger_Index_A").Find("R_Finger_Index_B").Find("R_Finger_Index_C");
                    }
                    else SuperController.LogMessage("A maximum of two UIAssist plugins can be active");
                }
                else SuperController.LogMessage("UIAssist can only be loaded as a Session Plugin");
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        protected void AtomNameUpdate(string oldName, string newName)
        {
            try
            {
                if (_lastViewedAtom == oldName) _lastViewedAtom = newName;
                if (_lastViewedPerson == oldName) _lastViewedPerson = newName;
                if (_lastViewedFemale == oldName) _lastViewedFemale = newName;
                if (_lastViewedMale == oldName) _lastViewedMale = newName;
                if (_lastSelectedAtom == oldName) _lastSelectedAtom = newName;
                if (_lastSelectedPerson == oldName) _lastSelectedPerson = newName;
                if (_lastSelectedFemale == oldName) _lastSelectedFemale = newName;
                if (_lastSelectedMale == oldName) _lastSelectedMale = newName;
                if (_atomVRMoveAtomName == oldName) _atomVRMoveAtomName = newName;

                foreach (ButtonGrid buttonGrid in _buttonGridsList)
                {
                    buttonGrid.AtomNameUpdate(oldName, newName);
                }
                _quickLaunchButtonGrid.AtomNameUpdate(oldName, newName);

                RefreshRightUIDisplay();
                if (_atomActiveClothingDictionary.ContainsKey(oldName))
                {
                    _atomActiveClothingDictionary.Add(newName, _atomActiveClothingDictionary[oldName]);
                    _atomActiveClothingDictionary.Remove(oldName);
                }
                _toolbox._heelAdjustTool.AtomNameUpdate(oldName, newName);
                if (_testSpawnAtom == oldName) _testSpawnAtom = newName;
                _uidMainButtonCanvasDisplay._uiActiveClothingEditor.UpdateAtomNames(oldName,newName);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void AtomRemovedUpdate (Atom atom)
        {
            try
            {
                string atomName = atom.name;
                if (_lastViewedAtom == atomName) _lastViewedAtom = "";
                if (_lastViewedPerson == atomName) _lastViewedPerson = "";
                if (_lastViewedFemale == atomName) _lastViewedFemale = "";
                if (_lastViewedMale == atomName) _lastViewedMale = "";
                if (_lastSelectedAtom == atomName) _lastSelectedAtom = "";
                if (_lastSelectedPerson == atomName) _lastSelectedPerson = "";
                if (_lastSelectedFemale == atomName) _lastSelectedFemale = "";
                if (_lastSelectedMale == atomName) _lastSelectedMale = "";
                if (_atomVRMoveAtomName == atomName) _atomVRMoveAtomName = "";

                foreach (ButtonGrid buttonGrid in _buttonGridsList)
                {
                    buttonGrid.AtomRemovedUpdate(atomName);
                }
                _quickLaunchButtonGrid.AtomRemovedUpdate(atomName);

                RefreshRightUIDisplay();
                if (_atomActiveClothingDictionary.ContainsKey(atomName))_atomActiveClothingDictionary.Remove(atomName);
                _toolbox._heelAdjustTool.AtomRemovedUpdate(atomName);
                if (_testSpawnAtom == atomName) DeactivateSpawnTest();

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public JSONClass GetJSON(JSONClass jc = null)
        {
            if (jc == null) { jc = new JSONClass(); }

            jc["PluginVersion"] = pluginVersion;
            jc["PluginVersionID"].AsFloat = pluginVersionFloat;
            jc["PatreonOnlyVersion"].AsBool = Constants._patreonContentEnabled;

            jc["vrHandControl"] = _vrHandControlJSON.val;
            jc["vrPinMode"] = _vrPinModeJSON.val;
            jc["gazeMenuSelectMode"] = _gazeMenuSelectModeJSON.val;

            jc["displayGazeSelection"].AsBool = _displayGazeSelectionsJSON.val;
            jc["displayScreenLabels"].AsBool = _displayScreenLabelsJSON.val;
            jc["gazeActivateVRScreen"].AsBool = _gazeActivateVRScreenJSON.val;
            jc["activateQuickLaunchBar"].AsBool = _activateQuickLaunchBarJSON.val;

            jc["vrHUDHorizontalOffset"].AsFloat = _vrHUDHorizontalOffsetJSON.val;
            jc["vrHUDVerticalOffset"].AsFloat = _vrHUDVerticalOffsetJSON.val;
            jc["vrHUDDepth"].AsFloat = _vrHUDDepthOffsetJSON.val;

            JSONArray gridArray = new JSONArray();
            _buttonGridsList.ForEach((grid) =>
            {
                gridArray.Add(grid.GetJSON());
            });

            jc["ButtonGrids"] = gridArray;
            jc["QuickLaunchButtonGrid"] = _quickLaunchButtonGrid.GetJSON();

            jc["DefaultButtonSkin"] = _defaultButtonSkin.GetJSON();
            jc["ToolBox"] = _toolbox.GetJSON();

            return jc;
        }
        public void LoadJSON(JSONClass jc)
        {
            _isRestoring = true;
            if (jc["vrHandControl"] != null) _vrHandControlJSON.val = jc["vrHandControl"];
            if (jc["vrPinMode"] != null) _vrPinModeJSON.val = jc["vrPinMode"];
            if (jc["gazeMenuSelectMode"] != null) _gazeMenuSelectModeJSON.val = jc["gazeMenuSelectMode"];

            //           if (jc["isLoaded"] != null) _isLoaded = jc["isLoaded"].AsBool;
            if (jc["displayGazeSelection"] != null) _displayGazeSelectionsJSON.val = jc["displayGazeSelection"].AsBool;
            if (jc["displayScreenLabels"] != null) _displayScreenLabelsJSON.val = jc["displayScreenLabels"].AsBool;
            if (jc["gazeActivateVRScreen"] != null) _gazeActivateVRScreenJSON.val = jc["gazeActivateVRScreen"].AsBool;
            if (jc["activateQuickLaunchBar"] != null) _activateQuickLaunchBarJSON.val = jc["activateQuickLaunchBar"].AsBool;
            else _activateQuickLaunchBarJSON.val = false;

            if (jc["vrHUDHorizontalOffset"] != null) _vrHUDHorizontalOffsetJSON.val = jc["vrHUDHorizontalOffset"].AsFloat;
            if (jc["vrHUDVerticalOffset"] != null) _vrHUDVerticalOffsetJSON.val = jc["vrHUDVerticalOffset"].AsFloat;
            if (jc["vrHUDDepth"] != null) _vrHUDDepthOffsetJSON.val = jc["vrHUDDepth"].AsFloat;


            if (jc["QuickLaunchButtonGrid"] != null) _quickLaunchButtonGrid.LoadJSON((JSONClass)jc["QuickLaunchButtonGrid"]);
            else
            {
                _quickLaunchButtonGrid = new ButtonGrid(6, 1, this, true);
                _quickLaunchButtonGrid._buttonSize = "Small";
            }
            if (jc["ButtonGrids"] != null)
            {
                JSONArray gridArrayJSON = jc["ButtonGrids"].AsArray;
                for (int i = 0; i < gridArrayJSON.Count; i++)
                {
                    JSONClass gridJSON = gridArrayJSON[i].AsObject;
                    ButtonGrid grid;
                    if (i >= _buttonGridsList.Count)
                    {
                        grid = new ButtonGrid(0, 0, this);
                        _buttonGridsList.Add(grid);
                    }
                    else { grid = _buttonGridsList[i]; }
                    grid.LoadJSON(gridJSON);
                }
            }
            if (jc["DefaultButtonSkin"] != null) { _defaultButtonSkin.LoadJSON((JSONClass)jc["DefaultButtonSkin"]); }
            if (jc["ToolBox"] != null) _toolbox.LoadJSON((JSONClass)jc["ToolBox"]);
            _isRestoring = false;
        }
        public string GetLocalPluginPath()
        {
            string pluginId = storeId.Split('_')[0];
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }) + 1);
            pathToScriptFolder = pathToScriptFolder.Replace('/', '\\');
            if (pathToScriptFolder.Contains(":")) pathToScriptFolder = pathToScriptFolder.Substring(pathToScriptFolder.IndexOf(':') + 1);
            char[] charsToTrim = { '/', '\\' };
            pathToScriptFolder = pathToScriptFolder.Trim(charsToTrim);
            bool exists = FileManagerSecure.DirectoryExists(pathToScriptFolder);
            if (!exists) pathToScriptFolder = "Custom\\Scripts";
            return pathToScriptFolder;
        }
        bool DefaultUIAPFileExists() {
            string[] fileNamesArray = SuperController.singleton.GetFilesAtPath(GetLocalPluginPath());
            List<string> fileNameList = new List<string>();
            foreach (string filePath in fileNamesArray) {
                string fileName = filePath.Substring(filePath.LastIndexOfAny(new char[] { '/', '\\' }) + 1);
                fileNameList.Add(fileName);
            }

            if (fileNameList.Contains("default.uiap")) { return (true); }
            return (false);
        }
        private static bool IsSelectableFreeController(string contollerName, bool isMale = false)
        {
            bool result = true;
            if (contollerName.StartsWith("hair")) result = false;
            else if (!isMale && contollerName.StartsWith("penis")) result = false;
            else if (!isMale && contollerName == "testesControl") result = false;

            return (result);
        }
        public UIDynamicTextField CreateTextInput(JSONStorableString jss, bool rightSide = false) {
            UIDynamicTextField textfield = CreateTextField(jss, rightSide); textfield.height = 1f; textfield.backgroundColor = Color.white; var input = textfield.gameObject.AddComponent<InputField>(); input.textComponent = textfield.UItext; input.lineType = InputField.LineType.MultiLineNewline; input.textComponent.resizeTextMaxSize = 30; input.textComponent.resizeTextForBestFit = true; jss.inputField = input; return (textfield);
        }
        public UIDynamicPopup CreateSetupPopup(JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool scrollable = true) {
            UIDynamicPopup dynamicPopup;
            if (scrollable) dynamicPopup = CreateScrollablePopup(stringChooserJSON, rightUI);
            else dynamicPopup = CreatePopup(stringChooserJSON, rightUI);
            dynamicPopup.popupPanelHeight = panelHeight;
           
            if (callbackMethod != null) { dynamicPopup.popup.onOpenPopupHandlers += callbackMethod; }
            if (label != null) { dynamicPopup.label = label; }
            return (dynamicPopup);
        }
        public void RefreshLeftUIDisplay()
        {
            CloseAllLeftUIPopups(null);

            if (_loadProfileButton != null) { _loadProfileButton.button.onClick.RemoveAllListeners(); RemoveButton(_loadProfileButton); }
            if (_saveProfileButton != null) { _saveProfileButton.button.onClick.RemoveAllListeners(); RemoveButton(_saveProfileButton); }
            if (_setDefaultProfileButton != null) { _setDefaultProfileButton.button.onClick.RemoveAllListeners(); RemoveButton(_setDefaultProfileButton); }
            if (_vrHandControlPopup != null) { RemovePopup(_vrHandControlPopup); }
            if (_vrPinModePopup != null) { RemovePopup(_vrPinModePopup); }
            if (_toolboxPopup != null) { RemovePopup(_toolboxPopup); }

            if (_gazeMenuSelectModePopup != null) { RemovePopup(_gazeMenuSelectModePopup); }
            if (_vrHUDHorizontalOffsetSlider != null) { RemoveSlider(_vrHUDHorizontalOffsetSlider); }
            if (_vrHUDVerticalOffsetSlider != null) { RemoveSlider(_vrHUDVerticalOffsetSlider); }
            if (_vrHUDDepthOffsetSlider != null) { RemoveSlider(_vrHUDDepthOffsetSlider); }


            if (_displayGazeSelectionsToggle != null) { RemoveToggle(_displayGazeSelectionsToggle); }
            if (_displayScreenLabelsToggle != null) { RemoveToggle(_displayScreenLabelsToggle); }
            if (_gazeActivateVRScreenToggle != null) { RemoveToggle(_gazeActivateVRScreenToggle); }
            if (_pluginVersionTextField != null) { RemoveTextField(_pluginVersionTextField); }
            if (_activateQuickLaunchBarToggle != null) RemoveToggle(_activateQuickLaunchBarToggle);
            if (_spacer1 != null) { RemoveSpacer(_spacer1); }
            if (_spacer2 != null) { RemoveSpacer(_spacer2); }
            if (_spacer2b != null) { RemoveSpacer(_spacer2b); }
            if (_spacer3 != null) { RemoveSpacer(_spacer3); }
            if (_spacer4 != null) { RemoveSpacer(_spacer4); }

            _isDisplayUpdating = true;
            RefreshPinModeChoices();
            if (!_vrPinModeJSON.choices.Contains(_vrPinModeJSON.val)) _vrPinModeJSON.val = _vrPinModeJSON.choices[0];
            RefreshGazeControlModeChoices();
            if (!_gazeMenuSelectModeJSON.choices.Contains(_gazeMenuSelectModeJSON.val)) _gazeMenuSelectModeJSON.val = _gazeMenuSelectModeJSON.choices[0];

            _isDisplayUpdating = false;

            _loadProfileButton = CreateButton("Load UI Profile ...");
            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Scripts\\JayJayWon", false, false, true, true);
            _loadProfileButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(LoadUIAPSelected, "uiap", GetLocalPluginPath(), false, true, false, null, false, shortcuts); });
            _saveProfileButton = CreateButton("Save UI Profile ...");
            _saveProfileButton.button.onClick.AddListener(() => {
                SuperController.singleton.GetMediaPathDialog(SaveUIAPSelected, "uiap", GetLocalPluginPath(), false, true, false);
                // Update the browser to be a Save browser
                uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                browser.SetTextEntry(true);
                browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), "uiap");
                browser.ActivateFileNameField();

            });
            _setDefaultProfileButton = CreateButton("Set current Profile as Default");
            _setDefaultProfileButton.button.onClick.AddListener(() => { SaveUIAPSelected(GetLocalPluginPath() + "\\default.uiap"); });
            _spacer1 = CreateSpacer();
            _spacer1.height = 10;
            _vrHandControlPopup = CreateSetupPopup(_vrHandControlJSON, false, 200f, VRHandSelectionPopupOpen, null, false);
            if (_vrHandControlJSON.val != "Disabled")
            {
                _gazeActivateVRScreenToggle = CreateToggle(_gazeActivateVRScreenJSON);
                _gazeActivateVRScreenToggle.label = "Gaze Open/Close Screens";
            }
            _vrPinModePopup = CreateSetupPopup(_vrPinModeJSON, false, 200f, RefreshPinModeChoices, null, false);
            if (_vrPinModeJSON.val == "HUD (Custom)")
            {
                _vrHUDHorizontalOffsetSlider = CreateSlider(_vrHUDHorizontalOffsetJSON);
                _vrHUDHorizontalOffsetSlider.label = "HUD Horizontal position";
                _vrHUDVerticalOffsetSlider = CreateSlider(_vrHUDVerticalOffsetJSON);
                _vrHUDVerticalOffsetSlider.label = "HUD Vertical position";
                _vrHUDDepthOffsetSlider = CreateSlider(_vrHUDDepthOffsetJSON);
                _vrHUDDepthOffsetSlider.label = "HUD Distance";
            }
            if (_vrHandControlJSON.val != "Disabled")
            {
                _gazeMenuSelectModePopup = CreateSetupPopup(_gazeMenuSelectModeJSON, false, 200f, RefreshGazeControlModeChoices, null, false);
            }

            _spacer2 = CreateSpacer();
            _spacer2.height = 10;
            _displayGazeSelectionsToggle = CreateToggle(_displayGazeSelectionsJSON);
            _displayGazeSelectionsToggle.label = "Show Gaze Targets on Wrist UI";

            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (Constants._patreonContentEnabled && test != null && test.Count == 19)
            {
                _displayScreenLabelsToggle = CreateToggle(_displayScreenLabelsJSON);
                _displayScreenLabelsToggle.label = "Show Screen Labels on Wrist UI";
                _activateQuickLaunchBarToggle = CreateToggle(_activateQuickLaunchBarJSON);
                _activateQuickLaunchBarToggle.label = "Enable Quick Launch Bar";
            }

            _spacer2b = CreateSpacer();
            _spacer2b.height = 10;
            _toolboxPopup = CreateSetupPopup(_toolboxJSON, false, 200f, ToolBoxPopupOpen, null, false);
            _toolboxPopup.labelTextColor = Color.blue;
            if (_toolboxJSON.val != "Select...") _toolbox.RefreshToolboxDisplay(_toolboxJSON.val);

            if (Constants._patreonContentEnabled && test != null && test.Count == 19)
            {
                _spacer3 = CreateSpacer();
                _spacer3.height = 10;
                _defaultButtonSkinDisplay.RefreshDisplay();
            }
            _spacer4 = CreateSpacer();
            _spacer4.height = 10;

            _pluginVersionTextField = CreateTextField(_pluginDetailsJSON, false);
            _pluginVersionTextField.height = 1;


        }
        public void RefreshRightUIDisplay()
        {
            CloseAllRightUIPopups(null);
            if (_buttonGridSelectorPopup != null) { RemovePopup(_buttonGridSelectorPopup); }

            _buttonGridSelectorPopup = CreateSetupPopup(_buttonGridSelectorJSON, true, 500f, ButtonGridSelectorPopupOpen, "Screen Selector");
            _buttonGridSelectorPopup.labelTextColor = Color.blue;
            _buttonGridDisplay.RefreshDisplay();

        }
        public void LoadUIAPSelected(string file)
        {
            try
            {
                if (file != "" && FileManagerSecure.FileExists(file))
                {
                    JSONClass jc = (JSONClass)SuperController.singleton.LoadJSON(file);
                    _buttonGridsList.Clear();

                    LoadJSON(jc);
                    foreach (ButtonGrid buttonGrid in _buttonGridsList)
                    {
                        buttonGrid.RecalcActiveButtonCount();
                        buttonGrid.RecalcGazeSelections();
                    }
                    _currentGridScreen = 0;

                    _buttonGridSelectorJSON.choices = GetGridChoices();
                    if (_buttonGridDisplay == null) _buttonGridDisplay = new ButtonGridDisplay(_buttonGridsList[0], this);

                    if (_buttonGridSelectorJSON.val == "1") SwitchDisplayGrid("1");
                    else _buttonGridSelectorJSON.val = "1";

                    if (_buttonGridsList[0]._activeButtonCount==0) _uiActivated = false;
                    _uiaEditMode = false;

                    _uiaActiveClothingMode = false;

                    _uidMainButtonCanvasDisplay.DestroyUIButtons();

                    RefreshLeftUIDisplay();
                    RefreshRightUIDisplay();
                    RefreshWristUIButtonGrid();
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SaveUIAPSelected(string file)
        {
            try
            {
                if (file != "")
                {
                    if (!file.EndsWith(".uiap")) { file = file + ".uiap"; }
                    SuperController.singleton.SaveJSON(GetJSON(), file);

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void DeleteCurrentScreen()
        {
            try
            {
                if (!_buttonGridDisplay._buttonDisplay._gridDeleteConfirm) _buttonGridDisplay._buttonDisplay._gridDeleteConfirm = true;
                RefreshRightUIDisplay();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void MoveScreenDown()
        {
            try
            {                
                if (_buttonGridSelectorJSON.val !="1" && _buttonGridSelectorJSON.val != "Quick Launch Bar")
                {
                    int currentGridRef = int.Parse(_buttonGridSelectorJSON.val);
                    ButtonGrid tempBG = _buttonGridsList[currentGridRef - 1];
                    _buttonGridsList[currentGridRef - 1] = _buttonGridsList[currentGridRef - 2];
                    _buttonGridsList[currentGridRef - 2] = tempBG;
                    _buttonGridSelectorJSON.val = (currentGridRef - 1).ToString();

                }
                RefreshWristUIButtonGrid();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void MoveScreenUp()
        {
            try
            {
                if (_buttonGridSelectorJSON.val != "Quick Launch Bar")
                {
                    
                    int currentGridRef = int.Parse(_buttonGridSelectorJSON.val);
                    if (currentGridRef<_buttonGridsList.Count)
                    {
                        ButtonGrid tempBG = _buttonGridsList[currentGridRef - 1];
                        _buttonGridsList[currentGridRef - 1] = _buttonGridsList[currentGridRef];
                        _buttonGridsList[currentGridRef] = tempBG;
                        _buttonGridSelectorJSON.val = (currentGridRef+1).ToString();
                    }
                    

                }
                RefreshWristUIButtonGrid();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void ConfirmDeleteCurrentScreen()
        {
            try
            {
                int gridRef = int.Parse(_buttonGridSelectorJSON.val);

                _buttonGridsList.RemoveAt(gridRef - 1);
                if (gridRef > _buttonGridsList.Count) { _buttonGridSelectorJSON.val = _buttonGridsList.Count.ToString(); }
                else
                {
                    SwitchDisplayGrid(gridRef.ToString());
                }

                if (_currentGridScreen == gridRef - 1)
                {
                    if (_currentGridScreen >= _buttonGridsList.Count)  _currentGridScreen--;
                    RefreshWristUIButtonGrid();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void ButtonGridSelectorPopupOpen()
        {
            _buttonGridSelectorJSON.choices = GetGridChoices();
            CloseAllRightUIPopups(_buttonGridSelectorPopup);
        }
        public void CloseAllRightUIPopups(UIDynamicPopup exception)
        {
            if (_buttonGridSelectorPopup != null && _buttonGridSelectorPopup != exception) { _buttonGridSelectorPopup.popup.visible = false; }
            if (_buttonGridDisplay != null) _buttonGridDisplay.CloseAllPopup(exception);
        }
        public void CloseAllLeftUIPopups(UIDynamicPopup exception)
        {
            if (_vrHandControlPopup != null && _vrHandControlPopup != exception) { _vrHandControlPopup.popup.visible = false; }
            if (_vrPinModePopup != null && _vrPinModePopup != exception) { _vrPinModePopup.popup.visible = false; }
            if (_gazeMenuSelectModePopup != null && _gazeMenuSelectModePopup != exception) { _gazeMenuSelectModePopup.popup.visible = false; }
            if (_toolboxPopup != null && _toolboxPopup != exception) { _toolboxPopup.popup.visible = false; }

            if (_defaultButtonSkinDisplay != null && _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup != null && _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup != exception) { _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup.popup.visible = false; }
            if (_defaultButtonSkinDisplay != null && _defaultButtonSkinDisplay._skinStringChooserParamPopup != null && _defaultButtonSkinDisplay._skinStringChooserParamPopup != exception) { _defaultButtonSkinDisplay._skinStringChooserParamPopup.popup.visible = false; }
        }
        protected void VRHandSelectionPopupOpen()
        {
            CloseAllLeftUIPopups(_vrHandControlPopup);
        }
        protected void RefreshPinModeChoices()
        {
            CloseAllLeftUIPopups(_vrPinModePopup);
            List<string> pinModeChoices = new List<string>();
            if (_vrHandControlJSON.val != "Disabled") pinModeChoices.Add("Static");
            pinModeChoices.Add("HUD (Left)");
            pinModeChoices.Add("HUD (Right)");
            pinModeChoices.Add("HUD (Custom)");
            if (_vrHandControlJSON.val != "Disabled") pinModeChoices.Add("Disabled");
            _vrPinModeJSON.choices = pinModeChoices;
        }
        protected void RefreshGazeControlModeChoices()
        {
            CloseAllLeftUIPopups(_gazeMenuSelectModePopup);
            List<string> gazeControlModeChoices = new List<string>();
            gazeControlModeChoices.Add("Wrist UI Only");
            if (_vrPinModeJSON.val == "Static") gazeControlModeChoices.Add("Wrist & Pinned UI");
            gazeControlModeChoices.Add("Disabled");
            _gazeMenuSelectModeJSON.choices = gazeControlModeChoices;
        }
        private List<string> GetGridChoices()
        {
            List<string> gridChoices = new List<string>();
            if (_activateQuickLaunchBarJSON.val) gridChoices.Add("Quick Launch Bar");
            for (int i = 1; i <= _buttonGridsList.Count; i++)
            {
                gridChoices.Add(i.ToString());
            }
            gridChoices.Add("Add Screen...");
            return (gridChoices);
        }
        protected void ToolBoxPopupOpen()
        {
            CloseAllLeftUIPopups(_toolboxPopup);
            List<string> toolboxChoices = new List<string>();
            toolboxChoices.Add("Auto Heel Adjust");
            toolboxChoices.Add("Gaze Assisted Select\n(Edit Mode only)");
            if (_isPrimaryUIAJSON.val)
            {
 //               toolboxChoices.Add("Auto Plugin Loader");
 //               toolboxChoices.Add("Auto Session Plugin Loader");
            }
            _toolboxJSON.choices = toolboxChoices;
        }
        public JSONStorable GetOtherActiveUIAPlugin()
        {
            JSONStorable otherUIA =null;
            foreach (string storableName in containingAtom.GetStorableIDs())
            {
                if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#"))
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist")
                    {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");

                        if (activeJSON.val)
                        {
                            otherUIA = pluginStorable;
                            break;
                        }
                    }

                }
            }
            return (otherUIA);
        }
        private int GetActiveUIAssistCount() {
            int pluginCount = 0;
            foreach (string storableName in containingAtom.GetStorableIDs()) {
                if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist") {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");

                        if (activeJSON.val) pluginCount++;
                    }

                }
            }
            return (pluginCount);
        }
        private void UpdatePrimaryUIA()
        {
            bool foundOtherActiveUIA = false;
            foreach (string storableName in containingAtom.GetStorableIDs())
            {
                if ((storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) && storableName!=this.name)
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist")
                    {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");
                        if (activeJSON.val)
                        {
                            JSONStorableBool primaryUIAJSON = pluginStorable.GetBoolJSONParam("PrimaryUIA");
                            _isPrimaryUIAJSON.val = !primaryUIAJSON.val;
                            foundOtherActiveUIA = true;
                            break;
                        }
                    }

                }
            }
            if (!foundOtherActiveUIA) _isPrimaryUIAJSON.val = true;
           
        }
        private void SwitchDisplayGrid(string grid) {
            try {
                int gridRef = -1;
                if (grid == "Add Screen...") {
                    ButtonGrid newBG = new ButtonGrid(3, 3, this);
                    _buttonGridsList.Add(newBG);
                    gridRef = _buttonGridsList.Count;
                    _buttonGridSelectorJSON.val = gridRef.ToString();
                }
                else if (grid=="Quick Launch Bar"){
                    _buttonGridDisplay._selectedButtonGrid = _quickLaunchButtonGrid;
                    if (_buttonGridDisplay._buttonSelectJSON.val == "1") { _buttonGridDisplay.SwitchSelectedButton("1"); }
                    else { _buttonGridDisplay._buttonSelectJSON.val = "1"; }
                    RefreshRightUIDisplay();
                }
                else {
                    gridRef = int.Parse(grid) - 1;
                    if (_buttonGridsList.Count > gridRef)
                    {
                        _buttonGridDisplay._selectedButtonGrid = _buttonGridsList[gridRef];
                        if (_buttonGridDisplay._buttonSelectJSON.val == "1") { _buttonGridDisplay.SwitchSelectedButton("1"); }
                        else { _buttonGridDisplay._buttonSelectJSON.val = "1"; }
                        RefreshRightUIDisplay();
                    }

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDisplayGazeSelections(bool display) {
            try {
                if (!_isRestoring && !_isDisplayUpdating) { RefreshWristUIButtonGrid(); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchActivateQuickLaunchBar(bool activate)
        {
            try
            {
                if (!_isRestoring && !_isDisplayUpdating) {
                    if (!activate && _buttonGridSelectorJSON.val == "Quick Launch Bar") _buttonGridSelectorJSON.val = "1";
                    RefreshWristUIButtonGrid();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }        
        private void SwitchDisplayScreenLabels(bool display) {
            try {
                if (!_isRestoring && !_isDisplayUpdating) {
                    RefreshWristUIButtonGrid();
                    RefreshRightUIDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void TogglePin()
        {
            try
            {
                _uiPinned = !_uiPinned;
                _autoDePinEnabled = false;

                if (_vrHandControlJSON.val == "Right Hand")
                {
                    _fixedPinPosition = SuperController.singleton.rightHand.position;
                    _fixedPinRotation = SuperController.singleton.rightHand.rotation;

                }
                else
                {
                    _fixedPinPosition = SuperController.singleton.leftHand.position;
                    _fixedPinRotation = SuperController.singleton.leftHand.rotation;
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshWristUIButtonGrid() {
           
            if (_uiActivated && _currentGridScreen >= 0) {
                _uidMainButtonCanvasDisplay.DestroyUIButtons();
                _uidMainButtonCanvasDisplay.CreateUIButtons();
            }
        }
        
        public void ButtonActivated(int buttonRef, ButtonGrid buttonGrid)
        {
            try
            {
                UIAButton button = buttonGrid._buttonList[buttonRef - 1];
                if (_testSpawnModeActive) DeactivateSpawnTest();
                string vrHand = "NONE";
                if (SuperController.singleton.GetRightSelect()) vrHand = "RIGHT";
                if (SuperController.singleton.GetLeftSelect()) vrHand = "LEFT";

                if (SuperController.singleton.mainHUD == null) _hudOnPreButtonAction = false;
                else _hudOnPreButtonAction = SuperController.singleton.mainHUD.gameObject.activeInHierarchy;
                
                if (!_uiaEditMode) button.StartAction(vrHand);
                
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void MainButtonClickDown() {
            try {
                _timeLastMainButtonClickDown = Time.time;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void MainButtonClickFinish()
        {
            if (_uiaActiveClothingMode)
            {
                _uiaActiveClothingMode = false;
                RefreshWristUIButtonGrid();
            }
            else
            {
                if (Time.time - _timeLastMainButtonClickDown > 0.6f)
                {
                    _uiaEditMode = true;
                    _uiActivated = true;
                }
                else if (_uiaEditMode) _uiaEditMode = false;
                else _uiActivated = !_uiActivated;

                if (!_uiActivated)
                {
                    _uidMainButtonCanvasDisplay.DestroyUIButtons();
                    _isUIGazeDeactivated = false;
                }
                else LaunchWristUIButtonGrid();

            }
            
        }

        private void LaunchWristUIButtonGrid()
        {
            if (_currentGridScreen < 0 || _buttonGridsList[_currentGridScreen]._activeButtonCount == 0)
            {
                _currentGridScreen = -1;
                for (int i = _buttonGridsList.Count - 1; i >= 0; i--)
                {
                    if (_buttonGridsList[i]._activeButtonCount > 0) { _currentGridScreen = i; }
                }
            }
            if (_currentGridScreen == -1)
            {
                _uiaEditMode = false;
                _uiActivated = false;
            }
            RefreshWristUIButtonGrid();
        }

        private void MainButtonClickUp()
        {
            try
            {
                if (_vrActive)
                {
                    MainButtonClickFinish();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void ToggleUIActivateButton() {
            try {
                if (!_vrActive)
                {
                    MainButtonClickFinish();
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private int GetNextLeftScreen() {
            int nextGridScreen = -1;
            for (int i = 0; i < _currentGridScreen; i++) {
                if (_buttonGridsList[i]._activeButtonCount > 0) { nextGridScreen = i; }
            }
            return (nextGridScreen);
        }
        private int GetNextRightScreen() {
            int nextGridScreen = -1;
            for (int i = _buttonGridsList.Count - 1; i > _currentGridScreen; i--) {
                if (_buttonGridsList[i]._activeButtonCount > 0) { nextGridScreen = i; }
            }
            return (nextGridScreen);
        }
        private void ScrollLeftButton() {
            try {
                _currentGridScreen = GetNextLeftScreen();
                _uidMainButtonCanvasDisplay.DestroyUIButtons();
                if (_currentGridScreen > -1) {
                    _uidMainButtonCanvasDisplay.CreateUIButtons();
                    _uiActivated = true;
                }
                else { _uiActivated = false; }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void ScrollRightButton() {
            try {
                _currentGridScreen = GetNextRightScreen();
                _uidMainButtonCanvasDisplay.DestroyUIButtons();
                if (_currentGridScreen > -1) {
                    _uidMainButtonCanvasDisplay.CreateUIButtons();
                    _uiActivated = true;
                }
                else { _uiActivated = false; }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void CreateDefaultButtonGrids() {
            ButtonGrid defaultBG = new ButtonGrid(3, 3, this);
            _buttonGridsList.Add(defaultBG);
            defaultBG._buttonList[0].UpdateButtonCategory("Clothing");
            defaultBG._buttonList[1].UpdateButtonCategory("Presets");
            defaultBG._buttonList[1].UpdateButtonType("Load Clothing Preset");
            defaultBG._buttonList[1].SetPreset1File("MeshedVR.PresetsPack.1:/Custom/Atom/Person/Clothing/MeshedVR/PresetsPack/Preset_VaM_Seductress_Sim.vap");
            defaultBG._buttonList[1]._autoLabel = false;
            defaultBG._buttonList[1].UpdateTargetName("Last viewed Female");
            defaultBG._buttonList[2].UpdateButtonCategory("Presets");
            defaultBG._buttonList[2].SetPreset1File("MeshedVR.PresetsPack.1:/Custom/Atom/Person/Appearance/MeshedVR/PresetsPack/Preset_Ren_Tara.vap");
            defaultBG._buttonList[2]._autoLabel = false;
            defaultBG._buttonList[2].UpdateTargetName("Last viewed Female");
            defaultBG.RecalcActiveButtonCount();
            defaultBG.RecalcGazeSelections();
            
        }
        void Start() {
            try {

                if (_isActivePlugin) {
                    
                    if (DefaultUIAPFileExists())
                    {
                        LoadUIAPSelected(GetLocalPluginPath() + "\\default.uiap");
                    }
                    else
                    {
                        CreateDefaultButtonGrids();
                        _buttonGridSelectorJSON.choices = GetGridChoices();
                        _buttonGridDisplay = new ButtonGridDisplay(_buttonGridsList[0], this);

                        RefreshLeftUIDisplay();
                        RefreshRightUIDisplay();
                    }
                    _lastGazeTargetUpateTime = Time.time;
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        void DisableHeelAdjust()
        {
            if (_isPrimaryUIAJSON.val)
            {
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "Person")
                    {
                        _toolbox._heelAdjustTool.FootPoseReset(atom);
                    }
                }
            }
        }
        void DisableHidePersonNodes()
        {
            SuperController.singleton.SetOnlyShowControllers(null);
        }
        void OnEnable()
        {
            try
            {
                if (_isActivePlugin)
                {
                    if (_uiMainCanvas != null) { _uiMainCanvas._canvas.enabled = true; }
                    if (_uidMainButtonCanvasDisplay._uiButtonCanvas != null) { _uidMainButtonCanvasDisplay._uiButtonCanvas._canvas.enabled = true; }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        void OnDisable()
        {
            try
            {
                if (_isActivePlugin)
                {
                    DisableHeelAdjust();
                    if (_uiMainCanvas != null) { _uiMainCanvas._canvas.enabled = false; }
                    if (_uidMainButtonCanvasDisplay._uiButtonCanvas != null) { _uidMainButtonCanvasDisplay._uiButtonCanvas._canvas.enabled = false; }
                    DeactivateSpawnTest();
                    DisableHidePersonNodes();
                    if (_atomVRMoveActive)
                    {
                        _atomVRMoveActive = false;
                        SuperController.singleton.helpHUDText.text = "";
                        Atom previousMoveAtom = SuperController.singleton.GetAtomByUid(_atomVRMoveAtomName);
                        if (previousMoveAtom != null)
                        {
                            previousMoveAtom.tempFreezePhysics = false;
                        }
                    }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        void OnDestroy() {
            try {
                if (_isActivePlugin)
                {
                    DisableHeelAdjust();
                    if (_uiMainCanvas != null) { _uiMainCanvas.OnDestroy(); }
                    if (_uidMainButtonCanvasDisplay._uiButtonCanvas != null) { _uidMainButtonCanvasDisplay._uiButtonCanvas.OnDestroy(); }
                    SuperController.singleton.onAtomUIDRenameHandlers -= _atomRenameHandle;
                    SuperController.singleton.onAtomRemovedHandlers -= _atomRemoveHandle;

                    DeactivateSpawnTest();
                    DisableHidePersonNodes();
                    if (_atomVRMoveActive)
                    {
                        _atomVRMoveActive = false;
                        SuperController.singleton.helpHUDText.text = "";
                        Atom previousMoveAtom = SuperController.singleton.GetAtomByUid(_atomVRMoveAtomName);
                        if (previousMoveAtom != null)
                        {
                            previousMoveAtom.tempFreezePhysics = false;
                        }
                    }
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public float GetButtonDistance(UIDynamicButton button, Ray ray)
        {
            float distance = 1000f;
            Vector3[] rectCornersArray = new Vector3[4];

            RectTransform rt = button.transform.GetComponent<RectTransform>();
            rt.GetWorldCorners(rectCornersArray);
            Vector3 buttonCentre = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
            distance = Vector3.Cross(ray.direction, buttonCentre - ray.origin).magnitude;
            return (distance);
        }

        private void UpdateUIAButtons(bool updateGazeButtons = true, bool updateGridButtons = true, bool enableGazeToggleScreens = false) {
            ButtonGrid buttonGrid=null;
            if (_currentGridScreen >= 0) buttonGrid = _buttonGridsList[_currentGridScreen];
            if (_currentGridScreen<0 || buttonGrid._activeButtonCount==0) updateGridButtons = false;
            int nearestButtonRef = -1;
            float nearestButtonDistance = 1000f;
            int nearestQLButtonRef = -1;
            float nearestQLButtonDistance = 1000f;
            Vector3[] rectCornersArray = new Vector3[4];
            Transform cameraT = SuperController.singleton.lookCamera.transform;
            Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);

            Ray ray = new Ray(eyePosition, cameraT.forward);

            if (updateGridButtons) {
                if (!_uiaActiveClothingMode) _uidMainButtonCanvasDisplay._uiMainButtonGridDisplay.UpdateUIAButtonGrid(updateGazeButtons || enableGazeToggleScreens, out nearestButtonRef, out nearestButtonDistance );
                if (_activateQuickLaunchBarJSON.val && !_uiaActiveClothingMode) _uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay.UpdateUIAButtonGrid(updateGazeButtons || enableGazeToggleScreens, out nearestQLButtonRef, out nearestQLButtonDistance);
            }
            float leftUIButtonDistance = 1000f;
            float rightUIButtonDistance = 1000f;
            float mainUIButtonDistance = 1000f;
            float pinUIButtonDistance = 1000f;
            float nearestGazeSelectButtonDistance = 1000f;
            int nearestGazeSelectButtonRef = -1;
            

            if (updateGazeButtons || enableGazeToggleScreens)
            {
                mainUIButtonDistance = GetButtonDistance(_mainButton, ray);

                if (_leftButton.gameObject.activeSelf) leftUIButtonDistance = GetButtonDistance(_leftButton, ray);
                if (_rightButton.gameObject.activeSelf) rightUIButtonDistance = GetButtonDistance(_rightButton, ray);
                if (_mainPinButton.gameObject.activeSelf) pinUIButtonDistance = GetButtonDistance(_mainPinButton, ray);
            }          

            if (_mainPinButton.gameObject.activeSelf && _uiPinned) _mainPinButton.label = "O";
            else if (_mainPinButton.gameObject.activeSelf && !_uiPinned) _mainPinButton.label = "X";

            if ( _uiActivated && !_uiaActiveClothingMode) _uidMainButtonCanvasDisplay.UpdateGASScrollList(ref nearestGazeSelectButtonDistance, ref  nearestGazeSelectButtonRef, updateGazeButtons || enableGazeToggleScreens);

           
            _gazeUIButton = -1;
            _gazeQLBUIButton = -1;
            _gazeNodeSelectButtonRef = -1;
            _gazeMainButton = false;
            _gazeLeftButton = false;
            _gazeRightButton = false;
            _gazePinButton = false;
            _leftButton.buttonColor = Color.grey;
            _rightButton.buttonColor = Color.grey;
            _mainPinButton.buttonColor = Color.grey;
            _mainButton.buttonColor = Color.grey;
            if (updateGazeButtons)
            {
                Color gazeButtonColor = new Color();
                gazeButtonColor = Color.HSVToRGB(_defaultButtonSkin._buttonGazeSelectColor.H, _defaultButtonSkin._buttonGazeSelectColor.S, _defaultButtonSkin._buttonGazeSelectColor.V);

                if (nearestQLButtonDistance < 0.1f
                    && nearestQLButtonDistance < nearestButtonDistance
                    && nearestQLButtonDistance < mainUIButtonDistance
                    && nearestQLButtonDistance < leftUIButtonDistance
                    && nearestQLButtonDistance < rightUIButtonDistance
                    && nearestQLButtonDistance < pinUIButtonDistance
                    && nearestQLButtonDistance < nearestGazeSelectButtonDistance)
                {
                    _uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay.UpdateButtonColor(nearestQLButtonRef, gazeButtonColor) ;
                    if (updateGazeButtons) _gazeQLBUIButton = nearestQLButtonRef;
                }
                else if (nearestButtonDistance < 0.1f
                    && nearestButtonDistance < mainUIButtonDistance
                    && nearestButtonDistance < leftUIButtonDistance
                    && nearestButtonDistance < rightUIButtonDistance
                    && nearestButtonDistance < pinUIButtonDistance
                    && nearestButtonDistance < nearestGazeSelectButtonDistance)
                {
                    _uidMainButtonCanvasDisplay._uiMainButtonGridDisplay.UpdateButtonColor(nearestButtonRef, gazeButtonColor);                    
                    if (updateGazeButtons) _gazeUIButton = nearestButtonRef;
                }
                else if (leftUIButtonDistance < 0.1f
                    && leftUIButtonDistance < mainUIButtonDistance
                    && leftUIButtonDistance < rightUIButtonDistance
                    && leftUIButtonDistance < pinUIButtonDistance
                    && leftUIButtonDistance < nearestGazeSelectButtonDistance)
                {
                    _leftButton.buttonColor = gazeButtonColor;
                    if (updateGazeButtons) _gazeLeftButton = true;
                }
                else if (rightUIButtonDistance < 0.1f
                    && rightUIButtonDistance < mainUIButtonDistance
                    && rightUIButtonDistance < pinUIButtonDistance
                    && rightUIButtonDistance < nearestGazeSelectButtonDistance)
                {
                    _rightButton.buttonColor = gazeButtonColor;
                    if (updateGazeButtons) _gazeRightButton = true;
                }
                else if (pinUIButtonDistance < 0.1f
                    && pinUIButtonDistance < mainUIButtonDistance
                    && pinUIButtonDistance < nearestGazeSelectButtonDistance)
                {
                    _mainPinButton.buttonColor = gazeButtonColor;
                    if (updateGazeButtons) _gazePinButton = true;
                }
                else if (nearestGazeSelectButtonDistance < 0.1f
                    && nearestGazeSelectButtonDistance < mainUIButtonDistance)
                {
                    _uidMainButtonCanvasDisplay.UpdateGASButtonColor(nearestGazeSelectButtonRef, gazeButtonColor);
                    
                    if (updateGazeButtons) _gazeNodeSelectButtonRef = nearestGazeSelectButtonRef;
                }
                else if (mainUIButtonDistance < 0.1f)
                {
                    _mainButton.buttonColor = gazeButtonColor;
                    if (updateGazeButtons) _gazeMainButton = true;
                }
            }
            
            if (updateGridButtons && !_uiaEditMode && !_uiaActiveClothingMode) {
                if (_displayGazeSelectionsJSON.val && (buttonGrid.IsSelectionTargetButton() || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid.IsSelectionTargetButton()))) {
                    string text = "";
                    float height = Constants._gazeStatusMaxHeight/5f;
                    _uidMainButtonCanvasDisplay._gazeStatusTextField.gameObject.SetActive(true);
                    if (buttonGrid._gazeSelectionFemales || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._gazeSelectionFemales)) {
                        if (text != "") text = text + "\n";
                        text = text + "Gaze Target Female: " + _lastViewedFemale;
                        height = height + Constants._gazeStatusMaxHeight / 5f; 
                    }
                    if (buttonGrid._gazeSelectionMales || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._gazeSelectionMales)) {
                        if (text != "") text = text + "\n";
                        text = text + "Gaze Target Male: " + _lastViewedMale;
                        height = height + Constants._gazeStatusMaxHeight / 5f; 
                    }
                    if (buttonGrid._gazeSelectionPerson || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._gazeSelectionPerson)) {
                        if (text != "") text = text + "\n";
                        text = text + "Gaze Target Person: " + _lastViewedPerson;
                        height = height + Constants._gazeStatusMaxHeight / 5f; 
                    }
                    if (buttonGrid._gazeSelectionAtoms || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._gazeSelectionAtoms))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Gaze Target Atom: " + _lastViewedAtom;
                        height = height + Constants._gazeStatusMaxHeight / 5f; 
                    }
                    if (buttonGrid._vamSelectionFemales || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._vamSelectionFemales))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Selected Target Female: " + _lastSelectedFemale;
                        height = height + Constants._gazeStatusMaxHeight / 5f;
                    }
                    if (buttonGrid._vamSelectionMales || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._vamSelectionMales))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Selected Target Male: " + _lastSelectedMale;
                        height = height + Constants._gazeStatusMaxHeight / 5f;
                    }
                    if (buttonGrid._vamSelectionPerson || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._vamSelectionPerson))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Selected Target Persom: " + _lastSelectedPerson;
                        height = height + Constants._gazeStatusMaxHeight / 5f;
                    }
                    if (buttonGrid._vamSelectionAtoms || (_activateQuickLaunchBarJSON.val && _quickLaunchButtonGrid._vamSelectionAtoms))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Selected Target Atom: " + _lastSelectedAtom;
                        height = height + Constants._gazeStatusMaxHeight / 5f;
                    }
                    _uidMainButtonCanvasDisplay._gazeStatusTextField.text = text;
                    _uidMainButtonCanvasDisplay._gazeStatusTextField.height = height;
                    _uidMainButtonCanvasDisplay._gazeStatusTextField.UItext. alignment = TextAnchor.MiddleLeft;
                 }
                else if (_uidMainButtonCanvasDisplay._gazeStatusTextField != null) { _uidMainButtonCanvasDisplay._gazeStatusTextField.gameObject.SetActive(false); }
            }

            Vector3 mainButtonToEyeDirection = _mainPinButton.transform.position - cameraT.position;
            float mainButtonToEyeAngle = Vector3.Angle(_mainPinButton.transform.forward, mainButtonToEyeDirection);
            if (!_uiActivated && mainUIButtonDistance < 0.2f && enableGazeToggleScreens && mainButtonToEyeAngle < 25f && _isUIGazeDeactivated)
            {
//                _timeLastMainButtonClickDown = Time.time;
                _uiActivated = true;
                LaunchWristUIButtonGrid();
            }
            if (_uiActivated && ((mainUIButtonDistance > 0.25f && nearestButtonDistance > 0.1f && nearestGazeSelectButtonDistance >0.1f && nearestQLButtonDistance>0.1f) || mainButtonToEyeAngle > 90f) && enableGazeToggleScreens)
            {
 //               _timeLastMainButtonClickDown = Time.time;
                //               MainButtonClickFinish();
                _uiActivated = false;
                _uidMainButtonCanvasDisplay.DestroyUIButtons();
                _isUIGazeDeactivated = true;
            }
            if (!_uiActivated && ((mainUIButtonDistance > 0.25f && nearestButtonDistance > 0.1f) || mainButtonToEyeAngle > 90f)) _isUIGazeDeactivated = true;
            
        }
        
        private void UpdateLeftRightButtons() {
            if (_uiActivated && !_uiaActiveClothingMode) {
                if (GetNextLeftScreen() >= 0) { _leftButton.gameObject.SetActive(true); }
                else { _leftButton.gameObject.SetActive(false); }
                if (GetNextRightScreen() >= 0) { _rightButton.gameObject.SetActive(true); }
                else { _rightButton.gameObject.SetActive(false); }
            }
            else {
                _leftButton.gameObject.SetActive(false);
                _rightButton.gameObject.SetActive(false);
            }
        }
        public bool IsMale(Atom atom)
        {
            return (atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian"));
        }
        public void UpdateGazeTargets() {
            List<GazeTarget> gazeAtomTargetsNarrowFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeAtomTargetsWideFOVList = new List<GazeTarget>();
            List<GazeTarget> gazePersonTargetsNarrowFOVList = new List<GazeTarget>();
            List<GazeTarget> gazePersonTargetsWideFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeFemaleTargetsNarrowFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeFemaleTargetsWideFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeMaleTargetsNarrowFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeMaleTargetsWideFOVList = new List<GazeTarget>();

            Dictionary<GazeTarget, float> gazeSelectTargetsNarrowFOVByDistance = new Dictionary<GazeTarget,float >();
            Dictionary<GazeTarget, float> gazeSelectTargetsWideFOVByAngle = new Dictionary<GazeTarget, float>();

            List<string> notHiddenAtomNames = SuperController.singleton.GetAtomUIDs();
            HashSet<FreeControllerV3> onlyShowControllers = new HashSet<FreeControllerV3>();

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                //Identify if atom is Male or Female (or neither)
                bool male = false;
                bool female = false;
                if (atom.type == "Person" && atom.on &&(IsMale(atom))) male = true;
                else if (atom.type == "Person" && atom.on) { female = true; }

                if (atom.name != "CoreControl" && atom.name != "[CameraRig]" && atom.on) {
                    // Assess Rigidbodies as Gaze Targets
                    if (!atom.hidden || SuperController.singleton.showHiddenAtoms)
                    {
                        foreach (Rigidbody rb in atom.rigidbodies.Where(rb => Constants._gazeRBs.Contains(rb.name)))
                        {
                            
                            Vector3 rbPosition = rb.transform.position;
                            Transform playerHeadTransform = CameraTarget.centerTarget.transform;
                            Vector3 playerLookingVector = playerHeadTransform.forward;
                            Vector3 directionToTarget = rbPosition - playerHeadTransform.position;
                            float angle = Vector3.Angle(playerLookingVector, directionToTarget);
                            float distance = Vector3.Distance(rbPosition, playerHeadTransform.position);
                            GazeTarget gazeTarget = new GazeTarget();
                            gazeTarget.atomName = atom.name;
                            gazeTarget.nodeName = rb.name;
                            gazeTarget.angle = angle;
                            gazeTarget.distance = distance;
                            if (male) gazeTarget.isMale = true;
                            if (female) gazeTarget.isFemale = true;
                            if (angle <= _gazeFOVNarrow / 2f)
                            {
                                gazeAtomTargetsNarrowFOVList.Add(gazeTarget);
                                if (male)
                                {
                                    gazeMaleTargetsNarrowFOVList.Add(gazeTarget);
                                    gazePersonTargetsNarrowFOVList.Add(gazeTarget);
                                }
                                else if (female)
                                {
                                    gazeFemaleTargetsNarrowFOVList.Add(gazeTarget);
                                    gazePersonTargetsNarrowFOVList.Add(gazeTarget);
                                }
                            }
                            if (angle <= _gazeFOVWide / 2f)
                            {
                                gazeAtomTargetsWideFOVList.Add(gazeTarget);
                                if (male)
                                {
                                    gazeMaleTargetsWideFOVList.Add(gazeTarget);
                                    gazePersonTargetsWideFOVList.Add(gazeTarget);
                                }
                                else if (female)
                                {
                                    gazeFemaleTargetsWideFOVList.Add(gazeTarget);
                                    gazePersonTargetsWideFOVList.Add(gazeTarget);
                                }
                            }
                        }
                    }
                    
                    foreach (FreeControllerV3 fc in atom.freeControllers)
                    {
                        // Assess FreeControllers to determine if they should be hidden or not
                        if (atom.type=="Person" && fc.name!="control" && IsSelectableFreeController(fc.name, male))
                        {
                            if ((fc.currentPositionState != FreeControllerV3.PositionState.Off || fc.currentRotationState != FreeControllerV3.RotationState.Off) && UserPreferences.singleton.hideInactiveTargets)
                            {
                                onlyShowControllers.Add(fc);
                            }
                        }
                        if (atom.type!="Person" && UserPreferences.singleton.hideInactiveTargets) onlyShowControllers.Add(fc);

                        //Assess Freecontrollers as Gaze Assisted Select nodes
                        if (IsSelectableFreeController(fc.name, male) && (!UserPreferences.singleton.hideInactiveTargets || onlyShowControllers.Contains(fc) || atom.type != "Person" || fc.name == "control") && atom.on && notHiddenAtomNames.Contains(atom.name))
                        {                            
                            Vector3 fcPosition = fc.transform.position;
                            Transform playerHeadTransform = CameraTarget.centerTarget.transform;
                            Vector3 playerLookingVector = playerHeadTransform.forward;
                            Vector3 directionToTarget = fcPosition - playerHeadTransform.position;
                            float angle = Vector3.Angle(playerLookingVector, directionToTarget);
                            float distance = Vector3.Distance(fcPosition, playerHeadTransform.position);
                            GazeTarget gazeTarget = new GazeTarget();
                            gazeTarget.atomName = atom.name;
                            gazeTarget.nodeName = fc.name;
                            gazeTarget.angle = angle;
                            gazeTarget.distance = distance;
                            if (male) gazeTarget.isMale = true;
                            if (female) gazeTarget.isFemale = true;

                            if (angle <= _gazeFOVNarrow / 2f)
                            {
                                gazeSelectTargetsNarrowFOVByDistance.Add(gazeTarget, distance);
                            }
                            else
                            {
                                gazeSelectTargetsWideFOVByAngle.Add(gazeTarget, angle);
                            }
                        }
                        
                    }
                    
                }
            }
            if (onlyShowControllers.Count > 0) SuperController.singleton.SetOnlyShowControllers(onlyShowControllers);
            else SuperController.singleton.SetOnlyShowControllers(null);

            string atomName = GetClosestGazeTarget(gazeAtomTargetsNarrowFOVList, gazeAtomTargetsWideFOVList);
            if (atomName != "") _lastViewedAtom = atomName;

            string personName = GetClosestGazeTarget(gazePersonTargetsNarrowFOVList, gazePersonTargetsWideFOVList);
            if (personName != "") _lastViewedPerson = personName;

            string femaleName = GetClosestGazeTarget(gazeFemaleTargetsNarrowFOVList, gazeFemaleTargetsWideFOVList);
            if (femaleName != "") _lastViewedFemale = femaleName;
           
            string maleName = GetClosestGazeTarget(gazeMaleTargetsNarrowFOVList, gazeMaleTargetsWideFOVList);
            if (maleName != "") _lastViewedMale = maleName;

            _uidMainButtonCanvasDisplay._orderedGazeAssistSelectionTargets.Clear();

            List<KeyValuePair<GazeTarget, float>> sortedListKVPs = new List<KeyValuePair<GazeTarget, float>>(gazeSelectTargetsNarrowFOVByDistance);
            sortedListKVPs.Sort(
                delegate (KeyValuePair<GazeTarget, float> firstPair,
                KeyValuePair<GazeTarget, float> nextPair)
                {
                    return firstPair.Value.CompareTo(nextPair.Value);
                }
            );
            foreach (KeyValuePair<GazeTarget, float> kvp in sortedListKVPs) _uidMainButtonCanvasDisplay._orderedGazeAssistSelectionTargets.Add(kvp.Key);

            sortedListKVPs = new List<KeyValuePair<GazeTarget, float>>(gazeSelectTargetsWideFOVByAngle);
            sortedListKVPs.Sort(
                delegate (KeyValuePair<GazeTarget, float> firstPair,
                KeyValuePair<GazeTarget, float> nextPair)
                {
                    return firstPair.Value.CompareTo(nextPair.Value);
                }
            );
            foreach (KeyValuePair<GazeTarget, float> kvp in sortedListKVPs) _uidMainButtonCanvasDisplay._orderedGazeAssistSelectionTargets.Add(kvp.Key);

        }
        public string GetClosestGazeTarget(List<GazeTarget> gazeTargetsNarrowFOVList, List<GazeTarget> gazeTargetsWideFOVList)
        {
            string atomName = "";

            if (gazeTargetsNarrowFOVList.Count > 0)
            {
                float distance = -1f;
                foreach (GazeTarget gazeTarget in gazeTargetsNarrowFOVList)
                {
                    if (distance == -1f || gazeTarget.distance < distance)
                    {
                        atomName = gazeTarget.atomName;
                        distance = gazeTarget.distance;
                    }
                }
            }
            else if (gazeTargetsWideFOVList.Count > 0)
            {
                float angle = -1f;
                foreach (GazeTarget gazeTarget in gazeTargetsWideFOVList)
                {
                    if (angle == -1f || gazeTarget.angle < angle)
                    {
                        atomName = gazeTarget.atomName;
                        angle = gazeTarget.angle;
                    }
                }
            }
            return (atomName);
        }
        public void SwitchVRHand(string pluginHand)
        {
            UpdateOtherUIAssistVRHand(pluginHand);
            
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
                RefreshWristUIButtonGrid();
            }
        }
        public void SwitchToolBox(string toolbox)
        {
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
            }
        }
        public void SwitchVRPinMode(string pinMode)
        {
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
            }
        }

        private void UpdateOtherUIAssistVRHand(string pluginHand) {
            foreach (string storableName in containingAtom.GetStorableIDs()) {
                if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);

                    if (pluginType == "JayJayWon.UIAssist" && storableName != this.name) {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");
                        if (activeJSON.val) {
                            JSONStorableStringChooser otherPluginVRHand = pluginStorable.GetStringChooserJSONParam("handControl");
                            if (otherPluginVRHand.val == _vrHandControlJSON.val) {
                                if (_vrHandControlJSON.val == "Left Hand") { otherPluginVRHand.val = "Right Hand"; }
                                else { otherPluginVRHand.val = "Left Hand"; }
                            }
                            return;
                        }
                    }

                }
            }
        }

        private void UpdateActiveClothingLists()
        {
            _atomActiveClothingDictionary.Clear();
            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "Person")
                {
                    ActiveClothingLists activeClothingLists = new ActiveClothingLists();
                    activeClothingLists.UpdateActiveClothing(atom);
                    _atomActiveClothingDictionary.Add(atom.name, activeClothingLists);

                    
                    if (_toolbox._heelAdjustTool._heelAdjustActiveJSON.val && _isPrimaryUIAJSON.val && !atom.isPhysicsFrozen && !SuperController.singleton.isLoading) _toolbox._heelAdjustTool.LoadFootPosesForClothing(atom, activeClothingLists._activeClothingHAFileNames);
                }

            }
        }
        private void UpdateHUDText()
        {
            if (_testSpawnModeActive)
            {
                string actionVerb = "Spawn";
                if (_testSpawnButton._buttonType == "Teleport Atom") actionVerb = "Teleport";
                SuperController.singleton.helpHUDText.text = "UIAssist: Atom "+actionVerb+" Test Mode Active";
            }
            else if (_atomSpawningActive)
            {
                SuperController.singleton.helpHUDText.text = "UIAssist: Atom Spawning In Progress";
            }
            else if (_atomVRMoveActive)
            {
                if (_atomVRMovePaused) SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Paused. Release Grab button to unpause.";
                else SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Active. Click Grab button to deactivate or hold Grab button to pause.";
            }
        }

 /*       private void CheckForRemovedGazeTargets()
        {
            List<string> atomNames = SuperController.singleton.GetAtomUIDs();

            if (!atomNames.Contains(_lastViewedAtom)) _lastViewedAtom = "";
            if (!atomNames.Contains(_lastViewedFemale)) _lastViewedFemale = "";
            if (!atomNames.Contains(_lastViewedMale)) _lastViewedMale = "";
            if (!atomNames.Contains(_lastViewedPerson)) _lastViewedPerson = "";
        }
*/
        private void UpdateSelectedAtom()
        {
            Atom atom = SuperController.singleton.GetSelectedAtom();
            if (atom != null)
            {
                SuperController.LogMessage(atom.name);
                _lastSelectedAtom = atom.name;
                if (atom.type == "Person")
                {
                    _lastSelectedPerson = atom.name;
                    if (IsMale(atom)) _lastSelectedMale = atom.name;
                    else _lastSelectedFemale = atom.name;
                }
            }
        }

        public string _lastLoadSceneFromVar = "";
        void Update() {
            try {
                // Detects when a scene has been loaded from VAR - used by ResetAppearance & ResetScale features
                string currentLoadDir  = SuperController.singleton.currentLoadDir;
                if (currentLoadDir != _lastLoadSceneFromVar && currentLoadDir.Contains(":/Saves/scene"))
                {
                    _lastLoadSceneFromVar = currentLoadDir;               
                }
               
                if (_isActivePlugin) {
                    if (SuperController.singleton.GetSelectedAtom() !=null && SuperController.singleton.GetSelectedAtom().name != _lastSelectedAtom) UpdateSelectedAtom();
                    

                    // Supports button repositioning when in UIA Edit mode
                    if (_uiaEditMode) _uidMainButtonCanvasDisplay.UpdateEditModeButtonPosition();

                    float currentTime = Time.time;
                    bool triggerUIButton = false;

                    // Update Gaze targets every 0.1 seconds
                    if (currentTime - _lastGazeTargetUpateTime >0.1f)
                    {
                        UpdateOtherUIAssistVRHand(_vrHandControlJSON.val);
 //                       CheckForRemovedGazeTargets();
                        _uidMainButtonCanvasDisplay._uiActiveClothingEditor.CheckForRemovedAtoms();

                        UpdateGazeTargets();
                        _lastGazeTargetUpateTime = currentTime;
                        UpdateSpawnTestAtomPosition();
                    }
                    if (SuperController.singleton.helpHUDText.text == "") UpdateHUDText();

                    // Enables or disable the Gaze Assist Select select box
                    if (_toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val && SuperController.singleton.gameMode == SuperController.GameMode.Edit && _uidMainButtonCanvasDisplay._gasButtonScrollListGO == null && !_uiaEditMode) RefreshWristUIButtonGrid();
                    if (_uidMainButtonCanvasDisplay._gasButtonScrollListGO != null && (!_toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val || SuperController.singleton.gameMode == SuperController.GameMode.Play)  ) RefreshWristUIButtonGrid();

                    string uiDisplayMode = "";
                    bool gazeControlActive = false;

                    if (_vrActive == false) uiDisplayMode = "HUD";
                    else if (_vrHandControlJSON.val == "Disabled" || (_uiPinned && (_vrPinModeJSON.val == "HUD (Left)" || _vrPinModeJSON.val == "HUD (Right)" || _vrPinModeJSON.val == "HUD (Custom)"))) uiDisplayMode = "VRHUD";
                    else if (_vrPinModeJSON.val == "Disabled")
                    {
                        uiDisplayMode = "WRIST";
                        if (_gazeMenuSelectModeJSON.val != "Disabled") gazeControlActive = true;
                    }
                    else if (_uiPinned)
                    {
                        uiDisplayMode = "STATIC";
                        if (_gazeMenuSelectModeJSON.val == "Wrist & Pinned UI") gazeControlActive = true;
                    }
                    else
                    {
                        uiDisplayMode = "WRIST";
                        if (_gazeMenuSelectModeJSON.val != "Disabled") gazeControlActive = true;
                    }

                    // Check if pin Mode is available
                    bool pinModeAvailable = false;
                    if (uiDisplayMode != "HUD" && _vrPinModeJSON.val != "Disabled" && _vrHandControlJSON.val != "Disabled" && _uiActivated)
                    {
                        _mainPinButton.gameObject.SetActive(true);
                        pinModeAvailable = true;
                    }
                    else _mainPinButton.gameObject.SetActive(false);

                    // Updates Heel Adjust
                    if (_timeLastHAFileListUpdated < 0f || (currentTime - _timeLastHAFileListUpdated > Constants._TimeBetweenHAFileUpdates))
                    {
                        if (_toolbox._heelAdjustTool._heelAdjustActiveJSON.val)_toolbox._heelAdjustTool.UpdateExistingHAFiles();
                        _timeLastHAFileListUpdated = currentTime;
                    }

                    // Upades Active Clothing Lists and Editor
                    if (_timeLastClothingUpdate < 0f || ((currentTime - _timeLastClothingUpdate) > Constants._TimeBetweenClothingUpdates))
                    {
                        UpdatePrimaryUIA();
                        _toolbox._heelAdjustTool.ResetAllSavedAtomRotationStates();
                        UpdateActiveClothingLists();
                        if (_uiaActiveClothingMode && _uiActivated )
                        {
                            string targetAtom = "";
                            if (_uiaActiveClothingModeTargetType == "Last viewed Person") targetAtom = _lastViewedPerson;
                            else if (_uiaActiveClothingModeTargetType == "Last viewed Male") targetAtom = _lastViewedMale;
                            else targetAtom = _lastViewedFemale;
                            if (targetAtom == "" || !_atomActiveClothingDictionary.ContainsKey(targetAtom))
                            {
                                _uidMainButtonCanvasDisplay._uiActiveClothingEditor.UpadateACE(targetAtom, null);
                            }
                            else _uidMainButtonCanvasDisplay._uiActiveClothingEditor.UpadateACE(targetAtom, _atomActiveClothingDictionary[targetAtom]);
                        }                        
                        _timeLastClothingUpdate = currentTime;
                    }

                    // Check if Grab activated and trigger gaze action 
                    if (SuperController.singleton.GetLeftGrab() && _vrHandControlJSON.val == "Left Hand" && gazeControlActive) triggerUIButton = true;
                    if (SuperController.singleton.GetRightGrab() && _vrHandControlJSON.val == "Right Hand" && gazeControlActive) triggerUIButton = true;
                    if (triggerUIButton && !_uiaEditMode) {
                        if (_gazeQLBUIButton > -1) ButtonActivated(_gazeQLBUIButton + 1, _quickLaunchButtonGrid);
                        else if (_gazeUIButton > -1) ButtonActivated(_gazeUIButton + 1, _buttonGridsList[_currentGridScreen]);
                        else if (_gazeNodeSelectButtonRef > -1) _uidMainButtonCanvasDisplay.GazeAssistSelectButtonActivated(_gazeNodeSelectButtonRef);
                        else if (_gazeMainButton)
                        {
                            _timeLastMainButtonClickDown = Time.time;
                            MainButtonClickFinish();
                        }
                        else if (_gazeLeftButton) ScrollLeftButton();
                        else if (_gazeRightButton) ScrollRightButton();
                        else if (_gazePinButton) TogglePin();
                    }

                    // Check and handle if VR Move Atom is active 
                    if (_atomVRMoveActive)
                    {
                        Transform vrMoveHandTransform = null;
                        Atom atom = SuperController.singleton.GetAtomByUid(_atomVRMoveAtomName);
                        if (_atomVRMoveControllerLorR == "LEFT") vrMoveHandTransform = SuperController.singleton.leftHand;
                        else if (_atomVRMoveControllerLorR == "RIGHT") vrMoveHandTransform = SuperController.singleton.rightHand;

                        if (atom != null && vrMoveHandTransform!=null)
                        {
                            bool buttonGrabActived = false;

                            if (_atomVRMoveControllerLorR == "LEFT" && SuperController.singleton.GetLeftGrab()) buttonGrabActived = true;
                            if (_atomVRMoveControllerLorR == "RIGHT" && SuperController.singleton.GetRightGrab()) buttonGrabActived = true;                            

                            if (buttonGrabActived)
                            {
                                _atomVRMovePaused = true;
                                _atomVRMovePauseStartTime = Time.time;
                            }
                            bool buttonGrabDeactived = false;
                            if (_atomVRMoveControllerLorR == "LEFT" && SuperController.singleton.GetLeftGrabRelease()) buttonGrabDeactived = true;
                            if (_atomVRMoveControllerLorR == "RIGHT" && SuperController.singleton.GetRightGrabRelease()) buttonGrabDeactived = true;
                            if (buttonGrabDeactived)
                            {
                                _atomVRMovePaused = false;
                                float timeSincePauseStarted = Time.time - _atomVRMovePauseStartTime;
                                if (timeSincePauseStarted < 0.5f)
                                {
                                    _atomVRMoveActive = false;
                                    SuperController.singleton.helpHUDText.text = "";
                                    atom.tempFreezePhysics = false;
                                }
                                else
                                {
                                    _atomVRMoveControllerLastPosition = vrMoveHandTransform.position;
                                    _atomVRMoveControllerLastRotation = vrMoveHandTransform.rotation;

                                    _atomVRMoveAtomLastPosition = atom.mainController.transform.position;
                                    _atomVRMoveAtomLastRotation = atom.mainController.transform.rotation;
                                }
                                _atomVRMovePauseStartTime = 0f;
                            }

                            if (!_atomVRMovePaused && _atomVRMoveActive)
                            {
                                float postitionGranularityMultiplier = 0f;
                                if (_currentVRMovevrMovePositionGranulatity == "Extra Fine") postitionGranularityMultiplier = 0.5f;
                                if (_currentVRMovevrMovePositionGranulatity == "Fine") postitionGranularityMultiplier = 1f;
                                if (_currentVRMovevrMovePositionGranulatity == "Medium") postitionGranularityMultiplier = 2f;
                                if (_currentVRMovevrMovePositionGranulatity == "Coarse") postitionGranularityMultiplier = 3f;
                                if (_currentVRMovevrMovePositionGranulatity == "Extra Coarse") postitionGranularityMultiplier = 4f;
                                Vector3 newPosition = new Vector3();
                                newPosition = _atomVRMoveAtomLastPosition + ((vrMoveHandTransform.position - _atomVRMoveControllerLastPosition) * postitionGranularityMultiplier);
                                if (_currentVRMovelockedPositionX) newPosition.x = _atomVRMoveAtomLastPosition.x;
                                if (_currentVRMovelockedPositionY) newPosition.y = _atomVRMoveAtomLastPosition.y;
                                if (_currentVRMovelockedPositionZ) newPosition.z = _atomVRMoveAtomLastPosition.z;
                                atom.mainController.transform.position = newPosition;


                                float rotationGranularityMultiplier = 0f;
                                if (_currentVRMovevrMoveRotationGranulatity == "Fine") rotationGranularityMultiplier = 1f;
                                if (_currentVRMovevrMoveRotationGranulatity == "Medium") rotationGranularityMultiplier = 2f;
                                if (_currentVRMovevrMoveRotationGranulatity == "Coarse") rotationGranularityMultiplier = 3f;

                                Quaternion angleDifference = vrMoveHandTransform.rotation * Quaternion.Inverse(_atomVRMoveControllerLastRotation);
                                Vector3 newRotationVector = new Vector3(_atomVRMoveAtomLastRotation.eulerAngles.x-(angleDifference.eulerAngles.x * rotationGranularityMultiplier), _atomVRMoveAtomLastRotation.eulerAngles.y + (angleDifference.eulerAngles.y * rotationGranularityMultiplier), _atomVRMoveAtomLastRotation.eulerAngles.z - (angleDifference.eulerAngles.z * rotationGranularityMultiplier));

                                if (_currentVRMovelockedRotationX) newRotationVector.x = _atomVRMoveAtomLastRotation.eulerAngles.x;
                                if (_currentVRMovelockedRotationY) newRotationVector.y = _atomVRMoveAtomLastRotation.eulerAngles.y;
                                if (_currentVRMovelockedRotationZ) newRotationVector.z = _atomVRMoveAtomLastRotation.eulerAngles.z;
                                Quaternion yRotation = Quaternion.Euler(newRotationVector.x, newRotationVector.y, newRotationVector.z);
                                atom.mainController.transform.rotation = yRotation;
                            }
                        }
                    }

                    // Check if a double click Select is starting or completing - to allow Toggle the pin
                    bool detachToggle = false;
                    if (pinModeAvailable)
                    {
                        if (SuperController.singleton.GetLeftSelect())
                        {
                            if (_timeLeftSelect == -1f) _timeLeftSelect = currentTime;
                            else
                            {
                                if (currentTime - _timeLeftSelect > 0.4f) _timeLeftSelect = currentTime;
                                else
                                {
                                    _timeLeftSelect = -1f;
                                    if (_vrHandControlJSON.val == "Left Hand") detachToggle = true;
                                }
                            }
                        }
                        if (SuperController.singleton.GetRightSelect())
                        {
                            if (_timeRightSelect == -1f) _timeRightSelect = currentTime;
                            else
                            {
                                if (currentTime - _timeRightSelect > 0.4f) _timeRightSelect = currentTime;
                                else
                                {
                                    _timeRightSelect = -1f;
                                    if (_vrHandControlJSON.val == "Right Hand") detachToggle = true;
                                }
                            }
                        }
                        if (detachToggle) TogglePin();
                    }

                    // Set the 'read only' parameters that are visible to other plugins
                    _isActivePluginJSON.val = true;
                    _pluginVersionJSON.val = pluginVersion;
                    _pluginVersionFloatJSON.val = pluginVersionFloat;
                    _isPatreonOnlyVersionJSON.val = Constants._patreonContentEnabled;

                    UpdateUIAButtons(gazeControlActive && !_uiaEditMode, _uiActivated , _gazeActivateVRScreenJSON.val && !(uiDisplayMode == "HUD") && !(uiDisplayMode == "VRHUD")&& !_uiaEditMode);

                    Transform vrHandTransform;
                    bool vrRightHand = _vrHandControlJSON.val == "Right Hand";
                    if (vrRightHand) vrHandTransform = SuperController.singleton.rightHand;
                    else vrHandTransform = SuperController.singleton.leftHand;
                    Vector3 canvasPosition;
                    Quaternion canvasRotation;
                    float horizontalHUDOffset = 0.9f;
                    float verticalHUDOffset = 0.065f;
                    float depthHUDOffset = 0.6f;

                    if (uiDisplayMode == "STATIC") {
                        canvasPosition = _fixedPinPosition;
                        canvasRotation = _fixedPinRotation;
                        float distance = Vector3.Distance(_fixedPinPosition, vrHandTransform.position);
                        if (distance > 0.3f) _autoDePinEnabled = true;
                        if (distance < 0.1f & _autoDePinEnabled) TogglePin();
                    }
                    else {
                        canvasPosition = vrHandTransform.position;
                        canvasRotation = vrHandTransform.rotation;
                    }
                    if (uiDisplayMode == "VRHUD")
                    {
                        if (_vrPinModeJSON.val == "HUD (Right)")
                        {
                            horizontalHUDOffset = 0.57f;
                            verticalHUDOffset = 0.3f;
                        }
                        else if (_vrPinModeJSON.val == "HUD (Left)")
                        {
                            horizontalHUDOffset = 0.43f;
                            verticalHUDOffset = 0.3f;
                        }
                        else if (_vrPinModeJSON.val == "HUD (Custom)")
                        {
                            horizontalHUDOffset = _vrHUDHorizontalOffsetJSON.val;
                            verticalHUDOffset = _vrHUDVerticalOffsetJSON.val;
                            depthHUDOffset = _vrHUDDepthOffsetJSON.val;
                        }
                    }
                    _uiMainCanvas.Update(canvasPosition, canvasRotation, uiDisplayMode, uiDisplayMode == "STATIC", vrRightHand, verticalHUDOffset, horizontalHUDOffset, depthHUDOffset);
                    _uidMainButtonCanvasDisplay._uiButtonCanvas.Update(canvasPosition, canvasRotation, uiDisplayMode, true, vrRightHand, verticalHUDOffset, horizontalHUDOffset, depthHUDOffset);

                    UpdateLeftRightButtons();
                    if (_uiaEditMode)
                    {
                        _mainButton.label = "EDIT MODE";
                        _mainButton.buttonText.fontSize = 14;
                        _mainButton.textColor = Color.white;
                        _mainButton.buttonColor = Color.red;
                    }
                    else if (_uiaActiveClothingMode)
                    {
                        _mainButton.label = "ACTIVE CLOTHING";
                        _mainButton.buttonText.fontSize = 12;
                        _mainButton.textColor = Color.black;
                        _mainButton.buttonColor = Color.green;
                    }
                    else if (_uiActivated && _buttonGridsList[_currentGridScreen]._gridLabel != "" && _displayScreenLabelsJSON.val && Constants._patreonContentEnabled)
                    {
                        _mainButton.label = _buttonGridsList[_currentGridScreen]._gridLabel;
                        _mainButton.buttonText.fontSize = 16;
                        if (_mainButton.textColor == Color.white)
                        {
                            _mainButton.textColor = Color.black;
                            _mainButton.buttonColor = Color.grey;
                        }
                        
                    }
                    else
                    {
                        if (!_defaultButtonSkin._24HourClock) _mainButton.label = System.DateTime.Now.ToString("h:mm tt");
                        else _mainButton.label = System.DateTime.Now.ToString("HH:mm");
                        _mainButton.buttonText.fontSize = 16;
                        if (_mainButton.textColor == Color.white)
                        {
                            _mainButton.textColor = Color.black;
                            _mainButton.buttonColor = Color.grey;
                        }
                    }
                    //Vector3 fingerTip = trans.position + (trans.up * 0.03f);
                    //float distance = Vector3.Distance(fingerTip,_mainButton.button.transform.position);
                    //_mainButton.label = distance.ToString("0.00");
                    
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public bool ActivateSpawnTest(string atomName, UIAButton button, bool collisionDisable)
        {
            if (_testSpawnModeActive) DeactivateSpawnTest();
            Atom atom = SuperController.singleton.GetAtomByUid(atomName);
            if (atom!=null)
            {
                _testSpawnModeActive = true;
                _testSpawnAtom = atomName;
                _testSpawnAtomOriginalPosition = atom.mainController.transform.position;
                _testSpawnAtomOriginalRotation = atom.mainController.transform.rotation;
                _testSpawnAtomOriginalPositionState = atom.mainController.currentPositionState;
                _testSpawnAtomOriginalRotationState = atom.mainController.currentRotationState;

                atom.mainController.currentPositionState = FreeControllerV3.PositionState.On;
                atom.mainController.currentRotationState = FreeControllerV3.RotationState.On;
                _testSpawnAtomOriginalCollision = atom.collisionEnabledJSON.val;
                ToggleSpawnTestCollisionEnabled(!collisionDisable);

                _testSpawnButton = button;
                Vector3 position = new Vector3(0f, 0f, 0f);
                Quaternion rotation = new Quaternion();

                string gazeTarget = "";
                switch (button._atomPositionRelativeTo)
                {
                    case "Last viewed Atom":
                        gazeTarget = _lastViewedAtom;
                        break;
                    case "Last viewed Person":
                        gazeTarget = _lastViewedPerson;
                        break;
                    case "Last viewed Female":
                        gazeTarget = _lastViewedFemale;
                        break;
                    case "Last viewed Male":
                        gazeTarget = _lastViewedMale;
                        break;
                    case "Last selected Atom":
                        gazeTarget = _lastSelectedAtom;
                        break;
                    case "Last selected Person":
                        gazeTarget = _lastSelectedPerson;
                        break;
                    case "Last selected Female":
                        gazeTarget = _lastSelectedFemale;
                        break;
                    case "Last selected Male":
                        gazeTarget = _lastSelectedMale;
                        break;
                }
                _testSpawnAtomRelativeAtom = gazeTarget;

                button.UpdateSpawnPosition(ref position, ref rotation, gazeTarget);
                button.TeleportAtomToSpawnPoint(atomName, position, rotation, false);
                _testSpawnAtomLastPosition = atom.mainController.transform.position;
                _testSpawnAtomLastRotation = atom.mainController.transform.rotation;
                return true;
            }
            return false;
        }
        public void ToggleSpawnTestCollisionEnabled (bool collisionEnabled)
        {
            if (_testSpawnModeActive)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(_testSpawnAtom);
                if (atom != null) atom.collisionEnabledJSON.val = collisionEnabled;
            }
        }

        public void DeactivateSpawnTest()
        {
            if (_testSpawnModeActive)
            {
                Vector3 position = new Vector3(0f, 0f, 0f);
                Quaternion rotation = new Quaternion();

                position = _testSpawnAtomOriginalPosition;
                rotation = _testSpawnAtomOriginalRotation;
                _testSpawnButton.TeleportAtomToSpawnPoint(_testSpawnAtom, position, rotation, false);
                ToggleSpawnTestCollisionEnabled(_testSpawnAtomOriginalCollision);
                _buttonGridDisplay._buttonDisplay._isTestSpawnPositionActive = false;
                _testSpawnModeActive = false;
                _buttonGridDisplay._buttonDisplay.RefreshDisplay();
                Atom atom = SuperController.singleton.GetAtomByUid(_testSpawnAtom);
                SuperController.singleton.helpHUDText.text = "";
                if (atom != null)
                {
                    atom.mainController.currentPositionState = _testSpawnAtomOriginalPositionState;
                    atom.mainController.currentRotationState = _testSpawnAtomOriginalRotationState;
                }
            }

        }
        private void UpdateSpawnTestAtomPosition()
        {
            if (_testSpawnModeActive)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(_testSpawnAtom);
                if (atom == null) _buttonGridDisplay._buttonDisplay.EndSpawnTest();
                else
                {
                    Vector3 currentPosition = atom.mainController.transform.position;
                    Quaternion currentRotation = atom.mainController.transform.rotation;
                    float distanceChange = Vector3.Distance(currentPosition, _testSpawnAtomLastPosition);

                    if (distanceChange > 0.000001f || currentRotation != _testSpawnAtomLastRotation)
                    {
                        Transform relativeToTransform = _testSpawnButton.GetSpawnRelativeToTransform(_testSpawnAtomRelativeAtom);
                        Vector3 relativePosition = relativeToTransform.InverseTransformDirection(currentPosition - relativeToTransform.position);
                        Quaternion relativeRotation = new Quaternion();

                        relativeRotation.eulerAngles = new Vector3(currentRotation.eulerAngles.x - relativeToTransform.rotation.eulerAngles.x, currentRotation.eulerAngles.y - relativeToTransform.rotation.eulerAngles.y, currentRotation.eulerAngles.z - relativeToTransform.rotation.eulerAngles.z);

                        _testSpawnButton._atomPositionX = relativePosition.x;
                        if (_testSpawnButton._atomAbsoluteHeight) _testSpawnButton._atomPositionY = currentPosition.y;
                        else _testSpawnButton._atomPositionY = relativePosition.y;
                        _testSpawnButton._atomPositionZ = relativePosition.z;
                        if (_testSpawnButton._atomAbsolutePitch) _testSpawnButton._atomRotationX = currentRotation.eulerAngles.x;
                        else _testSpawnButton._atomRotationX = relativeRotation.eulerAngles.x;
                        _testSpawnButton._atomRotationY = relativeRotation.eulerAngles.y;

                        if (_testSpawnButton._atomAbsoluteRoll) _testSpawnButton._atomRotationZ = currentRotation.eulerAngles.z;
                        else _testSpawnButton._atomRotationZ = relativeRotation.eulerAngles.z;

                        if (_buttonGridDisplay._buttonDisplay._selectedButton == _testSpawnButton) _buttonGridDisplay._buttonDisplay.RefreshDisplayValues();

                        _testSpawnAtomLastPosition = atom.mainController.transform.position;
                        _testSpawnAtomLastRotation = atom.mainController.transform.rotation;
                    }
                    else
                    {
                        Vector3 targetPosition = new Vector3();
                        Quaternion targetRotation = new Quaternion();

                        _testSpawnButton.UpdateSpawnPosition(ref targetPosition, ref targetRotation, _testSpawnAtomRelativeAtom);

                        if (targetPosition != _testSpawnAtomLastPosition || targetRotation != _testSpawnAtomLastRotation)
                        {
                            atom.mainController.transform.position = targetPosition;
                            atom.mainController.transform.rotation = targetRotation;
                            _testSpawnAtomLastPosition = targetPosition;
                            _testSpawnAtomLastRotation = targetRotation;
                        }
                    }
                }
            }
        }


        public bool _atomSpawningActive = false;
        public bool _atomVRMoveActive = false;
        public bool _atomVRMovePaused = false;
        public float _atomVRMovePauseStartTime = 0f;
        public string _atomVRMoveAtomName = "";
        public string _atomVRMoveControllerLorR = "";
        public UIAButton _atomVRMoveUIAButton = null;
        public Vector3 _atomVRMoveAtomLastPosition = new Vector3();
        public Quaternion _atomVRMoveAtomLastRotation = new Quaternion();
        public Vector3 _atomVRMoveControllerLastPosition = new Vector3();
        public Quaternion _atomVRMoveControllerLastRotation = new Quaternion();
        public bool _currentVRMovelockedPositionX = false;
        public bool _currentVRMovelockedPositionY = false;
        public bool _currentVRMovelockedPositionZ = false;
        public bool _currentVRMovelockedRotationX = false;
        public bool _currentVRMovelockedRotationY = false;
        public bool _currentVRMovelockedRotationZ = false;
        public string _currentVRMovevrMovePositionGranulatity = "Coarse";
        public string _currentVRMovevrMoveRotationGranulatity = "Coarse";

        public bool _restoreHUDStatePostButtonAction = false;
        public bool _hudOnPreButtonAction = false;

        private bool _testSpawnModeActive = false;
        private string _testSpawnAtom = "";
        private Vector3 _testSpawnAtomOriginalPosition = new Vector3();
        private Quaternion _testSpawnAtomOriginalRotation = new Quaternion();
        private bool _testSpawnAtomOriginalCollision;
        private UIAButton _testSpawnButton;
        private Vector3 _testSpawnAtomLastPosition = new Vector3();
        private Quaternion _testSpawnAtomLastRotation = new Quaternion();
        private string _testSpawnAtomRelativeAtom = "";
        private FreeControllerV3.PositionState _testSpawnAtomOriginalPositionState;
        private FreeControllerV3.RotationState _testSpawnAtomOriginalRotationState;

        protected WristUI _uiMainCanvas;
        public UIDButtonCanvasDisplay _uidMainButtonCanvasDisplay;

        private UIDynamicButton _mainButton;
        private UIDynamicButton _leftButton;
        private UIDynamicButton _rightButton;

        private UIDynamicButton _mainPinButton;
        private Vector3 _fixedPinPosition;
        private Quaternion _fixedPinRotation;

        private UIDynamicButton _loadProfileButton;
        private UIDynamicButton _saveProfileButton;
        private UIDynamicButton _setDefaultProfileButton;
        private UIDynamicPopup _vrHandControlPopup;
        private UIDynamicPopup _vrPinModePopup;
        private UIDynamicPopup _gazeMenuSelectModePopup;
        private UIDynamicPopup _toolboxPopup;
        private UIDynamicSlider _vrHUDHorizontalOffsetSlider;
        private UIDynamicSlider _vrHUDVerticalOffsetSlider;
        private UIDynamicSlider _vrHUDDepthOffsetSlider;

        private UIDynamicToggle _displayGazeSelectionsToggle;
        private UIDynamicToggle _displayScreenLabelsToggle;
        private UIDynamicToggle _gazeActivateVRScreenToggle;
        private UIDynamicTextField _pluginVersionTextField;
        private UIDynamic _spacer1;
        private UIDynamic _spacer2;
        private UIDynamic _spacer2b;
        private UIDynamic _spacer3;
        private UIDynamic _spacer4;

        public bool _uiActivated = false;
        public bool _isRestoring = false;
        public bool _vrActive = false;
        private bool _uiPinned = false;
        private int _gazeUIButton = -1;
        private int _gazeQLBUIButton = -1;
        private bool _gazeMainButton = false;
        private bool _gazeLeftButton = false;
        private bool _gazeRightButton = false;
        private bool _gazePinButton = false;
        private bool _autoDePinEnabled = false;
        private int _gazeNodeSelectButtonRef = -1;

        public int _currentGridScreen = 0;
        public string _lastViewedMale = "";
        public string _lastViewedFemale = "";
        public string _lastViewedPerson = "";
        public string _lastSelectedAtom = "";
        public string _lastSelectedMale = "";
        public string _lastSelectedFemale = "";
        public string _lastSelectedPerson = "";
        public string _lastViewedAtom = "";
        private float _lastGazeTargetUpateTime;
        public bool _secondValidation = false;
        private float _timeLeftSelect = -1f;
        private float _timeRightSelect = -1f;
        private float _timeLastClothingUpdate = -1f;
        private float _timeLastHAFileListUpdated = -1f;
        private float _timeLastMainButtonClickDown = 0f;
        public bool _uiaEditMode = false;
        public bool _uiaActiveClothingMode = false;
        public string _uiaActiveClothingModeTargetType = "NONE";
        private bool _isUIGazeDeactivated = true;
        public bool _suppressScaleLoad = false;

        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _buttonGridSelectorJSON;
        public JSONStorableBool _isPrimaryUIAJSON;

        public bool _isActivePlugin = false;
        public JSONStorableBool _isActivePluginJSON;
        public JSONStorableFloat _pluginVersionFloatJSON;
        public JSONStorableString _pluginVersionJSON;
        public JSONStorableBool _isPatreonOnlyVersionJSON;

        public JSONStorableStringChooser _vrHandControlJSON;
        public JSONStorableStringChooser _vrPinModeJSON;
        public JSONStorableStringChooser _gazeMenuSelectModeJSON;
        public JSONStorableStringChooser _toolboxJSON;

        public JSONStorableFloat _vrHUDHorizontalOffsetJSON;
        public JSONStorableFloat _vrHUDVerticalOffsetJSON;
        public JSONStorableFloat _vrHUDDepthOffsetJSON;
        public JSONStorableBool _gazeActivateVRScreenJSON;

        public JSONStorableBool _activateQuickLaunchBarJSON;
        private UIDynamicToggle _activateQuickLaunchBarToggle;

        public JSONStorableBool _displayGazeSelectionsJSON;
        public JSONStorableBool _displayScreenLabelsJSON;
        public JSONStorableString _pluginDetailsJSON;
        private UIDynamicPopup _buttonGridSelectorPopup;

        public ButtonSkin _defaultButtonSkin;
        public ButtonSkinDisplay _defaultButtonSkinDisplay;
        public List<ButtonGrid> _buttonGridsList = new List<ButtonGrid>();
        public ButtonGrid _quickLaunchButtonGrid ;
        public ButtonGridDisplay _buttonGridDisplay;
        
        public ToolBox _toolbox;

        public Dictionary<string, ActiveClothingLists> _atomActiveClothingDictionary = new Dictionary<string, ActiveClothingLists>();
        public Dictionary<string, List<string>> _presetMergeClothingGeometryIDs = new Dictionary<string, List<string>>();

        SuperController.OnAtomUIDRename _atomRenameHandle;
        SuperController.OnAtomRemoved _atomRemoveHandle;
        
    }

    public class GazeTarget {
        public string atomName;
        public string nodeName;
        public float angle;
        public float distance;
        public bool isMale = false;
        public bool isFemale = false;
    }

    public class ActiveClothingLists
    {
        public List<string> _activeClothingHAFileNames;
        public List<DAZClothingItem> _activeClothingDCIs;
        public Dictionary<DAZClothingItem, List<string>> _activeClothingTags;
        
        public Dictionary<DAZClothingItem, string> _activeClothingDisplayNames;
        public Dictionary<DAZClothingItem, ClothSimControl> _activeClothingCSC;
        public Dictionary<DAZClothingItem, bool> _activeClothingSelected;

        private int _selectedCount = 0;
        public ActiveClothingLists()
        {
            try
            {
                _activeClothingHAFileNames = new List<string>();
                _activeClothingTags = new Dictionary<DAZClothingItem, List<string>>();
                _activeClothingDCIs = new List<DAZClothingItem>();
                _activeClothingDisplayNames = new Dictionary<DAZClothingItem, string>();
                _activeClothingCSC = new Dictionary<DAZClothingItem, ClothSimControl>();
                _activeClothingSelected = new Dictionary<DAZClothingItem, bool>();

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void UpdateSelectedDCIState(DAZClothingItem dci, bool selectedState)
        {
            if (_activeClothingSelected.ContainsKey(dci))
            {
                if (_activeClothingSelected[dci] != selectedState)
                {
                    _activeClothingSelected[dci] = selectedState;
                    if (selectedState) _selectedCount++;
                    else _selectedCount--;
                }

            }
        }
        public bool GetSelectedDCIState(DAZClothingItem dci)
        {
            bool state = false;
            if (_activeClothingSelected.ContainsKey(dci))
            {
                state = _activeClothingSelected[dci] ;
            }
            return (state);
        }
        public int GetSelectedCount()
        {
            return (_selectedCount);
        }

        private string StripNonAlphaNum(string s)
        {
            return Regex.Replace(s, @"[^A-Za-z0-9]+", "");
        }
        private string GetItemNameFromClothingUID(string uid)
        {
            string itemName = uid;
            if (uid.Contains("/"))
            {
                itemName = uid.Substring(uid.LastIndexOf("/") + 1);
                if (uid.EndsWith(".vam"))
                {
                    itemName = itemName.Substring(0, itemName.Length - 4);
                }
            }
            return (itemName);
        }
        public void UpdateActiveClothing(Atom atom)
        {
            _activeClothingHAFileNames.Clear();
            _activeClothingTags.Clear();
            _activeClothingDCIs.Clear();
            _activeClothingDisplayNames.Clear();
            _activeClothingCSC.Clear();
            _activeClothingSelected.Clear();

            foreach (Transform child in atom.gameObject.transform.Find("rescale2/geometry/FemaleClothes"))
            {

                if (child.gameObject.name == "FemaleClothingPrefab(Clone)")
                {
                    foreach (Transform subchild in atom.gameObject.transform.Find("rescale2/geometry/FemaleClothes/FemaleClothingPrefab(Clone)"))
                    {
                        DAZClothingItem dci = subchild.GetComponent<DAZClothingItem>();
                        if (dci.isActiveAndEnabled)
                        {
                            _activeClothingDCIs.Add(dci);
                            string[] tags = dci.tagsArray;
                            if (tags.Length > 0)
                            {
                                List<string> newTagList = tags.ToList();
                                _activeClothingTags.Add(dci, newTagList);
                            }
                            
                            string haClothingItemName = StripNonAlphaNum(dci.creatorName) + "-" + StripNonAlphaNum(dci.uid);
                            _activeClothingHAFileNames.Add(haClothingItemName);
                            
                            _activeClothingDisplayNames.Add(dci, dci.uid + " (" + dci.creatorName + ")");
                            _activeClothingSelected.Add(dci, false);
                        }
                    }
                }
                else
                {
                    if (child.GetComponent<DAZClothingItem>().isActiveAndEnabled)
                    {
                        DAZClothingItem dci = child.GetComponent<DAZClothingItem>();
                        _activeClothingDCIs.Add(dci);

                        foreach (Transform grandchild in child)
                        {
                            ClothSimControl csc = grandchild.GetComponent<ClothSimControl>();
                            if (csc != null) _activeClothingCSC.Add(dci, csc);
                        }
                        string[] tags = dci.tagsArray;
                        if (tags.Length > 0)
                        {
                            List<string> newTagList = tags.ToList();
                            _activeClothingTags.Add(dci, newTagList);
                        }
                        string itemName = GetItemNameFromClothingUID(dci.uid);
                        string haClothingItemName = StripNonAlphaNum(dci.creatorName) + "-" + StripNonAlphaNum(itemName);
                        _activeClothingHAFileNames.Add(haClothingItemName);
                        
                        _activeClothingDisplayNames.Add(dci, itemName + " (" + dci.creatorName + ")");
                    }
                }
            }
        }

    }
    public class UIDButtonCanvasDisplay
    {
        public WristUI _uiButtonCanvas;

        private UIAssist _mvrScript;
        public GameObject _gasButtonScrollListGO;
        private GameObject _gasHideControlNodesToggleGO;
        private List<GameObject> _gasScrollListButtonListGOs;
        public List<GazeTarget> _orderedGazeAssistSelectionTargets = new List<GazeTarget>();

        public UIDButtonGridDisplay _uiMainButtonGridDisplay;
        public UIDButtonGridDisplay _uiQLBButtonGridDisplay;
        public UIDActiveClothingEditor _uiActiveClothingEditor;
        public UIDynamicTextField _gazeStatusTextField;

        public UIDButtonCanvasDisplay(UIAssist mvr = null)
        {
            try
            {
                _mvrScript = mvr;
                _uiButtonCanvas = new WristUI(mvr);
                _uiMainButtonGridDisplay = new UIDButtonGridDisplay(mvr, _uiButtonCanvas);
                _uiQLBButtonGridDisplay = new UIDButtonGridDisplay(mvr, _uiButtonCanvas);
                _uiActiveClothingEditor = new UIDActiveClothingEditor(mvr, _uiButtonCanvas);
                _gasScrollListButtonListGOs = new List<GameObject>();
                
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CreateUIButtons()
        {
            
            if (!_mvrScript._uiaActiveClothingMode)
            {
                ButtonGrid buttonGrid = _mvrScript._buttonGridsList[_mvrScript._currentGridScreen];

                Vector2 qlbSize = _mvrScript._quickLaunchButtonGrid.GetButtonDimensions();

                float verticalStartOffset = 30f;

                if (_mvrScript._activateQuickLaunchBarJSON.val)
                {
                    _uiQLBButtonGridDisplay.CreateUIButtons(_mvrScript._quickLaunchButtonGrid, 20f);
                    verticalStartOffset = (qlbSize.y * _mvrScript._quickLaunchButtonGrid._gridRows) + 20f;
                    _uiQLBButtonGridDisplay.UpdateButtonTextures();
                }

                _uiMainButtonGridDisplay.CreateUIButtons(buttonGrid, verticalStartOffset);
                _uiMainButtonGridDisplay.UpdateButtonTextures();

                float buttonsXMinWorld = _uiMainButtonGridDisplay.GetLeftEdgeXPosition();
                float buttonsYMaxWorld = _uiMainButtonGridDisplay.GetTopEdgeYPosition();
               
                float buttonsWidth = _uiMainButtonGridDisplay.GetWidth();
                float buttonsHeight = _uiMainButtonGridDisplay.GetHeight();
                float scrollList1Width = Constants._gasWidth;
                float scrollList1Height = Mathf.Max(buttonsHeight - Constants._toggleButtonHeight, Constants._gasMinHeight);

                if (!_mvrScript._uiaEditMode)
                {
                    _gazeStatusTextField = _uiButtonCanvas.CreateTextField(0f , buttonsYMaxWorld , 0f, Mathf.Max(buttonsWidth, Constants._gazeStatusMinWidth), Constants._gazeStatusMaxHeight);

                    _gazeStatusTextField.UItext.fontSize = 12;
                    _gazeStatusTextField.backgroundColor = new Color(173f / 255f, 216f / 255f, 230f / 255f, 1f);
                }

                if (_mvrScript._toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val && SuperController.singleton.gameMode == SuperController.GameMode.Edit && !_mvrScript._uiaEditMode)
                {
                    bool scrollListLeftNotRight = false;
                    if (_mvrScript._vrActive == false || _mvrScript._vrHandControlJSON.val == "Right Hand") scrollListLeftNotRight = true;

                    float listXoffset = (buttonsWidth / 2f) + (scrollList1Width / 2f);
                    float toggleXoffet = (buttonsWidth / 2f) ;
                    if (scrollListLeftNotRight)
                    {
                        listXoffset = -(buttonsWidth / 2f) - (scrollList1Width / 2f);
                        toggleXoffet = -(buttonsWidth / 2f) - (scrollList1Width) ;
                    }

                    _gasButtonScrollListGO = _uiButtonCanvas.CreateScrollableButtonList(listXoffset, (scrollList1Height / 2f) + verticalStartOffset + Constants._toggleButtonHeight, scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f));
                    _gasHideControlNodesToggleGO = _uiButtonCanvas.CreateToggle(_uiButtonCanvas._canvas.transform, toggleXoffet, verticalStartOffset , scrollList1Width, Constants._toggleButtonHeight, "Hide Inactive Targets", 11);

                    UIDynamicToggle uiToggle = _gasHideControlNodesToggleGO.transform.GetComponent<UIDynamicToggle>();
                    uiToggle.toggle.onValueChanged.AddListener((bool value) => { UserPreferences.singleton.hideInactiveTargets = value; });
                }
            }
            else
            {
                _uiActiveClothingEditor.CreateUI();
  


            }

        }
        public void DestroyUIButtons()
        {
            _uiMainButtonGridDisplay.DestroyUIButtons();
            _uiQLBButtonGridDisplay.DestroyUIButtons();

            if (_gazeStatusTextField != null) { GameObject.Destroy(_gazeStatusTextField.gameObject); }
            if (_gasButtonScrollListGO != null)
            {
                GameObject.Destroy(_gasButtonScrollListGO);
                GameObject.Destroy(_gasHideControlNodesToggleGO);
                _gasButtonScrollListGO = null;
                _gasHideControlNodesToggleGO = null;
                foreach (GameObject go in _gasScrollListButtonListGOs) GameObject.Destroy(go);
                _gasScrollListButtonListGOs.Clear();
            }
             _uiActiveClothingEditor.DestroyUI();

        }
        public void UpdateGASScrollList(ref float nearestGazeSelectButtonDistance, ref int nearestGazeSelectButtonRef, bool gazeUpdate)
        {
            if (_gasButtonScrollListGO != null)
            {
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);
                Ray ray = new Ray(eyePosition, cameraT.forward);
                Vector3[] rectCornersArray = new Vector3[4];

                SetGASButtonCount(_gasScrollListButtonListGOs, _orderedGazeAssistSelectionTargets.Count);

                for (int index = 0; index < _orderedGazeAssistSelectionTargets.Count; index++)
                {
                    int tempIndex = index;
                    UIDynamicButton selectTargetButton = _gasScrollListButtonListGOs[index].transform.GetComponent<UIDynamicButton>();
                    if (_orderedGazeAssistSelectionTargets[index].nodeName == "control") selectTargetButton.label = _orderedGazeAssistSelectionTargets[index].atomName;
                    else selectTargetButton.label = _orderedGazeAssistSelectionTargets[index].atomName + ":" + _orderedGazeAssistSelectionTargets[index].nodeName;
                    if (SuperController.singleton.GetSelectedAtom()?.name == _orderedGazeAssistSelectionTargets[index].atomName && SuperController.singleton.GetSelectedController()?.name == _orderedGazeAssistSelectionTargets[index].nodeName)
                    {
                        selectTargetButton.buttonColor = Color.yellow;
                    }
                    else selectTargetButton.buttonColor = new Color(0.8f, 0.8f, 0.8f);

                    if (gazeUpdate)
                    {
                        RectTransform rtButton = selectTargetButton.transform.GetComponent<RectTransform>();
                        rtButton.GetWorldCorners(rectCornersArray);
                        Vector3 buttonCentre1 = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                        float distance = Vector3.Cross(ray.direction, buttonCentre1 - ray.origin).magnitude;
                        if (distance < nearestGazeSelectButtonDistance || nearestGazeSelectButtonDistance < 0f)
                        {
                            nearestGazeSelectButtonDistance = distance;
                            nearestGazeSelectButtonRef = index;
                        }
                    }
                }
                UIDynamicToggle uiToggle = _gasHideControlNodesToggleGO.transform.GetComponent<UIDynamicToggle>();

                uiToggle.toggle.isOn = UserPreferences.singleton.hideInactiveTargets;
            }

        }
        public void UpdateGASButtonColor(int buttonIndex, Color color)
        {
            _gasScrollListButtonListGOs[buttonIndex].GetComponent<UIDynamicButton>().buttonColor = color;
        }
        private void SetGASButtonCount(List<GameObject> buttonListGOs, int buttonTargetCount)
        {
            if (buttonListGOs.Count > buttonTargetCount)
            {
                int buttonsToRemove = buttonListGOs.Count - buttonTargetCount;
                for (int counter = 1; counter <= buttonsToRemove; counter++)
                {
                    GameObject go = buttonListGOs[buttonListGOs.Count - 1];
                    GameObject.Destroy(go);
                    buttonListGOs.RemoveAt(buttonListGOs.Count - 1);
                }
                _mvrScript.UpdateGazeTargets();
            }
            else if (buttonListGOs.Count < buttonTargetCount)
            {
                GameObject buttonListContentGO = _gasButtonScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;
                int buttonsToAdd = buttonTargetCount - buttonListGOs.Count;
                int intialButtonListGOCount = buttonListGOs.Count;
                for (int counter = 1; counter <= buttonsToAdd; counter++)
                {

                    GameObject newButtonGO = _uiButtonCanvas.CreateUIButtonInParentGO(buttonListContentGO);
                    RectTransform childRect = newButtonGO.GetComponent<RectTransform>();
                    childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 24f);
                    childRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, -10f, 200f);


                    buttonListGOs.Add(newButtonGO);
                    UIDynamicButton selectTargetButton = newButtonGO.transform.GetComponent<UIDynamicButton>();
                    selectTargetButton.buttonText.alignment = TextAnchor.MiddleLeft;
                    RectTransform rt = selectTargetButton.buttonText.gameObject.GetComponent<RectTransform>();
                    rt.anchoredPosition = new Vector2(0, 0);
                    rt.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 20f, 400f);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 24f);
                    selectTargetButton.buttonText.fontSize = 12;

                    int index = intialButtonListGOCount + counter - 1;
                    newButtonGO.SetActive(true);

                    selectTargetButton.button.onClick.AddListener(delegate
                    {
                        try
                        {
                            GazeAssistSelectButtonActivated(index);
                        }
                        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

                    });

                }
                _mvrScript.UpdateGazeTargets();
            }
        }
        public void GazeAssistSelectButtonActivated(int index)
        {
            if (SuperController.singleton.GetSelectedAtom()?.name == _orderedGazeAssistSelectionTargets[index].atomName && SuperController.singleton.GetSelectedController()?.name == _orderedGazeAssistSelectionTargets[index].nodeName)
            {
                SuperController.singleton.SelectController(_orderedGazeAssistSelectionTargets[index].atomName, "control", false);
            }
            else SuperController.singleton.SelectController(_orderedGazeAssistSelectionTargets[index].atomName, _orderedGazeAssistSelectionTargets[index].nodeName, false);

        }
        public void UpdateEditModeButtonPosition()
        {
            _uiMainButtonGridDisplay.UpdateEditModeButtonPosition();
            _uiQLBButtonGridDisplay.UpdateEditModeButtonPosition();
        }

        public UIDButtonGridDisplay GetAlternateBGDInMoveRange(Vector3 buttonWorldPosition, UIDButtonGridDisplay uidBGD)
        {
            UIDButtonGridDisplay alternateUIDbgd=null;
            if (_mvrScript._activateQuickLaunchBarJSON.val)
            {
                float verticalOffset = _uiQLBButtonGridDisplay.GetHeight() + 10f;
                Vector3 localPosition = _uiButtonCanvas._canvas.transform.InverseTransformPoint(buttonWorldPosition);
                if (uidBGD == _uiQLBButtonGridDisplay)
                {
                    if (_uiMainButtonGridDisplay._buttonGrid.ContainsBlankButtons() && localPosition.y > verticalOffset+5f)
                    {
                        return (_uiMainButtonGridDisplay);
                    }
                }
                else if (_uiQLBButtonGridDisplay._buttonGrid.ContainsBlankButtons())
                {
                    if (localPosition.y < verticalOffset - 5f)
                    {
                        return (_uiQLBButtonGridDisplay);
                    }
                }
            }


            return (alternateUIDbgd);

        }
    }
    public class UIDButtonGridDisplay
    {
        private List<UIDynamicButton> _uiButtonList;

        private UIAssist _mvrScript;
        private WristUI _uiButtonCanvas;

        public ButtonGrid _buttonGrid ;

        private float _verticalStartOffset;
        private float _horizontalStartOffset;
        private Vector2 _buttonDimensions;
        private int _columns;
        private int _rows;
        public bool _buttonMoveActive = false;
        public int _buttonMoveIndex = -1;
        public float _buttonMoveStartTime = 0f;
        public string _buttonMoveController = "";

        public UIDButtonGridDisplay(UIAssist mvr, WristUI uiButtonCanvas)
        {
            try
            {
                _mvrScript = mvr;
                _uiButtonList = new List<UIDynamicButton>();
                _uiButtonCanvas = uiButtonCanvas;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshButtonPositions()
        {
            int buttonNum = 0;
            for (int row = 0; row < _rows; row++)
            {
                for (int col = 0; col < _columns; col++)
                {
                    UIDynamicButton gridButton = _uiButtonList[buttonNum];

                    gridButton.button.transform.localPosition = new Vector3((-_horizontalStartOffset + (col * _buttonDimensions.x)), (_verticalStartOffset + ((row + 0.5f) * _buttonDimensions.y)), 0f);

                    buttonNum++;
                }
            }
        }
        public void CreateUIButtons(ButtonGrid buttonGrid, float verticalStartOffset )
        {
            _buttonGrid = buttonGrid;
            _buttonDimensions = buttonGrid.GetButtonDimensions();           
            
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null)
            {
                _rows = Mathf.Min(buttonGrid._gridRows, 3);
                _columns = Mathf.Min(buttonGrid._gridColumns, 3);
            }
            else
            {
                _rows = buttonGrid._gridRows;
                _columns = buttonGrid._gridColumns;

            }
            _verticalStartOffset = verticalStartOffset;
            _horizontalStartOffset = _buttonDimensions.x * ((_columns - 1) / 2f);
            
            _uiButtonList.Clear();
          
            int buttonNum = 0;
            for (int row = 0; row < _rows; row++)
            {
                for (int col = 0; col < _columns; col++)
                {
                    buttonNum++;
                    UIDynamicButton gridButton = _uiButtonCanvas.CreateUIButtonInCanvas(buttonNum.ToString(), (-_horizontalStartOffset + (col * _buttonDimensions.x)) , (_verticalStartOffset + ((row + 0.5f) * _buttonDimensions.y)), 0f, _buttonDimensions.x, _buttonDimensions.y);
                    
                    int temp = buttonNum;
                    gridButton.button.onClick.AddListener(delegate {
                        _mvrScript.ButtonActivated(temp, buttonGrid);                       
                    });

                    EventTrigger trigger = gridButton.button.gameObject.AddComponent<EventTrigger>();
                    var pointerDown = new EventTrigger.Entry();
                    pointerDown.eventID = EventTriggerType.PointerDown;
                    pointerDown.callback.AddListener((e) => GridButtonClickDown(temp));
                    trigger.triggers.Add(pointerDown);

                    var pointerUp = new EventTrigger.Entry();
                    pointerUp.eventID = EventTriggerType.PointerUp;
                    pointerUp.callback.AddListener((e) => GridButtonClickUp(temp));
                    trigger.triggers.Add(pointerUp);

                    RectTransform rt = gridButton.buttonText.gameObject.GetComponent<RectTransform>();
                    rt.anchoredPosition = new Vector2(0, 0);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x * 0.8f);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y * 0.8f);
                  
                    _uiButtonList.Add(gridButton);
                }
            }
        }
        public void DestroyUIButtons()
        {
            foreach (UIDynamicButton uiButton in _uiButtonList)
            {
                if (uiButton != null) {
                    uiButton.button.onClick.RemoveAllListeners();

                    GameObject.Destroy(uiButton.gameObject);
                }
            }
        }
        public float GetLeftEdgeXPosition()
        {
            return (-_horizontalStartOffset - (_buttonDimensions.x / 2f));
        }
        public float GetTopEdgeYPosition()
        {
            return (_verticalStartOffset + ((_buttonGrid._maxButtonRow ) * _buttonDimensions.y) );
        }
        public float GetWidth()
        {
            return (_columns * _buttonDimensions.x);
        }
        public float GetHeight()
        {
            return (_rows * _buttonDimensions.y);
        }
        
        public void UpdateButtonTextures(int buttonRef)
        {
            UIDynamicButton uiButton = _uiButtonList[buttonRef];
            UIAButton button = _buttonGrid._buttonList[buttonRef];
            Texture2D thumbnailTexture;
            if (button._buttonType != "Blank" && button._buttonTexture != null && button._useButtonTexture && (button.isPresetType || button._buttonType == "Spawn Atom") || button.isSceneLoadType)
            {
                thumbnailTexture = button._buttonTexture;
            }
            else
            {
                int buttonState = 1;
                string targetAtom = "";
                if (button._buttonType != "Blank" && (button._targetCategory == "Gaze Selected Atom" || button._targetCategory == "VAM Selected Atom"))
                {
                    targetAtom = button.GetLastViewedOrSelectedAtomName();
                    if (targetAtom != "") { buttonState = button.GetButtonState(targetAtom); }
                }
                else { buttonState = button.GetButtonState(""); }
                if (buttonState == 1)
                {
                    if (button._buttonSkin == null || button._buttonSkin._customButtonOnTextureDefault) { thumbnailTexture = _mvrScript._defaultButtonSkin._buttonOnTexture; }
                    else { thumbnailTexture = button._buttonSkin._buttonOnTexture; }
                }
                else
                {
                    if (button._buttonSkin == null || button._buttonSkin._customButtonOffTextureDefault) { thumbnailTexture = _mvrScript._defaultButtonSkin._buttonOffTexture; }
                    else { thumbnailTexture = button._buttonSkin._buttonOffTexture; }

                }
                if (!_mvrScript._secondValidation) thumbnailTexture = null;
            }
            if (thumbnailTexture != null)
            {
                Sprite mySprite = Sprite.Create(thumbnailTexture, new Rect(0.0f, 0.0f, thumbnailTexture.width, thumbnailTexture.height), new Vector2(0.5f, 0.5f), 100.0f);
                uiButton.buttonImage.sprite = mySprite;
            }
        }

        public void UpdateButtonTextures()
        {

            int buttonCount = _uiButtonList.Count;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null)
            {
                buttonCount = Mathf.Min(buttonCount, 9);
            }

            for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++)
            {
                UpdateButtonTextures(buttonRef);
            }
        }

        public void UpdateButtonColor(int buttonIndex, Color color)
        {
            if (buttonIndex<_uiButtonList.Count && buttonIndex>=0)
            _uiButtonList[buttonIndex].buttonColor = color;
        }

        public void UpdateUIAButtonGrid( bool gazeUpate, out int nearestGazeButton, out float nearestGazeButtonDistance)
        {
            int buttonCount = _uiButtonList.Count;
            nearestGazeButton = -1;
            nearestGazeButtonDistance = -1f;

            for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++)
            {
                UIDynamicButton uiButton = _uiButtonList[buttonRef];
                UIAButton button = _buttonGrid._buttonList[buttonRef];
                ButtonSkin buttonSkin = button._buttonSkin;
                Vector3[] rectCornersArray = new Vector3[4];
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);
                Ray ray = new Ray(eyePosition, cameraT.forward);
                int buttonState = 1;
                string targetAtom = "";

                if (button._buttonType == "Blank" && !_mvrScript._uiaEditMode)
                {
                    uiButton.gameObject.SetActive(false);
                }
                else
                {
                    uiButton.gameObject.SetActive(true);
                    if (gazeUpate)
                    {
                        float distance = _mvrScript.GetButtonDistance(uiButton, ray);
                        if (distance < nearestGazeButtonDistance || nearestGazeButtonDistance < 0f)
                        {
                            nearestGazeButtonDistance = distance;
                            nearestGazeButton = buttonRef;
                        }
                    }
                    if ((button._targetCategory == "Gaze Selected Atom" || button._targetCategory == "VAM Selected Atom") && button.IsAtomTargetable())
                    {
                        targetAtom = button.GetLastViewedOrSelectedAtomName();
                        if (targetAtom != "") { buttonState = button.GetButtonState(targetAtom); }
                    }
                    else {
                        buttonState = button.GetButtonState("");
                        if (button._targetCategory == "Specific Atom") targetAtom = button._targetName;
                    }

                    string targetFont = "Arial";
                    if (buttonSkin == null || buttonSkin._buttonFontDefault || !_mvrScript._secondValidation) { targetFont = _mvrScript._defaultButtonSkin._buttonFont; }
                    else { targetFont = buttonSkin._buttonFont; }

                    Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
                    foreach (Font font in Resources.FindObjectsOfTypeAll(typeof(Font)) as Font[])
                    {
                        if (font.name == targetFont) selectedFont = font;
                    }
                    uiButton.buttonText.font = selectedFont;

                    HSVColor hsvTextColor = new HSVColor();
                    HSVColor hsvButtonColor = new HSVColor();
                    if (buttonState == 1)
                    {
                        if (button._autoLabel) uiButton.label = button.GetAutoLabel(1, targetAtom);
                        else uiButton.label = button.GetManualLabel(1, targetAtom);
                        if (buttonSkin == null || buttonSkin._textOnColorDefault || !_mvrScript._secondValidation) hsvTextColor = _mvrScript._defaultButtonSkin._textOnColor;
                        else hsvTextColor = buttonSkin._textOnColor;
                        if (buttonSkin == null || buttonSkin._buttonOnColorDefault || !_mvrScript._secondValidation) hsvButtonColor = _mvrScript._defaultButtonSkin._buttonOnColor;
                        else hsvButtonColor = buttonSkin._buttonOnColor;

                    }
                    else if (buttonState == 2)
                    {
                        if (button._autoLabel) uiButton.label = button.GetAutoLabel(2, targetAtom);
                        else uiButton.label = button.GetManualLabel(2, targetAtom);
                        if (buttonSkin == null || buttonSkin._textOffColorDefault || !_mvrScript._secondValidation) hsvTextColor = _mvrScript._defaultButtonSkin._textOffColor;
                        else hsvTextColor = buttonSkin._textOffColor;
                        if (buttonSkin == null || buttonSkin._buttonOffColorDefault || !_mvrScript._secondValidation) hsvButtonColor = _mvrScript._defaultButtonSkin._buttonOffColor;
                        else hsvButtonColor = buttonSkin._buttonOffColor;
                    }
                    if (buttonSkin == null || buttonSkin._fontSizeDefault || !_mvrScript._secondValidation)
                    {
                        if (_buttonGrid._buttonSize == "Micro")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)_mvrScript._defaultButtonSkin._maxMicroFontSize *2;
                            uiButton.buttonText.resizeTextMinSize = (int)_mvrScript._defaultButtonSkin._minMicroFontSize *2;
                        }
                        else if (_buttonGrid._buttonSize == "Mini")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)_mvrScript._defaultButtonSkin._maxMiniFontSize *2;
                            uiButton.buttonText.resizeTextMinSize = (int)_mvrScript._defaultButtonSkin._minMiniFontSize *2;
                        }
                        else if (_buttonGrid._buttonSize == "Small")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)_mvrScript._defaultButtonSkin._maxSmallFontSize*2;
                            uiButton.buttonText.resizeTextMinSize = (int)_mvrScript._defaultButtonSkin._minSmallFontSize*2;
                        }
                        else if (_buttonGrid._buttonSize == "Medium")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)_mvrScript._defaultButtonSkin._maxMediumFontSize*2;
                            uiButton.buttonText.resizeTextMinSize = (int)_mvrScript._defaultButtonSkin._minMediumFontSize*2;
                        }
                        else if (_buttonGrid._buttonSize == "Large")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)_mvrScript._defaultButtonSkin._maxLargeFontSize*2;
                            uiButton.buttonText.resizeTextMinSize = (int)_mvrScript._defaultButtonSkin._minLargeFontSize*2;
                        }
                    }
                    else
                    {
                        uiButton.buttonText.resizeTextMaxSize = (int)buttonSkin._maxFontSize*2;
                        uiButton.buttonText.resizeTextMinSize = (int)buttonSkin._minFontSize*2;
                    }
                    uiButton.buttonText.resizeTextForBestFit = true;
                    uiButton.buttonText.color = Color.HSVToRGB(hsvTextColor.H, hsvTextColor.S, hsvTextColor.V);
                    uiButton.buttonColor = Color.HSVToRGB(hsvButtonColor.H, hsvButtonColor.S, hsvButtonColor.V);

                    string textAlignment = "";
                    if (buttonSkin != null) textAlignment = buttonSkin._textAlignment;
                    if (buttonSkin == null || buttonSkin._textAlignmentDefault || !_mvrScript._secondValidation) { textAlignment = _mvrScript._defaultButtonSkin._textAlignment; }
                    if (textAlignment == "LowerCenter") { uiButton.buttonText.alignment = TextAnchor.LowerCenter; }
                    else if (textAlignment == "MiddleCenter") { uiButton.buttonText.alignment = TextAnchor.MiddleCenter; }
                    else if (textAlignment == "UpperCenter") { uiButton.buttonText.alignment = TextAnchor.UpperCenter; }
                    else if (textAlignment == "LowerLeft") { uiButton.buttonText.alignment = TextAnchor.LowerLeft; }
                    else if (textAlignment == "MiddleLeft") { uiButton.buttonText.alignment = TextAnchor.MiddleLeft; }
                    else if (textAlignment == "UpperLeft") { uiButton.buttonText.alignment = TextAnchor.UpperLeft; }
                    else if (textAlignment == "LowerRight") { uiButton.buttonText.alignment = TextAnchor.LowerRight; }
                    else if (textAlignment == "MiddleRight") { uiButton.buttonText.alignment = TextAnchor.MiddleRight; }
                    else if (textAlignment == "UpperRight") { uiButton.buttonText.alignment = TextAnchor.UpperRight; }
                }
            }
        }
        public void GridButtonClickDown(int buttonRef)
        {
            try
            {
                if (_mvrScript._uiaEditMode)
                {
                    _buttonMoveIndex = buttonRef - 1;
                    _buttonMoveActive = true;
                    _buttonMoveStartTime = Time.time;
                    if (_buttonGrid == _mvrScript._quickLaunchButtonGrid)   _mvrScript._buttonGridSelectorJSON.val = "Quick Launch Bar";
                    else
                    {
                        int index = _mvrScript._buttonGridsList.IndexOf(_buttonGrid);
                        _mvrScript._buttonGridSelectorJSON.val = (index+1).ToString();
                    }
                    _mvrScript._buttonGridDisplay._buttonSelectJSON.val = buttonRef.ToString();

                    SuperController.singleton.ShowMainHUD();
                    SuperController.singleton.activeUI = SuperController.ActiveUI.MainMenu;
                    
                    SuperController.singleton.mainMenuTabSelector.SetActiveTab("TabSessionPlugins");
                    IEnumerator enumerator1 = SuperController.singleton.mainMenuTabSelector.transform.GetEnumerator();
                    while (enumerator1.MoveNext())
                    {
                        UITab component = ((Component)enumerator1.Current).GetComponent<UITab>();
                        if ((UnityEngine.Object)component != (UnityEngine.Object)null)
                        {
                            foreach (var scriptController in component.GetComponentsInChildren<MVRScriptControllerUI>())
                            {
                                if (scriptController.label.text== _mvrScript.name)
                                {
                                    scriptController.openUIButton?.onClick?.Invoke();
                                    break;
                                }                                
                            }
                        }                            
                    }

                    if (_mvrScript._vrActive && _buttonMoveController == "")
                    {
                        if (SuperController.singleton.GetRightSelect()&&_mvrScript._uidMainButtonCanvasDisplay._uiMainButtonGridDisplay._buttonMoveController!="RIGHT" && _mvrScript._uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay._buttonMoveController != "RIGHT")
                        {
                            _buttonMoveController = "RIGHT";
                        }
                        else if (SuperController.singleton.GetLeftSelect())
                        {
                            _buttonMoveController = "LEFT";
                        }
                    }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void GridButtonClickUp(int buttonRef)
        {
            try
            {
                _buttonMoveActive = false;
                _buttonMoveController = "";
                if (_mvrScript._uiaEditMode) _mvrScript.RefreshWristUIButtonGrid();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void UpdateEditModeButtonPosition()
        {
            if (_buttonMoveActive && _buttonMoveIndex >= 0 && (Time.time-_buttonMoveStartTime)>0.2f)
            {
                UIDynamicButton uiButton = _uiButtonList[_buttonMoveIndex];
                Vector3 worldPosition = new Vector3();

                if (!_mvrScript._vrActive)
                {
                    Vector2 pos;
                    RectTransformUtility.ScreenPointToLocalPointInRectangle(_uiButtonCanvas._canvas.transform as RectTransform, Input.mousePosition, _uiButtonCanvas._canvas.worldCamera, out pos);
                    worldPosition = _uiButtonCanvas._canvas.transform.TransformPoint(pos);
                }
                else
                {
                    Plane plane = new Plane(_uiButtonCanvas._canvas.transform.forward, _uiButtonCanvas._canvas.transform.TransformPoint(new Vector3(0f, 0f, 0f)));

                    Transform lineLeft = SuperController.singleton.rayLineLeft.transform;
                    Transform lineRight = SuperController.singleton.rayLineRight.transform;

                    Ray ray;

                    if (_buttonMoveController =="LEFT") ray = new Ray(lineLeft.transform.position, lineLeft.transform.rotation * Vector3.forward);
                    else  ray = new Ray(lineRight.transform.position, lineRight.transform.rotation * Vector3.forward);
                   
                    float distanceA;

                    if (plane.Raycast(ray, out distanceA))
                    {
                        worldPosition = ray.GetPoint(distanceA * 0.9f);
                    }
                }
                uiButton.transform.position = worldPosition;
                uiButton.transform.SetAsLastSibling();
                int nearestButtonIndex = GetUIAEditNearestGridButtonIndex(worldPosition);

                if (nearestButtonIndex != _buttonMoveIndex)
                {
                    UIDynamicButton tempButton;
                    UIAButton tempUIAButton;

                    tempButton = _uiButtonList[_buttonMoveIndex];
                    tempUIAButton = _buttonGrid._buttonList[_buttonMoveIndex];

                    ButtonShift(_buttonMoveIndex, nearestButtonIndex);

                    _uiButtonList[nearestButtonIndex] = tempButton;                   
                    _buttonGrid._buttonList[nearestButtonIndex] = tempUIAButton;
                    _buttonGrid._buttonList[nearestButtonIndex]._buttonRef = nearestButtonIndex + 1;

                    _buttonMoveIndex = nearestButtonIndex;

                    if (_buttonGrid == _mvrScript._quickLaunchButtonGrid) _mvrScript._buttonGridSelectorJSON.val = "Quick Launch Bar";
                    else
                    {
                        int index = _mvrScript._buttonGridsList.IndexOf(_buttonGrid);
                        _mvrScript._buttonGridSelectorJSON.val = (index + 1).ToString();
                    }
                    _buttonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._buttonSelectJSON.val = (nearestButtonIndex+1).ToString();
                    RefreshButtonPositions();
                }
                UIDButtonGridDisplay alternateBGD = _mvrScript._uidMainButtonCanvasDisplay.GetAlternateBGDInMoveRange(worldPosition, this);
                if (alternateBGD != null)
                {
                    alternateBGD.InsertButtonFromBGD(_buttonGrid._buttonList[_buttonMoveIndex], _uiButtonList[_buttonMoveIndex], worldPosition, _buttonMoveStartTime, _buttonMoveController);
                    _buttonMoveActive = false;
                   
                    _buttonGrid._buttonList[_buttonMoveIndex] = new UIAButton(_buttonMoveIndex, _mvrScript);
                    _uiButtonList[_buttonMoveIndex] = null;
                    _buttonGrid.RecalcActiveButtonCount();
                    DestroyUIButtons();
                    CreateUIButtons(_buttonGrid,_verticalStartOffset);
                    UpdateButtonTextures();
                        //                   _mvrScript.RefreshWristUIButtonGrid();
                }
            }
            
        }
        public void InsertButtonFromBGD(UIAButton button, UIDynamicButton uidButton,Vector3 worldPosition, float startTime, string controller)
        {
            int nearestIndex = GetUIAEditNearestGridButtonIndex(worldPosition);
            int higherBlankButtonIndex = -1;
            int lowerBlankButtonIndex = -1;
            int nearestBlankButtonIndex = -1;
            for (int index = nearestIndex ; index < _buttonGrid._buttonList.Count();index++)
            {
                if (_buttonGrid._buttonList[index]._buttonType == "Blank")
                {
                    higherBlankButtonIndex = index;
                    break;
                }
            }
            for (int index = nearestIndex; index >= 0; index--)
            {
                if (_buttonGrid._buttonList[index]._buttonType == "Blank")
                {
                    lowerBlankButtonIndex = index;
                    break;
                }
            }

            if (lowerBlankButtonIndex!=-1 && (higherBlankButtonIndex == -1 || (higherBlankButtonIndex - nearestIndex) > (nearestIndex - lowerBlankButtonIndex))) nearestBlankButtonIndex = lowerBlankButtonIndex;
            else nearestBlankButtonIndex = higherBlankButtonIndex;

            UIDynamicButton nearestBlankUIDButton = _uiButtonList[nearestBlankButtonIndex];
            nearestBlankUIDButton.button.onClick.RemoveAllListeners();
            GameObject.Destroy(nearestBlankUIDButton.gameObject);

            ButtonShift(nearestBlankButtonIndex, nearestIndex);

            _uiButtonList[nearestIndex] = uidButton;
            _buttonGrid._buttonList[nearestIndex] = button;
            _buttonGrid._buttonList[nearestIndex]._buttonRef = nearestIndex + 1;

            _buttonMoveIndex = nearestIndex;
            _buttonMoveActive = true;
            _buttonMoveController = controller;
            _buttonMoveStartTime = startTime;

            EventTrigger trigger = uidButton.button.gameObject.GetComponent<EventTrigger>();
            List<EventTrigger.Entry> triggerEntryList = trigger.triggers;
            foreach(EventTrigger.Entry triggerEntry in triggerEntryList)
            {
                if (triggerEntry.eventID == EventTriggerType.PointerUp)
                {
                    triggerEntry.callback.RemoveAllListeners();
                    triggerEntry.callback.AddListener((e) => GridButtonClickUp(nearestIndex+1));
                }
            }
            uidButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x);
            uidButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y);

            RectTransform rt = uidButton.buttonText.gameObject.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(0, 0);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x * 0.8f);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y * 0.8f);

            _buttonGrid.RecalcActiveButtonCount();

            if (_buttonGrid == _mvrScript._quickLaunchButtonGrid) _mvrScript._buttonGridSelectorJSON.val = "Quick Launch Bar";
            else
            {
                int index = _mvrScript._buttonGridsList.IndexOf(_buttonGrid);
                _mvrScript._buttonGridSelectorJSON.val = (index + 1).ToString();
            }
            _mvrScript._buttonGridDisplay._buttonSelectJSON.val = (nearestIndex + 1).ToString();
            RefreshButtonPositions();

        }
        public void ButtonShift(int startIndex, int endIndex)
        {
            int increment = 1;
            if (endIndex < startIndex) increment = -1;

            for (int index = startIndex; index != endIndex; index += increment)
            {
                _uiButtonList[index] = _uiButtonList[index + increment];
                _buttonGrid._buttonList[index] = _buttonGrid._buttonList[index + increment];
                _buttonGrid._buttonList[index]._buttonRef = index + 1;
            }
        }
        private int GetUIAEditNearestGridButtonIndex(Vector3 worldPosition)
        {
            Vector3 localPosition = _uiButtonCanvas._canvas.transform.InverseTransformPoint(worldPosition);

            int xPos = (int)Math.Round((localPosition.x + _horizontalStartOffset) / _buttonDimensions.x);
            int yPos = (int)Math.Round((localPosition.y - _verticalStartOffset - 20f) / _buttonDimensions.y);
            int xPosLimited = Math.Max(0, Math.Min(xPos, _columns - 1));
            int yPosLimited = Math.Max(0, Math.Min(yPos, _rows - 1));
            return ((yPosLimited * _columns) + xPosLimited);
        }

    }
   
    public class UIDActiveClothingEditorItem
    {
        private UIAssist _mvrScript;
        private WristUI _uiButtonCanvas;
        public GameObject _aceItemContainerImageGO;
        private GameObject _aceItemRemoveButtonGO;
        private GameObject _aceItemUndressButtonGO;
        private GameObject _aceItemToggleGO;

        public UIDynamicToggle _aceItemSelectToggle;
        public UIDynamicButton _aceItemRemoveButton;
        public UIDynamicButton _aceItemUndressButton;


        public UIDActiveClothingEditorItem(UIAssist mvr, WristUI uiButtonCanvas, GameObject aceScrollListGO, Color panelColor)
        {
            try
            {
                _mvrScript = mvr;
                _uiButtonCanvas = uiButtonCanvas;

                float horizontalOffset = 5f;
                
                if (aceScrollListGO!=null)
                {
                    GameObject buttonListContentGO = aceScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

                    _aceItemContainerImageGO = _uiButtonCanvas.CreateUIContainerImageInParentGO(buttonListContentGO, 280f,25f, panelColor);
                    horizontalOffset = 0f;
                }
                else _aceItemContainerImageGO = _uiButtonCanvas.CreateUIContainerImageInParentGO(_uiButtonCanvas._canvas.gameObject, 300f, 25f,panelColor);


                _aceItemToggleGO = _uiButtonCanvas.CreateUIToggleInParentGO(_aceItemContainerImageGO, 0f, 0f, 180f, 25f);
                WristUI.SetAnchors(_aceItemContainerImageGO, _aceItemToggleGO, "left", horizontalOffset);
                _aceItemToggleGO.SetActive(true);
                _aceItemSelectToggle = _aceItemToggleGO.transform.GetComponent<UIDynamicToggle>();
                _aceItemSelectToggle.labelText.resizeTextMinSize = 8;
                _aceItemSelectToggle.labelText.resizeTextMaxSize = 13;
                _aceItemSelectToggle.labelText.resizeTextForBestFit = true;


                _aceItemRemoveButtonGO = _uiButtonCanvas.CreateUIButtonInParentGO(_aceItemContainerImageGO);
                WristUI.SetSizeAndAnchors(_aceItemContainerImageGO, _aceItemRemoveButtonGO, 50f, 20f, "right", -50f- (horizontalOffset*3));
                _aceItemRemoveButton = _aceItemRemoveButtonGO.transform.GetComponent<UIDynamicButton>();
                _aceItemRemoveButton.buttonText.fontSize = 11;
                _aceItemRemoveButton.label = "Remove";

                 _aceItemUndressButtonGO = _uiButtonCanvas.CreateUIButtonInParentGO(_aceItemContainerImageGO);
                WristUI.SetSizeAndAnchors(_aceItemContainerImageGO, _aceItemUndressButtonGO, 50f, 20f, "right", -horizontalOffset*3);
                 _aceItemUndressButton = _aceItemUndressButtonGO.transform.GetComponent<UIDynamicButton>();
                _aceItemUndressButton.buttonText.fontSize = 11;
                _aceItemUndressButton.label = "Undress";
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void DestroyClothingItem()
        {
            GameObject.Destroy(_aceItemContainerImageGO);
        }

    }

    public class ButtonGrid {
        public int _gridRows = 0;
        public int _gridColumns = 0;
        public string _buttonSize = "Medium";
        public int _activeButtonCount = 0;
        public int _maxButtonRow = 0;
        public bool _gazeSelectionAtoms = false;
        public bool _gazeSelectionFemales = false;
        public bool _gazeSelectionMales = false;
        public bool _gazeSelectionPerson = false;
        public bool _vamSelectionAtoms = false;
        public bool _vamSelectionFemales = false;
        public bool _vamSelectionMales = false;
        public bool _vamSelectionPerson = false;
        public string _gridLabel = "";
        public bool _quickLaunchBar = false;

        public List<UIAButton> _buttonList = new List<UIAButton>();
        private UIAssist _mvrScript;

        public ButtonGrid(int columns = 0, int rows = 0, UIAssist mvr = null, bool quickLaunchBar = false) {
            try {
                _gridRows = rows;
                _gridColumns = columns;
                _mvrScript = mvr;
                _quickLaunchBar = quickLaunchBar;
                UpdateGridButtons();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void AtomNameUpdate(string oldName, string newName)
        {
            foreach (UIAButton button in _buttonList)
            {

                foreach (string key in button._atomButtonStates.Keys.ToList())
                {

                    if (key == oldName)
                    {
                        button._atomButtonStates[newName] = button._atomButtonStates[key];
                        button._atomButtonStates.Remove(key);
                    }
                }
                foreach (string key in button._atomClothingPresets.Keys.ToList())
                {

                    if (key == oldName)
                    {
                        button._atomClothingPresets[newName] = button._atomClothingPresets[key];
                        button._atomClothingPresets.Remove(key);
                    }
                }
                if (button._targetCategory == "Specific Atom" && button._targetName == oldName) button._targetName = newName;
            }
        }
        public void AtomRemovedUpdate(string oldName)
        {
            foreach (UIAButton button in _buttonList)
            {
                foreach (string key in button._atomButtonStates.Keys.ToList())
                {
                    if (key == oldName) button._atomButtonStates.Remove(key);
                }
                foreach (string key in button._atomClothingPresets.Keys.ToList())
                {
                    if (key == oldName)button._atomClothingPresets.Remove(key);
                }
                if (button._targetCategory == "Specific Atom" && button._targetName == oldName) button._targetName = "";
            }
        }

        public UIAButton GetButtonBefore(int buttonRef) {
            UIAButton button = null;
            if (buttonRef > 1) {
                for (int i = buttonRef - 2; i >= 0; i--) {
                    if (_buttonList[i]._buttonType != "Blank") {
                        button = _buttonList[i];
                        break;
                    }
                }
            }
            return (button);
        }
        public bool ContainsBlankButtons()
        {
            bool blankButtons = false;
            foreach (UIAButton button in _buttonList)
            {
                if (button._buttonType == "Blank") return (true);
            }
            return (blankButtons);
        }
        public Vector2 GetButtonDimensions()
        {
            Vector2 dimensions = new Vector2();
            switch (_buttonSize)
            {
                case "Micro":
                    dimensions.y = 30;
                    dimensions.x = 30;
                    break;
                case "Mini":
                    dimensions.y = 40;
                    dimensions.x = 40;
                    break;
                case "Small":
                    dimensions.y = 60;
                    dimensions.x = 60;
                    break;
                case "Medium":
                    dimensions.y = 90;
                    dimensions.x = 90;
                    break;
                case "Large":
                    dimensions.y = 120;
                    dimensions.x = 120;
                    break;
            }

            return (dimensions);
        }
        public void UpdateGridButtons() {
            if (_buttonList.Count < (_gridRows * _gridColumns)) {
                for (int i = _buttonList.Count; i < (_gridRows * _gridColumns); i++) {
                    UIAButton newButton = new UIAButton(i + 1, _mvrScript);
                    _buttonList.Add(newButton);
                }
            }
        }

        public void RecalcActiveButtonCount() {
            _activeButtonCount = 0;
            _maxButtonRow = 0;
            foreach (UIAButton button in _buttonList) {
                if (button._buttonType != "Blank" && button._buttonRef <= (_gridRows * _gridColumns)) {
                    _activeButtonCount++;
                    _maxButtonRow = (button._buttonRef / _gridColumns);
                    if (button._buttonRef % _gridColumns > 0) { _maxButtonRow++; }
                }
            }
        }
        public bool IsSelectionTargetButton()
        {
            return (_gazeSelectionAtoms || _gazeSelectionMales || _gazeSelectionFemales || _gazeSelectionPerson || _vamSelectionAtoms || _vamSelectionMales || _vamSelectionFemales || _vamSelectionPerson);
        }
        public void RecalcGazeSelections() {

            _gazeSelectionAtoms = false;
            _gazeSelectionMales = false;
            _gazeSelectionFemales = false;
            _gazeSelectionPerson = false;
            _vamSelectionAtoms = false;
            _vamSelectionMales = false;
            _vamSelectionFemales = false;
            _vamSelectionPerson = false;
            foreach (UIAButton button in _buttonList) {
                if (button._targetCategory == "Gaze Selected Atom" && button.IsAtomTargetable()) {
                    if (button._targetName == "Last viewed Atom") {
                        _gazeSelectionAtoms = true;
                    }
                    if (button._targetName == "Last viewed Person") {
                        _gazeSelectionPerson = true;
                    }
                    if (button._targetName == "Last viewed Male") {
                        _gazeSelectionMales = true;
                    }
                    if (button._targetName == "Last viewed Female") {
                        _gazeSelectionFemales = true;
                    }
                }
                else if (button._targetCategory == "VAM Selected Atom" && button.IsAtomTargetable())
                {
                    if (button._targetName == "Last selected Atom")
                    {
                        _vamSelectionAtoms = true;
                    }
                    if (button._targetName == "Last selected Person")
                    {
                        _vamSelectionPerson = true;
                    }
                    if (button._targetName == "Last selected Male")
                    {
                        _vamSelectionMales = true;
                    }
                    if (button._targetName == "Last selected Female")
                    {
                        _vamSelectionFemales = true;
                    }
                }
            }

        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["gridColumns"].AsInt = _gridColumns;
            node["gridRows"].AsInt = _gridRows;
            node["buttonSize"] = _buttonSize;
            node["activeButtonCount"].AsInt = _activeButtonCount;
            node["maxButtonRow"].AsInt = _maxButtonRow;
            node["gridLabel"] = _gridLabel;
            node["quickLaunchBar"].AsBool = _quickLaunchBar;

            JSONArray buttonArray = new JSONArray();
            _buttonList.ForEach((button) =>
            {
                buttonArray.Add(button.GetJSON());
            });

            node["Buttons"] = buttonArray;

            return node;
        }
        public void LoadJSON(JSONClass gridJSON) {
            if (gridJSON["gridColumns"] != null) _gridColumns = gridJSON["gridColumns"].AsInt;
            if (gridJSON["gridRows"] != null) _gridRows = gridJSON["gridRows"].AsInt;
            if (gridJSON["buttonSize"] != null) _buttonSize = gridJSON["buttonSize"];
            if (gridJSON["activeButtonCount"] != null) _activeButtonCount = gridJSON["activeButtonCount"].AsInt;
            if (gridJSON["maxButtonRow"] != null) _maxButtonRow = gridJSON["maxButtonRow"].AsInt;
            if (gridJSON["gridLabel"] != null) _gridLabel = gridJSON["gridLabel"];
            if (gridJSON["quickLaunchBar"] != null) _quickLaunchBar = gridJSON["quickLaunchBar"].AsBool;

            _buttonList.Clear();
            if (gridJSON["Buttons"] != null)
            {
                JSONArray buttonArrayJSON = gridJSON["Buttons"].AsArray;
                for (int i = 0; i < buttonArrayJSON.Count; i++) {
                    JSONClass buttonJSON = buttonArrayJSON[i].AsObject;
                    UIAButton button = new UIAButton(i + 1, _mvrScript);
                    button.LoadJSON(buttonJSON);
                    _buttonList.Add(button);
                }
            }
        }
    }

    public class ButtonGridDisplay {
        public JSONStorableStringChooser _buttonLayoutJSON;
        public JSONStorableStringChooser _buttonSelectJSON;
        public JSONStorableStringChooser _buttonSizeJSON;
        public JSONStorableString _gridLabelJSON;

        public ButtonGrid _selectedButtonGrid;

        private UIDynamicTextField _gridLabelTextField;
        private UIAssist _mvrScript;
        private UIDynamicPopup _buttonLayoutPopup;
        private UIDynamicPopup _buttonSelectPopup;
        private UIDynamicPopup _buttonSizePopup;
        private UIDynamic _spacer1;
        public ButtonDisplay _buttonDisplay;
        bool _isDisplayUpdating = false;

        public ButtonGridDisplay(ButtonGrid bg, UIAssist mvrScript) {
            try {
                _selectedButtonGrid = bg;
                _mvrScript = mvrScript;
                _buttonLayoutJSON = new JSONStorableStringChooser("buttonLayout", GetButtonGridLayoutChoices(), "", "", SwitchButtonLayout);
                _buttonSelectJSON = new JSONStorableStringChooser("buttonSelection", null, "1", "", SwitchSelectedButton);
                
                _buttonSizeJSON = new JSONStorableStringChooser("buttonSize", null, "Medium", "", SwitchButtonSize);
                _gridLabelJSON = new JSONStorableString("gridLabel", "", SwitchGridLabel);
                _buttonDisplay = new ButtonDisplay(bg._buttonList[0], mvrScript);
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public List<string> GetButtonGridLayoutChoices() {
            List<string> buttonGridLayoutChoices = new List<string>();
            if (!_selectedButtonGrid._quickLaunchBar)
            {
                buttonGridLayoutChoices.Add("2x2"); buttonGridLayoutChoices.Add("2x3"); buttonGridLayoutChoices.Add("3x2"); buttonGridLayoutChoices.Add("3x3");
                if (Constants._patreonContentEnabled)
                {
                    PatreonFeatures.GridLayoutFeature(buttonGridLayoutChoices);
                }
            }
            else
            {
                buttonGridLayoutChoices.Add("4x1");
                buttonGridLayoutChoices.Add("5x1");
                buttonGridLayoutChoices.Add("6x1");
                buttonGridLayoutChoices.Add("7x1");
                buttonGridLayoutChoices.Add("8x1");
                buttonGridLayoutChoices.Add("9x1");
                buttonGridLayoutChoices.Add("4x2");
                buttonGridLayoutChoices.Add("5x2");
                buttonGridLayoutChoices.Add("6x2");
                buttonGridLayoutChoices.Add("7x2");
                buttonGridLayoutChoices.Add("8x2");
                buttonGridLayoutChoices.Add("9x2");

            }

            return (buttonGridLayoutChoices);
        }
        public List<string> GetButtonSizeChoices()
        {
            List<string> buttonSizeOptions = new List<string>();
            int buttonCols  = _selectedButtonGrid._gridColumns;
            buttonSizeOptions.Add("Micro");
            buttonSizeOptions.Add("Mini");
            if (buttonCols <=6) buttonSizeOptions.Add("Small");
            if (buttonCols <= 5 && !_selectedButtonGrid._quickLaunchBar)
            {
                buttonSizeOptions.Add("Medium");
                buttonSizeOptions.Add("Large");
            }
            return (buttonSizeOptions);
        }
        public void RefreshButtonSelectionChoices() {
            List<string> buttonSelectionChoices = new List<string>();
            _mvrScript.CloseAllRightUIPopups(_buttonSelectPopup);
            for (int i = 1; i <= (_selectedButtonGrid._gridRows * _selectedButtonGrid._gridColumns); i++) {
                buttonSelectionChoices.Add(i.ToString());
            }
            _buttonSelectJSON.choices = buttonSelectionChoices;
        }

        public void SwitchGridLabel(string newLabel) {
            try {
                if (!_isDisplayUpdating)
                {
                    _selectedButtonGrid._gridLabel = newLabel;
                    if (_mvrScript._uiActivated)
                    {
                        _mvrScript.RefreshWristUIButtonGrid();
                        _mvrScript.RefreshRightUIDisplay();
                    }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchSelectedButton(string newButton) {
            try {
                _buttonDisplay._selectedButton = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1];
                if (_buttonDisplay._buttonSkinDisplay != null) _buttonDisplay._buttonSkinDisplay._selectedButtonSkin = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._buttonSkin;
                if (_buttonDisplay._pluginVariablesDisplayOn != null)
                {
                    _buttonDisplay._pluginVariablesDisplayOn._selectButtonPluginVariables = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._pluginVariablesButtonOn;
                    _buttonDisplay._pluginVariablesDisplayOn._pluginVariableSelectorJSON.val = "Select Parameter...";
                }
                if (_buttonDisplay._pluginVariablesDisplayOff != null)
                {
                    _buttonDisplay._pluginVariablesDisplayOff._selectButtonPluginVariables = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._pluginVariablesButtonOff;
                    _buttonDisplay._pluginVariablesDisplayOff._pluginVariableSelectorJSON.val = "Select Parameter...";
                }
                _buttonDisplay._dmFaceDecalDisplay._selectDecalMakerTextureCollection = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._decalMakerTextureCollectionFace;
                _buttonDisplay._dmFaceDecalDisplay._dmTextureSelectorJSON.val = "Select Texture...";

                _buttonDisplay._dmTorsoDecalDisplay._selectDecalMakerTextureCollection = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._decalMakerTextureCollectionTorso;
                _buttonDisplay._dmTorsoDecalDisplay._dmTextureSelectorJSON.val = "Select Texture...";

                _buttonDisplay._dmLimbsDecalDisplay._selectDecalMakerTextureCollection = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._decalMakerTextureCollectionLimbs;
                _buttonDisplay._dmLimbsDecalDisplay._dmTextureSelectorJSON.val = "Select Texture...";

                _buttonDisplay._dmGenDecalDisplay._selectDecalMakerTextureCollection = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._decalMakerTextureCollectionGen;
                _buttonDisplay._dmGenDecalDisplay._dmTextureSelectorJSON.val = "Select Texture...";

                _buttonDisplay._testSpawnPositionTargetJSON.val = "Select...";
                if (_buttonDisplay._isTestSpawnPositionActive) _buttonDisplay.EndSpawnTest();
                RefreshDisplay();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchButtonSize(string newButtonSize) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButtonGrid._buttonSize = newButtonSize;
                    if (_mvrScript._uiActivated)
                    {
                        _mvrScript.RefreshWristUIButtonGrid();
                    }
                }
                
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchButtonLayout(string newLayout) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButtonGrid._gridColumns = int.Parse(newLayout.Substring(0, 1));
                    _selectedButtonGrid._gridRows = int.Parse(newLayout.Substring(2, 1));

                    int buttonCols = _selectedButtonGrid._gridColumns;
                    if (buttonCols > 6 && _selectedButtonGrid._buttonSize != "Micro" && _selectedButtonGrid._buttonSize != "Mini") _buttonSizeJSON.val = "Mini";
                    else if (buttonCols >5 && (_selectedButtonGrid._buttonSize == "Medium" || _selectedButtonGrid._buttonSize == "Large") ) _buttonSizeJSON.val = "Small";
                    
                    _selectedButtonGrid.UpdateGridButtons();
                    _selectedButtonGrid.RecalcActiveButtonCount();
                    _selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplayValues();
                    if (_mvrScript._uiActivated)
                    {
                        _mvrScript.RefreshWristUIButtonGrid();
                    }
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshDisplay() {
            CloseAllPopup(null);
            if (_buttonLayoutPopup != null) { _mvrScript.RemovePopup(_buttonLayoutPopup); }
            if (_buttonSelectPopup != null) { _mvrScript.RemovePopup(_buttonSelectPopup); }
            if (_buttonSizePopup != null) { _mvrScript.RemovePopup(_buttonSizePopup); }
            if (_gridLabelTextField != null) { _mvrScript.RemoveTextField(_gridLabelTextField); }
            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (_mvrScript._displayScreenLabelsJSON.val && Constants._patreonContentEnabled && test != null && test.Count == 19 && !_selectedButtonGrid._quickLaunchBar) _gridLabelTextField = _mvrScript.CreateTextInput(_gridLabelJSON, true);

            string layoutLabel =  "Screen " + _mvrScript._buttonGridSelectorJSON.val + " Layout";
            string sizeLabel = "Screen " + _mvrScript._buttonGridSelectorJSON.val + " Button Size";
            if (_selectedButtonGrid._quickLaunchBar)
            {
                layoutLabel = "Quick Launch Bar Buttons";
                sizeLabel = "Button Size";
            }
            _buttonLayoutPopup = _mvrScript.CreateSetupPopup(_buttonLayoutJSON, true, 500f, ButtonLayoutPopupOpen, layoutLabel);
             _buttonSizePopup = _mvrScript.CreateSetupPopup(_buttonSizeJSON, true, 300f, ButtonSizePopupOpen, sizeLabel, false);

            _spacer1 = _mvrScript.CreateSpacer(true);
            _spacer1.height = 10;
            _buttonSelectPopup = _mvrScript.CreateSetupPopup(_buttonSelectJSON, true, 500f, RefreshButtonSelectionChoices, "Button Selector");
            _buttonSelectPopup.labelTextColor = Color.blue;

            RefreshDisplayValues();
            _buttonDisplay.RefreshDisplay();
        }
        public void ButtonLayoutPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_buttonLayoutPopup);
            _buttonLayoutJSON.choices = GetButtonGridLayoutChoices();
        }
        public void ButtonSizePopupOpen()
        {
            _mvrScript.CloseAllRightUIPopups(_buttonSizePopup);
            _buttonSizeJSON.choices = GetButtonSizeChoices();
        }
        public void CloseAllPopup(UIDynamicPopup exception) {
            if (_buttonLayoutPopup != null && _buttonLayoutPopup != exception) { _buttonLayoutPopup.popup.visible = false; }
            if (_buttonSelectPopup != null && _buttonSelectPopup != exception) { _buttonSelectPopup.popup.visible = false; }
            if (_buttonSizePopup != null && _buttonSizePopup != exception) { _buttonSizePopup.popup.visible = false; }

            _buttonDisplay.CloseAllPopup(exception);
        }
        public void RefreshDisplayValues() {
            _isDisplayUpdating = true;
            int cols = _selectedButtonGrid._gridColumns;
            int rows = _selectedButtonGrid._gridRows;
            if (_selectedButtonGrid._gridLabel == "") _gridLabelJSON.val = "<Screen " + _mvrScript._buttonGridSelectorJSON.val + " Label>";
            else _gridLabelJSON.val = _selectedButtonGrid._gridLabel;
            _buttonSizeJSON.val = _selectedButtonGrid._buttonSize;
            _buttonLayoutJSON.val = cols.ToString() + " x " + rows.ToString();

            if (int.Parse(_buttonSelectJSON.val) > rows * cols) { _buttonSelectJSON.val = (rows * cols).ToString(); }

            _isDisplayUpdating = false;
        }
    }

    public class UIAButton {

        public int _buttonRef;
        public string _buttonCategory = "";
        public string _buttonType = "";
        private int _buttonState = 1;
        public int _maxButtonStates = 1;
        public string _targetCategory = "";
        public string _targetName = "";
        public string _targetParamNameOn = "";
        public string _targetParamNameOff = "";
        public bool _suppressClothingLoad = false;
        public bool _suppressPersonScaleLoad = false;
        public bool _mergeLoadPreset = false;
        public bool _autoLabel = true;
        public string _pluginType = "";
        public bool _pluginForceMergeLoad = false;
        public bool _closeGameUIonPlayMode = true;
        public bool _openGameUIonEditMode = true;
        public bool _decalFaceLoad = false;
        public bool _decalBodyLoad = false;
        public bool _decalLimbsLoad = false;
        public bool _decalGenLoad = false;

        public string _atomCategory = "People";
        public string _atomType = "Person";
        public string _atomPositionRelativeTo = "";
        public string _targetNode = "control";
        public bool _parentLinkToTarget = false;

        public float _atomPositionX = 0;
        public float _atomPositionY = 0;
        public float _atomPositionZ = 0;
        public float _atomRotationX = 0;
        public float _atomRotationY = 0;
        public float _atomRotationZ = 0;
        public bool _atomAbsoluteHeight = false;
        public bool _atomAbsolutePitch = false;
        public bool _atomAbsoluteRoll = false;

        public string _preset1Type = "None";
        public string _preset1FilePath = "";
        public string _preset2Type = "None";
        public string _preset2FilePath = "";
        public string _preset3Type = "None";
        public string _preset3FilePath = "";
        public string _buttonLabelOn = "";
        public string _buttonLabelOff = "";

        public bool _lockedPositionX =false;
        public bool _lockedPositionY = true;
        public bool _lockedPositionZ = false;
        public bool _lockedRotationX = true;
        public bool _lockedRotationY = false;
        public bool _lockedRotationZ = true;
        public string _vrMovePositionGranulatity = "Coarse";
        public string _vrMoveRotationGranulatity = "Medium";

        public bool _vrMoveDisablePhysics = false;

        public Texture2D _buttonTexture = null;
        public bool _useButtonTexture = true;
        public ButtonSkin _buttonSkin;

        public DecalMakerTextureCollection _decalMakerTextureCollectionFace;
        public DecalMakerTextureCollection _decalMakerTextureCollectionTorso;
        public DecalMakerTextureCollection _decalMakerTextureCollectionLimbs;
        public DecalMakerTextureCollection _decalMakerTextureCollectionGen;

        public PluginVariables _pluginVariablesButtonOn;
        public PluginVariables _pluginVariablesButtonOff;

        private UIAssist _mvrScript;

        public JSONClass _hairScalpControl;
        public JSONClass _hairSimControl;

        public List<ButtonPluginLoad> _pluginLoadList;
        public Dictionary<string, int> _atomButtonStates;
        public Dictionary<string, string> _atomClothingPresets;
        public List<string> _tagsToRemoveOnMerge;

        // These data items are deliberatly not persisted in save file
        public string _selectedPreset1File = "";
        public string _selectedPreset2File = "";
        public string _selectedPreset3File = "";
        private string _selectedTarget = "";
        private bool _hideMenuOnFileSelect = false;
        public bool _thumbnailImageLoading = false;

        public bool isLegacyPresetLoadSaveType
        {
            get { return (_buttonType.StartsWith("Load Legacy") || _buttonType.StartsWith("Save Legacy")); }
        }
        public bool isLegacyPresetSaveType
        {
            get { return (_buttonType.StartsWith("Save Legacy")); }
        }
        public bool isNonLegacyPresetLoadType
        {
            get { return ((_buttonType.StartsWith("Load") && _buttonType.EndsWith("Preset") && _buttonType!="Load Legacy Preset") ); }
        }
        public bool isClothingPresetLoadType
        {
            get { return ((_buttonCategory == "Clothing" && _buttonType.EndsWith("Preset")) || _buttonType=="Load Clothing Preset"); }
        }
        public bool isPresetType
        {
            get { return (isLegacyPresetLoadSaveType || isNonLegacyPresetLoadType || (_buttonCategory == "Clothing" && _buttonType.EndsWith("Preset"))); }
        }
        public bool isSceneLoadType
        {
            get { return (_buttonType.EndsWith("Load Scene")); }
        }
        public bool isUIAPLoadType
        {
            get { return (_buttonType == "Load UI Assist Profile"); }
        }

        //Button is of a toggle type (e.g. 2 states) and uses a preset file
        public bool isPresetToggleType
        {
            get { return (_buttonType== "Undress All Clothing" || _buttonType == "Merge Clothing Preset" || _buttonType == "Undress Cloth Preset"); }
        }
        public bool isPersonAtomOnlyType
        {
            get{return ((isPresetType && _buttonType!="Load General Preset") || _buttonType.EndsWith("Clothing") || _buttonType == "Detach Root Control" || (_buttonCategory=="Appearance" && _buttonType!="Suppress Scale Load")); }
        }
        public UIAButton(int buttonRef, UIAssist mvr)
        {
            try
            {
                _buttonRef = buttonRef;
                _buttonCategory = "Miscellaneous";
                _buttonType = "Blank";
                _targetCategory = "Gaze Selected Atom";
                _mvrScript = mvr;
                _pluginLoadList = new List<ButtonPluginLoad>();
                ButtonPluginLoad tempPIL = new ButtonPluginLoad();
                _pluginLoadList.Add(tempPIL);
                _atomButtonStates = new Dictionary<string, int>();
                _atomClothingPresets = new Dictionary<string, string>();
                _decalMakerTextureCollectionFace = new DecalMakerTextureCollection(this);
                _decalMakerTextureCollectionTorso = new DecalMakerTextureCollection(this);
                _decalMakerTextureCollectionLimbs = new DecalMakerTextureCollection(this);
                _decalMakerTextureCollectionGen = new DecalMakerTextureCollection(this);

                _hairScalpControl = null;
                _hairSimControl = null;


                _tagsToRemoveOnMerge = new List<string>();
                if (Constants._patreonContentEnabled) _buttonSkin = new ButtonSkin(_maxButtonStates, mvr);
                _pluginVariablesButtonOn = new PluginVariables(this);
                _pluginVariablesButtonOff = new PluginVariables(this);

                List<string> test = new List<string>();
                test.Add("Temp");
                PatreonFeatures.GridLayoutFeature(test);
                if (!Constants._patreonContentEnabled || test == null || test.Count != 19)
                {
                    _preset1Type = "Single File";
                    _preset2Type = "Single File";
                    _preset3Type = "Single File";
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["buttonCategory"] = _buttonCategory;
            node["buttonType"] = _buttonType;
            if (_buttonType != "Blank") {
                node["targetCategory"] = _targetCategory;
                node["targetName"] = _targetName;
                node["targetParamNameOn"] = _targetParamNameOn;
                node["targetParamNameOff"] = _targetParamNameOff;
                node["pluginType"] = _pluginType;
                node["presetFilePath"] = _preset1FilePath;
                node["preset2FilePath"] = _preset2FilePath;
                node["preset3FilePath"] = _preset3FilePath;
                node["preset1Type"] = _preset1Type;
                node["preset2Type"] = _preset2Type;
                node["preset3Type"] = _preset3Type;
                node["parentLinkToTarget"].AsBool = _parentLinkToTarget;

                node["buttonLabelOn"] = _buttonLabelOn;
                node["buttonLabelOff"] = _buttonLabelOff;

                node["suppressClothingLoad"].AsBool = _suppressClothingLoad;
                node["suppressPersonScaleLoad"].AsBool = _suppressPersonScaleLoad;
                node["mergeLoadPreset"].AsBool = _mergeLoadPreset;
                
                node["autoLabel"].AsBool = _autoLabel;
                node["useButtonTexture"].AsBool = _useButtonTexture;
                node["pluginForceMergeLoad"].AsBool = _pluginForceMergeLoad;
                node["closeGameUIonPlayMode"].AsBool = _closeGameUIonPlayMode;
                node["openGameUIonEditMode"].AsBool = _openGameUIonEditMode;

                node["decalFaceLoad"].AsBool = _decalFaceLoad;
                node["decalBodyLoad"].AsBool = _decalBodyLoad;
                node["decalLimbsLoad"].AsBool = _decalLimbsLoad;
                node["decalGenLoad"].AsBool = _decalGenLoad;

                node["atomCategory"] = _atomCategory;
                node["atomType"] = _atomType;

                node["atomPositionRelativeTo"] = _atomPositionRelativeTo;
                node["targetNode"] = _targetNode;

                node["atomPositionX"].AsFloat = _atomPositionX;
                node["atomPositionY"].AsFloat = _atomPositionY;
                node["atomPositionZ"].AsFloat = _atomPositionZ;
                node["atomRotationX"].AsFloat = _atomRotationX;
                node["atomRotationY"].AsFloat = _atomRotationY;
                node["atomRotationZ"].AsFloat = _atomRotationZ;
                node["atomAbsoluteHeight"].AsBool = _atomAbsoluteHeight;
                node["atomAbsolutePitch"].AsBool = _atomAbsolutePitch;
                node["atomAbsoluteRoll"].AsBool = _atomAbsoluteRoll;

                node["lockedPositionX"].AsBool = _lockedPositionX;
                node["lockedPositionY"].AsBool = _lockedPositionY;
                node["lockedPositionZ"].AsBool = _lockedPositionZ;
                node["lockedRotationX"].AsBool = _lockedRotationX;
                node["lockedRotationY"].AsBool = _lockedRotationY;
                node["lockedRotationZ"].AsBool = _lockedRotationZ;
                node["vrMoveDisablePhysics"].AsBool = _vrMoveDisablePhysics;                

                node["vrMovePositionGranulatity"] = _vrMovePositionGranulatity;
                node["vrMoveRotationGranulatity"] = _vrMoveRotationGranulatity;

                node["maxButtonStates"].AsInt = _maxButtonStates;

                if (_hairSimControl != null) node["hairSimControl"] = _hairSimControl;
                if (_hairScalpControl != null) node["hairScalpControl"] = _hairScalpControl;

                if (Constants._patreonContentEnabled) node["ButtonSkin"] = _buttonSkin.GetJSON();

                node["pluginVariablesOn"] = _pluginVariablesButtonOn.GetJSON();
                node["pluginVariablesOff"] = _pluginVariablesButtonOff.GetJSON();

                if (_buttonType == "Load Plugin(s)" || _buttonType == "Spawn Atom") {
                    JSONArray pluginLoadArray = new JSONArray();
                    _pluginLoadList.ForEach((pluginLoad) =>
                    {
                        if (pluginLoad._filePath != "") pluginLoadArray.Add(pluginLoad.GetJSON());
                    });

                    node["pluginLoads"] = pluginLoadArray;
                }
                if (_buttonType == "Decal Maker Toggle")
                {
                    node["decalMakerTextureCollectionFace"] = _decalMakerTextureCollectionFace.GetJSON();
                    node["decalMakerTextureCollectionTorso"] = _decalMakerTextureCollectionTorso.GetJSON();
                    node["decalMakerTextureCollectionLimbs"] = _decalMakerTextureCollectionLimbs.GetJSON();
                    node["decalMakerTextureCollectionGen"] = _decalMakerTextureCollectionGen.GetJSON();
                }
                if (_tagsToRemoveOnMerge.Count > 0)
                {
                    JSONArray tagsToRemoveArray = new JSONArray();
                    foreach (string tag in _tagsToRemoveOnMerge) node["tagsToRemove"][-1] = tag;
                }
            }
            return node;
        }
        public void LoadJSON(JSONClass buttonJSON) {
            if (buttonJSON["buttonCategory"] != null) _buttonCategory = buttonJSON["buttonCategory"];
            if (buttonJSON["buttonType"] != null)
            {
                if (buttonJSON["buttonType"]== "Hide 'Off' Control Nodes") _buttonType = "Hide Inactive Targets";
                else _buttonType = buttonJSON["buttonType"];
            }
            if (buttonJSON["targetCategory"] != null) _targetCategory = buttonJSON["targetCategory"];
            if (buttonJSON["targetName"] != null)
            {
                if (_buttonType == "Spawn Atom") _targetName = "";
                else _targetName = buttonJSON["targetName"];
            }
            if (buttonJSON["targetParamNameOn"] != null) _targetParamNameOn = buttonJSON["targetParamNameOn"];
            if (buttonJSON["targetParamNameOff"] != null) _targetParamNameOff = buttonJSON["targetParamNameOff"];
            if (buttonJSON["pluginType"] != null) _pluginType = buttonJSON["pluginType"];
            if (buttonJSON["presetFilePath"] != null) _preset1FilePath = buttonJSON["presetFilePath"];
            if (buttonJSON["preset2FilePath"] != null) _preset2FilePath = buttonJSON["preset2FilePath"];
            if (buttonJSON["preset3FilePath"] != null) _preset3FilePath = buttonJSON["preset3FilePath"];
            if (buttonJSON["preset1Type"] != null) _preset1Type = buttonJSON["preset1Type"];
            else if (_buttonType.EndsWith("Preset") || _buttonType.StartsWith("Load Legacy")) _preset1Type = "Single File";
            // Fix to correct bug in 1.4.0 that introduce preset1Type of "None" for Preset button types
            if (_buttonType.EndsWith("Preset") && _preset1Type=="None") _preset1Type = "Single File";
            if (buttonJSON["preset2Type"] != null) _preset2Type = buttonJSON["preset2Type"];
            if (buttonJSON["preset3Type"] != null) _preset3Type = buttonJSON["preset3Type"];
            if (buttonJSON["parentLinkToTarget"] != null) _parentLinkToTarget = buttonJSON["parentLinkToTarget"].AsBool ;

            if (buttonJSON["buttonLabelOn"] != null) _buttonLabelOn = buttonJSON["buttonLabelOn"];
            if (buttonJSON["buttonLabelOff"] != null) _buttonLabelOff = buttonJSON["buttonLabelOff"];

            if (buttonJSON["suppressClothingLoad"] != null) _suppressClothingLoad = buttonJSON["suppressClothingLoad"].AsBool;
            if (buttonJSON["suppressPersonScaleLoad"] != null) _suppressPersonScaleLoad = buttonJSON["suppressPersonScaleLoad"].AsBool;
            if (buttonJSON["mergeLoadPreset"] != null) _mergeLoadPreset = buttonJSON["mergeLoadPreset"].AsBool;
            
            if (buttonJSON["autoLabel"] != null) _autoLabel = buttonJSON["autoLabel"].AsBool;
            if (buttonJSON["useButtonTexture"] != null) _useButtonTexture = buttonJSON["useButtonTexture"].AsBool;
            if (buttonJSON["pluginForceMergeLoad"] != null) _pluginForceMergeLoad = buttonJSON["pluginForceMergeLoad"].AsBool;
            if (buttonJSON["closeGameUIonPlayMode"] != null) _closeGameUIonPlayMode = buttonJSON["closeGameUIonPlayMode"].AsBool;
            if (buttonJSON["openGameUIonEditMode"] != null) _openGameUIonEditMode = buttonJSON["openGameUIonEditMode"].AsBool;

            if (buttonJSON["decalFaceLoad"] != null) _decalFaceLoad = buttonJSON["decalFaceLoad"].AsBool;
            if (buttonJSON["decalBodyLoad"] != null) _decalBodyLoad = buttonJSON["decalBodyLoad"].AsBool;
            if (buttonJSON["decalLimbsLoad"] != null) _decalLimbsLoad = buttonJSON["decalLimbsLoad"].AsBool;
            if (buttonJSON["decalGenLoad"] != null) _decalGenLoad = buttonJSON["decalGenLoad"].AsBool;

            if (buttonJSON["hairSimControl"] != null) _hairSimControl=(JSONClass)buttonJSON["hairSimControl"];
            else _hairSimControl = null;
            if (buttonJSON["hairScalpControl"] != null) _hairScalpControl=(JSONClass)buttonJSON["hairScalpControl"];
            else _hairScalpControl = null;

            if (buttonJSON["maxButtonStates"] != null) _maxButtonStates = buttonJSON["maxButtonStates"].AsInt;
            if (buttonJSON["ButtonSkin"] != null && Constants._patreonContentEnabled) _buttonSkin.LoadJSON((JSONClass)buttonJSON["ButtonSkin"]);
            if (buttonJSON["pluginVariablesOn"] != null) _pluginVariablesButtonOn.LoadJSON((JSONClass)buttonJSON["pluginVariablesOn"]);
            if (buttonJSON["pluginVariablesOff"] != null) _pluginVariablesButtonOff.LoadJSON((JSONClass)buttonJSON["pluginVariablesOff"]);

            if (buttonJSON["atomCategory"] != null) _atomCategory = buttonJSON["atomCategory"];
            if (buttonJSON["atomType"] != null) _atomType = buttonJSON["atomType"];
            if (buttonJSON["targetNode"] != null) _targetNode = buttonJSON["targetNode"];
            if (buttonJSON["atomPositionRelativeTo"] != null) _atomPositionRelativeTo = buttonJSON["atomPositionRelativeTo"];
            else if (_buttonType == "Spawn Atom") _atomPositionRelativeTo = buttonJSON["targetName"];

            if (buttonJSON["atomPositionX"] != null) _atomPositionX = buttonJSON["atomPositionX"].AsFloat;
            if (buttonJSON["atomPositionY"] != null) _atomPositionY = buttonJSON["atomPositionY"].AsFloat;
            if (buttonJSON["atomPositionZ"] != null) _atomPositionZ = buttonJSON["atomPositionZ"].AsFloat;
            if (buttonJSON["atomRotationX"] != null) _atomRotationX = buttonJSON["atomRotationX"].AsFloat;
            if (buttonJSON["atomRotationY"] != null) _atomRotationY = buttonJSON["atomRotationY"].AsFloat;
            if (buttonJSON["atomRotationZ"] != null) _atomRotationZ = buttonJSON["atomRotationZ"].AsFloat;
            if (buttonJSON["atomAbsoluteHeight"] != null) _atomAbsoluteHeight = buttonJSON["atomAbsoluteHeight"].AsBool;
            if (buttonJSON["atomAbsolutePitch"] != null) _atomAbsolutePitch = buttonJSON["atomAbsolutePitch"].AsBool;
            if (buttonJSON["atomAbsoluteRoll"] != null) _atomAbsoluteRoll = buttonJSON["atomAbsoluteRoll"].AsBool;

            if (buttonJSON["lockedPositionX"] != null) _lockedPositionX= buttonJSON["lockedPositionX"].AsBool;
            if (buttonJSON["lockedPositionY"] != null) _lockedPositionY = buttonJSON["lockedPositionY"].AsBool ;
            if (buttonJSON["lockedPositionZ"] != null) _lockedPositionZ =buttonJSON["lockedPositionZ"].AsBool ;
            if (buttonJSON["lockedRotationX"] != null) _lockedRotationX = buttonJSON["lockedRotationX"].AsBool ;
            if (buttonJSON["lockedRotationY"] != null) _lockedRotationY = buttonJSON["lockedRotationY"].AsBool;
            if (buttonJSON["lockedRotationZ"] != null) _lockedRotationZ = buttonJSON["lockedRotationZ"].AsBool ;
            if (buttonJSON["vrMoveDisablePhysics"] != null) _vrMoveDisablePhysics = buttonJSON["vrMoveDisablePhysics"].AsBool;

            
            if (buttonJSON["vrMovePositionGranulatity"] != null) _vrMovePositionGranulatity= buttonJSON["vrMovePositionGranulatity"] ;
            if (buttonJSON["vrMoveRotationGranulatity"] != null) _vrMoveRotationGranulatity=buttonJSON["vrMoveRotationGranulatity"] ;

            if (buttonJSON["pluginLoads"] != null)
            {
                JSONArray pluginLoadsArrayJSON = buttonJSON["pluginLoads"].AsArray;
                for (int i = 0; i < pluginLoadsArrayJSON.Count; i++) {
                    JSONClass pluginLoadJSON = pluginLoadsArrayJSON[i].AsObject;
                    ButtonPluginLoad pluginLoad;
                    if (i >= _pluginLoadList.Count) { pluginLoad = new ButtonPluginLoad(); _pluginLoadList.Add(pluginLoad); }
                    else { pluginLoad = _pluginLoadList[i]; }

                    pluginLoad.LoadJSON(pluginLoadJSON);
                }
            }
            
            if (_buttonType == "Decal Maker Toggle")
            {
                if (buttonJSON["decalMakerTextureCollectionFace"] != null) _decalMakerTextureCollectionFace.LoadJSON((JSONClass) buttonJSON["decalMakerTextureCollectionFace"]);
                if (buttonJSON["decalMakerTextureCollectionTorso"] != null) _decalMakerTextureCollectionTorso.LoadJSON((JSONClass)buttonJSON["decalMakerTextureCollectionTorso"]);
                if (buttonJSON["decalMakerTextureCollectionLimbs"] != null) _decalMakerTextureCollectionLimbs.LoadJSON((JSONClass)buttonJSON["decalMakerTextureCollectionLimbs"]);
                if (buttonJSON["decalMakerTextureCollectionGen"] != null) _decalMakerTextureCollectionGen.LoadJSON((JSONClass)buttonJSON["decalMakerTextureCollectionGen"]);

            }
            if (_preset1FilePath != ""  && _preset1FilePath.Contains(".")) QueueLoadTexture(_preset1FilePath.Substring(0, _preset1FilePath.LastIndexOf('.')) + ".jpg");
            
            if (buttonJSON["tagsToRemove"] != null)
            {
                _tagsToRemoveOnMerge = new List<string>();
                JSONArray tagsToRemoveArray = buttonJSON["tagsToRemove"].AsArray;
                for (int i = 0; i < tagsToRemoveArray.Count; i++)
                {
                    _tagsToRemoveOnMerge.Add(tagsToRemoveArray[i]);
                }
            }

        }
        public void CopyFrom(UIAButton bn) {
            if (bn != null) {
                _buttonCategory = bn._buttonCategory;
                _buttonType = bn._buttonType;
                _buttonState = 1;
                _maxButtonStates = bn._maxButtonStates;
                _targetCategory = bn._targetCategory;
                _targetName = bn._targetName;
                _targetParamNameOn = bn._targetParamNameOn;
                _targetParamNameOff = bn._targetParamNameOff;
                _suppressClothingLoad = bn._suppressClothingLoad;
                _suppressPersonScaleLoad = bn._suppressPersonScaleLoad;
                _mergeLoadPreset = bn._mergeLoadPreset;
                _autoLabel = bn._autoLabel;
                _pluginType = bn._pluginType;
                _pluginForceMergeLoad = bn._pluginForceMergeLoad;
                _preset1FilePath = bn._preset1FilePath;
                _preset2FilePath = bn._preset2FilePath;
                _preset3FilePath = bn._preset3FilePath;
                _preset1Type = bn._preset1Type;
                _preset2Type = bn._preset2Type;
                _preset3Type = bn._preset3Type;
                _parentLinkToTarget = bn._parentLinkToTarget;

                _buttonLabelOn = bn._buttonLabelOn;
                _buttonLabelOff = bn._buttonLabelOff;
                _buttonTexture = bn._buttonTexture;
                _useButtonTexture = bn._useButtonTexture;
                _closeGameUIonPlayMode = bn._closeGameUIonPlayMode;
                _openGameUIonEditMode = bn._openGameUIonEditMode;

                _decalFaceLoad = bn._decalFaceLoad;
                _decalBodyLoad = bn._decalBodyLoad;
                _decalLimbsLoad = bn._decalLimbsLoad;
                _decalGenLoad = bn._decalGenLoad;

                if (bn._hairScalpControl != null)
                {
                    _hairScalpControl = (JSONClass) JSONClass.Parse(bn._hairScalpControl.ToString());
                    
                }
                else _hairScalpControl = null;

                if (bn._hairSimControl != null)
                {
                    _hairSimControl = (JSONClass)JSONClass.Parse(bn._hairSimControl.ToString());                    
                }
                else _hairSimControl = null;

                _atomCategory = bn._atomCategory;
                _atomType = bn._atomType;
                _targetNode = bn._targetNode;
                _atomPositionRelativeTo = bn._atomPositionRelativeTo;
                
                _atomPositionX = bn._atomPositionX;
                _atomPositionY = bn._atomPositionY;
                _atomPositionZ = bn._atomPositionZ;
                _atomRotationX = bn._atomRotationX;
                _atomRotationY = bn._atomRotationY;
                _atomRotationZ = bn._atomRotationZ;
                _atomAbsoluteHeight = bn._atomAbsoluteHeight;
                _atomAbsolutePitch = bn._atomAbsolutePitch;
                _atomAbsoluteRoll = bn._atomAbsoluteRoll;
                _lockedPositionX = bn._lockedPositionX;
                _lockedPositionY = bn._lockedPositionY;
                _lockedPositionZ = bn._lockedPositionZ;
                _lockedRotationX = bn._lockedRotationX;
                _lockedRotationY = bn._lockedRotationY;
                _lockedRotationZ = bn._lockedRotationZ;
                _vrMoveDisablePhysics = bn._vrMoveDisablePhysics;
                _vrMovePositionGranulatity = bn._vrMovePositionGranulatity;
                _vrMoveRotationGranulatity = bn._vrMoveRotationGranulatity;

                _buttonSkin.CopyFrom(bn._buttonSkin);
                
                _atomButtonStates = new Dictionary<string, int>();
                _atomClothingPresets = new Dictionary<string, string>();

                _tagsToRemoveOnMerge = new List<string>();
                foreach (string tag in bn._tagsToRemoveOnMerge) _tagsToRemoveOnMerge.Add(tag);
                _pluginLoadList.Clear();

                for (int i = 0; i < bn._pluginLoadList.Count; i++)
                {
                    ButtonPluginLoad newPluginLoad = new ButtonPluginLoad();
                    newPluginLoad.CopyFrom(bn._pluginLoadList[i]);
                    _pluginLoadList.Add(newPluginLoad);
                }
                _pluginVariablesButtonOn.CopyFrom(bn._pluginVariablesButtonOn);
                _pluginVariablesButtonOff.CopyFrom(bn._pluginVariablesButtonOff);

                _decalMakerTextureCollectionFace.CopyFrom(bn._decalMakerTextureCollectionFace);

                _decalMakerTextureCollectionTorso.CopyFrom(bn._decalMakerTextureCollectionTorso);
                _decalMakerTextureCollectionLimbs.CopyFrom(bn._decalMakerTextureCollectionLimbs);
                _decalMakerTextureCollectionGen.CopyFrom(bn._decalMakerTextureCollectionGen);                
            }
        }
        public string GetManualLabel(int buttonState, string targetAtomName) {
            string buttonLabel = "";
            if (buttonState == 1) buttonLabel = _buttonLabelOn;
            else buttonLabel = _buttonLabelOff;
            if (_targetCategory == "Gaze Selected Atom" && buttonLabel.ToUpper().Contains("<GAZE>")) {
                int index = buttonLabel.ToUpper().IndexOf("<GAZE>");
                if (targetAtomName == "") buttonLabel = buttonLabel.Substring(0, index) + "<TARGET>" + buttonLabel.Substring(index + 6);
                else buttonLabel = buttonLabel.Substring(0, index) + targetAtomName + buttonLabel.Substring(index + 6);
            }
            else if (_targetCategory == "VAM Selected Atom" && buttonLabel.ToUpper().Contains("<SELECTED>"))
            {
                int index = buttonLabel.ToUpper().IndexOf("<SELECTED>");
                if (targetAtomName == "") buttonLabel = buttonLabel.Substring(0, index) + "<TARGET>" + buttonLabel.Substring(index + 10);
                else buttonLabel = buttonLabel.Substring(0, index) + targetAtomName + buttonLabel.Substring(index + 10);
            }
            else if ((_targetCategory == "VAM Selected Atom"|| _targetCategory == "Gaze Selected Atom") && buttonLabel.ToUpper().Contains("<TARGET>"))
            {
                int index = buttonLabel.ToUpper().IndexOf("<TARGET>");
                if (targetAtomName != "") buttonLabel = buttonLabel.Substring(0, index) + targetAtomName + buttonLabel.Substring(index + 8);
            }
            return (buttonLabel);
        }
        public string GetAutoLabel(int buttonState, string targetAtomName) {
            string buttonLabel = "";

            if (_buttonType == "Blank") return ("");
            else if (_buttonType == "Freeze Motion/Sound")
            {
                if (buttonState == 1) return ("Feeze Motion / Sound");
                else return ("Unfeeze Motion / Sound");
            }
            else if (_buttonType == "Toggle Play/Edit Mode")
            {
                if (buttonState == 1) return ("Edit Mode");
                else return ("Play Mode");
            }
            else if (_buttonType == "Toggle Soft Body Physics")
            {
                if (buttonState == 1) return ("Enable Soft Body Physics");
                else return ("Disable Soft Body Physics");
            }
            else if (_buttonType == "Toggle Desktop VSync")
            {
                if (buttonState == 1) return ("Enable Desktop VSync");
                else return ("Disable Desktop VSync");
            }
            else if (_buttonType == "Toggle Performance Monitor")
            {
                if (buttonState == 1) return ("Enable Perf Monitor");
                else return ("Disable Perf Monitor");
            }
            else if (_buttonType == "Toggle Show Hidden Atoms")
            {
                if (buttonState == 1) return ("Show Hidden Atoms");
                else return ("Disable Show Hidden Atoms");
            }
            else if (_buttonType == "Load Scene")
            {
                buttonLabel = "Load Scene";
                return (buttonLabel);
            }
 
            else if (_buttonType == "Merge Load Scene")
            {
                buttonLabel = "Merge Scene";
                return (buttonLabel);
            }
            else if (_buttonType == "Load UI Assist Profile")
            {
                buttonLabel = "Load UI Assist Profile";
                return (buttonLabel);
            }
            else if (_buttonType == "Gaze Assisted Select")
            {
                if (buttonState == 1) buttonLabel = "Enable Gaze Assisted Select";
                else buttonLabel = "Disable Gaze Assisted Select";
                return (buttonLabel);
            }
            else if (_buttonType == "Hide Inactive Targets")
            {
                if (buttonState == 1) buttonLabel = "Hide Inactive Targets";
                else buttonLabel = "Show Inactive Targets";
                return (buttonLabel);
            }
            else if (_buttonType == "Active Clothing Editor")
            {
                if (buttonState == 1) buttonLabel = "Open Active Clothing";
                else buttonLabel = "Close Active Clothing";
                return (buttonLabel);
            }
            else if (_buttonType == "Raise People by Heel Adjust Height")
            {
                if (buttonState == 1) buttonLabel = "Enable Heel Adjust Height raise";
                else buttonLabel = "Disable Heel Adjust Height raise";
                return (buttonLabel);
            }
            else if (_buttonType == "Spawn Atom")
            {
                buttonLabel = "Spawn " + _atomType;
                if (_atomPositionRelativeTo.StartsWith("Last"))
                {
                    if (targetAtomName == "") buttonLabel = buttonLabel + " on <TARGET>";
                    else buttonLabel = buttonLabel + " on " + targetAtomName;
                }
                if (_atomPositionRelativeTo == "VR Left Hand") buttonLabel = buttonLabel + " at Left Hand";
                if (_atomPositionRelativeTo == "VR Right Hand") buttonLabel = buttonLabel + " at Right Hand";
                if (_atomPositionRelativeTo == "VR Head") buttonLabel = buttonLabel + " here";
                return (buttonLabel);
            }
            else if (_buttonType == "Teleport Atom")
            {
                buttonLabel = "Teleport";
                if (_atomPositionRelativeTo.StartsWith("Last"))
                {
                    if (targetAtomName == "") buttonLabel = buttonLabel + " to <TARGET>";
                    else buttonLabel = buttonLabel + " to " + targetAtomName;
                }
                if (_atomPositionRelativeTo == "VR Left Hand") buttonLabel = buttonLabel + " to Left Hand";
                if (_atomPositionRelativeTo == "VR Right Hand") buttonLabel = buttonLabel + " to Right Hand";
                if (_atomPositionRelativeTo == "VR Head") buttonLabel = buttonLabel + " here";
                return (buttonLabel);
            }
            else if (_buttonType == "Hide Atom")
            {
                if (buttonState == 1) return ("Hide " + targetAtomName);
                else return ("Show " + targetAtomName);
            }
            else if (_buttonType == "Detach Root Control")
            {
                if (buttonState == 1) return ("Detach Root Control from " + targetAtomName);
                else return ("Reattach Root Control to  " + targetAtomName);
            }
            else if (_buttonType == "Toggle Atom On/Off")
            {
                if (buttonState == 1) return ("Turn " +targetAtomName +" atom Off");
                else return ("Turn " + targetAtomName + " atom On");
            }
            else if (_buttonType == "Delete Atom") return ("Delete " + targetAtomName);
            else if (_buttonType == "Select Atom") return ("Select " + targetAtomName);
            else if (_buttonType == "Move Atom (VR Only)")
            {
                if (buttonState == 1)
                {
                    if (targetAtomName!="") return ("Move " + targetAtomName);
                    else return ("Move Atom");
                } 
                else return ("Stop Move Atom");
            }
            else if (_buttonType == "Toggle Atom Collision")
            {
                if (buttonState == 1) return ("Enable Collision for " + targetAtomName);
                else return ("Disable Collision for " + targetAtomName);
                
            }

            if (_buttonCategory == "Presets" || _buttonCategory == "Clothing" || _buttonCategory == "Appearance")
            {
                if (isPresetType) {
                    buttonLabel = _buttonType.Substring(0, _buttonType.Length - 7);
                    if (buttonLabel.EndsWith("App")) buttonLabel = buttonLabel + "earance";
                    string presetName = " <PRESET>";
                    if (_preset1FilePath != "") {
                        int filenameIndex = _preset1FilePath.LastIndexOf('/');
                        presetName = " '" + _preset1FilePath.Substring(filenameIndex + 8, _preset1FilePath.Length - (filenameIndex + 8) - 4) + "'";
                    }
                    buttonLabel = buttonLabel + presetName;
                }
                else buttonLabel = _buttonType;
                if (buttonLabel.StartsWith("Undress") && buttonState == 1) { buttonLabel = "Set Undress for " + buttonLabel.Substring(8); }
                if (buttonLabel.StartsWith("Undress") && buttonState == 2) { buttonLabel = "Unset Undress for " + buttonLabel.Substring(8); }
                if (buttonLabel.StartsWith("Merge") && buttonState == 2) { buttonLabel = "Unmerge " + buttonLabel.Substring(6); }
                if (_buttonType == "Suppress Scale Load" && buttonState == 2) buttonLabel = "Unsuppress Scale Load";

            }
            else if (_buttonCategory == "Plugins") {
                string targetParamOn = _targetParamNameOn;
                string targetParamOff = _targetParamNameOff;
                if (targetParamOn == "No Actions available") targetParamOn = "<ACTION>";
                if (targetParamOff == "No Actions available") targetParamOff = "<ACTION>";
                string pluginType = _pluginType;
                if (pluginType.Contains(".")) {
                    pluginType = pluginType.Substring(pluginType.IndexOf('.') + 1);
                }

                if (_buttonType == "Load Plugin(s)") buttonLabel = "Load Plugins";
                else if (_buttonType == "Plugin Action") buttonLabel = pluginType + ": " + targetParamOn;
                else if (_buttonType == "Plugin Action Toggle" && buttonState == 1) buttonLabel = pluginType + ": " + targetParamOn;
                else if (_buttonType == "Plugin Action Toggle" && buttonState == 2) buttonLabel = pluginType + ": " + targetParamOff;
                else if (_buttonType == "Plugin Bool Toggle" && buttonState == 1) buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to TRUE";
                else if (_buttonType == "Plugin Bool Toggle" && buttonState == 2) buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to FALSE";
                else if (_buttonType == "Set Plugin Bool True") buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to TRUE";
                else if (_buttonType == "Set Plugin Bool False") buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to FALSE";
                else if (_buttonType == "Unload All Plugins") buttonLabel = "Unload all Plugins";
            }

            string preposition = " ";
            if (buttonLabel.StartsWith("Load") || buttonLabel.StartsWith("Merge") || buttonLabel.StartsWith("Unmerge")) preposition = " to ";
            if (buttonLabel.StartsWith("Remove") || buttonLabel.StartsWith("Undress") || buttonLabel.StartsWith("Set") || buttonLabel.StartsWith("Unset")) preposition = " from ";
            if (buttonLabel.StartsWith("Reset") || buttonLabel.StartsWith("Set") || buttonLabel.EndsWith("Scale")) preposition = " on ";
            if (_buttonCategory == "Plugins") preposition = " for ";

            if (_buttonType == "Suppress Scale Load")  buttonLabel = buttonLabel;
            else if (((_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom") && targetAtomName == "") || _targetName == "<None available>") buttonLabel = buttonLabel + preposition + "<TARGET>";
            else if (_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom") buttonLabel = buttonLabel + preposition + "'" + targetAtomName + "'";
            else buttonLabel = buttonLabel + preposition + "'" + _targetName + "'";

            return (buttonLabel);
        }
        public bool IsAtomTargetable()
        {
            return (_buttonType != "Suppress Scale Load" &&
                _buttonCategory != "Miscellaneous" &&
                _buttonCategory != "UI Assist"
  /*              &&
                _buttonType !="Spawn Atom" &&
                _buttonType != "Teleport Atom"*/);
        }

        public void ToggleButtonState(string atomName) {
            if (_maxButtonStates == 2) {
                if ((_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom")&& atomName!=null) {
                    int currentState = 0;
                    if (_atomButtonStates.TryGetValue(atomName, out currentState)) {
                        if (currentState == 1) { _atomButtonStates[atomName] = 2; }
                        else { _atomButtonStates[atomName] = 1; }
                    }
                    else {
                        _atomButtonStates.Add(atomName, 2);
                    }
                }
                else {
                    if (_buttonState == 1) { _buttonState = 2; }
                    else { _buttonState = 1; }
                }
                _mvrScript.RefreshWristUIButtonGrid();
            }
            else _buttonState = 1;
        }
        public int GetButtonState(string atomName) {
            int buttonState = 1;
            if (_buttonType=="Suppress Scale Load")
            {
                if (_mvrScript._suppressScaleLoad) buttonState = 2;
                else buttonState = 1;
                
            }
            else if (_buttonType == "Freeze Motion/Sound")
            {
                if (SuperController.singleton.freezeAnimation) _buttonState = 2;
                else _buttonState = 1;
                buttonState = _buttonState;
            }
            else if (_buttonType == "Toggle Play/Edit Mode")
            {
                if (SuperController.singleton.gameMode== SuperController.GameMode.Edit) _buttonState = 2;
                else _buttonState = 1;
                return(_buttonState);
            }
            else if (_buttonType == "Detach Root Control")
            {
                if (SuperController.singleton.GetAtomUIDs().Contains(atomName))
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                    JSONStorable storable = atom.GetStorableByID("control");
                    if (storable != null)
                    {
                        JSONStorableBool detach = storable.GetBoolJSONParam("detachControl");
                        if (detach != null)
                        {
                            if (detach.val) _buttonState = 2;
                            else _buttonState = 1;
                        }
                    }
                }
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Hide Atom")
            {
                if (SuperController.singleton.GetAtomUIDs().Contains(atomName))
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);

                    if (atom.hidden) _buttonState = 2;
                    else _buttonState = 1;
                    
                }
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Move Atom (VR Only)")
            {
                if (_mvrScript._atomVRMoveActive && _mvrScript._atomVRMoveUIAButton==this) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Toggle Atom On/Off")
            {
                if (SuperController.singleton.GetAtomUIDs().Contains(atomName))
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);

                    if (atom.on) _buttonState = 2;
                    else _buttonState = 1;

                }
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Toggle Atom Collision")
            {
                if (SuperController.singleton.GetAtomUIDs().Contains(atomName))
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);

                    if (atom.collisionEnabledJSON.val) _buttonState = 2;
                    else _buttonState = 1;                    
                }
                else _buttonState = 1;
                return (_buttonState);

            }
            else if (_buttonType == "Toggle Soft Body Physics")
            {
                if (UserPreferences.singleton.softPhysics) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Toggle Desktop VSync")
            {
                if (UserPreferences.singleton.desktopVsync) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Toggle Performance Monitor")
            {
                PerfMon pm = UserPreferences.singleton.transform.Find("PerfMon").GetComponent<PerfMon>();
                if (pm.onToggle.isOn) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Toggle Show Hidden Atoms")
            {
                if (SuperController.singleton.showHiddenAtoms) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }

            else if (_buttonType == "Gaze Assisted Select")
            {
                if (_mvrScript._toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Hide Inactive Targets")
            {
                if (UserPreferences.singleton.hideInactiveTargets) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Active Clothing Editor")
            {
                if (_mvrScript._uiaActiveClothingMode) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if (_buttonType == "Raise People by Heel Adjust Height")
            {
                if (_mvrScript._toolbox._heelAdjustTool._heelAdjustRaisePeopleJSON.val) _buttonState = 2;
                else _buttonState = 1;
                return (_buttonState);
            }
            else if ((_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom") && atomName != "") {
                int currentState = 1;
                if (_atomButtonStates.TryGetValue(atomName, out currentState)) {
                    buttonState = currentState;
                }
                else {
                    _atomButtonStates.Add(atomName, 1);
                    buttonState = 1;
                }
            }
            else {
                buttonState = _buttonState;
            }

            if (_buttonType == "Merge Clothing Preset" && _preset1Type.EndsWith("Folder") && _selectedPreset1File == "")
            {
                buttonState = 1;
            }
            else if (_buttonType == "Merge Clothing Preset" || _buttonType == "Decal Maker Toggle") {
                    
                List<string> atomNames = new List<string>();
                if (_targetCategory == "Atom Group")
                {
                    foreach (Atom atomFromGroup in SuperController.singleton.GetAtoms())
                    {
                        bool isMale = false;
                        if (atomFromGroup.type == "Person" && _mvrScript.IsMale(atomFromGroup)) isMale = true;

                        if (atomFromGroup.name != "CoreControl" && atomFromGroup.name != "[CameraRig]" && atomFromGroup.on)
                        {
                            switch (_targetName)
                            {
                                case "All Atoms":
                                    atomNames.Add(atomFromGroup.name);
                                    break;
                                case "All People":
                                    if (atomFromGroup.type == "Person") { atomNames.Add(atomFromGroup.name); }
                                    break;
                                case "All Females":
                                    if (atomFromGroup.type == "Person" && !isMale) { atomNames.Add(atomFromGroup.name); }
                                    break;
                                case "All Males":
                                    if (atomFromGroup.type == "Person" && isMale) { atomNames.Add(atomFromGroup.name); }
                                    break;
                            }
                        }
                    }
                }
                else if (_targetCategory == "Specific Atom" && SuperController.singleton.GetAtomByUid(_targetName)!=null && SuperController.singleton.GetAtomByUid(_targetName).on) atomNames.Add(_targetName);
                else if ((_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom" ) && SuperController.singleton.GetAtomByUid(atomName) != null && SuperController.singleton.GetAtomByUid(atomName).on) atomNames.Add(atomName);

                if (_buttonType == "Decal Maker Toggle")
                {
                    bool allDecalsLoaded = true;
                    foreach (string atomNameForDecalToggleAction in atomNames)
                    {
                        if (!allDecalsLoaded) break;
                        else
                        {
                            Atom atom = SuperController.singleton.GetAtomByUid(atomNameForDecalToggleAction);

                            string pluginKey = GetDecalMakerPluginKey(atom);

                            if (pluginKey != "")
                            {
                                foreach (string storableName in atom.GetStorableIDs())
                                {
                                    if (storableName.StartsWith(pluginKey))
                                    {
                                        JSONStorable dmStorable = atom.GetStorableByID(storableName);
                                        JSONClass decalMakerJSON;
                                        try
                                        {
                                            // Causes an exception if DecalMaker is loading - so we just catch and ignore.
                                            decalMakerJSON = dmStorable.GetJSON(true, true);
                                        }
                                        catch (Exception e) { decalMakerJSON = null;  }
                                        if (dmStorable != null && decalMakerJSON != null)
                                        {
                                            List<string> loadedFaceDecalTextures = GetLoadedDecalTextures((JSONArray)decalMakerJSON["Decal"]["face"]);
                                            List<string> loadedTorsoDecalTextures = GetLoadedDecalTextures((JSONArray)decalMakerJSON["Decal"]["torso"]);
                                            List<string> loadedGenDecalTextures = GetLoadedDecalTextures((JSONArray)decalMakerJSON["Decal"]["genitals"]);
                                            List<string> loadedLimbsDecalTextures = GetLoadedDecalTextures((JSONArray)decalMakerJSON["Decal"]["limbs"]);

                                            foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionFace._decalMakerTexturesList)
                                            {
                                                if (dmt._decalTexturePath != "" && !loadedFaceDecalTextures.Contains(dmt._decalTexturePath))
                                                {
                                                    allDecalsLoaded = false;
                                                    break;
                                                }
                                            }
                                            if (allDecalsLoaded)
                                            {
                                                foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionTorso._decalMakerTexturesList)
                                                {
                                                    if (dmt._decalTexturePath != "" && !loadedTorsoDecalTextures.Contains(dmt._decalTexturePath))
                                                    {
                                                        allDecalsLoaded = false;
                                                        break;
                                                    }
                                                }
                                            }

                                            if (allDecalsLoaded)
                                            {
                                                foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionGen._decalMakerTexturesList)
                                                {
                                                    if (dmt._decalTexturePath != "" && !loadedGenDecalTextures.Contains(dmt._decalTexturePath))
                                                    {
                                                        allDecalsLoaded = false;
                                                        break;
                                                    }
                                                }
                                            }

                                            if (allDecalsLoaded)
                                            {
                                                foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionLimbs._decalMakerTexturesList)
                                                {
                                                    if (dmt._decalTexturePath != "" && !loadedLimbsDecalTextures.Contains(dmt._decalTexturePath))
                                                    {
                                                        allDecalsLoaded = false;
                                                        break;
                                                    }
                                                }
                                            }

                                        }
                                        else allDecalsLoaded = false;
                                        break;
                                    }
                                }
                            }
                            else allDecalsLoaded = false;
                        }


                        
                    }
                    if (allDecalsLoaded && atomNames.Count > 0)
                    {
                        if ((_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom") && atomName != null)
                        {
                            if (_atomButtonStates.ContainsKey(atomName)) _atomButtonStates[atomName] = 2;
                            else _atomButtonStates.Add(atomName, 2);
                        }
                        else _buttonState = 2;

                        buttonState = 2;
                    }
                    else buttonState = 1;
                }
                else if (_buttonType == "Merge Clothing Preset")
                {
                    bool allAtomsClothingItemInExpectedState = false;
                    foreach (string atomNameForMergeAction in atomNames)
                    {
                        Atom atom = SuperController.singleton.GetAtomByUid(atomNameForMergeAction);
                        if (!_atomClothingPresets.ContainsKey(atomNameForMergeAction))
                        {
                            if (_preset1Type.EndsWith("Folder")) return (1);
                            _atomClothingPresets.Add(atomNameForMergeAction, _preset1FilePath);
                        }

                        string presetFileName = _atomClothingPresets[atomNameForMergeAction];
                        List<string> mergeClothingGeometryIDs = null;


                        if (!_mvrScript._presetMergeClothingGeometryIDs.ContainsKey(presetFileName))
                        {
                            mergeClothingGeometryIDs = new List<string>();
                            PatreonFeatures.PopulateMergeClothingGeometryIDs(_mvrScript, mergeClothingGeometryIDs, presetFileName, atom);

                            if (_mvrScript._presetMergeClothingGeometryIDs.ContainsKey(presetFileName)) _mvrScript._presetMergeClothingGeometryIDs.Remove(presetFileName);
                            _mvrScript._presetMergeClothingGeometryIDs.Add(presetFileName, mergeClothingGeometryIDs);
                        }
                        else mergeClothingGeometryIDs = _mvrScript._presetMergeClothingGeometryIDs[presetFileName];
                        JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");

                        bool clothingItemInExpectedState = false;
                        foreach (string clothingItem in mergeClothingGeometryIDs)
                        {
                            JSONStorableBool clothingItemStateJSON = atomGeometryStorable.GetBoolJSONParam(clothingItem);
                            if (clothingItemStateJSON != null)
                            {
                                bool clothingItemCurrentState = clothingItemStateJSON.val;
                                if (buttonState == 2 && clothingItemCurrentState) clothingItemInExpectedState = true;
                                if (buttonState == 1 && !clothingItemCurrentState) clothingItemInExpectedState = true;
                                if (clothingItemInExpectedState) break;
                            }
                            else { clothingItemInExpectedState = true; }
                        }
                        if (clothingItemInExpectedState) { allAtomsClothingItemInExpectedState = true; break; }
                    }
                    if (!allAtomsClothingItemInExpectedState && atomNames.Count > 0)
                    {
                        if ((_targetCategory == "Gaze Selected Atom"|| _targetCategory == "VAM Selected Atom") && atomName != null)
                        {
                            if (_atomButtonStates.ContainsKey(atomName))
                            {
                                if (_atomButtonStates[atomName] == 1) { _atomButtonStates[atomName] = 2; }
                                else { _atomButtonStates[atomName] = 1; }
                            }
                            else
                            {
                                _atomButtonStates.Add(atomName, 2);
                            }
                            buttonState = _atomButtonStates[atomName];
                        }
                        else
                        {
                            if (_buttonState == 1) { _buttonState = 2; }
                            else { _buttonState = 1; }
                            buttonState = _buttonState;
                        }

                    }
                }


            }
            return (buttonState);
        }
        private string GetLastViewedOrSelectedAtomName(string targetName)
        {
            string atomName = "";
            switch (targetName)
            {
                case "Last viewed Person":
                    atomName = _mvrScript._lastViewedPerson;
                    break;
                case "Last viewed Female":
                    atomName = _mvrScript._lastViewedFemale;
                    break;
                case "Last viewed Male":
                    atomName = _mvrScript._lastViewedMale;
                    break;
                case "Last viewed Atom":
                    atomName = _mvrScript._lastViewedAtom;
                    break;
                case "Last selected Person":
                    atomName = _mvrScript._lastSelectedPerson;
                    break;
                case "Last selected Female":
                    atomName = _mvrScript._lastSelectedFemale;
                    break;
                case "Last selected Male":
                    atomName = _mvrScript._lastSelectedMale;
                    break;
                case "Last selected Atom":
                    atomName = _mvrScript._lastSelectedAtom;
                    break;
            }
            return (atomName);
        }
        public string GetLastViewedOrSelectedAtomName()
        {
            string atomName = "";
            if (_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom")
            {
                GetLastViewedOrSelectedAtomName(_targetName);
            }

            return (atomName);
        }
        public string GetDefaultPresetFolder(int presetNum = 0)
        {
            string presetFolder = "";
            switch (_buttonType)
            {
                case "Load Legacy Preset":
                case "Save Legacy Preset":
                    presetFolder = "Saves\\Person\\full";
                    break;
                case "Load Legacy Appearance":
                case "Save Legacy Appearance":
                    presetFolder = "Saves\\Person\\appearance";
                    break;
                case "Load Legacy Pose":
                case "Save Legacy Pose":
                    presetFolder = "Saves\\Person\\pose";
                    break;
                case "Load Plugins Preset":
                    presetFolder = "Custom\\Atom\\Person\\Plugins";
                    break;
                case "Load App Preset":
                    presetFolder = "Custom\\Atom\\Person\\Appearance";
                    break;
                case "Load Pose Preset":
                    presetFolder = "Custom\\Atom\\Person\\Pose";
                    break;
                case "Load Hair Preset":
                    presetFolder = "Custom\\Atom\\Person\\Hair";
                    break;
                case "Load Glute Preset":
                    presetFolder = "Custom\\Atom\\Person\\GlutePhysics";
                    break;
                case "Load Breast Preset":
                    presetFolder = "Custom\\Atom\\Person\\BreastPhysics";
                    break;
                case "Load General Preset":
                    presetFolder = "Custom\\Atom\\Person\\General";
                    break;
                case "Load Skin Preset":
                    presetFolder = "Custom\\Atom\\Person\\Skin";
                    break;
                case "Load Morph Preset":
                    presetFolder = "Custom\\Atom\\Person\\Morphs";
                    break;
                case "Load Scene":
                case "Merge Load Scene":
                    presetFolder = "Saves\\scene";
                    break;
                case "Load UI Assist Profile":
                    presetFolder = "Custom\\Scripts";
                    break;
                case "Spawn Atom":
                    if (_atomType != "Person") presetFolder = "Custom\\Atom\\" + _atomType;
                    else if (presetNum == 1) presetFolder = "Custom\\Atom\\Person\\Appearance";
                    else if (presetNum == 2) presetFolder = "Custom\\Atom\\Person\\Clothing";
                    else if (presetNum == 3) presetFolder = "Custom\\Atom\\Person\\Pose";
                    break;
                case "Load Clothing Preset":
                case "Merge Clothing Preset":
                case "Remove Clothing Preset":
                case "Undress Cloth Preset":
                case "Reset Sim Cloth Preset":
                    presetFolder = "Custom\\Atom\\Person\\Clothing";
                    break;
            }
            return (presetFolder);
        }
        public List<string> GetButtonCategoryChoices() {
            List<string> buttonCategoryChoices = new List<string>();
            buttonCategoryChoices.Add("Atom Control");
            buttonCategoryChoices.Add("Appearance");
            buttonCategoryChoices.Add("Clothing");
            buttonCategoryChoices.Add("Presets");
            buttonCategoryChoices.Add("Plugins");
            buttonCategoryChoices.Add("UI Assist");
            buttonCategoryChoices.Add("Miscellaneous");
            return (buttonCategoryChoices);
        }
        public List<string> GetButtonTypeChoices() {
            List<string> buttonTypeChoices = new List<string>();
            switch (_buttonCategory) {
                case "Appearance":
                    buttonTypeChoices.Add("Load App Preset");
                    buttonTypeChoices.Add("Reset Scale");
                    buttonTypeChoices.Add("Reset Appearance");
                    buttonTypeChoices.Add("Decal Maker Toggle");
                    buttonTypeChoices.Add("Set Hair Color");

                    if (Constants._patreonContentEnabled)buttonTypeChoices.Add("Suppress Scale Load");
                    break;
                case "Clothing":
                    buttonTypeChoices.Add("Remove All Clothing");
                    buttonTypeChoices.Add("Undress All Clothing");
                    buttonTypeChoices.Add("Reset Sim All Clothing");
                    buttonTypeChoices.Add("Load Clothing Preset");
                    if (Constants._patreonContentEnabled) {
                        
                        buttonTypeChoices.Add("Remove Clothing Preset");
                        buttonTypeChoices.Add("Merge Clothing Preset");
                        buttonTypeChoices.Add("Undress Cloth Preset");
                        buttonTypeChoices.Add("Reset Sim Cloth Preset");
                        buttonTypeChoices.Add("Active Clothing Editor");
                    }
                    break;
                case "Presets":
                    buttonTypeChoices.Add("Load App Preset");
                    buttonTypeChoices.Add("Load Clothing Preset");
                    buttonTypeChoices.Add("Load Pose Preset");
                    buttonTypeChoices.Add("Load Hair Preset");
                    buttonTypeChoices.Add("Load Plugins Preset");
                    buttonTypeChoices.Add("Load Morph Preset");
                    buttonTypeChoices.Add("Load Skin Preset");
                    buttonTypeChoices.Add("Load Glute Preset");
                    buttonTypeChoices.Add("Load Breast Preset");
                    buttonTypeChoices.Add("Load General Preset");
                    buttonTypeChoices.Add("Load Legacy Preset");
                    buttonTypeChoices.Add("Save Legacy Preset");
                    buttonTypeChoices.Add("Load Legacy Appearance");
                    buttonTypeChoices.Add("Save Legacy Appearance");
                    buttonTypeChoices.Add("Load Legacy Pose");
                    buttonTypeChoices.Add("Save Legacy Pose");
                    break;
                case "Plugins":
                    buttonTypeChoices.Add("Load Plugins Preset");
                    buttonTypeChoices.Add("Plugin Action");
                    buttonTypeChoices.Add("Plugin Action Toggle");
                    buttonTypeChoices.Add("Plugin Bool Toggle");
                    buttonTypeChoices.Add("Set Plugin Bool True");
                    buttonTypeChoices.Add("Set Plugin Bool False");
                    buttonTypeChoices.Add("Unload All Plugins");
                    buttonTypeChoices.Add("Load Plugin(s)");
                    break;
                case "Miscellaneous":                   
                    buttonTypeChoices.Add("Freeze Motion/Sound");
                    buttonTypeChoices.Add("Toggle Play/Edit Mode");
                    buttonTypeChoices.Add("Load Scene");
                    buttonTypeChoices.Add("Merge Load Scene");
                    buttonTypeChoices.Add("Toggle Soft Body Physics");
                    buttonTypeChoices.Add("Toggle Desktop VSync");
                    buttonTypeChoices.Add("Toggle Performance Monitor");
                    buttonTypeChoices.Add("Toggle Show Hidden Atoms");
                    buttonTypeChoices.Add("Blank");
                    break;
                case "Atom Control":
                    buttonTypeChoices.Add("Spawn Atom");
//                    buttonTypeChoices.Add("Toggle Atom On/Off");
                    buttonTypeChoices.Add("Hide Atom");
                    buttonTypeChoices.Add("Delete Atom");
                    buttonTypeChoices.Add("Select Atom");
                    buttonTypeChoices.Add("Move Atom (VR Only)");
                    buttonTypeChoices.Add("Teleport Atom");

                    buttonTypeChoices.Add("Toggle Atom Collision");
                    buttonTypeChoices.Add("Detach Root Control");
                    break;
                case "UI Assist":
                    buttonTypeChoices.Add("Load UI Assist Profile");
                    buttonTypeChoices.Add("Gaze Assisted Select");
                    buttonTypeChoices.Add("Hide Inactive Targets");
                    buttonTypeChoices.Add("Raise People by Heel Adjust Height");
                    break;
            }
            return (buttonTypeChoices);
        }
        public List<string> GetAtomCategoryChoices()
        {
            List<string> choices = new List<string>();
            choices.Add("People");
            choices.Add("Misc");
            choices.Add("Animation");
            choices.Add("Force");
            choices.Add("Light");
            choices.Add("Toys");
            choices.Add("Sound");
            choices.Add("Triggers");
            return (choices);
        }
        public List<string> GetPresetTypeChoices()
        {
            List<string> choices = new List<string>();
            if (_buttonType=="Spawn Atom")choices.Add("None");
            choices.Add("Single File");
            if (!isUIAPLoadType && !isLegacyPresetSaveType)  choices.Add("Random from Folder");
            choices.Add("Choose from Folder");
            return (choices);
        }
        public List<string> GetAtomTypeChoices()
        {
            List<string> choices = new List<string>();
            switch (_atomCategory)
            {
                case "People":
                    choices.Add("Person");
                    break;
                case "Misc":
                    choices.Add("CustomUnityAsset");
                    choices.Add("ClothGrabSphere");
                    choices.Add("Empty");
                    choices.Add("ImagePanel");
                    choices.Add("ImagePanelEmissive");
                    choices.Add("ImagePanelTransparent");
                    choices.Add("ImagePanelTransparentEmissive");
                    choices.Add("SimpleSign");
                    choices.Add("UIText");
                    choices.Add("WebBrowser");
                    choices.Add("WebPanel");
                    choices.Add("WebPanelEmissive");
                    break;
                case "Animation":
                    choices.Add("AnimationPattern");
                    break;
                case "Force":
                    choices.Add("CycleForce");
                    choices.Add("GrabPoint");
                    choices.Add("RhythmForce");
                    choices.Add("SyncForce");
                    break;
                case "Light":
                    choices.Add("InvisibleLight");
                    choices.Add("Torch");
                    break;
                case "Toys":
                    choices.Add("Dildo");
                    choices.Add("ToyAH");
                    choices.Add("ToyBP");
                    choices.Add("Paddle");
                    break;
                case "Sound":
                    choices.Add("AptSpeaker");
                    choices.Add("AudioSource");
                    choices.Add("RhythmAudioSource");
                    break;
                case "Triggers":
                    choices.Add("Button");
                    choices.Add("CollisionTrigger");
                    choices.Add("LookAtTrigger");
                    choices.Add("UIButton");
                    choices.Add("UISlider");
                    choices.Add("UIToggle");
                    choices.Add("VariableTrigger");
                    break;
            }
            return (choices);
        }
        public List<string> GetTargetCategoryChoices() {
            List<string> atomCategoryChoices = new List<string>();
            atomCategoryChoices.Add("Gaze Selected Atom");
            atomCategoryChoices.Add("VAM Selected Atom");
            if (_buttonCategory!="Atom Control" ) atomCategoryChoices.Add("Atom Group");
            atomCategoryChoices.Add("Specific Atom");
            return (atomCategoryChoices);
        }
        public List<string> GetTargetNodeChoices()
        {
            List<string> nodeChoices = new List<string>();
            nodeChoices.Add("control");
            nodeChoices.Add("headControl");
            nodeChoices.Add("head");
            nodeChoices.Add("Lips");
            nodeChoices.Add("Mouth");
            nodeChoices.Add("neckControl");
            nodeChoices.Add("neck");
            nodeChoices.Add("lShoulderControl");
            nodeChoices.Add("lCollar");
            nodeChoices.Add("lPectoral");
            nodeChoices.Add("lArmControl");
            nodeChoices.Add("lShldr");
            nodeChoices.Add("lElbowControl");
            nodeChoices.Add("lForeArm");
            nodeChoices.Add("lHandControl");
            nodeChoices.Add("lHand");
            nodeChoices.Add("rShoulderControl");
            nodeChoices.Add("rCollar");
            nodeChoices.Add("rPectoral");
            nodeChoices.Add("rArmControl");
            nodeChoices.Add("rShldr");
            nodeChoices.Add("rElbowControl");
            nodeChoices.Add("rForeArm");
            nodeChoices.Add("rHandControl");
            nodeChoices.Add("rHand");

            nodeChoices.Add("chestControl");
            nodeChoices.Add("chest");
            nodeChoices.Add("lNippleControl");
            nodeChoices.Add("lNipple");
            nodeChoices.Add("rNippleControl");
            nodeChoices.Add("rNipple");

            nodeChoices.Add("abdomen2Control");
            nodeChoices.Add("abdomen2");
            nodeChoices.Add("pelvisControl");
            nodeChoices.Add("pelvis");
            nodeChoices.Add("hipControl");
            nodeChoices.Add("hip");
            nodeChoices.Add("abdomenControl");
            nodeChoices.Add("abdomen");
            nodeChoices.Add("LGlute");
            nodeChoices.Add("RGlute");

            nodeChoices.Add("testesControl");
            nodeChoices.Add("penisBaseControl");
            nodeChoices.Add("penisMidControl");
            nodeChoices.Add("penisTipControl");

            nodeChoices.Add("Labia");
            nodeChoices.Add("Vagina");

            nodeChoices.Add("lThighControl");
            nodeChoices.Add("lThigh");
            nodeChoices.Add("lKneeControl");
            nodeChoices.Add("lShin");
            nodeChoices.Add("lFootControl");
            nodeChoices.Add("lFoot");
            nodeChoices.Add("lToeControl");

            nodeChoices.Add("rThighControl");
            nodeChoices.Add("rThigh");
            nodeChoices.Add("rKneeControl");
            nodeChoices.Add("rShin");
            nodeChoices.Add("rFootControl");
            nodeChoices.Add("rFoot");
            nodeChoices.Add("rToeControl");

            return nodeChoices;
        }
        public List<string> GetMoveRelativeToChoices()
        {
            List<string> atomChoices = new List<string>();
            if (_buttonType=="Spawn Atom")
            {
                atomChoices.Add("Last viewed Person");
                atomChoices.Add("Last viewed Female");
                atomChoices.Add("Last viewed Male");
            }

            atomChoices.Add("Scene Origin");
            atomChoices.Add("VR Head");
            atomChoices.Add("VR Left Hand");
            atomChoices.Add("VR Right Hand");
            return (atomChoices);
        }
        public List<string> GetTargetChoices() {
            List<string> atomChoices = new List<string>();
  
            switch (_targetCategory)
            {
                case "Gaze Selected Atom":
                    atomChoices.Add("Last viewed Person"); atomChoices.Add("Last viewed Female"); atomChoices.Add("Last viewed Male");
                    if (_buttonCategory == "Plugins" || _buttonType == "Load General Preset" || _buttonCategory == "Atom Control") { atomChoices.Add("Last viewed Atom"); }
                    break;
                case "VAM Selected Atom":
                    atomChoices.Add("Last selected Person"); atomChoices.Add("Last selected Female"); atomChoices.Add("Last selected Male");
                    if (_buttonCategory == "Plugins" || _buttonType == "Load General Preset" || _buttonCategory == "Atom Control") { atomChoices.Add("Last selected Atom"); }
                    break;
                case "Atom Group":
                    atomChoices.Add("All Females"); atomChoices.Add("All Males"); atomChoices.Add("All People");
                    if (_buttonCategory == "Plugins" || _buttonType == "Load General Preset") { atomChoices.Add("All Atoms"); }
                    break;
                case "Specific Atom":
                    foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
                    {
                        Atom atom = SuperController.singleton.GetAtomByUid(atomUID);
                        if (atomUID != "[CameraRig]" && (atomUID != "CoreControl" || _buttonType == "Load Plugin(s)") && (atom.type == "Person" || !isPersonAtomOnlyType) ) { atomChoices.Add(atomUID); }
//                         if (atomUID != "[CameraRig]" && (atomUID != "CoreControl" || _buttonType == "Load Plugin(s)") && ((_buttonCategory != "Clothing" && _buttonType.Substring(_buttonType.Length - 6, 6) != "Preset") || atom.type == "Person" || _buttonType == "Load General Preset")) { atomChoices.Add(atomUID); }
                    }
                    break;
            }
            if (atomChoices.Count == 0) { atomChoices.Add("<None available>"); }
            return (atomChoices);
        }
        public List<string> GetPluginTypeChoices() {
            List<string> popupChoices = new List<string>();

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                foreach (string storableName in atom.GetStorableIDs()) {
                    if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) {
                        string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                        if (!popupChoices.Contains(pluginType)) { popupChoices.Add(pluginType); }
                    }
                }
            }

            if (popupChoices.Count == 0) { popupChoices.Add("No plugins in Scene"); }
            popupChoices.Sort();
            return (popupChoices);
        }
        public List<string> GetPositionGranularityChoices()
        {
            List<string> choices = new List<string>();
            choices.Add("Extra Fine");
            choices.Add("Fine");
            choices.Add("Medium");
            choices.Add("Coarse");
            choices.Add("Extra Coarse");
            return (choices);
        }
        public List<string> GetRotationGranularityChoices()
        {
            List<string> choices = new List<string>();
            choices.Add("Fine");
            choices.Add("Medium");
            choices.Add("Coarse");
            return (choices);
        }
        public string GetVarTypeFromPlugin(string varName)
        {
            string varType = "NONE";
            bool foundVar = false;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {               
                    if ((storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);
                        if (pluginStorable.GetBoolParamNames().Contains(varName)) varType = "BOOL";
                        else if (pluginStorable.GetFloatParamNames().Contains(varName)) varType = "FLOAT";
                        else if (pluginStorable.GetStringParamNames().Contains(varName)) varType = "STRING";
                        else if (pluginStorable.GetStringChooserParamNames().Contains(varName)) varType = "STRINGCHOOSER";
                        else if (pluginStorable.GetUrlParamNames().Contains(varName)) varType = "URL";
                        else if (pluginStorable.GetColorParamNames().Contains(varName)) varType = "COLOR";
                        if (varType != "NONE")
                        {
                            foundVar = true;
                            break;
                        }
                    }
                }
                if (foundVar) break;
            }
            return (varType);
        }
        public float GetMinFloatValue(string varName)
        {
            float minFloat = 100000000000f;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if ((storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);

                        if (pluginStorable.GetFloatParamNames().Contains(varName))
                        {
                            JSONStorableFloat floatJSON = pluginStorable.GetFloatJSONParam(varName);
                            if (floatJSON.min < minFloat) minFloat = floatJSON.min;
                            
                        }

                    }
                }
            }
            if (minFloat == 100000000000f) minFloat = 0f;
            return (minFloat);
        }
        public float GetMaxFloatValue(string varName)
        {
            float maxFloat = -100000000000f;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if ((storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);

                        if (pluginStorable.GetFloatParamNames().Contains(varName))
                        {
                            JSONStorableFloat floatJSON = pluginStorable.GetFloatJSONParam(varName);
                            if (floatJSON.max > maxFloat) maxFloat = floatJSON.max;

                        }
                    }
                }
            }
            if (maxFloat == -100000000000f) maxFloat = 1f;
            return (maxFloat);
        }
        public List<string> GetStringChooserChoices(string varName)
        {
            List<string> choices = new List<string>();

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if ((storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);
                        if (pluginStorable.GetStringChooserParamNames().Contains(varName))
                        {
                            List<string> pluginSCChoices = pluginStorable.GetStringChooserJSONParamChoices(varName);
                            if (pluginSCChoices != null)
                            {
                                foreach (string choice in pluginSCChoices)
                                {
                                    if (!choices.Contains(choice)) choices.Add(choice);
                                }
                            }
                        }
                    }
                }
            }
            return (choices);
        }

        public void UpdateButtonCategory(string category) {
            _buttonCategory = category;
            if (!GetButtonTypeChoices().Contains(_buttonType)) { UpdateButtonType(GetButtonTypeChoices()[0]); }
            if (!GetTargetChoices().Contains(_targetName)) { UpdateTargetName(GetTargetChoices()[0]); }
            _preset1FilePath = "";
            _preset2FilePath = "";
            _preset3FilePath = "";
            _selectedPreset1File = "";
            _selectedPreset2File = "";
            _selectedPreset3File = "";
            _atomButtonStates.Clear();
            List<string> presetTypeChoices = GetPresetTypeChoices();
            if (!presetTypeChoices.Contains(_preset1Type)) _preset1Type = presetTypeChoices[0];
            if (!presetTypeChoices.Contains(_preset2Type)) _preset2Type = presetTypeChoices[0];
            if (!presetTypeChoices.Contains(_preset3Type)) _preset3Type = presetTypeChoices[0];
        }
        public void UpdateButtonType(string type) {
            _buttonType = type;

            switch (_buttonType) {
                case "Plugin Action Toggle":
                case "Undress All Clothing":
                case "Merge Clothing Preset":
                case "Undress Cloth Preset":
                case "Plugin Bool Toggle":
                case "Suppress Scale Load":
                case "Freeze Motion/Sound":
                case "Toggle Play/Edit Mode":
                case "Gaze Assisted Select":
                case "Hide Inactive Targets":
                case "Active Clothing Editor":
                case "Raise People by Heel Adjust Height":
                case "Decal Maker Toggle":
                case "Toggle Soft Body Physics":
                case "Toggle Desktop VSync":
                case "Toggle Performance Monitor":
                case "Toggle Show Hidden Atoms":
                case "Hide Atom":
                case "Detach Root Control":
                case "Toggle Atom On/Off":
                case "Move Atom (VR Only)":
                case "Toggle Atom Collision":
                    _maxButtonStates = 2;
                    if (_buttonSkin != null) _buttonSkin._maxButtonStates = 2;
                    break;
                default:
                    _maxButtonStates = 1;
                    _buttonState = 1;
                    if (_buttonSkin != null) _buttonSkin._maxButtonStates = 1;
                    break;
            }
            if (_buttonCategory != "Clothing") {
                _preset1FilePath = "";
                _selectedPreset1File = "";
                _atomButtonStates.Clear();
                _buttonTexture = null;
                
            }
            if (_buttonType == "Active Clothing Editor")
            {
                _targetCategory = "Gaze Selected Atom";
            }
            
            _preset2FilePath = "";
            _preset3FilePath = "";
            _selectedPreset2File = "";
            _selectedPreset3File = "";

            List<string> presetTypeChoices = GetPresetTypeChoices();
            if (!presetTypeChoices.Contains(_preset1Type)) _preset1Type = presetTypeChoices[0];
            if (!presetTypeChoices.Contains(_preset2Type)) _preset2Type = presetTypeChoices[0];
            if (!presetTypeChoices.Contains(_preset3Type)) _preset3Type = presetTypeChoices[0];

            List<string> targetChoices = new List<string>();
            targetChoices = GetTargetChoices();
            if (!targetChoices.Contains(_targetName)) {
                if (targetChoices.Contains(_mvrScript._defaultButtonSkin._defaultGazeTarget)) _targetName = _mvrScript._defaultButtonSkin._defaultGazeTarget;
                else _targetName = targetChoices[0];
            }

            if (_buttonCategory == "Plugins" && _buttonType != "Load Plugin(s)") {
                List<string> pluginTypeChoices = new List<string>();
                pluginTypeChoices = GetPluginTypeChoices();
                if (!pluginTypeChoices.Contains(_pluginType)) { _pluginType = pluginTypeChoices[0]; }
                List<string> targetParamNameChoices = new List<string>();
                if (_buttonType.Contains("Bool")) { targetParamNameChoices = GetTargetPluginParamNameChoices("BOOL"); }
                else { targetParamNameChoices = GetTargetPluginParamNameChoices("ACTION"); }
                if (!targetParamNameChoices.Contains(_targetParamNameOn)) { _targetParamNameOn = targetParamNameChoices[0]; }
                if (_buttonType.Contains("Toggle")) {
                    if (!targetParamNameChoices.Contains(_targetParamNameOff)) { _targetParamNameOff = targetParamNameChoices[0]; }
                }
                else { _targetParamNameOff = ""; }
            }
            if (_buttonType == "Blank") _targetName = "";
        }
        public void UpdateTargetCategory(string category) {
            _targetCategory = category;
            if (!GetTargetChoices().Contains(_targetName)) { UpdateTargetName(GetTargetChoices()[0]); }
        }
        public void UpdateTargetName(string atom) {
            _targetName = atom;
        }
        public void UpdateTagItemToRemove(string tag)
        {
            if (_tagsToRemoveOnMerge.Contains(tag)) _tagsToRemoveOnMerge.Remove(tag);
            else _tagsToRemoveOnMerge.Add(tag);
        }

        public void SetPreset1File(string fileName) {
            _preset1FilePath = fileName;
            _mvrScript._presetMergeClothingGeometryIDs.Remove(fileName);
            if (fileName !="" && fileName!=string.Empty && fileName.Contains(".")) QueueLoadTexture(fileName.Substring(0, fileName.LastIndexOf('.')) + ".jpg");
        }
        public void SetPreset2File(string fileName)
        {
            _preset2FilePath = fileName;
        }
        public void SetPreset3File(string fileName)
        {
            _preset3FilePath = fileName;
        }
        public List<string> GetTargetPluginParamNameChoices(string paramType) {
            List<string> choices = new List<string>();
            bool pluginOfTypeFound = false;

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                foreach (string storableName in atom.GetStorableIDs()) {
                    if (storableName.EndsWith(_pluginType)) {
                        pluginOfTypeFound = true;
                        JSONStorable storable = atom.GetStorableByID(storableName);
                        if (paramType == "ACTION")
                        {
                            foreach (string actionName in storable.GetActionNames())
                            {
                                if (!UIAssist.VAMSaveRestoreActions.Contains(actionName) && !choices.Contains(actionName)) { choices.Add(actionName); }
                            }
                        }
                        else
                        {
                            foreach (string boolName in storable.GetBoolParamNames())
                            {
                                if (!choices.Contains(boolName)) { choices.Add(boolName); }
                            }
                            if (paramType == "ALL")
                            {
                                foreach (string floatName in storable.GetFloatParamNames())
                                {
                                    if (!choices.Contains(floatName)) { choices.Add(floatName); }
                                }
                                foreach (string stringName in storable.GetStringParamNames())
                                {
                                    if (!choices.Contains(stringName)) { choices.Add(stringName); }
                                }
                                foreach (string stringChooserName in storable.GetStringChooserParamNames())
                                {
                                    if (!choices.Contains(stringChooserName)) { choices.Add(stringChooserName); }
                                }
                                foreach (string urlName in storable.GetUrlParamNames())
                                {
 //                                   if (!choices.Contains(urlName)) { choices.Add(urlName); }
                                }
                                foreach (string colorName in storable.GetColorParamNames())
                                {
                                    if (!choices.Contains(colorName)) { choices.Add(colorName); }
                                }
                            }
                        }
                    }

                }
            }

            if (choices.Count < 1) {
                if (!pluginOfTypeFound) { choices.Add("No Plugin of type loaded"); }
                else {
                    if (paramType == "ACTION") { choices.Add("No Actions available"); }
                    else { choices.Add("No Bools available"); }
                }
            }
            choices.Sort();
            return (choices);
        }

        public JSONArray GetPluginDataByName(string atomName, string name) {
            Atom atom = SuperController.singleton.GetAtomByUid(atomName);
            JSONArray pluginData = new JSONArray();
            if (atom != null) {
                foreach (string storableName in atom.GetStorableIDs()) {

                    if (storableName.StartsWith(name)) {
                        JSONClass storableJSON = atom.GetStorableByID(storableName).GetJSON();
                        pluginData[-1] = storableJSON;

                    }
                }
            }
            return (pluginData);
        }
        public List<string> GetPluginRefsOfType(Atom atom, string type) {
            List<string> pluginRefs = new List<string>();
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            if (manager != null) {
                JSONClass currentPlugins = null;
                currentPlugins = manager.GetJSON(true, true, true);
                if (currentPlugins != null) {
                    foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"]) {
                        if (kvp.Value.ToString().Substring(kvp.Value.ToString().LastIndexOfAny(new char[] { '/', '\\' }) + 1) == type.Substring(type.LastIndexOfAny(new char[] { '/', '\\' }) + 1) + "\"") { pluginRefs.Add(kvp.Key); }
                    }
                }
            }
            return (pluginRefs);
        }
        public void QueueLoadTexture(string url)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url))
                return;
            if (!FileManagerSecure.FileExists(url)) {
                _buttonTexture = null;
                return;
            }

            _thumbnailImageLoading = true;
            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = true;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = false;
            queuedImage.linear = false;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.createNormalFromBump = false;
            queuedImage.bumpStrength = 0f;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
            {
                try
                {
                    _thumbnailImageLoading = false;
                    Texture2D tex = qi.tex;
                    _buttonTexture = tex;

                    if (_mvrScript._uidMainButtonCanvasDisplay._uiMainButtonGridDisplay._buttonGrid !=null && _mvrScript._uidMainButtonCanvasDisplay._uiMainButtonGridDisplay._buttonGrid._buttonList.Contains(this))
                    {
                        _mvrScript._uidMainButtonCanvasDisplay._uiMainButtonGridDisplay.UpdateButtonTextures(this._buttonRef-1);
                    }
                    if (_mvrScript._uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay._buttonGrid != null && _mvrScript._uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay._buttonGrid._buttonList.Contains(this))
                    {
                        _mvrScript._uidMainButtonCanvasDisplay._uiQLBButtonGridDisplay.UpdateButtonTextures(this._buttonRef-1);
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

            };

            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }
        private string GetRandomPresetFromFolder (string folder, string type)
        {
            string[] fileArray= FileManagerSecure.GetFiles(folder, "*."+ type);
            if (fileArray.Count() > 0)
            {
                int index = UnityEngine.Random.Range(0, fileArray.Count()-1);
                return (fileArray[index]);
            }
            return "";
        }
        private void PlayActionPreset1FileSelected (string presetFileName)
        {
            try
            {
                if (presetFileName == "") return;
                else
                {
                    _selectedPreset1File = presetFileName;
                    if (isClothingPresetLoadType)_mvrScript._presetMergeClothingGeometryIDs.Remove(presetFileName);

                    string fileExtension = "vap";
                    string prefixRemoval = "Preset_";
                    if (isLegacyPresetLoadSaveType) fileExtension = "json"; prefixRemoval = null;

                    string folderName = _preset3FilePath;
                    if (folderName == "") folderName = GetDefaultPresetFolder(3);
                    
                    if (_buttonType == "Spawn Atom")
                    {
                        if (_preset3Type == "Random from Folder" && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                        {
                            _selectedPreset3File = GetRandomPresetFromFolder(folderName, fileExtension);
                        }
                        else if (_preset3Type == "Choose from Folder" && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                        {
                            SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                            SuperController.singleton.ShowMainHUDAuto();
                            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(folderName);
                            SuperController.singleton.GetMediaPathDialog(PlayActionPreset3FileSelected, fileExtension, folderName, false, true, false, prefixRemoval, false, shortcuts);
                            return;
                        }
                        else _selectedPreset3File = _preset3FilePath;
                    }
                    PlayAction("NONE");
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void PlayActionPreset3FileSelected(string presetFileName)
        {
            try
            {
                if (presetFileName == "") return;
                else
                {
                    _selectedPreset3File = presetFileName;
                    PlayAction("None");
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void StartAction (string vrHand)
        {
            string targetName = "";
            if (_buttonType == "Spawn Atom") targetName = _atomPositionRelativeTo;
            else targetName = _targetName;

            _selectedTarget = GetLastViewedOrSelectedAtomName(targetName);
           
            if (!isPresetToggleType || GetButtonState(_selectedTarget) == 1)
            {
                _selectedPreset1File = "";
                _selectedPreset2File = "";
                _selectedPreset3File = "";
            }

            if (_buttonType == "Spawn Atom" || isPresetType || isSceneLoadType || isUIAPLoadType)
            {
                string fileType = "vap";
                string prefixRemoval = null;
                if (isNonLegacyPresetLoadType) prefixRemoval = "Preset_";

                if (isSceneLoadType || isLegacyPresetLoadSaveType) fileType = "json";
                else if (isUIAPLoadType) fileType = "uiap";

                string folderName = _preset1FilePath;
                if (folderName == "") folderName = GetDefaultPresetFolder(1);

                if (_preset1Type == "Random from Folder" && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                {
                    if (GetButtonState(_selectedTarget) == 1)
                    {
                        _selectedPreset1File = GetRandomPresetFromFolder(folderName, fileType);
                        _mvrScript._presetMergeClothingGeometryIDs.Remove(_selectedPreset1File);
                    }
                }
                else if (_preset1Type == "Choose from Folder"  && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                {
                    if (GetButtonState(_selectedTarget) == 1)
                    {
                        SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                        SuperController.singleton.ShowMainHUDAuto();
                        List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(folderName);
                        List<ShortCut> filteredShortcuts = new List<ShortCut>();
                        string filterPath = "AddonPackages";
                        if (_preset2FilePath != "") filterPath = SuperController.singleton.NormalizeDirectoryPath(_preset2FilePath);
                        foreach (ShortCut sc in shortcuts)
                        {
                            if (sc.path.StartsWith(filterPath)) filteredShortcuts.Add(sc);
                            else if (!sc.path.StartsWith("AddonPackages")) filteredShortcuts.Add(sc);
                        }

                        SuperController.singleton.GetMediaPathDialog(PlayActionPreset1FileSelected, fileType, folderName, false, true, false, prefixRemoval, false, filteredShortcuts);

                        if (isLegacyPresetSaveType)
                        {
                            // Update the browser to be a Save browser
                            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                            browser.SetTextEntry(true);
                            browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), "json");
                            browser.ActivateFileNameField();
                        }
                        return;
                    }
                }
                else _selectedPreset1File = _preset1FilePath;

                if (_buttonType == "Spawn Atom")
                {
                    folderName = _preset3FilePath;
                    if (folderName == "") folderName = GetDefaultPresetFolder(3);
                    if (_preset3Type == "Random from Folder" && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                    {
                        _selectedPreset3File = GetRandomPresetFromFolder(folderName,"vap");
                    }
                    else if (_preset3Type == "Choose from Folder" && Constants._patreonContentEnabled && _mvrScript._secondValidation)
                    {
                        SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                        SuperController.singleton.ShowMainHUDAuto();
                        List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(folderName);
                        SuperController.singleton.GetMediaPathDialog(PlayActionPreset3FileSelected, "vap", folderName, false, true, false, prefixRemoval, false, shortcuts);
                        return;
                    }
                    else _selectedPreset3File = _preset3FilePath;
                }
            }
            PlayAction(vrHand);
        }
        public Transform GetSpawnRelativeToTransform(string gazeTargetAtomName)
        {
            GameObject emptyGO = new GameObject();
            Transform relativeToTransform = emptyGO.transform;
            relativeToTransform.position = new Vector3(0f, 0f, 0f);
            relativeToTransform.rotation = new Quaternion(0f,0f,0f,1f);

            if (_atomPositionRelativeTo.StartsWith("Last"))
            {
                Atom targetRelativePositionAtom = SuperController.singleton.GetAtomByUid(gazeTargetAtomName);
                if (targetRelativePositionAtom != null)
                {
                    if (_targetNode == "Lips" || _targetNode == "Mouth" || _targetNode == "Labia" || _targetNode == "Vagina")
                    {
                        string targetNode = _targetNode;
                        if (_targetNode == "Lips") targetNode = "LipTrigger";
                        else if (_targetNode == "Mouth") targetNode = "MouthTrigger";
                        else if (_targetNode == "Labia") targetNode = "LabiaTrigger";
                        else if (_targetNode == "Vagina") targetNode = "VaginaTrigger";
                        Rigidbody nodeRB = targetRelativePositionAtom.rigidbodies.First(rb => rb.name == targetNode);
                        if (nodeRB != null) relativeToTransform = nodeRB.transform;
                    }
                    else if (_targetNode.Contains("control") || _targetNode.Contains("Control"))
                    {
                        FreeControllerV3 nodeFCV3 = targetRelativePositionAtom.freeControllers.First(fc => fc.name == _targetNode);
                        if (nodeFCV3 != null) relativeToTransform = nodeFCV3.transform;
                    }
                    else
                    {
                        Rigidbody nodeRB = targetRelativePositionAtom.rigidbodies.First(rb => rb.name == _targetNode);
                        if (nodeRB != null) relativeToTransform = nodeRB.transform;
                    }
                }
            }
            else if (_atomPositionRelativeTo == "VR Head" || (!_mvrScript._vrActive && (_atomPositionRelativeTo == "VR Left Hand" || _atomPositionRelativeTo == "VR Right Hand"))) relativeToTransform = SuperController.singleton.lookCamera.transform;
            else if (_atomPositionRelativeTo == "VR Left Hand") relativeToTransform =SuperController.singleton.leftHand.transform;
            else if (_atomPositionRelativeTo == "VR Right Hand") relativeToTransform = SuperController.singleton.rightHand.transform;
            
            return relativeToTransform;
        }
        public void UpdateSpawnPosition(ref Vector3 position,ref Quaternion rotation, string gazeTargetAtomName)
        {
            Vector3 offsetPosition = new Vector3(_atomPositionX, _atomPositionY, _atomPositionZ);
            Vector3 basePosition = new Vector3(0f, 0f, 0f);
            Quaternion baseRotation = new Quaternion();
            if (_atomAbsoluteHeight) offsetPosition.y = 0f;
            Transform relativeToTransform = GetSpawnRelativeToTransform(gazeTargetAtomName);
            if (relativeToTransform != null)
            {
                basePosition = relativeToTransform.position;
                baseRotation = relativeToTransform.rotation;
                offsetPosition = relativeToTransform.rotation * offsetPosition;
            }

            position.x = basePosition.x + offsetPosition.x;
            if (!_atomAbsoluteHeight) position.y = basePosition.y + offsetPosition.y;
            else position.y = _atomPositionY;
            position.z = basePosition.z + offsetPosition.z;
            float xRotation = baseRotation.eulerAngles.x + _atomRotationX;
            if (_atomAbsolutePitch) xRotation = _atomRotationX;
            float zRotation = baseRotation.eulerAngles.z + _atomRotationZ;
            if (_atomAbsoluteRoll) zRotation = _atomRotationZ;
            rotation.eulerAngles = new Vector3(xRotation, baseRotation.eulerAngles.y + _atomRotationY, zRotation);
        }
        public void TeleportAtomToSpawnPoint(string teleportAtomName, Vector3 position, Quaternion rotation, bool applyPresets = false)
        {
            Atom atom = SuperController.singleton.GetAtomByUid(teleportAtomName);
            if (atom != null)
            {
                JSONClass atomsJSON = SuperController.singleton.GetSaveJSON(atom);
                JSONArray atomsArrayJSON = atomsJSON["atoms"].AsArray;
                JSONClass atomJSON = (JSONClass)atomsArrayJSON[0];

                JSONArray storablesArrayJSON = atomJSON["storables"].AsArray;

                atomJSON["position"]["x"].AsFloat = position.x;
                atomJSON["position"]["y"].AsFloat = position.y;
                atomJSON["position"]["z"].AsFloat = position.z;
                atomJSON["containerPosition"]["x"].AsFloat = position.x;
                atomJSON["containerPosition"]["y"].AsFloat = position.y;
                atomJSON["containerPosition"]["z"].AsFloat = position.z;
                atomJSON["rotation"]["x"].AsFloat = rotation.eulerAngles.x;
                atomJSON["rotation"]["y"].AsFloat = rotation.eulerAngles.y;
                atomJSON["rotation"]["z"].AsFloat = rotation.eulerAngles.z;
                atomJSON["containerRotation"]["x"].AsFloat = rotation.eulerAngles.x;
                atomJSON["containerRotation"]["y"].AsFloat = rotation.eulerAngles.y;
                atomJSON["containerRotation"]["z"].AsFloat = rotation.eulerAngles.z;

                Dictionary<string, int> storablesIndexDict = new Dictionary<string, int>();

                for (int i = 0; i < storablesArrayJSON.Count; i++)
                {
                    JSONClass storableJSON = (JSONClass)storablesArrayJSON[i];
                    storablesIndexDict.Add(storableJSON["id"].Value, i);
                    if (storableJSON["id"].Value == "control")
                    {
                        atomJSON["storables"][i]["position"]["x"].AsFloat = position.x;
                        atomJSON["storables"][i]["position"]["y"].AsFloat = position.y;
                        atomJSON["storables"][i]["position"]["z"].AsFloat = position.z;
                    }
                    if (storableJSON["id"].Value == "hip")
                    {
                        atomJSON["storables"][i]["rootPosition"]["x"].AsFloat = position.x;
                        atomJSON["storables"][i]["rootPosition"]["y"].AsFloat = atomJSON["storables"][i]["rootPosition"]["y"].AsFloat + position.y;
                        atomJSON["storables"][i]["rootPosition"]["z"].AsFloat = position.z;
                    }
                    if (storableJSON["id"].Value.StartsWith("hair"))
                    {
                        atomJSON["storables"][i]["position"]["x"].AsFloat = atomJSON["storables"][i]["position"]["x"].AsFloat + position.x;
                        atomJSON["storables"][i]["position"]["y"].AsFloat = atomJSON["storables"][i]["position"]["y"].AsFloat + position.y;
                        atomJSON["storables"][i]["position"]["z"].AsFloat = atomJSON["storables"][i]["position"]["z"].AsFloat + position.z;
                    }

                    if (storableJSON["id"].Value == "control" || storableJSON["id"].Value == "hip" || storableJSON["id"].Value.StartsWith("hair"))
                    {
                        string rotationKeyName = "";
                        if (storableJSON["id"].Value == "hip") rotationKeyName = "rootRotation";
                        else rotationKeyName = "rotation";
                        atomJSON["storables"][i][rotationKeyName]["x"].AsFloat = rotation.eulerAngles.x;
                        atomJSON["storables"][i][rotationKeyName]["y"].AsFloat = rotation.eulerAngles.y;
                        atomJSON["storables"][i][rotationKeyName]["z"].AsFloat = rotation.eulerAngles.z;
                    }
                }

                atom.PreRestore();
                atom.RestoreTransform(atomJSON);
                atom.Restore(atomJSON);
                atom.LateRestore(atomJSON);
                atom.PostRestore();
            }

        }
        private void PlayAction(string vrHand)
        {
            if (!_mvrScript._hudOnPreButtonAction) SuperController.singleton.HideMainHUD();
            if (_buttonType == "Suppress Scale Load")
            {
                if (_mvrScript._suppressScaleLoad) _mvrScript._suppressScaleLoad = false;
                else _mvrScript._suppressScaleLoad = true;
            }
            else if (_buttonType == "Freeze Motion/Sound")
            {
                if (SuperController.singleton.freezeAnimation) SuperController.singleton.SetFreezeAnimation(false);
                else SuperController.singleton.SetFreezeAnimation(true);
            }
            else if (_buttonType == "Toggle Play/Edit Mode")
            {
                if (SuperController.singleton.gameMode == SuperController.GameMode.Play)
                {
                    SuperController.singleton.gameMode = SuperController.GameMode.Edit;
                    if (_openGameUIonEditMode) SuperController.singleton.ShowMainHUDAuto();
                }
                else
                {
                    SuperController.singleton.gameMode = SuperController.GameMode.Play;
                    if (_closeGameUIonPlayMode) SuperController.singleton.HideMainHUD();
                }
            }

            else if (_buttonType == "Load Scene")
            {
                SuperController.singleton.Load(_selectedPreset1File);
            }
            else if (_buttonType == "Merge Load Scene")
            {
                SuperController.singleton.LoadMerge(_selectedPreset1File);
            }
            else if (_buttonType == "Toggle Soft Body Physics") UserPreferences.singleton.softPhysics = !UserPreferences.singleton.softPhysics;
            else if (_buttonType == "Toggle Desktop VSync") UserPreferences.singleton.desktopVsync = !UserPreferences.singleton.desktopVsync;
            else if (_buttonType == "Toggle Performance Monitor")
            {
                PerfMon pm = UserPreferences.singleton.transform.Find("PerfMon").GetComponent<PerfMon>();
                pm.onToggle.isOn = !pm.onToggle.isOn;
            }

            else if (_buttonType == "Toggle Show Hidden Atoms") SuperController.singleton.showHiddenAtoms = !SuperController.singleton.showHiddenAtoms;
            else if (_buttonType == "Load UI Assist Profile")
            {
                _mvrScript.LoadUIAPSelected(_selectedPreset1File);
            }
            else if (_buttonType == "Gaze Assisted Select")
            {
                int buttonState = GetButtonState("");
                if (buttonState == 1)
                {
                    _mvrScript._toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val = true;
                    SuperController.singleton.gameMode = SuperController.GameMode.Edit;
                }
                else _mvrScript._toolbox._gazeAssistedSelectTool._gazeAssistActiveJSON.val = false;
            }
            else if (_buttonType == "Hide Inactive Targets") UserPreferences.singleton.hideInactiveTargets = !UserPreferences.singleton.hideInactiveTargets;
            else if (_buttonType == "Active Clothing Editor" && Constants._patreonContentEnabled)
            {
                int buttonState = GetButtonState("");
                if (buttonState == 1)
                {
                    _mvrScript._uiaActiveClothingMode = true;
                    _mvrScript._uiaActiveClothingModeTargetType = _targetName;
                }
                else _mvrScript._uiaActiveClothingMode = false;
                _mvrScript.RefreshWristUIButtonGrid();
            }
            else if (_buttonType == "Raise People by Heel Adjust Height")
            {
                int buttonState = GetButtonState("");
                if (buttonState == 1) _mvrScript._toolbox._heelAdjustTool._heelAdjustRaisePeopleJSON.val = true;

                else _mvrScript._toolbox._heelAdjustTool._heelAdjustRaisePeopleJSON.val = false;
            }

            else if (_buttonType == "Spawn Atom")
            {
                Vector3 position = new Vector3(0f, 0f, 0f);
                Quaternion rotation = new Quaternion();

                _mvrScript._atomSpawningActive = true;
                UpdateSpawnPosition(ref position, ref rotation, _selectedTarget);

                _mvrScript.StartCoroutine(CreateAtom(_atomType, (atomName) =>
                {
                    try
                    {
                        TeleportAtomToSpawnPoint(atomName, position, rotation, false);

                        _mvrScript.StartCoroutine(RestoreAtomPreset(atomName, _selectedTarget));
                    }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                }));

            }
            else
            {

                if (_targetCategory == "Specific Atom")
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(_targetName);

                    if (atom != null)
                    {
                        int buttonState = GetButtonState(atom.name);
                        PlayActionTargetAtom(atom, buttonState, vrHand);
                        if (_buttonType != "Delete Atom") ToggleButtonState(atom.name);
                    }
                }
                else if (_targetCategory == "Gaze Selected Atom" || _targetCategory == "VAM Selected Atom")
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(_selectedTarget);
                    if (atom != null)
                    {
                        int buttonState = GetButtonState(atom.name);
                        PlayActionTargetAtom(atom, buttonState, vrHand);
                        if (_buttonType != "Delete Atom") ToggleButtonState(atom.name);
                    }
                }
                else if (_targetCategory == "Atom Group")
                {
                    int buttonState = GetButtonState("");
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.name != "CoreControl" && atom.name != "[CameraRig]")
                        {
                            bool isMale = false;
                            if (atom.type == "Person" && _mvrScript.IsMale(atom)) isMale = true;
                            switch (_targetName)
                            {
                                case "All Atoms":
                                    PlayActionTargetAtom(atom, buttonState, vrHand);
                                    break;
                                case "All People":
                                    if (atom.type == "Person") { PlayActionTargetAtom(atom, buttonState, vrHand); }
                                    break;
                                case "All Females":
                                    if (atom.type == "Person" && !isMale) { PlayActionTargetAtom(atom, buttonState, vrHand); }
                                    break;
                                case "All Males":
                                    if (atom.type == "Person" && isMale) { PlayActionTargetAtom(atom, buttonState, vrHand); }
                                    break;
                            }
                        }
                    }
                    if (_buttonType != "Delete Atom") ToggleButtonState(_targetName);
                }
            }
        }
        private IEnumerator RestoreAtomPreset(string atomName,string targetAtomName)
        {
            yield return new WaitForEndOfFrame();

            Atom atom = SuperController.singleton.GetAtomByUid(atomName);

            if (_parentLinkToTarget && _atomPositionRelativeTo.StartsWith("Last"))
            {
                string targetNode = _targetNode;
                if (_targetNode == "Lips" || _targetNode == "Mouth") targetNode = "head";
                else if ( _targetNode == "Labia" || _targetNode == "Vagina") targetNode = "pelvis";
                Atom targetAtom = SuperController.singleton.GetAtomByUid(targetAtomName);
                atom.mainController.SetLinkToAtom(targetAtomName);
                Rigidbody linkRB = targetAtom.rigidbodies.First(rb => rb.name == targetNode);
                atom.mainController.linkToRB = linkRB;
                atom.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
                atom.mainController.currentRotationState = FreeControllerV3.RotationState.ParentLink;
            }
            int maxPlugins = 100;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null || test.Count != 19)
            {
                maxPlugins = 1;
            }
            
            if (_atomType == "Person")
            {
                JSONStorable js = atom.GetStorableByID("PosePresets");
                JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");

                presetPathJSON.val = "";
                if (_preset3Type!="None" && _selectedPreset3File != "")
                {
                    presetPathJSON.val = SuperController.singleton.NormalizePath(_selectedPreset3File);
                    js.CallAction("LoadPreset");
                }
                js = atom.GetStorableByID("AppearancePresets");
                presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");
                presetPathJSON.val = "";
                if (_preset1Type != "None" && _selectedPreset1File != "")
                {
                    presetPathJSON.val = SuperController.singleton.NormalizePath(_selectedPreset1File);
                    js.CallAction("LoadPreset");
                }

               
            }
            
            LoadPlugins(atom, maxPlugins);
            _mvrScript._atomSpawningActive = false;
            SuperController.singleton.helpHUDText.text = "";
        }
        private void PlayActionTargetAtom(Atom atom, int buttonState, string vrHand) {
            JSONStorable js = null;
            string actionType = "";
            bool reverse = false;
            string actionMode = "";
            

            switch (_buttonType) {
                case "Load Plugins Preset":
                    js = atom.GetStorableByID("PluginPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load App Preset":
                    js = atom.GetStorableByID("AppearancePresets");
                    actionType = "LOADPRESET";
                    
                    if (!_suppressClothingLoad) actionMode = "RESETACE";
                    break;
                case "Load Pose Preset":
                    js = atom.GetStorableByID("PosePresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Hair Preset":
                    js = atom.GetStorableByID("HairPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Glute Preset":
                    js = atom.GetStorableByID("FemaleGlutePhysicsPresets");
                    if (js == null) { js = atom.GetStorableByID("MaleGlutePhysicsPresets"); }
                    actionType = "LOADPRESET";
                    break;
                case "Load Breast Preset":
                    js = atom.GetStorableByID("FemaleBreastPhysicsPresets");
                    if (js == null) { js = atom.GetStorableByID("MaleBreastPhysicsPresets"); }
                    actionType = "LOADPRESET";
                    break;
                case "Load Morph Preset":
                    js = atom.GetStorableByID("MorphPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Skin Preset":
                    js = atom.GetStorableByID("SkinPresets");
                    actionType = "LOADPRESET";
                    if (_decalGenLoad || _decalFaceLoad || _decalLimbsLoad || _decalBodyLoad) actionMode = "SKINDECALS";
                    break;
                case "Load General Preset":
                    js = atom.GetStorableByID("Preset");
                    actionType = "LOADPRESET";
                    actionMode = "RESETACE";
                    break;
                case "Load Clothing Preset":
                    js = atom.GetStorableByID("ClothingPresets");
                    actionType = "LOADPRESET";
                    actionMode = "RESETACE";
                    break;
                case "Remove All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "REMOVE";
                    break;
                case "Undress All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "UNDRESS";
                    if (buttonState == 2) { reverse = true; }
                    break;
                case "Reset Sim All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "RESETSIM";
                    break;
                case "Remove Clothing Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "REMOVE";
                    break;
                case "Merge Clothing Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "MERGE";
                    if (buttonState == 2) { reverse = true; }
                    break;
                case "Undress Cloth Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "UNDRESS";
                    if (buttonState == 2) { reverse = true; }
                    break;
                case "Reset Sim Cloth Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "RESETSIM";
                    break;
                case "Load Plugin(s)":
                    actionType = "LOADPLUGINS";
                    break;
                case "Unload All Plugins":
                    actionType = "UNLOADPLUGINS";
                    break;
                case "Plugin Action":
                    actionType = "PLUGINACTION";
                    actionMode = "SINGLE";
                    break;
                case "Plugin Action Toggle":
                    actionType = "PLUGINACTION";
                    actionMode = "TOGGLE";
                    break;
                case "Plugin Bool Toggle":
                    actionType = "PLUGINBOOL";
                    actionMode = "TOGGLE";
                    break;
                case "Set Plugin Bool True":
                    actionType = "PLUGINBOOL";
                    actionMode = "TRUE";
                    break;
                case "Set Plugin Bool False":
                    actionType = "PLUGINBOOL";
                    actionMode = "FALSE";
                    break;
                case "Reset Scale":
                    actionType = "RESET";
                    actionMode = "SCALE";
                    break;
                case "Reset Appearance":
                    actionType = "RESET";
                    actionMode = "APP";
                    break;
                case "Decal Maker Toggle":
                    actionType = "DECALMAKER";
                    break;
                case "Set Hair Color":
                    actionType = "SETHAIRCOLOR";
                    break;
                case "Hide Atom":
                    actionType = "HIDEATOM";
                    break;
                case "Detach Root Control":
                    actionType = "DETACHCONTROLNODE";
                    break;
                case "Delete Atom":
                    actionType = "DELETEATOM";
                    break;
                case "Select Atom":
                    actionType = "SELECTATOM";
                    break;
                case "Move Atom (VR Only)":
                    actionType = "MOVEATOM";
                    break;
                case "Teleport Atom":
                    actionType = "TELEPORTATOM";
                    break;
                case "Toggle Atom Collision":
                    actionType = "ATOMCOLLISION";
                    break;
                case "Toggle Atom On/Off":
                    actionType = "ATOMON";
                    break;
                case "Load Legacy Preset":
                    actionType = "LOADLEGACYPRESET";
                    break;
                case "Load Legacy Appearance":
                    actionType = "LOADLEGACYAPP";
                    break;
                case "Load Legacy Pose":
                    actionType = "LOADLEGACYPOSE";
                    break;
                case "Save Legacy Preset":
                    actionType = "SAVELEGACYPRESET";
                    break;
                case "Save Legacy Appearance":
                    actionType = "SAVELEGACYAPP";
                    break;
                case "Save Legacy Pose":
                    actionType = "SAVELEGACYPOSE";
                    break;
            }
            switch (actionType) {
                case "LOADLEGACYPRESET":
                    if (_selectedPreset1File != "" && FileManagerSecure.FileExists(_selectedPreset1File)) atom.LoadPreset(_selectedPreset1File);
                    break;
                case "LOADLEGACYAPP":
                    if (_selectedPreset1File != "" && FileManagerSecure.FileExists(_selectedPreset1File)) atom.LoadAppearancePreset(_selectedPreset1File);
                    break;
                case "LOADLEGACYPOSE":
                    if (_selectedPreset1File != "" && FileManagerSecure.FileExists(_selectedPreset1File)) atom.LoadPhysicalPreset(_selectedPreset1File);
                    break;
                case "SAVELEGACYPRESET":
                    if (_selectedPreset1File != "") SuperController.singleton.Save(_selectedPreset1File, atom, true, true);
                    break;
                case "SAVELEGACYAPP":
                    if (_selectedPreset1File != "") SuperController.singleton.Save(_selectedPreset1File, atom, false, true);
                    break;
                case "SAVELEGACYPOSE":
                    if (_selectedPreset1File != "") SuperController.singleton.Save(_selectedPreset1File, atom, false, true);                    
                    break;
                case "ATOMON":
                    atom.ToggleOn();
                    break;
                case "HIDEATOM":
                    atom.hidden = !atom.hidden;
                    break;
                case "DELETEATOM":
                    atom.Remove();
                    break;

                case "TELEPORTATOM":
                    Vector3 position = new Vector3(0f, 0f, 0f);
                    Quaternion rotation = new Quaternion();

                    UpdateSpawnPosition(ref position, ref rotation, "");
                    TeleportAtomToSpawnPoint(atom.name, position, rotation, false);

                    break;
                case "MOVEATOM":
                    if (_mvrScript._vrActive && vrHand !="NONE")
                    {
                        if (!_mvrScript._atomVRMoveActive || _mvrScript._atomVRMoveUIAButton != this)
                        {
                            if (_mvrScript._atomVRMoveActive)
                            {
                                Atom previousMoveAtom = SuperController.singleton.GetAtomByUid(_mvrScript._atomVRMoveAtomName);
                                if (previousMoveAtom != null)
                                {
                                    previousMoveAtom.tempFreezePhysics = false;
                                }
                            }

                            if (_vrMoveDisablePhysics) atom.tempFreezePhysics = true;
                            _mvrScript._atomVRMoveActive = true;
                            _mvrScript._atomVRMoveAtomName = atom.name;
                            _mvrScript._atomVRMoveUIAButton = this;
                            FreeControllerV3 atomFC = atom.mainController;
                            Transform vrHandTransform = null;
                            if (vrHand == "" && _mvrScript._vrHandControlJSON.val == "Left Hand") vrHand = "LEFT";
                            if (vrHand == "" && _mvrScript._vrHandControlJSON.val == "Right Hand") vrHand = "RIGHT";
                            if (vrHand == "" && _mvrScript._vrHandControlJSON.val == "Disabled") vrHand = "RIGHT";

                            if (vrHand == "LEFT") vrHandTransform = SuperController.singleton.leftHand;
                            else if (vrHand == "RIGHT") vrHandTransform = SuperController.singleton.rightHand;
                            _mvrScript._atomVRMoveControllerLorR = vrHand;
                            _mvrScript._atomVRMoveControllerLastPosition = vrHandTransform.position;
                            _mvrScript._atomVRMoveControllerLastRotation = vrHandTransform.rotation;

                            _mvrScript._atomVRMoveAtomLastPosition = atomFC.transform.position;
                            _mvrScript._atomVRMoveAtomLastRotation = atomFC.transform.rotation;

                            _mvrScript._currentVRMovelockedPositionX = _lockedPositionX;
                            _mvrScript._currentVRMovelockedPositionY = _lockedPositionY;
                            _mvrScript._currentVRMovelockedPositionZ = _lockedPositionZ;
                            _mvrScript._currentVRMovelockedRotationX = _lockedRotationX;
                            _mvrScript._currentVRMovelockedRotationY = _lockedRotationY;
                            _mvrScript._currentVRMovelockedRotationZ = _lockedRotationZ;
                            _mvrScript._currentVRMovevrMovePositionGranulatity = _vrMovePositionGranulatity;
                            _mvrScript._currentVRMovevrMoveRotationGranulatity = _vrMoveRotationGranulatity;
                        }
                        else
                        {
                            _mvrScript._atomVRMoveActive = false;
                            atom.tempFreezePhysics = false;
                            SuperController.singleton.helpHUDText.text = "";
                        }
                        
                    }                   
                    break;
                case "DETACHCONTROLNODE":                   
                    JSONStorable storable = atom.GetStorableByID("control");
                    if (storable != null)
                    {
                        JSONStorableBool detach = storable.GetBoolJSONParam("detachControl");
                        if (detach != null)detach.val = !detach.val;
                    }
                    break;
                case "SELECTATOM":
                    FreeControllerV3 fcV3 = atom.freeControllers.First(fc => fc.name == "control");
                    SuperController.singleton.SelectController(fcV3);                    
                    break;
                case "ATOMCOLLISION":
                    atom.collisionEnabledJSON.val = !atom.collisionEnabledJSON.val;
                    break;
                case "LOADPRESET":
                    if (js != null && _selectedPreset1File != "" && FileManagerSecure.FileExists(_selectedPreset1File)) {
                        
                        string fileName = "";
                        if (actionMode == "SKINDECALS" && Constants._patreonContentEnabled)
                        {
                            PatreonFeatures.SkinPresetDecalLoad(_mvrScript, atom, _selectedPreset1File, _decalFaceLoad, _decalBodyLoad, _decalLimbsLoad, _decalGenLoad);
                        }
                        else
                        {

                            if ((_suppressPersonScaleLoad || _mvrScript._suppressScaleLoad) && Constants._patreonContentEnabled) fileName = PatreonFeatures.AppPresetSuppressLoad(_mvrScript, atom, _selectedPreset1File); 
                            else fileName = _selectedPreset1File;
                            
                            bool tempLockParams = false;
                            PresetManagerControl tempPMC=null;
                            if (_suppressClothingLoad)
                            {
                                List<PresetManagerControl> pmControlList = atom.presetManagerControls;
                                foreach (PresetManagerControl pmc in pmControlList)
                                {
                                    if (pmc.name == "ClothingPresets")
                                    {
                                        tempLockParams = pmc.lockParams;
                                        tempPMC = pmc;
                                        pmc.lockParams = true;
                                    }
                                    
                                }
                            }
                            JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");
                            presetPathJSON.val = SuperController.singleton.NormalizePath(fileName);

                            if (_mergeLoadPreset) js.CallAction("MergeLoadPreset");
                            else js.CallAction("LoadPreset");

                            if (actionMode == "SUPPRESS")
                            {
                                FileManagerSecure.DeleteFile("Custom/Atom/Person/Appearance/Preset_UIAssistTemp1.vap");
                                FileManagerSecure.DeleteFile("Custom/Atom/Person/Appearance/Preset_UIAssistTemp2.vap");
                            }
                            if (_suppressClothingLoad) tempPMC.lockParams = tempLockParams;
                            if (actionMode == "RESETACE") _mvrScript._uidMainButtonCanvasDisplay._uiActiveClothingEditor.RefreshClothing(atom.name);
                        }
                    }
                    break;
                case "CLOTHING":
                    ClothingActions(atom, actionMode, reverse);
                    break;
                case "CLOTHINGPRESET":
                    if (Constants._patreonContentEnabled)
                    {
                        if (actionMode=="MERGE" && _tagsToRemoveOnMerge.Count > 0 && _mvrScript._atomActiveClothingDictionary.ContainsKey(atom.name))
                        {
                            PatreonFeatures.RemoveTagClothingItems(atom, _mvrScript._atomActiveClothingDictionary[atom.name]._activeClothingTags,_tagsToRemoveOnMerge);
                        }
                        List<string> mergeClothingGeometryIDs = null;
                        if (actionMode == "MERGE" && _mvrScript._presetMergeClothingGeometryIDs.ContainsKey(_selectedPreset1File)) mergeClothingGeometryIDs = _mvrScript._presetMergeClothingGeometryIDs[_selectedPreset1File];
                        else if (actionMode == "MERGE")
                        {
                            mergeClothingGeometryIDs = new List<string>();
                            if (_mvrScript._presetMergeClothingGeometryIDs.ContainsKey(_selectedPreset1File)) _mvrScript._presetMergeClothingGeometryIDs.Remove(_selectedPreset1File);
                            _mvrScript._presetMergeClothingGeometryIDs.Add(_selectedPreset1File, mergeClothingGeometryIDs);
                        }
                        
                        PatreonFeatures.ClothingPresetActions(_mvrScript, mergeClothingGeometryIDs, atom, actionMode, _selectedPreset1File, reverse);

                        if (!reverse)
                        {
                            if (_atomClothingPresets.ContainsKey(atom.name)) _atomClothingPresets[atom.name] = _selectedPreset1File;
                            else _atomClothingPresets.Add(atom.name, _selectedPreset1File);
                        }
                        else if (_atomClothingPresets.ContainsKey(atom.name)) _atomClothingPresets.Remove(atom.name);

                    }
                    break;
                case "LOADPLUGINS":
                    if (atom.name != "CoreControl") StartLoadPlugins(atom, _pluginForceMergeLoad);
                    break;
                case "DECALMAKER":
                    if (atom.type == "Person")
                    {
                        DecalMakerToggle(atom);
                    }
                    break;
                case "UNLOADPLUGINS":
                    if (atom.name != "CoreControl") UnloadPlugins(atom);
                    break;
                case "PLUGINACTION":
                    PluginAction(atom, actionMode);
                    break;
                case "PLUGINBOOL":
                    PluginBool(atom, actionMode);
                    break;
                case "RESET":
                    ResetAppearance(atom, actionMode);
                    break;
                case "SETHAIRCOLOR":
                    if (_hairScalpControl != null) UpdateScalpMaterials(atom, _hairScalpControl);
                    if (_hairSimControl != null) UpdateHairMaterials(atom, _hairSimControl);
                    break;
            }
        }
        private void UpdateHairMaterials(Atom atom, JSONNode hairProperties)
        {
            foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
            {
                HairSimControl hairControl = hairGroup.GetComponentInChildren<HairSimControl>();

                if (hairControl != null)
                {
                    hairControl.SetFloatParamValue("primarySpecularSharpness", hairProperties["primarySpecularSharpness"].AsFloat);
                    hairControl.SetFloatParamValue("secondarySpecularSharpness", hairProperties["secondarySpecularSharpness"].AsFloat);
                    hairControl.SetFloatParamValue("specularShift", hairProperties["specularShift"].AsFloat);
                    hairControl.SetFloatParamValue("randomColorPower", hairProperties["randomColorPower"].AsFloat);
                    hairControl.SetFloatParamValue("randomColorOffset", hairProperties["randomColorOffset"].AsFloat);
                    hairControl.SetFloatParamValue("diffuseSoftness", hairProperties["diffuseSoftness"].AsFloat);
                    hairControl.SetFloatParamValue("fresnelPower", hairProperties["fresnelPower"].AsFloat);
                    hairControl.SetFloatParamValue("fresnelAttenuation", hairProperties["fresnelAttenuation"].AsFloat);
                    hairControl.SetFloatParamValue("IBLFactor", hairProperties["IBLFactor"].AsFloat);
                    hairControl.SetFloatParamValue("normalRandomize", hairProperties["normalRandomize"].AsFloat);

                    JSONNode rootColorNode = hairProperties["rootColor"];
                    JSONNode tipColorNode = hairProperties["tipColor"];
                    JSONNode specColorNode = hairProperties["specularColor"];

                    HSVColor rootColor; rootColor.H = rootColorNode["h"].AsFloat; rootColor.S = rootColorNode["s"].AsFloat; rootColor.V = rootColorNode["v"].AsFloat;
                    HSVColor tipColor; tipColor.H = tipColorNode["h"].AsFloat; tipColor.S = tipColorNode["s"].AsFloat; tipColor.V = tipColorNode["v"].AsFloat;
                    HSVColor specColor; specColor.H = specColorNode["h"].AsFloat; specColor.S = specColorNode["s"].AsFloat; specColor.V = specColorNode["v"].AsFloat;

                    hairControl.SetColorParamValue("rootColor", rootColor);
                    hairControl.SetColorParamValue("tipColor", tipColor);
                    hairControl.SetFloatParamValue("colorRolloff", hairProperties["colorRolloff"].AsFloat);
                    hairControl.SetColorParamValue("specularColor", specColor);
                }
            }
        }
        private void UpdateScalpMaterials(Atom atom, JSONNode scalpProperties)
        {
            foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
            {
                DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
                HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();
                
                if (hairControl != null && simControl != null)
                {
                    DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

                    float currentScalpAlpha = scalpControl.GetFloatParamValue("Alpha Adjust");

                    // Don't update anything if the current alpha is almost transparent.
                    if (currentScalpAlpha > -0.97f)
                    {   

                        float specularTextureOffset = scalpProperties["Specular Texture Offset"].AsFloat;
                        float specularIntensity = scalpProperties["Specular Intensity"].AsFloat;
                        float gloss = scalpProperties["Gloss"].AsFloat;
                        float specularFresnel = scalpProperties["Specular Fresnel"].AsFloat;
                        float glossTextureOffset = scalpProperties["Gloss Texture Offset"].AsFloat;
                        float giFilter = scalpProperties["Global Illumination Filter"].AsFloat;
                        float alphaAdjust = scalpProperties["Alpha Adjust"].AsFloat;
                        float diffuseTextureOffset = scalpProperties["Diffuse Texture Offset"].AsFloat;

                        JSONNode diffuseColorNode = scalpProperties["Diffuse Color"];
                        JSONNode specularColorNode = scalpProperties["Specular Color"];
                        JSONNode subsurfaceColorNode = scalpProperties["Subsurface Color"];

                        HSVColor diffuseColor; diffuseColor.H = diffuseColorNode["h"].AsFloat; diffuseColor.S = diffuseColorNode["s"].AsFloat; diffuseColor.V = diffuseColorNode["v"].AsFloat;
                        HSVColor specularColor; specularColor.H = specularColorNode["h"].AsFloat; specularColor.S = specularColorNode["s"].AsFloat; specularColor.V = specularColorNode["v"].AsFloat;
                        HSVColor subsurfaceColor; subsurfaceColor.H = subsurfaceColorNode["h"].AsFloat; subsurfaceColor.S = subsurfaceColorNode["s"].AsFloat; subsurfaceColor.V = subsurfaceColorNode["v"].AsFloat;

                        scalpControl.SetFloatParamValue("Specular Texture Offset", specularTextureOffset);
                        scalpControl.SetFloatParamValue("Specular Intensity", specularIntensity);
                        scalpControl.SetFloatParamValue("Gloss", gloss);
                        scalpControl.SetFloatParamValue("Specular Fresnel", specularFresnel);
                        scalpControl.SetFloatParamValue("Gloss Texture Offset", glossTextureOffset);
                        scalpControl.SetFloatParamValue("Global Illumination Filter", giFilter);
                        scalpControl.SetFloatParamValue("Alpha Adjust", alphaAdjust);
                        scalpControl.SetFloatParamValue("Diffuse Texture Offset", diffuseTextureOffset);

                        scalpControl.SetColorParamValue("Diffuse Color", diffuseColor);
                        scalpControl.SetColorParamValue("Specular Color", specularColor);
                        scalpControl.SetColorParamValue("Subsurface Color", subsurfaceColor);
                    }
                }
            }
        }
        private List<string> GetLoadedDecalTextures(JSONArray decalsJSONArray)
        {
            List<string> decalsList = new List<string>();

            foreach (JSONClass decalJSON in decalsJSONArray)
            {
                decalsList.Add(decalJSON["Path"].Value);        
            }

            return (decalsList);
        }
        private Dictionary<string,int> GetLoadedDecalArrayIndicies(JSONArray decalsJSONArray)
        {
            Dictionary<string, int> decalsDict = new Dictionary<string, int>();
            int index = 0;
            foreach (JSONClass decalJSON in decalsJSONArray)
            {
                string path = decalJSON["Path"].Value;
                if (!decalsDict.ContainsKey(path)) decalsDict.Add(path, index);
                index++;
            }

            return (decalsDict);
        }
        private string GetDecalMakerPluginKey (Atom atom)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            
            JSONClass currentPlugins = manager.GetJSON(true, true, false);
            
            string pluginKey = "";
            foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"])
            {
                string pluginPath = kvp.Value.ToString().TrimStart('"').TrimEnd('"');
                if (pluginPath.StartsWith("Chokaphi.DecalMaker."))
                {
                    JSONStorable pluginStorable = atom.GetStorableByID(kvp.Key + "_VAM_Decal_Maker.Decal_Maker");
                    if (pluginStorable!=null && pluginStorable.GetActionNames().Contains("ClearAll"))
                    {
                        pluginKey = kvp.Key;
                        break;
                    }
                }
            }
            
            return (pluginKey);
        }

        private void AddDMDecal(JSONArray decalArray, DecalMakerTexture dmt, Dictionary<string, int> loadedDecalIndicies)
        {
            if (loadedDecalIndicies.ContainsKey(dmt._decalTexturePath))
            {
                JSONClass existingDecalSaveJSON = (JSONClass) decalArray[loadedDecalIndicies[dmt._decalTexturePath]];
                existingDecalSaveJSON["H"].AsFloat = dmt._decalTextureColor.H;
                existingDecalSaveJSON["S"].AsFloat = dmt._decalTextureColor.S;
                existingDecalSaveJSON["V"].AsFloat = dmt._decalTextureColor.V;
                existingDecalSaveJSON["Alpha"].AsFloat = dmt._decalTextureAlpha;
                existingDecalSaveJSON["Path"] = dmt._decalTexturePath;
            }
            else
            {

                JSONClass newDecalSaveJSON = new JSONClass();
                newDecalSaveJSON["H"].AsFloat = dmt._decalTextureColor.H;
                newDecalSaveJSON["S"].AsFloat = dmt._decalTextureColor.S;
                newDecalSaveJSON["V"].AsFloat = dmt._decalTextureColor.V;
                newDecalSaveJSON["Alpha"].AsFloat = dmt._decalTextureAlpha;
                newDecalSaveJSON["Path"] = dmt._decalTexturePath;
                decalArray.Add(newDecalSaveJSON);
            }
        }
        private void RemoveDMDecal(JSONArray decalArray, DecalMakerTexture dmt, Dictionary<string, int> loadedDecalIndicies)
        {
            if (loadedDecalIndicies.ContainsKey(dmt._decalTexturePath))
            {
                decalArray.Remove(loadedDecalIndicies[dmt._decalTexturePath]);
            }
        }
        private void DecalMakerToggle(Atom atom)
        {
            string pluginKey = GetDecalMakerPluginKey(atom);
            if (pluginKey != "")
            {
                Dictionary<string, int> loadedFaceDecalIndicies = null;
                Dictionary<string, int> loadedTorsoDecalIndicies = null;
                Dictionary<string, int> loadedGenDecalIndicies = null;
                Dictionary<string, int> loadedLimbsDecalIndicies = null;
                JSONClass decalJSON = null;
                bool loadDM = false;

                JSONStorable decalPluginStorable = atom.GetStorableByID(pluginKey + "_VAM_Decal_Maker.Decal_Maker");

                if (decalPluginStorable != null)
                {
                    decalJSON = decalPluginStorable.GetJSON(true, true);
                    
                }
                else
                {
                    loadDM = true;
                    decalJSON = new JSONClass();
                    decalJSON["SaveVersion"] = "1";
                    decalJSON["Nipple Cutouts ON"] = "true";
                    decalJSON["Genital Cutouts ON"] = "true";
                    decalJSON["enabled"] = "true";

                    decalJSON["Decal"] = new JSONClass();
                    decalJSON["Decal"]["face"] = new JSONArray();
                    decalJSON["Decal"]["torso"] = new JSONArray();
                    decalJSON["Decal"]["genitals"] = new JSONArray();
                    decalJSON["Decal"]["limbs"] = new JSONArray();
                    decalJSON["Specular"] = new JSONClass();
                    decalJSON["Specular"]["face"] = new JSONArray();
                    decalJSON["Specular"]["torso"] = new JSONArray();
                    decalJSON["Specular"]["genitals"] = new JSONArray();
                    decalJSON["Specular"]["limbs"] = new JSONArray();

                    decalJSON["Gloss"] = new JSONClass();
                    decalJSON["Gloss"]["face"] = new JSONArray();
                    decalJSON["Gloss"]["torso"] = new JSONArray();
                    decalJSON["Gloss"]["genitals"] = new JSONArray();
                    decalJSON["Gloss"]["limbs"] = new JSONArray();

                    decalJSON["Normal"] = new JSONClass();
                    decalJSON["Normal"]["face"] = new JSONArray();
                    decalJSON["Normal"]["torso"] = new JSONArray();
                    decalJSON["Normal"]["genitals"] = new JSONArray();
                    decalJSON["Normal"]["limbs"] = new JSONArray();

                }
                loadedFaceDecalIndicies = GetLoadedDecalArrayIndicies((JSONArray)decalJSON["Decal"]["face"]);
                loadedTorsoDecalIndicies = GetLoadedDecalArrayIndicies((JSONArray)decalJSON["Decal"]["torso"]);
                loadedGenDecalIndicies = GetLoadedDecalArrayIndicies((JSONArray)decalJSON["Decal"]["genitals"]);
                loadedLimbsDecalIndicies = GetLoadedDecalArrayIndicies((JSONArray)decalJSON["Decal"]["limbs"]);

                if (decalJSON != null)
                {
                    int state = GetButtonState(atom.name);

                    JSONStorableUrl faceDecalUrl = atom.GetStorableByID("textures").GetUrlJSONParam("faceDecalUrl");
                    foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionFace._decalMakerTexturesList)
                    {
                        if (dmt._decalTexturePath != "" && dmt._decalTexturePath != faceDecalUrl.val)
                        {
                            if (state == 1)  AddDMDecal((JSONArray)decalJSON["Decal"]["face"], dmt, loadedFaceDecalIndicies);
                            else RemoveDMDecal((JSONArray)decalJSON["Decal"]["face"], dmt, loadedFaceDecalIndicies);
                        }
                    }                    
                    if (faceDecalUrl.val!="")
                    {
                        if (!loadedFaceDecalIndicies.ContainsKey(faceDecalUrl.val))
                        {
                            JSONClass newDecalSaveJSON = new JSONClass();
                            newDecalSaveJSON["H"].AsFloat = 0f;
                            newDecalSaveJSON["S"].AsFloat = 0f;
                            newDecalSaveJSON["V"].AsFloat = 1f;
                            newDecalSaveJSON["Alpha"].AsFloat = 1f;
                            newDecalSaveJSON["Path"] = faceDecalUrl.val;
                            JSONArray decalArray = (JSONArray)decalJSON["Decal"]["face"];
                            decalArray.Add(newDecalSaveJSON);
                        }
                    }

                    JSONStorableUrl torsoDecalUrl = atom.GetStorableByID("textures").GetUrlJSONParam("torsoDecalUrl");
                    foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionTorso._decalMakerTexturesList)
                    {
                        if (dmt._decalTexturePath != "" && dmt._decalTexturePath != torsoDecalUrl.val)
                        {
                            if (state == 1)AddDMDecal((JSONArray)decalJSON["Decal"]["torso"], dmt, loadedTorsoDecalIndicies);
                            else RemoveDMDecal((JSONArray)decalJSON["Decal"]["torso"], dmt, loadedTorsoDecalIndicies);
                        }
                    }
                    if (torsoDecalUrl.val != "")
                    {
                        if (!loadedTorsoDecalIndicies.ContainsKey(torsoDecalUrl.val))
                        {
                            JSONClass newDecalSaveJSON = new JSONClass();
                            newDecalSaveJSON["H"].AsFloat = 0f;
                            newDecalSaveJSON["S"].AsFloat = 0f;
                            newDecalSaveJSON["V"].AsFloat = 1f;
                            newDecalSaveJSON["Alpha"].AsFloat = 1f;
                            newDecalSaveJSON["Path"] = torsoDecalUrl.val;
                            JSONArray decalArray = (JSONArray)decalJSON["Decal"]["torso"];
                            decalArray.Add(newDecalSaveJSON);
                        }
                    }

                    JSONStorableUrl genitalsDecalUrl = atom.GetStorableByID("textures").GetUrlJSONParam("genitalsDecalUrl");
                    foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionGen._decalMakerTexturesList)
                    {
                        if (dmt._decalTexturePath != "" && dmt._decalTexturePath != genitalsDecalUrl.val)
                        {
                            if (state == 1)AddDMDecal((JSONArray)decalJSON["Decal"]["genitals"], dmt, loadedGenDecalIndicies);
                            else RemoveDMDecal((JSONArray)decalJSON["Decal"]["genitals"], dmt, loadedGenDecalIndicies);
                        }
                    }
                    if (genitalsDecalUrl.val != "")
                    {
                        if (!loadedGenDecalIndicies.ContainsKey(genitalsDecalUrl.val))
                        {
                            JSONClass newDecalSaveJSON = new JSONClass();
                            newDecalSaveJSON["H"].AsFloat = 0f;
                            newDecalSaveJSON["S"].AsFloat = 0f;
                            newDecalSaveJSON["V"].AsFloat = 1f;
                            newDecalSaveJSON["Alpha"].AsFloat = 1f;
                            newDecalSaveJSON["Path"] = genitalsDecalUrl.val;
                            JSONArray decalArray = (JSONArray)decalJSON["Decal"]["genitals"];
                            decalArray.Add(newDecalSaveJSON);
                        }
                    }

                    JSONStorableUrl limbsDecalUrl = atom.GetStorableByID("textures").GetUrlJSONParam("limbsDecalUrl");
                    foreach (DecalMakerTexture dmt in _decalMakerTextureCollectionLimbs._decalMakerTexturesList)
                    {
                        if (dmt._decalTexturePath != "" && dmt._decalTexturePath != limbsDecalUrl.val)
                        {
                            if (state == 1)AddDMDecal((JSONArray)decalJSON["Decal"]["limbs"], dmt, loadedLimbsDecalIndicies);
                            else RemoveDMDecal((JSONArray)decalJSON["Decal"]["limbs"], dmt, loadedLimbsDecalIndicies);
                        }
                    }
                    if (limbsDecalUrl.val != "")
                    {
                        if (!loadedLimbsDecalIndicies.ContainsKey(limbsDecalUrl.val))
                        {
                            JSONClass newDecalSaveJSON = new JSONClass();
                            newDecalSaveJSON["H"].AsFloat = 0f;
                            newDecalSaveJSON["S"].AsFloat = 0f;
                            newDecalSaveJSON["V"].AsFloat = 1f;
                            newDecalSaveJSON["Alpha"].AsFloat = 1f;
                            newDecalSaveJSON["Path"] = limbsDecalUrl.val;
                            JSONArray decalArray = (JSONArray)decalJSON["Decal"]["limbs"];
                            decalArray.Add(newDecalSaveJSON);
                        }
                    }
                }
 //               if (loadDM)StartLoadPlugins(atom, false, true, null);

                _mvrScript.StartCoroutine(DMToggleByCharSkinChange(atom, decalPluginStorable, decalJSON, false));                
            }
        }
        private IEnumerator DMToggleByCharSkinChange(Atom atom, JSONStorable decalPluginStorable, JSONClass decalJSON, bool waitForDMLoad)
        {

            DAZCharacterSelector dazCharacterSelector = atom.GetComponentInChildren<DAZCharacterSelector>();
            DAZCharacter dazCharacter = dazCharacterSelector.selectedCharacter;
            JSONStorable receiver = atom.GetStorableByID("geometry");
            JSONStorableStringChooser characterJSON = receiver.GetStringChooserJSONParam("characterSelection");

            string currentCharacterSkinName = characterJSON.val;
            string tempCharacterSkinName = "";

            switch (dazCharacter.UVname)
            {
                case "UV: Base Female":
                    if (currentCharacterSkinName == "Female 1") tempCharacterSkinName = "Female 7";
                    else tempCharacterSkinName = "Female 1";
                    break;
                case "UV: Victoria 6":
                    if (currentCharacterSkinName == "Female 2") tempCharacterSkinName = "Female 4";
                    else tempCharacterSkinName = "Female 2";
                    break;
                case "UV: Base Male":
                    if (currentCharacterSkinName == "Male 4") tempCharacterSkinName = "Male 5";
                    else tempCharacterSkinName = "Male 4";
                    break;
                case "UV: Michael 6":
                   tempCharacterSkinName = "Male 4";
                    break;
                case "UV: Gianni 6":
                    tempCharacterSkinName = "Male 4";
                    break;
                case "UV: Darius 6":
                    tempCharacterSkinName = "Male 4";
                    break;
                case "UV: Lee 6":
                    tempCharacterSkinName = "Male 4";
                    break;
                case "UV: Scott 6":
                    tempCharacterSkinName = "Male 4";
                    break;
                default:
                    tempCharacterSkinName = "Female 1";                    
                    break;
            }

            if (waitForDMLoad) yield return new WaitForSeconds(3f);
            JSONStorableAction clearAllAction = decalPluginStorable.GetAction("ClearAll");
            if (clearAllAction != null)
            {
                decalPluginStorable.CallAction("ClearAll");
                yield return new WaitForSeconds(0.5f);
                yield return new WaitUntil(() => dazCharacter.ready);
                decalPluginStorable.RestoreFromJSON(decalJSON);
                yield return new WaitForSeconds(0.5f);
                yield return new WaitUntil(() => dazCharacter.ready);
                characterJSON.val = tempCharacterSkinName;
                yield return new WaitForSeconds(0.5f);
                yield return new WaitUntil(() => dazCharacter.ready);
                characterJSON.val = currentCharacterSkinName;
            }
        }

        private void ResetAppearance(Atom atom, string mode)
        {
            JSONArray atomsArray = SuperController.singleton.loadJson["atoms"].AsArray;
            bool atomExistedAtLoad = false;
            bool varReferenceAvailable = true;
            JSONClass atomAtLoadJSON =null;
                      
            foreach (JSONClass atomJSON in atomsArray)
            {
                if ((string)atomJSON["id"] == atom.name)
                {
                    atomExistedAtLoad = true;
                    atomAtLoadJSON = atomJSON;
                    string atomAtLoadString = atomAtLoadJSON.ToString();
                    if (atomAtLoadString.Contains("SELF:"))
                    {
                        if (_mvrScript._lastLoadSceneFromVar.Contains(":/Saves/scene"))
                        {
                            string varPackageName = _mvrScript._lastLoadSceneFromVar.Substring(0, _mvrScript._lastLoadSceneFromVar.IndexOf(':'));
                            atomAtLoadJSON = (JSONClass) JSON.Parse(atomAtLoadString.Replace("SELF", varPackageName));
                        }
                        else varReferenceAvailable = false;
                    }
                    break;
                }
            }

            if (mode == "SCALE")
            {

                JSONStorable scaleStorableJSON = atom.GetStorableByID("rescaleObject");
                JSONStorableFloat scaleJSON = scaleStorableJSON.GetFloatJSONParam("scale");
                if (atomExistedAtLoad)  
                {
                    float resetScale = 1f;
                    foreach (JSONClass storable in atomAtLoadJSON["storables"].AsArray)
                    {
                        if (storable["id"].Value == "rescaleObject")
                        {
                            resetScale = storable["scale"].AsFloat;
                            break;
                        }
                    }
                    scaleJSON.val = resetScale;
                }
                else { scaleJSON.val = 1f; }
            }

            else if (mode == "APP" && atomExistedAtLoad && varReferenceAvailable)
            {
                JSONArray newAtomsArray = new JSONArray();
                JSONClass newSave = new JSONClass();
                newAtomsArray.Add(atomAtLoadJSON);
                newSave["atoms"] = newAtomsArray;
                SuperController.singleton.SaveJSON(newSave, "Saves\\UIAtemp.json");
                atom.LoadAppearancePreset("Saves\\UIAtemp.json");
                FileManagerSecure.DeleteFile("Saves\\UIAtemp.json");               
            }
            else if (mode == "APP" && !atomExistedAtLoad) atom.ResetAppearance();
        }
        private void PluginBool(Atom atom, string mode) {
            bool boolState = true;
            if (mode == "FALSE" || (mode == "TOGGLE" && GetButtonState(atom.name) == 2)) { boolState = false; }

            foreach (string storableName in atom.GetStorableIDs()) {
                if (storableName.EndsWith(_pluginType)) {
                    JSONStorable storable = atom.GetStorableByID(storableName);
                    if (storable != null && storable.GetBoolParamNames().Contains(_targetParamNameOn)) {
                        storable.SetBoolParamValue(_targetParamNameOn, boolState);
                    }
                }
            }
        }
        private void PluginAction(Atom atom, string mode) {
            string actionName = "";
            PluginVariables preActionVariables;

            if (mode == "SINGLE" || GetButtonState(atom.name) == 1) {
                actionName = _targetParamNameOn;
                preActionVariables = _pluginVariablesButtonOn;
            }
            else {
                actionName = _targetParamNameOff;
                preActionVariables = _pluginVariablesButtonOff; 
            }

            
            foreach (string storableName in atom.GetStorableIDs()) {
                if (storableName.EndsWith(_pluginType)) {
                    JSONStorable storable = atom.GetStorableByID(storableName);
                    if (storable != null && storable.GetActionNames().Contains(actionName)) {
                        // Set any pre-action plugin parameters
                        foreach (KeyValuePair<string,PluginVariable> kvp in preActionVariables._pluginVarDict)
                        {
                            if (storable.GetBoolParamNames().Contains(kvp.Key))
                            {
                                JSONStorableBool boolJSON = storable.GetBoolJSONParam(kvp.Key);
                                boolJSON.val = kvp.Value._pluginBoolValue;
                            }
                            else if (storable.GetFloatParamNames().Contains(kvp.Key))
                            {
                                JSONStorableFloat floatJSON = storable.GetFloatJSONParam(kvp.Key);
                                floatJSON.val = kvp.Value._pluginFloatValue;
                            }
                            else if (storable.GetStringParamNames().Contains(kvp.Key))
                            {
                                JSONStorableString stringJSON = storable.GetStringJSONParam(kvp.Key);
                                stringJSON.val = kvp.Value._pluginStringValue;
                            }
                            else if (storable.GetStringChooserParamNames().Contains(kvp.Key))
                            {
                                JSONStorableStringChooser stringChooserJSON = storable.GetStringChooserJSONParam(kvp.Key);
                                stringChooserJSON.val = kvp.Value._pluginStringValue;
                                
                            }
                            else if (storable.GetUrlParamNames().Contains(kvp.Key))
                            {
                                JSONStorableUrl urlJSON = storable.GetUrlJSONParam(kvp.Key);
                                urlJSON.val = kvp.Value._pluginStringValue;
                            }
                            else if (storable.GetColorParamNames().Contains(kvp.Key))
                            {
                                JSONStorableColor colorJSON = storable.GetColorJSONParam(kvp.Key);
                                colorJSON.val = kvp.Value._pluginColorValue;
                            }

                        }

                        storable.CallAction(actionName);
                    }
                }
            }
        }

        private void ClothingActions(Atom atom, string mode, bool reverse) {
            foreach (string receiverName in atom.GetStorableIDs()) {
                JSONStorable receiver = atom.GetStorableByID(receiverName);

                if (receiver != null) {
                    if (receiver.storeId.Length >= 3 && receiver.storeId.Substring(receiver.storeId.Length - 3) == "Sim") {
                        JSONStorableBool unDressBool = receiver.GetBoolJSONParam("allowDetach");
                        JSONStorableBool simEnabledBool = receiver.GetBoolJSONParam("simEnabled");

                        if (mode == "UNDRESS" && unDressBool != null) { unDressBool.val = !reverse; }
                        if (mode == "RESETSIM" && simEnabledBool != null) { if (simEnabledBool.val) { receiver.CallAction("Reset"); } }
                    }
                    if (receiver.storeId == "geometry" && mode == "REMOVE") {
                        foreach (string target in receiver.GetBoolParamNames()) {
                            if (target.Length > 8 && target.Substring(0, 9) == "clothing:") {
                                JSONStorableBool boolTarget = receiver.GetBoolJSONParam(target);
                                if (boolTarget != null) { boolTarget.val = false; }
                            }
                        }
                    }
                }
            }
        }

        private bool IsPluginLoaded(JSONClass pluginMgrJSON, string pluginFileName) {
            bool loaded = false;
            for (int index = 0; index < pluginMgrJSON["plugins"].Count; ++index) {
                if (pluginMgrJSON["plugins"][index] == null) { break; }
                else if (pluginMgrJSON["plugins"][index].Value == pluginFileName) {
                    loaded = true;
                    break;
                }
            }
            return (loaded);
        }
        private int GetNextFreePluginID(JSONClass pluginMgrJSON) {
            int pluginID = -1;
            for (int index = 0; index < 1000; ++index) {
                if (pluginMgrJSON["plugins"]["plugin#" + index.ToString()] == null) {
                    pluginID = index;
                    break;
                }
            }
            return (pluginID);
        }
        private int GetPluginMgrIndex(JSONClass atomJSON) {
            int pluginMgrIndex = -1;
            int storablesIndex = 0;

            foreach (JSONClass storableJSON in atomJSON["storables"].AsArray) {
                if (storableJSON["id"].Value == "PluginManager") { pluginMgrIndex = storablesIndex; break; }
                storablesIndex++;
            }
            return (pluginMgrIndex);
        }

        private void StartLoadPlugins(Atom atom, bool merge, bool dm = false, JSONClass decalJSON =null)
        {
            Dictionary<string, bool> timelineIsPlayingDic = new Dictionary<string, bool>();
            Dictionary<string, float> timelineScrubberDic = new Dictionary<string, float>();
            Dictionary<string, string> timelineAnimationDic = new Dictionary<string, string>();

            // Check for active Timeline plugins
            foreach (string storableID in atom.GetStorableIDs())
            {
                if ((storableID.StartsWith("plugin#") || storableID.StartsWith("pluginOrig#")) && storableID.EndsWith("Timeline.AtomPlugin"))
                {
                    JSONStorable pluginReceiver = atom.GetStorableByID(storableID);
                    JSONStorableFloat timelineScrubberJSON = pluginReceiver.GetFloatJSONParam("Scrubber");
                    JSONStorableBool timelineIsPlayJSON = pluginReceiver.GetBoolJSONParam("Is Playing");
                    JSONStorableStringChooser timelineAnimationJSON = pluginReceiver.GetStringChooserJSONParam("Animation");
                    if (timelineIsPlayJSON != null && timelineScrubberJSON != null && timelineAnimationJSON != null )
                    {
                        timelineIsPlayingDic.Add(storableID, timelineIsPlayJSON.val);
                        timelineScrubberDic.Add(storableID, timelineScrubberJSON.val);
                        timelineAnimationDic.Add(storableID, timelineAnimationJSON.val);
                    }
                }
            }

            int maxPlugins = 100;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null || test.Count != 19)
            {
                maxPlugins = 1;
            }

            if (merge) MergeLoadPlugins(atom, maxPlugins);
            else LoadPlugins(atom, maxPlugins,dm, decalJSON);

            // Reset active Timeline plugins
            _mvrScript.StartCoroutine(RestoreTimeLineState(atom, timelineIsPlayingDic, timelineScrubberDic, timelineAnimationDic,merge));

        }
        private IEnumerator RestoreTimeLineState(Atom atom, Dictionary<string, bool> timelineIsPlayingDic, Dictionary<string, float> timelineScrubberDic, Dictionary<string, string> timelineAnimationDic, bool merge)
        {
            if (merge) yield return new WaitUntil(() => !SuperController.singleton.isLoading);
            else yield return new WaitForEndOfFrame();
            if (timelineIsPlayingDic.Count > 0)
            {
                foreach (KeyValuePair<string, bool> kvp in timelineIsPlayingDic)
                {
                    JSONStorable pluginReceiver = atom.GetStorableByID(kvp.Key);
                    if (pluginReceiver != null && timelineAnimationDic[kvp.Key]!="(Playing...)")
                    {
                        JSONStorableAction timelineStopActionJSON = pluginReceiver.GetAction("Stop");
                        if (timelineStopActionJSON != null) pluginReceiver.CallAction("Stop");

                        JSONStorableFloat timelineScrubberJSON = pluginReceiver.GetFloatJSONParam("Scrubber");
                        JSONStorableBool timelineIsPlayJSON = pluginReceiver.GetBoolJSONParam("Is Playing");
                        JSONStorableStringChooser timelineAnimationJSON = pluginReceiver.GetStringChooserJSONParam("Animation");
                        timelineAnimationJSON.val = timelineAnimationDic[kvp.Key];
                        timelineScrubberJSON.val = timelineScrubberDic[kvp.Key];
                        if (kvp.Value)
                        {
                            JSONStorableAction timelinePlayActionJSON = pluginReceiver.GetAction("Play");
                            if (timelinePlayActionJSON != null) pluginReceiver.CallAction("Play");
                        }

                    }
                }
            }
        }
        private void MergeLoadPlugins(Atom atom, int maxPlugins)
        {
            JSONClass atomsJSON = SuperController.singleton.GetSaveJSON(atom, true, true);
            JSONClass atomJSON = (JSONClass)atomsJSON["atoms"][0];
            JSONClass containingAtomJSON = atom.GetJSON();
            int pluginMgrIndex = GetPluginMgrIndex(atomJSON);
            bool editModeOn = SuperController.singleton.editModeToggle.isOn;      

            int pluginCount = 1;
            foreach (ButtonPluginLoad buttonPluginToLoad in _pluginLoadList)
            {
                if (buttonPluginToLoad._filePath != "" && buttonPluginToLoad._filePath != null && pluginCount <= maxPlugins)
                {
                    string normalFilePath = SuperController.singleton.NormalizePath(buttonPluginToLoad._filePath);
                    if (!IsPluginLoaded((JSONClass)atomJSON["storables"][pluginMgrIndex], normalFilePath) || !buttonPluginToLoad._singleInstancePerAtom || _buttonType == "Spawn Atom")
                    {
                        int nextFreePluginID = GetNextFreePluginID((JSONClass)atomJSON["storables"][pluginMgrIndex]);
                        atomJSON["storables"][pluginMgrIndex]["plugins"]["plugin#" + nextFreePluginID.ToString()] = normalFilePath;
                        if (buttonPluginToLoad._pluginSaveData != "" && buttonPluginToLoad._pluginSaveData != null && maxPlugins > 1)
                        {
                            PatreonFeatures.MergePluginStorables(atomJSON, buttonPluginToLoad._pluginSaveJSON, nextFreePluginID);
                        }
                    }
                }
                pluginCount++;
            }

            JSONClass mergeSceneJSON = new JSONClass();
            mergeSceneJSON["atoms"][-1] = atomJSON;
            SuperController.singleton.SaveJSON(mergeSceneJSON, "Custom\\Atom\\Person\\UIAmergeTemp.json");
            SuperController.singleton.LoadMerge("Custom\\Atom\\Person\\UIAmergeTemp.json");
            FileManagerSecure.DeleteFile("Custom\\Atom\\Person\\UIAmergeTemp.json");
            SuperController.singleton.editModeToggle.isOn = editModeOn;
        }
        private void UnloadPlugins(Atom atom)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;

            JSONClass emptyManager = new JSONClass();
            emptyManager["id"] = "PluginManager";
            manager.LateRestoreFromJSON(emptyManager);
        }
        private void LoadPlugins(Atom atom, int maxPlugins, bool decalMaker = false, JSONClass dmJSON = null)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            
            JSONClass currentPlugins = manager.GetJSON(true, true, true);
            Dictionary<string, JSONClass> pluginSettings = new Dictionary<string, JSONClass>();

            foreach (string sid in atom.GetStorableIDs())
            {
                foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"])
                {
                    if (sid.StartsWith(kvp.Key + "_"))
                    {
                        JSONClass pluginData = atom.GetStorableByID(sid).GetJSON(true, true, true);
                        pluginSettings[kvp.Key + "_"] = pluginData;
                    }
                }
            }

            int pluginCount = 1;
            if (decalMaker)
            {
                int nextFreePluginID = GetNextFreePluginID(currentPlugins);
                currentPlugins["plugins"]["plugin#" + nextFreePluginID.ToString()] = "Chokaphi.DecalMaker.latest:/Custom/Scripts/Chokaphi/VAM_Decal_Maker/VAM_Decal_Maker.cs";
                if (dmJSON != null)
                {
                    dmJSON["id"].Value = "plugin#" + nextFreePluginID.ToString() + "_VAM_Decal_Maker.Decal_Maker";
                    pluginSettings["plugin#" + nextFreePluginID.ToString() + "_"] = dmJSON;
                }
                else
                {
                    dmJSON = new JSONClass();
                    dmJSON["id"].Value = "plugin#" + nextFreePluginID.ToString() + "_VAM_Decal_Maker.Decal_Maker";
                    pluginSettings["plugin#" + nextFreePluginID.ToString() + "_"] = dmJSON;
                }
                
            }
            else
            {
                foreach (ButtonPluginLoad buttonPluginToLoad in _pluginLoadList)
                {
                    if (buttonPluginToLoad._filePath != "" && buttonPluginToLoad._filePath != null && pluginCount <= maxPlugins)
                    {
                        string normalFilePath = SuperController.singleton.NormalizePath(buttonPluginToLoad._filePath);
                        if (!IsPluginLoaded(currentPlugins, normalFilePath) || !buttonPluginToLoad._singleInstancePerAtom || _buttonType == "Spawn Atom")
                        {

                            int nextFreePluginID = GetNextFreePluginID(currentPlugins);
                            currentPlugins["plugins"]["plugin#" + nextFreePluginID.ToString()] = normalFilePath;

                            if (buttonPluginToLoad._pluginSaveData != "" && buttonPluginToLoad._pluginSaveData != null && maxPlugins > 1)
                            {
                                PatreonFeatures.MergePluginSettings(pluginSettings, buttonPluginToLoad._pluginSaveJSON, nextFreePluginID);

                            }
                        }
                    }
                    pluginCount++;
                }
            }
            manager.LateRestoreFromJSON(currentPlugins);

            foreach (KeyValuePair<string, JSONClass> kvp in pluginSettings)
            {                 
                string sid = atom.GetStorableIDs().FirstOrDefault(id => id.StartsWith(kvp.Key));
                JSONStorable pluginStorable = atom.GetStorableByID(sid);
                pluginStorable.RestoreFromJSON(kvp.Value);
            }

        }
        private IEnumerator CreateAtom(string atomType, Action<string> onAtomCreated)
        {
            List<Atom> atoms = SuperController.singleton.GetAtoms();
            List<string> atomNames = new List<string>();
            foreach (Atom tempAtom in atoms) atomNames.Add(tempAtom.uid);
            string newAtomName = atomType+"#";

            for (int i =1; i<1000; i++)
            {
                if (!atomNames.Contains(newAtomName + i.ToString()))
                {
                    newAtomName = newAtomName + i.ToString();
                    break;
                }
            }
            
            yield return SuperController.singleton.AddAtomByType(atomType, newAtomName);
           
            onAtomCreated(newAtomName);
        }

    }

    public class ButtonDisplay {
        private UIAssist _mvrScript;
        public UIAButton _selectedButton;
        public JSONStorableStringChooser _buttonCategoryJSON;
        public JSONStorableStringChooser _buttonTypeJSON;
        public JSONStorableString _buttonLabelOnJSON;
        public JSONStorableString _buttonLabelOffJSON;
        public JSONStorableStringChooser _targetCategoryJSON;
        public JSONStorableStringChooser _targetNameJSON;
        public JSONStorableStringChooser _pluginLoadSelectorJSON;
        public JSONStorableStringChooser _pluginSaveDataSelectorJSON;
        public JSONStorableStringChooser _tagItemsToRemoveSelectorJSON;
        public JSONStorableBool _pluginForceMergeLoadJSON;
        public JSONStorableBool _closeGameUIonPlayModeJSON;
        public JSONStorableBool _openGameUIonEditModeJSON;
        public JSONStorableStringChooser _atomCategoryJSON;
        public JSONStorableStringChooser _atomTypeJSON;
        public JSONStorableStringChooser _targetNodeJSON;
        public JSONStorableStringChooser _atomPositionRelativeToJSON;
        
        public JSONStorableFloat _atomPositionXJSON;
        public JSONStorableFloat _atomPositionYJSON;
        public JSONStorableFloat _atomPositionZJSON;
        public JSONStorableFloat _atomRotationXJSON;
        public JSONStorableFloat _atomRotationYJSON;
        public JSONStorableFloat _atomRotationZJSON;
        public JSONStorableBool _atomAbsoluteHeightJSON;
        public JSONStorableBool _atomAbsolutePitchJSON;
        public JSONStorableBool _atomAbsoluteRollJSON;
        public JSONStorableBool _parentLinkToTargetJSON;

        public JSONStorableBool _lockedPositionXJSON;
        public JSONStorableBool _lockedPositionYJSON;
        public JSONStorableBool _lockedPositionZJSON;
        public JSONStorableBool _lockedRotationXJSON;
        public JSONStorableBool _lockedRotationYJSON;
        public JSONStorableBool _lockedRotationZJSON;
        public JSONStorableBool _vrMoveDisablePhysicsJSON;
        public JSONStorableStringChooser _vrMovePositionGranulatityJSON;
        public JSONStorableStringChooser _vrMoveRotationGranulatityJSON;

        public JSONStorableBool _singleInstancePerAtomJSON;
        public JSONStorableBool _suppressClothingLoadJSON;
        public JSONStorableBool _suppressPersonScaleLoadJSON;
        public JSONStorableBool _mergeLoadPresetJSON;
        public JSONStorableBool _autoLabelJSON;

        public JSONStorableBool _decalFaceLoadJSON;
        public JSONStorableBool _decalBodyLoadJSON;
        public JSONStorableBool _decalLimbsLoadJSON;
        public JSONStorableBool _decalGenLoadJSON;

        public JSONStorableString _preset1FilePathJSON;
        public JSONStorableString _preset2FilePathJSON;
        public JSONStorableString _preset3FilePathJSON;
        public JSONStorableStringChooser _preset1TypeJSON;
        public JSONStorableStringChooser _preset2TypeJSON;
        public JSONStorableStringChooser _preset3TypeJSON;

        public JSONStorableStringChooser _hairSimSelectorJSON;
        public JSONStorableStringChooser _hairScalpSelectorJSON;
        public JSONStorableString _hairSimControlJSON;
        public JSONStorableString _hairScalpControlJSON;
        public JSONStorableString _pluginFilePathJSON;
        public JSONStorableString _pluginSaveDataJSON;     
        public JSONStorableStringChooser _targetPluginTypeJSON;

        public JSONStorableStringChooser _targetParamNameOnJSON;
        public JSONStorableStringChooser _targetParamNameOffJSON;
        public JSONStorableBool _useButtonTextureJSON;

        public JSONStorableBool _testSpawnPositionDisableCollisionJSON;
        public JSONStorableStringChooser _testSpawnPositionTargetJSON;

        private UIDynamicPopup _buttonCategoryPopup;
        private UIDynamicPopup _buttonTypePopup;
        private UIDynamicPopup _targetCategoryPopup;
        private UIDynamicPopup _targetNamePopup;
        private UIDynamicPopup _pluginLoadSelectorPopup;
        private UIDynamicPopup _pluginSaveDataSelectorPopup;
        private UIDynamicPopup _targetPluginTypePopup;
        private UIDynamicPopup _tagItemsToRemoveSelectorPopup;
        private UIDynamicPopup _preset1TypePopup;
        private UIDynamicPopup _preset2TypePopup;
        private UIDynamicPopup _preset3TypePopup;
        private UIDynamicTextField _filePath1TextField;
        private UIDynamicTextField _filePath2TextField;
        private UIDynamicTextField _filePath3TextField;
        private UIDynamicTextField _buttonLabelOnTextField;
        private UIDynamicTextField _buttonLabelOffTextField;
        private UIDynamicTextField _pluginFilePathTextField;
        private UIDynamicTextField _pluginSaveDataTextField;
        private UIDynamicTextField _hairSimControlTextField;
        private UIDynamicTextField _hairScalpControlTextField;
        private UIDynamicPopup _hairSimSelectorPopup;
        private UIDynamicPopup _hairScalpSelectorPopup;

        private UIDynamicPopup _targetParamNameOnPopup;
        private UIDynamicPopup _targetParamNameOffPopup;
        private UIDynamicButton _selectFile1Button;
        private UIDynamicButton _selectFile2Button;
        private UIDynamicButton _selectFile3Button;
        private UIDynamicButton _selectPluginFileButton;
        private UIDynamicButton _copyButtonButton;
        private UIDynamicToggle _singleInstancePerAtomToggle;
        private UIDynamicToggle _useButtonTextureToggle;
        private UIDynamicToggle _suppressClothingLoadToggle;
        private UIDynamicToggle _suppressPersonScaleLoadToggle;
        private UIDynamicToggle _mergeLoadPresetToggle;
        private UIDynamicToggle _autoLabelToggle;
        private UIDynamicToggle _pluginForceMergeLoadToggle;
        private UIDynamicToggle _closeGameUIonPlayModeToggle;
        private UIDynamicToggle _openGameUIonEditModeToggle;
        private UIDynamicToggle _parentLinkToTargetToggle;
        private UIDynamicToggle _testSpawnPositionDisableCollisionToggle;
        private UIDynamicPopup _testSpawnPositionTargetPopup;
        private UIDynamicButton _testSpawnPositionButton;

        private UIDynamicPopup _atomCategoryPopup;
        private UIDynamicPopup _atomTypePopup;
        private UIDynamicPopup _targetNodePopup;
        private UIDynamicPopup _atomPositionRelativeToPopup;
        
        private UIDynamicSlider _atomPositionXSlider;
        private UIDynamicSlider _atomPositionYSlider;
        private UIDynamicSlider _atomPositionZSlider;
        private UIDynamicSlider _atomRotationXSlider;
        private UIDynamicSlider _atomRotationYSlider;
        private UIDynamicSlider _atomRotationZSlider;
        private UIDynamicToggle _atomAbsoluteHeightToggle;
        private UIDynamicToggle _atomAbsolutePitchToggle;
        private UIDynamicToggle _atomAbsoluteRollToggle;

        public UIDynamicToggle _lockedPositionXToggle;
        public UIDynamicToggle _lockedPositionYToggle;
        public UIDynamicToggle _lockedPositionZToggle;
        public UIDynamicToggle _lockedRotationXToggle;
        public UIDynamicToggle _lockedRotationYToggle;
        public UIDynamicToggle _lockedRotationZToggle;
        public UIDynamicToggle _vrMoveDisablePhysicsToggle;
        public UIDynamicPopup _vrMovePositionGranulatityPopup;
        public UIDynamicPopup _vrMoveRotationGranulatityPopup;

        private UIDynamicToggle _decalFaceLoadToggle;
        private UIDynamicToggle _decalBodyLoadToggle;
        private UIDynamicToggle _decalLimbsLoadToggle;
        private UIDynamicToggle _decalGenLoadToggle;

        private UIDynamicButton _gridDeleteButton;
        private UIDynamicButton _gridDeleteConfirmButton;
        private UIDynamicButton _gridMoveUpButton;
        private UIDynamicButton _gridMoveDownButton;

        private UIDynamic _spacer1;
        private UIDynamic _spacer1b;
        private UIDynamic _spacer1c;
        private UIDynamic _spacer1d;
        private UIDynamic _spacer2;
        private UIDynamic _spacer2b;
        private UIDynamic _spacer3;

        public ButtonSkinDisplay _buttonSkinDisplay;
        public PluginVariablesDisplay _pluginVariablesDisplayOn;
        public PluginVariablesDisplay _pluginVariablesDisplayOff;
        public DecalMakerTexturesDisplay _dmFaceDecalDisplay;
        public DecalMakerTexturesDisplay _dmTorsoDecalDisplay;
        public DecalMakerTexturesDisplay _dmLimbsDecalDisplay;
        public DecalMakerTexturesDisplay _dmGenDecalDisplay;

        bool _isDisplayUpdating = false;
        public bool _isTestSpawnPositionActive = false;
        public bool _gridDeleteConfirm = false;

        public readonly string[] _ClothingTags = { "Accessory", "Bodysuit", "Bottom", "Bra", "Dress", "Glasses", "Gloves", "Hat", "Jewelry", "Mask", "Panties", "Pants", "Shirt", "Shoes", "Shorts", "Skirt", "Socks", "Stockings", "Sweater", "Top", "Underwear", "Arms", "Feet", "Full Body", "Hands", "Head", "Hip", "Legs", "Neck", "Torso" };

        public ButtonDisplay(UIAButton button, UIAssist mvrScript) {
            try {
                _selectedButton = button;
                _mvrScript = mvrScript;

                _buttonCategoryJSON = new JSONStorableStringChooser("buttonCategory", _selectedButton.GetButtonCategoryChoices(), "", "", SwitchButonCategory);
                _buttonTypeJSON = new JSONStorableStringChooser("buttonType", null, "", "", SwitchButonType);
                _targetCategoryJSON = new JSONStorableStringChooser("atomCategory", null, "", "", SwitchTargetCategory);
                _targetNameJSON = new JSONStorableStringChooser("atomName", null, "", "", SwitchTargetName);
                _preset1FilePathJSON = new JSONStorableString("filePath1", "", SwitchFilePath1);
                _preset2FilePathJSON = new JSONStorableString("filePath2", "", SwitchFilePath2);
                _preset3FilePathJSON = new JSONStorableString("filePath3", "", SwitchFilePath3);
                _preset1TypeJSON = new JSONStorableStringChooser("preset1Type", null, "None", null, SwitchPreset1Type);
                _preset2TypeJSON = new JSONStorableStringChooser("preset2Type", null, "None", null, SwitchPreset2Type);
                _preset3TypeJSON = new JSONStorableStringChooser("preset3Type", null, "None", null, SwitchPreset3Type);

                _pluginLoadSelectorJSON = new JSONStorableStringChooser("pluginLoadSelector", null, "1", "Plugin Selector", SwitchPluginLoad);
                _targetParamNameOnJSON = new JSONStorableStringChooser("targetNameOn", null, "", "", SwitchTargetParamNameOn);
                _targetParamNameOffJSON = new JSONStorableStringChooser("targetNameOff", null, "", "", SwitchTargetParamNameOff);
                _buttonLabelOnJSON = new JSONStorableString("buttonLabelOn", "", SwitchButtonLabelOn);
                _buttonLabelOffJSON = new JSONStorableString("buttonLabelOff", "", SwitchButtonLabelOff);

                _hairSimControlJSON = new JSONStorableString("hairSimControl", "");
                _hairScalpControlJSON = new JSONStorableString("hairScalpControl", "");
                _hairSimSelectorJSON = new JSONStorableStringChooser("hairSimSelector", null,"Select...","Get Hair Color from:",SwitchHairColor);
                _hairScalpSelectorJSON = new JSONStorableStringChooser("hairScalpSelector", null, "Select...", "Get Scalp Color from:",SwitchScalpColor);

                _pluginFilePathJSON = new JSONStorableString("pluginFilePath", "", SwitchPluginFilePath);
                _pluginSaveDataJSON = new JSONStorableString("pluginSaveData", "");
                _singleInstancePerAtomJSON = new JSONStorableBool("singleInstancePerAtom", true, SwitchSingleInstancePerAtom);
                _pluginSaveDataSelectorJSON = new JSONStorableStringChooser("pluginSaveDataSelector", null, "", "", SwitchSaveDataPlugin);
                _targetPluginTypeJSON = new JSONStorableStringChooser("targetPluginType", null, "", "", SwitchTargetPluginType);
                _useButtonTextureJSON = new JSONStorableBool("useButtonTexture", true, SwitchUseButtonTexture);
                _suppressClothingLoadJSON = new JSONStorableBool("suppressClothingLoad", false, SwitchSurpressClothingLoad);
                _suppressPersonScaleLoadJSON = new JSONStorableBool("suppressPersonScaleLoad", false, SwitchSurpressPersonScaleLoad);
                _mergeLoadPresetJSON = new JSONStorableBool("mergeLoadPresetJSON", false, SwitchMergeLoadPreset);
                _autoLabelJSON = new JSONStorableBool("autoLabel", true, SwitchAutoLabel);
                _pluginForceMergeLoadJSON = new JSONStorableBool("pluginForceMergeLoad", false, SwitchForceMergeLoad);
                _closeGameUIonPlayModeJSON = new JSONStorableBool("closeGameUIonPlayMode", true, SwitchCloseGameUIonPlayMode);
                _openGameUIonEditModeJSON = new JSONStorableBool("openGameUIonEditMode", true, SwitchOpenGameUIonEditMode);
                _parentLinkToTargetJSON = new JSONStorableBool("parentLinkToTarget", false, SwitchParentLinkToTarget);

                _atomCategoryJSON = new JSONStorableStringChooser("atomCategory", null, "People", "New Atom Category", SwitchAtomCategory);
                _atomTypeJSON = new JSONStorableStringChooser("atomType", null, "Person", "New Atom Type", SwitchAtomType);
                _targetNodeJSON = new JSONStorableStringChooser("targetNode", null, "control", "Target Node", SwitchTargetNode);
                _atomPositionRelativeToJSON = new JSONStorableStringChooser("atomPositionRelativeTo", null, "", "", SwitchPositionRelativeTo);

                _atomPositionXJSON = new JSONStorableFloat("atomPostionX", 0f, SwitchAtomPositionX, -3f, 3f, false,true);
                _atomPositionYJSON = new JSONStorableFloat("atomPostionY", 0f, SwitchAtomPositionY, -3f, 3f, false, true);
                _atomPositionZJSON = new JSONStorableFloat("atomPostionZ", 0f, SwitchAtomPositionZ, -3f, 3f, false, true);
                _atomRotationXJSON = new JSONStorableFloat("atomRotationX", 0f, SwitchAtomRotationX, -180f, 180f, true, true);
                _atomRotationYJSON = new JSONStorableFloat("atomRotationY", 0f, SwitchAtomRotationY, -180f, 180f, true, true);
                _atomRotationZJSON = new JSONStorableFloat("atomRotationZ", 0f, SwitchAtomRotationZ, -180f, 180f, true, true);
                _atomAbsoluteHeightJSON = new JSONStorableBool("atomAbsoluteHeight",false,SwitchAtomAbsHeight);
                _atomAbsolutePitchJSON = new JSONStorableBool("atomAbsolutePitch", false, SwitchAtomAbsPitch);
                _atomAbsoluteRollJSON = new JSONStorableBool("atomAbsoluteRoll", false, SwitchAtomAbsRoll);
                _testSpawnPositionDisableCollisionJSON = new JSONStorableBool("testSpawnPositionDisableCollision", false, SwitchAtomSpawnTestCollision);
                _testSpawnPositionTargetJSON = new JSONStorableStringChooser("testSpawnPositionTarget", null, "Select...", "Test Atom Spawn Position on", SwitchAtomSpawnTestTarget);

                _lockedPositionXJSON = new JSONStorableBool("lockedPositionX", false, SwitchLockedPositionX);
                _lockedPositionYJSON = new JSONStorableBool("lockedPositionY", false, SwitchLockedPositionY);
                _lockedPositionZJSON = new JSONStorableBool("lockedPositionZ", false, SwitchLockedPositionZ);
                _lockedRotationXJSON = new JSONStorableBool("lockedRotationX", false, SwitchLockedRotationX);
                _lockedRotationYJSON = new JSONStorableBool("lockedRotationY", false, SwitchLockedRotationY);
                _lockedRotationZJSON = new JSONStorableBool("lockedRotationZ", false, SwitchLockedRotationZ);
                _vrMoveDisablePhysicsJSON = new JSONStorableBool("vrMoveDisablePhysics", false, SwitchVRMoveDisablePhysics);

                _vrMovePositionGranulatityJSON = new JSONStorableStringChooser("vrMovePositionGranulatity", null, "", "", SwitchVRMovePositionGranularity);
                _vrMoveRotationGranulatityJSON = new JSONStorableStringChooser("vrMoveRotationGranulatity", null, "", "", SwitchVRMoveRotationGranularity);

                _decalFaceLoadJSON = new JSONStorableBool("decalFaceLoad", false, SwitchDecalFaceLoad);
                _decalBodyLoadJSON = new JSONStorableBool("decalBodyLoad", false, SwitchDecalBodyLoad);
                _decalLimbsLoadJSON = new JSONStorableBool("decalLimbsLoad", false, SwitchDecalLimbsLoad);
                _decalGenLoadJSON = new JSONStorableBool("decalGenLoad", false, SwitchDecalGenLoad);

                if (_selectedButton._buttonSkin != null) _buttonSkinDisplay = new ButtonSkinDisplay(_selectedButton._buttonSkin, _mvrScript, false);
                if (_selectedButton._pluginVariablesButtonOn != null) _pluginVariablesDisplayOn = new PluginVariablesDisplay(_mvrScript,_selectedButton._pluginVariablesButtonOn);
                if (_selectedButton._pluginVariablesButtonOff != null) _pluginVariablesDisplayOff = new PluginVariablesDisplay(_mvrScript, _selectedButton._pluginVariablesButtonOff);

                _dmFaceDecalDisplay = new DecalMakerTexturesDisplay(_mvrScript, _selectedButton._decalMakerTextureCollectionFace);
                _dmTorsoDecalDisplay = new DecalMakerTexturesDisplay(_mvrScript, _selectedButton._decalMakerTextureCollectionTorso);
                _dmLimbsDecalDisplay = new DecalMakerTexturesDisplay(_mvrScript, _selectedButton._decalMakerTextureCollectionLimbs);
                _dmGenDecalDisplay = new DecalMakerTexturesDisplay(_mvrScript, _selectedButton._decalMakerTextureCollectionGen);

                _tagItemsToRemoveSelectorJSON = new JSONStorableStringChooser("tagItemsToRemoveSelector", null, "Toggle Clothing Tag...", "Clothing Tags to Remove", SwitchClothingTagsToggle);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchVRMoveDisablePhysics(bool disablePhys)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._vrMoveDisablePhysics = disablePhys;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedPositionX(bool lockedPos)
        {
            try
            {
                if (!_isDisplayUpdating)_selectedButton._lockedPositionX = lockedPos;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedPositionY(bool lockedPos)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._lockedPositionY = lockedPos;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedPositionZ(bool lockedPos)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._lockedPositionZ = lockedPos;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedRotationX(bool lockedRot)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._lockedRotationX = lockedRot;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedRotationY(bool lockedRot)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._lockedRotationY = lockedRot;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchLockedRotationZ(bool lockedRot)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._lockedRotationZ = lockedRot;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchVRMovePositionGranularity(string granularity)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._vrMovePositionGranulatity = granularity;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchVRMoveRotationGranularity(string granularity)
        {
            try
            {
                if (!_isDisplayUpdating) _selectedButton._vrMoveRotationGranulatity = granularity;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void StartSpawnTest()
        {
            bool success = _mvrScript.ActivateSpawnTest(_testSpawnPositionTargetJSON.val, _selectedButton,_testSpawnPositionDisableCollisionJSON.val);
            if (success) _isTestSpawnPositionActive = true;
            RefreshDisplay();
        }
        public void EndSpawnTest()
        {
            _mvrScript.DeactivateSpawnTest();
        }
        private void SwitchAtomSpawnTestCollision(bool collisionDisabled)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _mvrScript.ToggleSpawnTestCollisionEnabled(!collisionDisabled);
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomSpawnTestTarget(string targetAtom)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    if (targetAtom == "No Atoms of Spawn Type available") _testSpawnPositionTargetJSON.val = "Select...";
                    if (_isTestSpawnPositionActive) EndSpawnTest();
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchParentLinkToTarget(bool parentLink)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._parentLinkToTarget = parentLink;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomAbsPitch(bool absPitch)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomAbsolutePitch = absPitch;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomAbsRoll(bool absRoll)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomAbsoluteRoll = absRoll;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomAbsHeight(bool absHeight)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomAbsoluteHeight = absHeight;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        
        private void SwitchAtomRotationX(float rotation)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomRotationX = rotation;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomRotationY(float rotation)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomRotationY = rotation;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomRotationZ(float rotation)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomRotationZ = rotation;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomPositionX(float position)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomPositionX = position;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomPositionY(float position)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomPositionY = position;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomPositionZ(float position)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomPositionZ = position;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPositionRelativeTo(string referencePoint)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomPositionRelativeTo = referencePoint;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetNode(string node)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._targetNode = node;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomType(string type)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomType = type;
                    _selectedButton._preset1FilePath = "";
                    _selectedButton._preset1Type = "Single File";
                    _selectedButton._buttonTexture = null;
                    _selectedButton._preset2FilePath = "";
                    _selectedButton._preset2Type = "Single File";
                    _selectedButton._preset3FilePath = "";
                    _selectedButton._preset3Type = "Single File";
                    _testSpawnPositionTargetJSON.val = "Select...";
                    _selectedButton._selectedPreset1File = "";
                    _selectedButton._selectedPreset2File = "";
                    _selectedButton._selectedPreset3File = "";

                    if (_isTestSpawnPositionActive) EndSpawnTest();
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAtomCategory(string category)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._atomCategory = category;
                    List<string> atomTypeChoice = _selectedButton.GetAtomTypeChoices();
                    if (!atomTypeChoice.Contains(_atomTypeJSON.val))
                    {
                        _atomTypeJSON.val = atomTypeChoice[0];
                    }
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchCloseGameUIonPlayMode(bool toggle)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._closeGameUIonPlayMode = toggle;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchOpenGameUIonEditMode(bool toggle)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._openGameUIonEditMode = toggle;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchForceMergeLoad(bool pluginForceMergeLoad)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._pluginForceMergeLoad = pluginForceMergeLoad;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchClothingTagsToggle(string clothingTag)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    string tag = clothingTag.Substring(2, clothingTag.Length-4);
                    _selectedButton.UpdateTagItemToRemove(tag);
                    RefreshDisplay();
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAutoLabel(bool switchAutoLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._autoLabel = switchAutoLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSurpressClothingLoad(bool suppressClothingLoad) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._suppressClothingLoad = suppressClothingLoad;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSurpressPersonScaleLoad(bool suppressPersonScaleLoad) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._suppressPersonScaleLoad = suppressPersonScaleLoad;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchMergeLoadPreset(bool mergeLoadPreset)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._mergeLoadPreset = mergeLoadPreset;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetPluginType(string targetPluginType) {
            try {
                if (!_isDisplayUpdating) {
                    if (targetPluginType != "No plugins in Scene") {
                        _selectedButton._pluginType = targetPluginType;
                        _selectedButton.UpdateButtonType(_selectedButton._buttonType);
                        _selectedButton._pluginVariablesButtonOn._pluginVarDict.Clear();
                        _selectedButton._pluginVariablesButtonOff._pluginVarDict.Clear();
                        _pluginVariablesDisplayOn._pluginVariableSelectorJSON.val = "Select Parameter...";
                        _pluginVariablesDisplayOff._pluginVariableSelectorJSON.val = "Select Parameter...";
                        RefreshDisplay();
                    }
                    else { _selectedButton._pluginType = ""; }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetParamNameOn(string targetParamChoice) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._targetParamNameOn = targetParamChoice;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetParamNameOff(string targetParamChoice) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._targetParamNameOff = targetParamChoice;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSaveDataPlugin(string pluginChoice) {
            try {
                if (!_isDisplayUpdating) {
                    if (pluginChoice == "<Clear Plugin Data>") {
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = "";
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = null;
                    }
                    else if (pluginChoice != "<None available>") {
                        string pluginName = pluginChoice.Substring(0, pluginChoice.IndexOf('['));
                        string atomName = pluginChoice.Substring(pluginChoice.IndexOf('[') + 1);
                        JSONArray pluginDataJSON = _selectedButton.GetPluginDataByName(atomName.Remove(atomName.Length - 1), pluginName);
                        string pluginData = pluginDataJSON.ToString();
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = pluginData.Substring(0, Math.Min(pluginData.Length, 500));
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = pluginDataJSON;

                    }
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchUseButtonTexture(bool useTexture) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._useButtonTexture = useTexture;
                    _mvrScript.RefreshWristUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchSingleInstancePerAtom(bool singleInstance) {
            try {
                if (!_isDisplayUpdating) {
                    if (_selectedButton._buttonType == "Load Plugin(s)") {
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._singleInstancePerAtom = singleInstance;
                    }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPluginFilePath(string filePath)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath = filePath;
                    _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = "";
                    _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = null;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFilePath1(string filePath) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.SetPreset1File(filePath);
                    
                    
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFilePath2(string filePath)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton.SetPreset2File(filePath);
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFilePath3(string filePath)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton.SetPreset3File(filePath);
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPreset1Type(string presetType)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._preset1Type = presetType;

                    if (presetType.EndsWith("Folder"))
                    {
                        _selectedButton._buttonTexture = null;
                        if (FileManagerSecure.FileExists(_selectedButton._preset1FilePath))_preset1FilePathJSON.val = FileManagerSecure.GetDirectoryName(_selectedButton._preset1FilePath);
                        if (!FileManagerSecure.DirectoryExists(_preset1FilePathJSON.val)) _preset1FilePathJSON.val = "";
                    }
                    else if (presetType == "Single File")
                    {
                        if (!FileManagerSecure.FileExists(_selectedButton._preset1FilePath)) _preset1FilePathJSON.val = "";
                    }

                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPreset2Type(string presetType)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._preset2Type = presetType;
                    if (presetType.EndsWith("Folder"))
                    {
                        if (FileManagerSecure.FileExists(_selectedButton._preset2FilePath)) _preset2FilePathJSON.val = FileManagerSecure.GetDirectoryName(_selectedButton._preset2FilePath);
                        if (!FileManagerSecure.DirectoryExists(_preset2FilePathJSON.val)) _preset2FilePathJSON.val = "";
                    }
                    else if (presetType == "Single File")
                    {
                        if (!FileManagerSecure.FileExists(_selectedButton._preset2FilePath)) _preset2FilePathJSON.val = "";
                    }
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPreset3Type(string presetType)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._preset3Type = presetType;
                    if (presetType.EndsWith("Folder"))
                    {
                        if (FileManagerSecure.FileExists(_selectedButton._preset3FilePath)) _preset3FilePathJSON.val = FileManagerSecure.GetDirectoryName(_selectedButton._preset3FilePath);
                        if (!FileManagerSecure.DirectoryExists(_preset3FilePathJSON.val)) _preset3FilePathJSON.val = "";
                    }
                    else if (presetType == "Single File")
                    {
                        if (!FileManagerSecure.FileExists(_selectedButton._preset3FilePath)) _preset3FilePathJSON.val = "";
                    }
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPluginLoad(string pluginLoad) {
            try {
                if (!_isDisplayUpdating) {
                    if (pluginLoad=="Add Plugin...")
                    {
                        ButtonPluginLoad newPluginLoad = new ButtonPluginLoad();
                        _selectedButton._pluginLoadList.Add(newPluginLoad);
                        _isDisplayUpdating = true;
                        _pluginLoadSelectorJSON.val = _selectedButton._pluginLoadList.Count.ToString();
                        _isDisplayUpdating = false ;
                    }
                    else if (pluginLoad.StartsWith("Remove Plugin"))
                    {
                        _selectedButton._pluginLoadList.RemoveAt(int.Parse(pluginLoad.Substring(14)) -1);
                        _isDisplayUpdating = true;
                        _pluginLoadSelectorJSON.val = "1";
                        _isDisplayUpdating = false;
                    }
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButtonLabelOn(string buttonLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._buttonLabelOn = buttonLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButtonLabelOff(string buttonLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._buttonLabelOff = buttonLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButonCategory(string buttonCategory) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateButtonCategory(buttonCategory);
                    _selectedButton._buttonTexture = null;
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshWristUIButtonGrid();
                    _testSpawnPositionTargetJSON.val = "Select...";
                    if (_isTestSpawnPositionActive) EndSpawnTest();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButonType(string buttonType) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateButtonType(buttonType);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshWristUIButtonGrid();
                    _testSpawnPositionTargetJSON.val = "Select...";
                    if (_isTestSpawnPositionActive) EndSpawnTest();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetCategory(string atomCategory) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateTargetCategory(atomCategory);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetName(string atomName) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateTargetName(atomName);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDecalFaceLoad(bool toggleValue)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._decalFaceLoad = toggleValue;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDecalBodyLoad(bool toggleValue)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._decalBodyLoad = toggleValue;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDecalLimbsLoad(bool toggleValue)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._decalLimbsLoad = toggleValue;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDecalGenLoad(bool toggleValue)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._decalGenLoad = toggleValue;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchHairColor(string stringValue)
        {
            try
            {
               if (!_isDisplayUpdating)
                {
                    if (stringValue=="Clear Hair Color")
                    {
                        _selectedButton._hairSimControl = null;
                        _hairSimControlJSON.val = "";
                        RefreshDisplay();
                    }
                    else if (stringValue!="No Sim Hair available")
                    {
                        string atomName = stringValue.Substring(0, stringValue.IndexOf('>'));
                        string hairName = stringValue.Substring(atomName.Length+1);
                        
                        Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                        if (atom != null)
                        {
                            foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                            {
                                HairSimControl hsc = hairGroup.GetComponentInChildren<HairSimControl>();
                                if (hairGroup.displayName == hairName && hsc != null)
                                {
                                    
                                    if (_selectedButton._hairSimControl==null)_selectedButton._hairSimControl = new JSONClass();
                                    _selectedButton._hairSimControl=hsc.GetJSON(true, true, true);
                                   
                                    
                                    break;
                                }
                            }
                            RefreshDisplay();
                        }
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchScalpColor(string stringValue)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    if (!_isDisplayUpdating)
                    {
                        if (stringValue == "Clear Scalp Color")
                        {
                            _selectedButton._hairScalpControl = null;
                            _hairSimControlJSON.valNoCallback = "";
                            RefreshDisplay();
                        }
                        else if (stringValue != "No Sim Hair available")
                        {
                            string atomName = stringValue.Substring(0, stringValue.IndexOf('>'));
                            string hairName = stringValue.Substring(atomName.Length+1);
                            
                            Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                            if (atom != null)
                            {
                                foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                                {
                                    if (hairGroup.displayName == hairName && hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>() != null)
                                    {
                                        _selectedButton._hairScalpControl = new JSONClass();
                                        _selectedButton._hairScalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>().GetJSON(true, true, true);
                                        RefreshDisplay();
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void RefreshDisplay() {
            CloseAllPopup(null);
            RefreshDisplayValues();
            if (_buttonCategoryPopup != null) { _mvrScript.RemovePopup(_buttonCategoryPopup); }
            if (_buttonTypePopup != null) { _mvrScript.RemovePopup(_buttonTypePopup); }
            if (_targetCategoryPopup != null) { _mvrScript.RemovePopup(_targetCategoryPopup); }
            if (_targetNamePopup != null) { _mvrScript.RemovePopup(_targetNamePopup); }
            if (_tagItemsToRemoveSelectorJSON != null) { _mvrScript.RemovePopup(_tagItemsToRemoveSelectorJSON); }
            if (_pluginLoadSelectorPopup != null) { _mvrScript.RemovePopup(_pluginLoadSelectorPopup); }
            if (_filePath1TextField != null) { _mvrScript.RemoveTextField(_filePath1TextField); }
            if (_filePath2TextField != null) { _mvrScript.RemoveTextField(_filePath2TextField); }
            if (_filePath3TextField != null) { _mvrScript.RemoveTextField(_filePath3TextField); }
            if (_pluginFilePathTextField != null) { _mvrScript.RemoveTextField(_pluginFilePathTextField); }
            if (_preset1TypePopup != null) _mvrScript.RemovePopup(_preset1TypePopup);
            if (_preset2TypePopup != null) _mvrScript.RemovePopup(_preset2TypePopup);
            if (_preset3TypePopup != null) _mvrScript.RemovePopup(_preset3TypePopup);
            if (_testSpawnPositionTargetPopup != null) _mvrScript.RemovePopup(_testSpawnPositionTargetPopup);

            if (_selectFile1Button != null) { _selectFile1Button.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_selectFile1Button); }
            if (_selectFile2Button != null) { _selectFile2Button.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_selectFile2Button); }
            if (_selectFile3Button != null) { _selectFile3Button.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_selectFile3Button); }
            if (_testSpawnPositionButton != null) { _testSpawnPositionButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_testSpawnPositionButton); }
            if (_copyButtonButton != null) { _copyButtonButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_copyButtonButton); }
            if (_selectPluginFileButton != null) { _selectPluginFileButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_selectPluginFileButton); }
            
            if (_buttonLabelOnTextField != null) { _mvrScript.RemoveTextField(_buttonLabelOnTextField); }
            if (_buttonLabelOffTextField != null) { _mvrScript.RemoveTextField(_buttonLabelOffTextField); }
            if (_singleInstancePerAtomToggle != null) { _mvrScript.RemoveToggle(_singleInstancePerAtomToggle); }
            if (_pluginSaveDataSelectorPopup != null) { _mvrScript.RemovePopup(_pluginSaveDataSelectorPopup); }
            if (_pluginForceMergeLoadToggle != null) { _mvrScript.RemoveToggle(_pluginForceMergeLoadToggle); }
            if (_pluginSaveDataTextField != null) { _mvrScript.RemoveTextField(_pluginSaveDataTextField); }
            if (_useButtonTextureToggle != null) { _mvrScript.RemoveToggle(_useButtonTextureToggle); }
            if (_targetParamNameOnPopup != null) { _mvrScript.RemovePopup(_targetParamNameOnPopup); }
            if (_targetParamNameOffPopup != null) { _mvrScript.RemovePopup(_targetParamNameOffPopup); }
            if (_pluginSaveDataSelectorPopup != null) { _mvrScript.RemovePopup(_pluginSaveDataSelectorPopup); }
            if (_targetPluginTypePopup != null) { _mvrScript.RemovePopup(_targetPluginTypePopup); }
            if (_suppressClothingLoadToggle != null) { _mvrScript.RemoveToggle(_suppressClothingLoadToggle); }
            if (_suppressPersonScaleLoadToggle != null) { _mvrScript.RemoveToggle(_suppressPersonScaleLoadToggle); }
            if (_mergeLoadPresetToggle != null) { _mvrScript.RemoveToggle(_mergeLoadPresetToggle); }
            if (_autoLabelToggle != null) { _mvrScript.RemoveToggle(_autoLabelToggle); }
            if (_closeGameUIonPlayModeToggle != null) { _mvrScript.RemoveToggle(_closeGameUIonPlayModeToggle); }
            if (_openGameUIonEditModeToggle != null) { _mvrScript.RemoveToggle(_openGameUIonEditModeToggle); }
            if (_parentLinkToTargetToggle != null) { _mvrScript.RemoveToggle(_parentLinkToTargetToggle); }
            if (_testSpawnPositionDisableCollisionToggle != null) { _mvrScript.RemoveToggle(_testSpawnPositionDisableCollisionToggle); }

            if (_decalBodyLoadToggle != null) { _mvrScript.RemoveToggle(_decalBodyLoadToggle); }
            if (_decalFaceLoadToggle != null) { _mvrScript.RemoveToggle(_decalFaceLoadToggle); }
            if (_decalGenLoadToggle != null) { _mvrScript.RemoveToggle(_decalGenLoadToggle); }
            if (_decalLimbsLoadToggle != null) { _mvrScript.RemoveToggle(_decalLimbsLoadToggle); }

            if (_atomCategoryPopup != null) { _mvrScript.RemovePopup(_atomCategoryPopup); }
            if (_atomTypePopup != null) { _mvrScript.RemovePopup(_atomTypePopup); }
            if (_targetNodePopup != null) { _mvrScript.RemovePopup(_targetNodePopup); }
            if (_atomPositionRelativeToPopup != null) { _mvrScript.RemovePopup(_atomPositionRelativeToPopup); }
            
            if (_atomPositionXSlider != null) { _mvrScript.RemoveSlider(_atomPositionXSlider); }
            if (_atomPositionYSlider != null) { _mvrScript.RemoveSlider(_atomPositionYSlider); }
            if (_atomPositionZSlider != null) { _mvrScript.RemoveSlider(_atomPositionZSlider); }
            if (_atomRotationXSlider != null) { _mvrScript.RemoveSlider(_atomRotationXSlider); }
            if (_atomRotationYSlider != null) { _mvrScript.RemoveSlider(_atomRotationYSlider); }
            if (_atomRotationZSlider != null) { _mvrScript.RemoveSlider(_atomRotationZSlider); }
            if (_atomAbsoluteHeightToggle != null) { _mvrScript.RemoveToggle(_atomAbsoluteHeightToggle); }
            if (_atomAbsolutePitchToggle != null) { _mvrScript.RemoveToggle(_atomAbsolutePitchToggle); }
            if (_atomAbsoluteRollToggle != null) { _mvrScript.RemoveToggle(_atomAbsoluteRollToggle); }

            if (_lockedPositionXToggle != null) _mvrScript.RemoveToggle(_lockedPositionXToggle);
            if (_lockedPositionYToggle != null) { _mvrScript.RemoveToggle(_lockedPositionYToggle); }
            if (_lockedPositionZToggle != null) { _mvrScript.RemoveToggle(_lockedPositionZToggle); }
            if (_lockedRotationXToggle != null) { _mvrScript.RemoveToggle(_lockedRotationXToggle); }
            if (_lockedRotationYToggle != null) { _mvrScript.RemoveToggle(_lockedRotationYToggle); }
            if (_lockedRotationZToggle != null) { _mvrScript.RemoveToggle(_lockedRotationZToggle); }
            if (_vrMoveDisablePhysicsToggle != null) { _mvrScript.RemoveToggle(_vrMoveDisablePhysicsToggle); }            

            if (_vrMovePositionGranulatityPopup != null) _mvrScript.RemovePopup(_vrMovePositionGranulatityPopup);
            if (_vrMoveRotationGranulatityPopup != null) _mvrScript.RemovePopup(_vrMoveRotationGranulatityPopup);

            if (_hairSimSelectorPopup != null) _mvrScript.RemovePopup(_hairSimSelectorPopup);
            if (_hairScalpSelectorPopup != null) _mvrScript.RemovePopup(_hairScalpSelectorPopup);
            if (_hairSimControlTextField != null) { _mvrScript.RemoveTextField(_hairSimControlTextField); }
            if (_hairScalpControlTextField != null) { _mvrScript.RemoveTextField(_hairScalpControlTextField); }

            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            if (_spacer1b != null) { _mvrScript.RemoveSpacer(_spacer1b); }
            if (_spacer1c != null) { _mvrScript.RemoveSpacer(_spacer1c); }
            if (_spacer1d != null) { _mvrScript.RemoveSpacer(_spacer1d); }
            if (_spacer2 != null) { _mvrScript.RemoveSpacer(_spacer2); }
            if (_spacer2b != null) { _mvrScript.RemoveSpacer(_spacer2b); }
            if (_spacer3 != null) { _mvrScript.RemoveSpacer(_spacer3); }
            if (_gridDeleteButton != null) { _gridDeleteButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_gridDeleteButton); }
            if (_gridDeleteConfirmButton != null) { _gridDeleteConfirmButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_gridDeleteConfirmButton); }
            if (_gridMoveUpButton != null) { _gridMoveUpButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_gridMoveUpButton); }
            if (_gridMoveDownButton != null) { _gridMoveDownButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_gridMoveDownButton); }
            if (_buttonSkinDisplay != null) _buttonSkinDisplay.ClearDisplay();
            if (_pluginVariablesDisplayOn != null) _pluginVariablesDisplayOn.ClearDisplay();
            if (_pluginVariablesDisplayOff != null) _pluginVariablesDisplayOff.ClearDisplay();
            _dmFaceDecalDisplay.ClearDisplay();
            _dmTorsoDecalDisplay.ClearDisplay();
            _dmLimbsDecalDisplay.ClearDisplay();
            _dmGenDecalDisplay.ClearDisplay();

            _buttonCategoryPopup = _mvrScript.CreateSetupPopup(_buttonCategoryJSON, true, 500f, ButtonCategoryPopupOpen, "Button " + _selectedButton._buttonRef.ToString() + " Category",false);
            _buttonTypePopup = _mvrScript.CreateSetupPopup(_buttonTypeJSON, true, 500f, UpdateButtonTypeChoices, "Button " + _selectedButton._buttonRef.ToString() + " Operation",false);

            if (_selectedButton._buttonType != "Blank") {
                _autoLabelToggle = _mvrScript.CreateToggle(_autoLabelJSON, true);
                _autoLabelToggle.label = "Auto Label Button " + _selectedButton._buttonRef.ToString();
                if (!_selectedButton._autoLabel) {
                    _buttonLabelOnTextField = _mvrScript.CreateTextInput(_buttonLabelOnJSON, true);
                    if (_selectedButton._maxButtonStates > 1) {
                        _buttonLabelOffTextField = _mvrScript.CreateTextInput(_buttonLabelOffJSON, true);
                    }
                }

                if (_selectedButton._buttonType == "Toggle Play/Edit Mode")
                {
                    _spacer1 = _mvrScript.CreateSpacer(true);
                    _spacer1.height = 10;
                    _closeGameUIonPlayModeToggle = _mvrScript.CreateToggle(_closeGameUIonPlayModeJSON, true);
                    _closeGameUIonPlayModeToggle.label = "Close Game UI for Play Mode";
                    _openGameUIonEditModeToggle = _mvrScript.CreateToggle(_openGameUIonEditModeJSON, true);
                    _openGameUIonEditModeToggle.label = "Open Game UI for Edit Mode";
                }
                if (_selectedButton._buttonType.StartsWith("Spawn Atom") )
                {
                    _atomCategoryPopup = _mvrScript.CreateSetupPopup(_atomCategoryJSON, true, 400, UpdateAtomCategoryChoices, "New Atom Category", false);
                    if (_selectedButton._atomCategory!="People") _atomTypePopup = _mvrScript.CreateSetupPopup(_atomTypeJSON, true, 400, UpdateAtomTypeChoices, "New Atom Type");
                }

                // Target Selection
                if (_selectedButton.IsAtomTargetable())
                {
                    _spacer1 = _mvrScript.CreateSpacer(true);
                    _spacer1.height = 10;
                    if (!_selectedButton._buttonType.StartsWith("Spawn Atom") && _selectedButton._buttonType != "Active Clothing Editor")
                    {
                        _targetCategoryPopup = _mvrScript.CreateSetupPopup(_targetCategoryJSON, true, 500f, UpdateTargetCategoryChoices, "Target Category",false);
                        _targetNamePopup = _mvrScript.CreateSetupPopup(_targetNameJSON, true, 500f, UpdateTargetChoices, "Button Target");
                    }
                    else if (_selectedButton._buttonType =="Active Clothing Editor")
                    {
                        _targetNamePopup = _mvrScript.CreateSetupPopup(_targetNameJSON, true, 500f, UpdateTargetChoices, "Button Target");
                    }

                    if (_selectedButton._buttonType.StartsWith("Spawn Atom") || _selectedButton._buttonType.StartsWith("Teleport Atom"))
                    {
                        string actionVerb = "Spawn";
                        if (_selectedButton._buttonType.StartsWith("Teleport Atom")) actionVerb = "Teleport";
                        _spacer1b = _mvrScript.CreateSpacer(true);
                        _spacer1b.height = 10;
                        _atomPositionRelativeToPopup = _mvrScript.CreateSetupPopup(_atomPositionRelativeToJSON, true, 500f, UpdateMoveRelativeToChoices, actionVerb+" relative to:", false);
                        if (_atomPositionRelativeToJSON.val.StartsWith("Last"))
                        {
                            _targetNodePopup = _mvrScript.CreateSetupPopup(_targetNodeJSON, true, 400, UpdateTargetNodeChoices, actionVerb+" relative to node:", true);
                            _parentLinkToTargetToggle = _mvrScript.CreateToggle(_parentLinkToTargetJSON, true);
                            _parentLinkToTargetToggle.label = "Parent Link new Atom to node";
                        }
                        _atomPositionXSlider = _mvrScript.CreateSlider(_atomPositionXJSON, true);
                        _atomPositionXSlider.label = "Position X (relative)";
                        
                        if (_selectedButton._atomPositionRelativeTo != "Scene Origin")
                        {
                            _atomAbsoluteHeightToggle = _mvrScript.CreateToggle(_atomAbsoluteHeightJSON, true);
                            _atomAbsoluteHeightToggle.label = "Fixed Height (Y position)";
                        }
                        _atomPositionYSlider = _mvrScript.CreateSlider(_atomPositionYJSON, true);
                        if (_selectedButton._atomAbsoluteHeight && _selectedButton._atomPositionRelativeTo != "Scene Origin") _atomPositionYSlider.label = "Position Y (absolute)";
                        else _atomPositionYSlider.label = "Position Y (relative)";
                        
                        _atomPositionZSlider = _mvrScript.CreateSlider(_atomPositionZJSON, true);
                        _atomPositionZSlider.label = "Position Z (relative)";

                        if (_selectedButton._atomPositionRelativeTo != "Scene Origin")
                        {
                            _atomAbsolutePitchToggle = _mvrScript.CreateToggle(_atomAbsolutePitchJSON, true);
                            _atomAbsolutePitchToggle.label = "Fixed Pitch (X rotation)";
                        }
                        _atomRotationXSlider = _mvrScript.CreateSlider(_atomRotationXJSON, true);
                        if (_selectedButton._atomAbsolutePitch && _selectedButton._atomPositionRelativeTo != "Scene Origin") _atomRotationXSlider.label = "Rotation X (absolute)";
                        else _atomRotationXSlider.label = "Rotation X (relative)";

                        _atomRotationYSlider = _mvrScript.CreateSlider(_atomRotationYJSON, true);
                        _atomRotationYSlider.label = "Rotation Y (relative)";

                        if (_selectedButton._atomPositionRelativeTo != "Scene Origin")
                        {
                            _atomAbsoluteRollToggle = _mvrScript.CreateToggle(_atomAbsoluteRollJSON, true);
                            _atomAbsoluteRollToggle.label = "Fixed Roll (Z rotation)";
                        }
                        _atomRotationZSlider = _mvrScript.CreateSlider(_atomRotationZJSON, true);
                        if (_selectedButton._atomAbsoluteRoll && _selectedButton._atomPositionRelativeTo != "Scene Origin") _atomRotationZSlider.label = "Rotation Z (absolute)";
                        else _atomRotationZSlider.label = "Rotation Z (relative)";
                        _testSpawnPositionTargetPopup = _mvrScript.CreateSetupPopup(_testSpawnPositionTargetJSON, true,300f,UpdateTestSpawnPosTargetChoices,"Test Atom "+actionVerb+" Position on" );

                        if (_testSpawnPositionTargetJSON.val!="Select...")
                        {
                            _testSpawnPositionDisableCollisionToggle = _mvrScript.CreateToggle(_testSpawnPositionDisableCollisionJSON, true);
                            _testSpawnPositionDisableCollisionToggle.label = "Disable Atom Collision for test";
                            if (_isTestSpawnPositionActive)
                            {
                                _testSpawnPositionButton = _mvrScript.CreateButton("End Test", true);
                                _testSpawnPositionButton.button.onClick.AddListener(() => { EndSpawnTest(); });
                            }

                            else
                            {
                                _testSpawnPositionButton = _mvrScript.CreateButton("Start Test", true);
                                _testSpawnPositionButton.button.onClick.AddListener(() => { StartSpawnTest(); });
                            }

                        }
                    }
                }

                if (_selectedButton._buttonType == "Move Atom (VR Only)")
                {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _vrMovePositionGranulatityPopup = _mvrScript.CreateSetupPopup(_vrMovePositionGranulatityJSON, true, 300f, UpdateVRMovePositionGranulatityChoices, "Position Granularity", false);
                    

                    _lockedPositionXToggle = _mvrScript.CreateToggle(_lockedPositionXJSON, true);
                    _lockedPositionXToggle.label = "Lock X Axis Position";
                    _lockedPositionYToggle = _mvrScript.CreateToggle(_lockedPositionYJSON, true);
                    _lockedPositionYToggle.label = "Lock Y Axis Position (Height)";
                    _lockedPositionZToggle = _mvrScript.CreateToggle(_lockedPositionZJSON, true);
                    _lockedPositionZToggle.label = "Lock Z Axis Position";

                    _vrMoveRotationGranulatityPopup = _mvrScript.CreateSetupPopup(_vrMoveRotationGranulatityJSON, true, 300f, UpdateVRMoveRotationGranulatityChoices, "Rotation Granularity", false);
                    _lockedRotationXToggle = _mvrScript.CreateToggle(_lockedRotationXJSON, true);
                    _lockedRotationXToggle.label = "Lock X Axis Rotation (Pitch)";
                    _lockedRotationYToggle = _mvrScript.CreateToggle(_lockedRotationYJSON, true);
                    _lockedRotationYToggle.label = "Lock Y Axis Rotation (Yaw)";
                    _lockedRotationZToggle = _mvrScript.CreateToggle(_lockedRotationZJSON, true);
                    _lockedRotationZToggle.label = "Lock Z Axis Rotation (Roll)";

                    _vrMoveDisablePhysicsToggle = _mvrScript.CreateToggle(_vrMoveDisablePhysicsJSON, true);
                    _vrMoveDisablePhysicsToggle.label = "Disable Atom Physics while moving";
                    
                }

                if (_selectedButton._buttonType == "Decal Maker Toggle")
                {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _dmFaceDecalDisplay.RefreshDisplay("Face");
                    _dmTorsoDecalDisplay.RefreshDisplay("Torso");
                    _dmLimbsDecalDisplay.RefreshDisplay("Limb");
                    _dmGenDecalDisplay.RefreshDisplay("Genital");
                }
                if (_selectedButton._buttonType=="Set Hair Color")
                {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _hairSimSelectorPopup = _mvrScript.CreateSetupPopup(_hairSimSelectorJSON,true,300f,UpdateHairSimChoices);
                    if (_selectedButton._hairSimControl != null) _hairSimControlTextField = _mvrScript.CreateTextField(_hairSimControlJSON, true);
                    
                    _hairScalpSelectorPopup = _mvrScript.CreateSetupPopup (_hairScalpSelectorJSON, true, 300f, UpdateHairScalpChoices);
                    if (_selectedButton._hairScalpControl != null) _hairScalpControlTextField= _mvrScript.CreateTextField(_hairScalpControlJSON, true);

                }

                    // Plugin Settings
                if (_selectedButton._buttonCategory == "Plugins" && _selectedButton._buttonType != "Load Plugin(s)" && _selectedButton._buttonType != "Unload All Plugins" && _selectedButton._buttonType != "Load Plugins Preset") {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _targetPluginTypePopup = _mvrScript.CreateSetupPopup(_targetPluginTypeJSON, true, 300f, UpdatePluginTypeChoices, "Plugin Type");
                    if (_selectedButton._pluginType != "") {
                        if (_selectedButton._buttonType == "Plugin Action Toggle" || _selectedButton._buttonType == "Plugin Action") {
                            if (_selectedButton._buttonType == "Plugin Action Toggle")
                            {
                                _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Button ON Plugin Action");
                                if (_targetParamNameOnJSON.val !="No Actions available") _pluginVariablesDisplayOn.RefreshDisplay("Button ON ");
                            }
                            else
                            {
                                _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Select Plugin Action");
                                if (_targetParamNameOnJSON.val != "No Actions available") _pluginVariablesDisplayOn.RefreshDisplay("");
                            }                                                       
                        }

                        else if (_selectedButton._buttonType == "Plugin Bool Toggle" || _selectedButton._buttonType == "Set Plugin Bool True" || _selectedButton._buttonType == "Set Plugin Bool False") {
                            _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Select Plugin Bool");
                        }
                        if (_selectedButton._buttonType == "Plugin Action Toggle") {
                            _targetParamNameOffPopup = _mvrScript.CreateSetupPopup(_targetParamNameOffJSON, true, 300f, UpdatePluginParamOffSelectorChoices, "Button OFF Plugin Action");
                            if (_targetParamNameOffJSON.val != "No Actions available") _pluginVariablesDisplayOff.RefreshDisplay("Button OFF ");
                        }
                    }
                }

                // Preset and File Selection
                if (_selectedButton.isPresetType || _selectedButton._buttonType == "Spawn Atom" || _selectedButton.isSceneLoadType || _selectedButton.isUIAPLoadType) {
                    _spacer1d = _mvrScript.CreateSpacer(true);
                    _spacer1d.height = 10;
                    
                    string selectFile1ButtonLabel = "Select Preset File:";
                    string selectPresetTypeLabel = "Preset Type";
                    string prefixRemoval = null;
                    string fileType = "vap";                   
                    
                    if (_selectedButton._buttonType == "Spawn Atom" && _selectedButton._atomType == "Person")
                    {
                        selectFile1ButtonLabel = "Select Appearance Preset";
                        selectPresetTypeLabel = "App Preset Type";
                    }
                    else if (_selectedButton.isSceneLoadType)
                    {
                        selectFile1ButtonLabel = "Select Scene";
                        selectPresetTypeLabel = "Scene Selection Type";
                        fileType = "json";
                    }
                    else if (_selectedButton.isLegacyPresetLoadSaveType)
                    {
                        fileType = "json";
                    }
                    else if (_selectedButton.isUIAPLoadType)
                    {
                        selectFile1ButtonLabel = "Select UIAP Profile";
                        selectPresetTypeLabel = "UIAP Selection Type";
                        fileType = "uiap";
                    }
                    else if (_selectedButton.isNonLegacyPresetLoadType)
                    {
                        prefixRemoval = "Preset_";
                    }
                    if (Constants._patreonContentEnabled) _preset1TypePopup = _mvrScript.CreateSetupPopup(_preset1TypeJSON, true, 300, UpdatePresetType1Choices, selectPresetTypeLabel,false);
                    if (_selectedButton._preset1Type != "None")
                    {
                        
                        _selectFile1Button = _mvrScript.CreateButton(selectFile1ButtonLabel, true);
                        List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(_selectedButton.GetDefaultPresetFolder(1));
                        if (_selectedButton._preset1Type.EndsWith("Folder"))
                        {
                            _selectFile1Button.label = selectFile1ButtonLabel + " Folder:";
                            _selectFile1Button.button.onClick.AddListener(() => { SuperController.singleton.GetDirectoryPathDialog(File1Selected, _selectedButton.GetDefaultPresetFolder(1), shortcuts, false); });
                        }
                        else _selectFile1Button.button.onClick.AddListener(() => {
                            SuperController.singleton.GetMediaPathDialog(File1Selected, fileType, _selectedButton.GetDefaultPresetFolder(1), false, true, false, prefixRemoval, false, shortcuts);
                            if (_selectedButton.isLegacyPresetSaveType)
                            {
                                // Update the browser to be a Save browser
                                uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                                browser.SetTextEntry(true);
                                browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), "json");
                                browser.ActivateFileNameField();
                            }
                        });
                        _filePath1TextField = _mvrScript.CreateTextField(_preset1FilePathJSON, true);
                        _filePath1TextField.height = 1;
                        if (_selectedButton.isNonLegacyPresetLoadType)
                        {
                            _mergeLoadPresetToggle = _mvrScript.CreateToggle(_mergeLoadPresetJSON, true);
                            _mergeLoadPresetToggle.label = "Use Merge Load";
                        }
                        if (_selectedButton._buttonType == "Load App Preset")
                        {
                            _suppressClothingLoadToggle = _mvrScript.CreateToggle(_suppressClothingLoadJSON, true);
                            _suppressClothingLoadToggle.label = "Suppress Clothing Load";
                        }
                        if (Constants._patreonContentEnabled)
                        {
                            if (_selectedButton._buttonType == "Load Skin Preset")
                            {
                                _decalFaceLoadToggle = _mvrScript.CreateToggle(_decalFaceLoadJSON, true);
                                _decalFaceLoadToggle.label = "Limit preset load to Face decal";
                                _decalBodyLoadToggle = _mvrScript.CreateToggle(_decalBodyLoadJSON, true);
                                _decalBodyLoadToggle.label = "Limit preset load to Body decal";
                                _decalLimbsLoadToggle = _mvrScript.CreateToggle(_decalLimbsLoadJSON, true);
                                _decalLimbsLoadToggle.label = "Limit preset load to Limbs decal";
                                _decalGenLoadToggle = _mvrScript.CreateToggle(_decalGenLoadJSON, true);
                                _decalGenLoadToggle.label = "Limit preset load to Gen decal";

                            }
                            if (_selectedButton._buttonType == "Load App Preset")
                            {
                                _suppressPersonScaleLoadToggle = _mvrScript.CreateToggle(_suppressPersonScaleLoadJSON, true);
                                _suppressPersonScaleLoadToggle.label = "Suppress Person Scale Load";
                            }
                            if (_selectedButton._preset1Type != "Choose from Folder" && _selectedButton._preset1Type != "Random from Folder" && _selectedButton._buttonType != "Load UI Assist Profile")
                            {
                                _useButtonTextureToggle = _mvrScript.CreateToggle(_useButtonTextureJSON, true);
                                _useButtonTextureToggle.label = "Use Preset Thumbnail image";
                            }
                            if (_selectedButton._buttonType == "Merge Clothing Preset")
                            {
                                _tagItemsToRemoveSelectorPopup = _mvrScript.CreateSetupPopup(_tagItemsToRemoveSelectorJSON, true, 600f, UpdateTagItemChoices, null);
                            }
                        }
                    }
                    if (_selectedButton.isSceneLoadType && _selectedButton._preset1Type=="Choose from Folder")
                    {
                        _selectFile2Button = _mvrScript.CreateButton("Filter Shortcuts by Addons Folder:", true);
                        _selectFile2Button.button.onClick.AddListener(() => {
                            
                            SuperController.singleton.GetDirectoryPathDialog(File2Selected, "AddonPackages", null, false);
                            uFileBrowser.FileBrowser browser = SuperController.singleton.directoryBrowserUI;
                            browser.browseVarFilesAsDirectories = false;

                        });
                        _filePath2TextField = _mvrScript.CreateTextField(_preset2FilePathJSON, true);
                        _filePath2TextField.height = 1;
                    }
                    
                    if (_selectedButton._buttonType == "Spawn Atom" && _selectedButton._atomType == "Person")
                    {
                        if (Constants._patreonContentEnabled)
                        {
                            _preset3TypePopup = _mvrScript.CreateSetupPopup(_preset3TypeJSON, true, 300, UpdatePresetType3Choices, "Pose Preset Type", false);
                            _preset3TypePopup.popup.labelText.resizeTextForBestFit = true;
                        }

                        if (_selectedButton._preset3Type != "None")
                        {
                            if (_selectedButton._preset3Type.EndsWith("Folder")) _selectFile3Button = _mvrScript.CreateButton("Select Pose Preset Folder", true);
                            else _selectFile3Button = _mvrScript.CreateButton("Select Pose Preset", true);
                            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(_selectedButton.GetDefaultPresetFolder(3));
                            if (_selectedButton._preset3Type.EndsWith("Folder")) _selectFile3Button.button.onClick.AddListener(() => { SuperController.singleton.GetDirectoryPathDialog(File3Selected, _selectedButton.GetDefaultPresetFolder(3), shortcuts, false); });
                            else _selectFile3Button.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(File3Selected, "vap", _selectedButton.GetDefaultPresetFolder(3), false, true, false, prefixRemoval, false, shortcuts); });
                            _filePath3TextField = _mvrScript.CreateTextField(_preset3FilePathJSON, true);
                            _filePath3TextField.height = 1;
                        }
                    }
                }
                // Plugin Loading
                if (_selectedButton._buttonType == "Load Plugin(s)" || _selectedButton._buttonType == "Spawn Atom")
                {
                    _spacer1c = _mvrScript.CreateSpacer(true);
                    _spacer1c.height = 10;
                    if (Constants._patreonContentEnabled)
                    {
                        _pluginLoadSelectorPopup = _mvrScript.CreateSetupPopup(_pluginLoadSelectorJSON, true, 500f, PluginLoadSelectorPopupOpen, null);
                        _pluginLoadSelectorPopup.labelTextColor = Color.blue;
                    }

                    _selectPluginFileButton = _mvrScript.CreateButton("Select Plugin File", true);
                    List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Scripts");
                    _selectPluginFileButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(PluginFileSelected, "cs|cslist", "Custom\\Scripts", false, true, false, null, false, shortcuts); });
                    _pluginFilePathTextField = _mvrScript.CreateTextField(_pluginFilePathJSON, true);
                    _pluginFilePathTextField.height = 1;
                    if (_selectedButton._buttonType == "Load Plugin(s)")
                    {
                        _pluginForceMergeLoadToggle = _mvrScript.CreateToggle(_pluginForceMergeLoadJSON, true);
                        _pluginForceMergeLoadToggle.label = "Force Merge Load";
                    }
                    if (_selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath != "")
                    {
                        if (_selectedButton._buttonType == "Load Plugin(s)")
                        {
                            _singleInstancePerAtomToggle = _mvrScript.CreateToggle(_singleInstancePerAtomJSON, true);
                            _singleInstancePerAtomToggle.label = "Limit to 1 plugin instance per Atom";
                        }
                        if (Constants._patreonContentEnabled && _selectedButton._targetName != "CoreControl")
                        {
                            _pluginSaveDataSelectorPopup = _mvrScript.CreateSetupPopup(_pluginSaveDataSelectorJSON, true, 300f, UpdateSaveDataSelectorChoices, "Inital plugin data on Load");

                            _pluginSaveDataTextField = _mvrScript.CreateTextField(_pluginSaveDataJSON, true);
                        }
                    }
                }
                if (Constants._patreonContentEnabled) { _buttonSkinDisplay.RefreshDisplay(); }
            }
            UIAButton prevButton = _mvrScript._buttonGridDisplay._selectedButtonGrid.GetButtonBefore(_selectedButton._buttonRef);
            if (prevButton != null) {
                _spacer2 = _mvrScript.CreateSpacer(true);
                _spacer2.height = 10;
                _copyButtonButton = _mvrScript.CreateButton("Copy all Button " + prevButton._buttonRef + " values to Button " + _selectedButton._buttonRef, true);
                _copyButtonButton.buttonColor = Color.yellow;
                _copyButtonButton.textColor = Color.black;
                _copyButtonButton.button.onClick.AddListener(() => {
                    _selectedButton.CopyFrom(prevButton);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshWristUIButtonGrid();
                    RefreshDisplay(); });
            }


            if (_mvrScript._buttonGridsList.Count > 1 && _mvrScript._buttonGridSelectorJSON.val!="Quick Launch Bar") {
                _spacer2b = _mvrScript.CreateSpacer(true);
                _spacer2b.height = 25;

                int currentGrid = int.Parse(_mvrScript._buttonGridSelectorJSON.val);
                if (currentGrid > 1)
                {
                    _gridMoveDownButton = _mvrScript.CreateButton("Move Screen "+ currentGrid+" Down", true);
                    _gridMoveDownButton.button.onClick.AddListener(() => { _mvrScript.MoveScreenDown(); });
                }
                if (currentGrid < _mvrScript._buttonGridsList.Count)
                {
                    _gridMoveUpButton = _mvrScript.CreateButton("Move Screen " + currentGrid + " Up", true);
                    _gridMoveUpButton.button.onClick.AddListener(() => { _mvrScript.MoveScreenUp(); });
                }

                if (_gridDeleteConfirm)
                {
                    _gridDeleteButton = _mvrScript.CreateButton("Cancel Delete Screen " + currentGrid, true);
                    _gridDeleteButton.buttonColor = Color.green;
                    _gridDeleteButton.textColor = Color.black;
                    _gridDeleteButton.button.onClick.AddListener(() => { _mvrScript.RefreshRightUIDisplay(); });

                    _gridDeleteConfirmButton = _mvrScript.CreateButton("Confirm Delete Screen " + currentGrid, true);
                    _gridDeleteConfirmButton.buttonColor = Color.red;
                    _gridDeleteConfirmButton.textColor = Color.white;
                    _gridDeleteConfirmButton.button.onClick.AddListener(() => { _mvrScript.ConfirmDeleteCurrentScreen(); });

                    
                    _gridDeleteConfirm = false;
                }
                else
                {
                    _gridDeleteButton = _mvrScript.CreateButton("Delete Screen " + currentGrid, true);
                    _gridDeleteButton.buttonColor = Color.red;
                    _gridDeleteButton.textColor = Color.white;

                    _gridDeleteButton.button.onClick.AddListener(() => { _mvrScript.DeleteCurrentScreen(); });
                }
                
            }
            _spacer3 = _mvrScript.CreateSpacer(true);
            _spacer3.height = 400;
            
        }
        public void ButtonCategoryPopupOpen() { _mvrScript.CloseAllRightUIPopups(_buttonCategoryPopup); }
        public void PluginLoadSelectorPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_pluginLoadSelectorPopup);
            List<string> popupChoices = new List<string>();

            for (int i = 1; i <= _selectedButton._pluginLoadList.Count; i++)
            {
                popupChoices.Add(i.ToString());
            }

            popupChoices.Add("Add Plugin...");
            if (_selectedButton._pluginLoadList.Count>1) popupChoices.Add("Remove Plugin "+ _pluginLoadSelectorJSON.val);
            _pluginLoadSelectorJSON.choices = popupChoices;
        }

        public void CloseAllPopup(UIDynamicPopup exception) {
            if (_buttonCategoryPopup != null && _buttonCategoryPopup != exception) { _buttonCategoryPopup.popup.visible = false; }
            if (_buttonTypePopup != null && _buttonTypePopup != exception) { _buttonTypePopup.popup.visible = false; }
            if (_targetCategoryPopup != null && _targetCategoryPopup != exception) { _targetCategoryPopup.popup.visible = false; }
            if (_targetNamePopup != null && _targetNamePopup != exception) { _targetNamePopup.popup.visible = false; }
            if (_pluginLoadSelectorPopup != null && _pluginLoadSelectorPopup != exception) { _pluginLoadSelectorPopup.popup.visible = false; }
            if (_pluginSaveDataSelectorPopup != null && _pluginSaveDataSelectorPopup != exception) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            if (_targetParamNameOnPopup != null && _targetParamNameOnPopup != exception) { _targetParamNameOnPopup.popup.visible = false; }
            if (_targetParamNameOffPopup != null && _targetParamNameOffPopup != exception) { _targetParamNameOffPopup.popup.visible = false; }
            if (_tagItemsToRemoveSelectorPopup != null && _tagItemsToRemoveSelectorPopup != exception) { _tagItemsToRemoveSelectorPopup.popup.visible = false; }
            if (_atomCategoryPopup != null && _atomCategoryPopup != exception) { _atomCategoryPopup.popup.visible = false; }
            if (_atomTypePopup != null && _atomTypePopup != exception) { _atomTypePopup.popup.visible = false; }
            if (_targetNodePopup != null && _targetNodePopup != exception) { _targetNodePopup.popup.visible = false; }
            if (_atomPositionRelativeToPopup != null && _atomPositionRelativeToPopup != exception) { _atomPositionRelativeToPopup.popup.visible = false; }
            
            if (_preset1TypePopup != null && _preset1TypePopup != exception)  _preset1TypePopup.popup.visible = false;
            if (_preset2TypePopup != null && _preset2TypePopup != exception) _preset2TypePopup.popup.visible = false;
            if (_preset3TypePopup != null && _preset3TypePopup != exception) _preset3TypePopup.popup.visible = false;
            if (_testSpawnPositionTargetPopup!=null && _testSpawnPositionTargetPopup!=exception) _testSpawnPositionTargetPopup.popup.visible = false;

            if (_hairScalpSelectorPopup != null && _hairScalpSelectorPopup != exception) _hairScalpSelectorPopup.popup.visible = false;
            if (_hairSimSelectorPopup != null && _hairSimSelectorPopup != exception) _hairSimSelectorPopup.popup.visible = false;

            if (_targetPluginTypePopup != null && _targetPluginTypePopup != exception) { _targetPluginTypePopup.popup.visible = false; }
            if (_pluginSaveDataSelectorPopup != null && _pluginSaveDataSelectorPopup != exception) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            if (_buttonSkinDisplay != null && _buttonSkinDisplay._buttonSkinParamSelectorPopup != null && _buttonSkinDisplay._buttonSkinParamSelectorPopup != exception) { _buttonSkinDisplay._buttonSkinParamSelectorPopup.popup.visible = false; }
            if (_buttonSkinDisplay != null && _buttonSkinDisplay._skinStringChooserParamPopup != null && _buttonSkinDisplay._skinStringChooserParamPopup != exception) { _buttonSkinDisplay._skinStringChooserParamPopup.popup.visible = false; }

            if (_vrMovePositionGranulatityPopup != null && _vrMovePositionGranulatityPopup != exception) _vrMovePositionGranulatityPopup.popup.visible = false;
            if (_vrMoveRotationGranulatityPopup != null && _vrMoveRotationGranulatityPopup != exception) _vrMoveRotationGranulatityPopup.popup.visible = false;

            _dmFaceDecalDisplay.CloseAllPopup(exception);
            _dmTorsoDecalDisplay.CloseAllPopup(exception);
            _dmLimbsDecalDisplay.CloseAllPopup(exception);
            _dmGenDecalDisplay.CloseAllPopup(exception);
            if (_pluginVariablesDisplayOn != null) _pluginVariablesDisplayOn.CloseAllPopup(exception);
            if (_pluginVariablesDisplayOff != null) _pluginVariablesDisplayOff.CloseAllPopup(exception);
        }
       
        protected void PluginFileSelected(string fileName)
        {
            try
            {
                if (fileName != "")
                {
                    if (fileName.IndexOf(':') == 1)
                    {
                        _pluginFilePathJSON.val = fileName.Substring(Constants._vamFilePath.Length, fileName.Length - Constants._vamFilePath.Length);
                    }
                    else { _pluginFilePathJSON.val = fileName; }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void File1Selected(string fileName) {
            try {
                if (fileName != "") {
                    if (fileName.IndexOf(':') == 1) {
                        _preset1FilePathJSON.val = fileName.Substring(Constants._vamFilePath.Length, fileName.Length - Constants._vamFilePath.Length); }
                    else { _preset1FilePathJSON.val = fileName; }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void File2Selected(string fileName)
        {
            try
            {

                if (fileName != "")
                {
                    if (fileName.IndexOf(':') == 1)
                    {
                        _preset2FilePathJSON.val = fileName.Substring(Constants._vamFilePath.Length, fileName.Length - Constants._vamFilePath.Length);
                    }
                    else { _preset2FilePathJSON.val = fileName; }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void File3Selected(string fileName)
        {
            try
            {
                if (fileName != "")
                {
                    if (fileName.IndexOf(':') == 1)
                    {
                        _preset3FilePathJSON.val = fileName.Substring(Constants._vamFilePath.Length, fileName.Length - Constants._vamFilePath.Length);
                    }
                    else { _preset3FilePathJSON.val = fileName; }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateTestSpawnPosTargetChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_testSpawnPositionTargetPopup);
                List<string> popupChoices = new List<string>();
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == _selectedButton._atomType && _selectedButton._buttonType=="Spawn Atom") popupChoices.Add(atom.name);
                    else if (_selectedButton._buttonType == "Teleport Atom") popupChoices.Add(atom.name);

                }
                if (popupChoices.Count == 0) popupChoices.Add("No Atoms of Spawn Type available");

                _testSpawnPositionTargetJSON.choices = popupChoices;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateTagItemChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_tagItemsToRemoveSelectorPopup);
                List<string> popupChoices = new List<string>();
                foreach (string tag in _ClothingTags)
                {
                    if (_selectedButton._tagsToRemoveOnMerge.Contains(tag))
                    {
                        popupChoices.Add("+ "+tag+" +");
                    }
                    else popupChoices.Add("  "+tag+"  ");

                }


                _tagItemsToRemoveSelectorJSON.choices = popupChoices;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void UpdateHairSimChoices()
        {
            try
            {
//                _mvrScript.CloseAllRightUIPopups(_hairSimSelectorPopup);
                List<string> popupChoices = new List<string>();

                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                    {
                        DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
                        HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

                        if (hairControl != null && simControl != null)
                        {
                            popupChoices.Add(atom.name+">"+hairGroup.displayName);
                        }
                    }
                }

                if (popupChoices.Count == 0) popupChoices.Add("No Sim Hair available");
                if (_selectedButton._hairSimControl!=null) popupChoices.Add("Clear Hair Color");
                _hairSimSelectorJSON.choices = popupChoices;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateHairScalpChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_hairScalpSelectorPopup);
                List<string> popupChoices = new List<string>();

                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                    {
                        DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
                        HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

                        if (hairControl != null && simControl != null)
                        {
                            popupChoices.Add(atom.name + ">" + hairGroup.displayName);
                        }
                    }
                }

                if (popupChoices.Count == 0) { popupChoices.Add("No Sim Hair available"); }
                if (_selectedButton._hairScalpControl
                    != null) popupChoices.Add("Clear Scalp Color");
                _hairScalpSelectorJSON.choices = popupChoices;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePluginTypeChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetPluginTypePopup);
                List<string> popupChoices = new List<string>();

                foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                    foreach (string storableName in atom.GetStorableIDs()) {
                        if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#")) {
                            string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                            if (!popupChoices.Contains(pluginType)) { popupChoices.Add(pluginType); }
                        }
                    }
                }

                if (popupChoices.Count == 0) { popupChoices.Add("No plugins in Scene"); }
                _targetPluginTypeJSON.choices = popupChoices;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePluginParamOnSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetParamNameOnPopup);
                if (_buttonTypeJSON.val.Contains("Bool")) { _targetParamNameOnJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("BOOL"); }
                else { _targetParamNameOnJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("ACTION"); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePluginParamOffSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetParamNameOffPopup);
                if (_buttonTypeJSON.val.Contains("Bool")) { _targetParamNameOffJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("BOOL"); }
                else { _targetParamNameOffJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("ACTION"); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateSaveDataSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_pluginSaveDataSelectorPopup);
                List<string> pluginChoices = new List<string>();
                foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                    List<string> pluginRefsInAtom = new List<string>();
                    pluginRefsInAtom = _selectedButton.GetPluginRefsOfType(atom, _pluginFilePathJSON.val);
                    foreach (string pluginRef in pluginRefsInAtom) {
                        pluginChoices.Add(pluginRef + "[" + atom.name + "]");
                    }
                }
                if (_pluginSaveDataJSON.val != "" && _pluginSaveDataJSON.val != null) { pluginChoices.Add("<Clear Plugin Data>"); }
                if (pluginChoices.Count == 0) { pluginChoices.Add("<None available>"); }
                _pluginSaveDataSelectorJSON.choices = pluginChoices;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void UpdateButtonTypeChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_buttonTypePopup);
                _buttonTypeJSON.choices = _selectedButton.GetButtonTypeChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateVRMovePositionGranulatityChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_vrMovePositionGranulatityPopup);
                _vrMovePositionGranulatityJSON.choices = _selectedButton.GetPositionGranularityChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateVRMoveRotationGranulatityChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_vrMoveRotationGranulatityPopup);
                _vrMoveRotationGranulatityJSON.choices = _selectedButton.GetRotationGranularityChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }


        private void UpdateAtomCategoryChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_atomCategoryPopup);

                _atomCategoryJSON.choices = _selectedButton.GetAtomCategoryChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateAtomTypeChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_atomTypePopup);

                _atomTypeJSON.choices = _selectedButton.GetAtomTypeChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateTargetCategoryChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetCategoryPopup);
                _targetCategoryJSON.choices = _selectedButton.GetTargetCategoryChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateTargetChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetNamePopup);
                _targetNameJSON.choices = _selectedButton.GetTargetChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateMoveRelativeToChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_atomPositionRelativeToPopup);
                _atomPositionRelativeToJSON.choices = _selectedButton.GetMoveRelativeToChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateTargetNodeChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_targetNodePopup);
                _targetNodeJSON.choices = _selectedButton.GetTargetNodeChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePresetType1Choices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_preset1TypePopup);
                _preset1TypeJSON.choices = _selectedButton.GetPresetTypeChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePresetType2Choices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_preset2TypePopup);
                _preset2TypeJSON.choices = _selectedButton.GetPresetTypeChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePresetType3Choices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_preset3TypePopup);
                _preset3TypeJSON.choices = _selectedButton.GetPresetTypeChoices();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshDisplayValues() {
            _isDisplayUpdating = true;
            _buttonCategoryJSON.val = _selectedButton._buttonCategory;
            _buttonTypeJSON.val = _selectedButton._buttonType;
            if (_selectedButton._buttonLabelOn == "") {
                if (_selectedButton._maxButtonStates > 1) { _buttonLabelOnJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Toggle On Label>"; }
                else { _buttonLabelOnJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Label>"; }
            }
            else { _buttonLabelOnJSON.val = _selectedButton._buttonLabelOn; }
            if (_selectedButton._buttonLabelOff == "") { _buttonLabelOffJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Toggle Off Label>"; }
            else { _buttonLabelOffJSON.val = _selectedButton._buttonLabelOff; }
            _targetCategoryJSON.val = _selectedButton._targetCategory;
            _targetNameJSON.val = _selectedButton._targetName;
            if (_selectedButton._buttonType == "Load Plugin(s)" || _selectedButton._buttonType == "Spawn Atom") {
                if (int.Parse(_pluginLoadSelectorJSON.val) > _selectedButton._pluginLoadList.Count) _pluginLoadSelectorJSON.val = _selectedButton._pluginLoadList.Count.ToString();
                _pluginFilePathJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath;
                _pluginSaveDataJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData;
                _singleInstancePerAtomJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._singleInstancePerAtom;
            }

            if (_pluginSaveDataSelectorPopup != null) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            _pluginSaveDataSelectorJSON.val = "Select...";
            _targetPluginTypeJSON.val = _selectedButton._pluginType;
            _pluginForceMergeLoadJSON.val = _selectedButton._pluginForceMergeLoad;

            _targetParamNameOnJSON.val = _selectedButton._targetParamNameOn;
            _targetParamNameOffJSON.val = _selectedButton._targetParamNameOff;
            _suppressClothingLoadJSON.val = _selectedButton._suppressClothingLoad;
            _suppressPersonScaleLoadJSON.val = _selectedButton._suppressPersonScaleLoad;
            _mergeLoadPresetJSON.val = _selectedButton._mergeLoadPreset;
            _autoLabelJSON.val = _selectedButton._autoLabel;
            _useButtonTextureJSON.val = _selectedButton._useButtonTexture;

            _atomCategoryJSON.val = _selectedButton._atomCategory;
            _atomTypeJSON.val = _selectedButton._atomType;
            _targetNodeJSON.val = _selectedButton._targetNode;
            _atomPositionRelativeToJSON.val = _selectedButton._atomPositionRelativeTo;

            _atomPositionXJSON.val = _selectedButton._atomPositionX;
            _atomPositionYJSON.val = _selectedButton._atomPositionY;
            _atomPositionZJSON.val = _selectedButton._atomPositionZ;
            if (_selectedButton._atomRotationX > 180f) _atomRotationXJSON.val = -180f + (_selectedButton._atomRotationX -180f); 
            else _atomRotationXJSON.val = _selectedButton._atomRotationX;
            if (_selectedButton._atomRotationY > 180f) _atomRotationYJSON.val = -180f + (_selectedButton._atomRotationY - 180f);
            else _atomRotationYJSON.val = _selectedButton._atomRotationY;
            if (_selectedButton._atomRotationZ > 180f) _atomRotationZJSON.val = -180f + (_selectedButton._atomRotationZ - 180f);
            else _atomRotationZJSON.val = _selectedButton._atomRotationZ;
            _lockedPositionXJSON.val = _selectedButton._lockedPositionX;
            _lockedPositionYJSON.val = _selectedButton._lockedPositionY;
            _lockedPositionZJSON.val = _selectedButton._lockedPositionZ;
            _lockedRotationXJSON.val = _selectedButton._lockedRotationX;
            _lockedRotationYJSON.val = _selectedButton._lockedRotationY;
            _lockedRotationZJSON.val = _selectedButton._lockedRotationZ;
            _vrMoveDisablePhysicsJSON.val = _selectedButton._vrMoveDisablePhysics;
            
            _vrMovePositionGranulatityJSON.val = _selectedButton._vrMovePositionGranulatity;
            _vrMoveRotationGranulatityJSON.val = _selectedButton._vrMoveRotationGranulatity;


            _atomAbsoluteHeightJSON.val = _selectedButton._atomAbsoluteHeight;
            _atomAbsolutePitchJSON.val = _selectedButton._atomAbsolutePitch;
            _atomAbsoluteRollJSON.val = _selectedButton._atomAbsoluteRoll;

            _tagItemsToRemoveSelectorJSON.val = "Toggle Clothing Tag...";
            _preset1FilePathJSON.val = _selectedButton._preset1FilePath;
            _preset2FilePathJSON.val = _selectedButton._preset2FilePath;
            _preset3FilePathJSON.val = _selectedButton._preset3FilePath;
            _preset1TypeJSON.val = _selectedButton._preset1Type;
            _preset2TypeJSON.val = _selectedButton._preset2Type;
            _preset3TypeJSON.val = _selectedButton._preset3Type;
            _parentLinkToTargetJSON.val = _selectedButton._parentLinkToTarget;

            _decalBodyLoadJSON.val = _selectedButton._decalBodyLoad;
            _decalFaceLoadJSON.val = _selectedButton._decalFaceLoad;
            _decalLimbsLoadJSON.val = _selectedButton._decalLimbsLoad;
            _decalGenLoadJSON.val = _selectedButton._decalGenLoad;

            if (_selectedButton._hairScalpControl != null) _hairScalpControlJSON.val = _selectedButton._hairScalpControl.ToString();
            else _hairScalpControlJSON.val = "";
            if (_selectedButton._hairSimControl != null) _hairSimControlJSON.val = _selectedButton._hairSimControl.ToString();
            else _hairSimControlJSON.val = "";
            _hairSimSelectorJSON.val = "Select...";
            _hairScalpSelectorJSON.val = "Select...";
            _isDisplayUpdating = false;
        }
    }

    public class ButtonPluginLoad {
        public string _filePath = "";
        public bool _singleInstancePerAtom = true;
        public string _pluginSaveData = "";
        public JSONArray _pluginSaveJSON;
        
        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["filePath"] = _filePath;
            node["singleInstancePerAtom"].AsBool = _singleInstancePerAtom;
            if (_pluginSaveJSON!=null) node["pluginSaveJSON"] = _pluginSaveJSON;
            return node;
        }
        public void CopyFrom(ButtonPluginLoad bpl) {
            _filePath = bpl._filePath;
            _singleInstancePerAtom = bpl._singleInstancePerAtom;
            _pluginSaveData = bpl._pluginSaveData;
            _pluginSaveJSON = bpl._pluginSaveJSON;
        }

        public void LoadJSON(JSONClass pluginLoadJSON) {

            _pluginSaveJSON = new JSONArray();

            if (pluginLoadJSON["filePath"] != null) _filePath = pluginLoadJSON["filePath"];
            if (pluginLoadJSON["pluginSaveData"] != null)
            {
                string tempJSONString = pluginLoadJSON["pluginSaveData"];
                if (tempJSONString == "")
                {
                    _pluginSaveJSON = null;
                    _pluginSaveData = "";
                }
                else
                {
                    _pluginSaveJSON = (JSONArray)JSON.Parse("[" + tempJSONString + "]");
                    _pluginSaveData = tempJSONString.Substring(0, Math.Min(tempJSONString.Length, 500));
                }

            }
            else
            {
                if (pluginLoadJSON["pluginSaveJSON"] != null)
                {
                    _pluginSaveJSON = (JSONArray)pluginLoadJSON["pluginSaveJSON"];
                    _pluginSaveData = _pluginSaveJSON.ToString().Substring(0, Math.Min(_pluginSaveJSON.ToString().Length, 500));
                }

            }
            if (pluginLoadJSON["singleInstancePerAtom"] != null) _singleInstancePerAtom = pluginLoadJSON["singleInstancePerAtom"].AsBool;

        }

    }
    public class DecalMakerTextureCollection
    {
        public List<DecalMakerTexture> _decalMakerTexturesList;
        private UIAButton _parentButton;

        public DecalMakerTextureCollection(UIAButton parent)
        {
            try
            {
                _decalMakerTexturesList = new List<DecalMakerTexture>();
                _parentButton = parent;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CopyFrom(DecalMakerTextureCollection dmtc)
        {
            if (dmtc != null)
            {
                _decalMakerTexturesList = new List<DecalMakerTexture>();
                foreach (DecalMakerTexture dmt in dmtc._decalMakerTexturesList)
                {
                    DecalMakerTexture newDMT = new DecalMakerTexture();
                    newDMT._decalTexturePath = dmt._decalTexturePath;
                    newDMT._decalTextureColor = dmt._decalTextureColor;
                    newDMT._decalTextureAlpha = dmt._decalTextureAlpha;

                    _decalMakerTexturesList.Add(newDMT);
                }
            }
        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            JSONArray decalMakerTexturesArray = new JSONArray();

            foreach (DecalMakerTexture dmt in _decalMakerTexturesList)
            {
                JSONClass dmtNode = new JSONClass();
                dmtNode["decalTexturePath"] = dmt._decalTexturePath;
                dmtNode["decalTextureAlpha"].AsFloat = dmt._decalTextureAlpha;
                dmtNode["decalTextureColorHValue"].AsFloat = dmt._decalTextureColor.H;
                dmtNode["decalTextureColorSValue"].AsFloat = dmt._decalTextureColor.S;
                dmtNode["decalTextureColorVValue"].AsFloat = dmt._decalTextureColor.V;

                decalMakerTexturesArray.Add(dmtNode);
            }

            node["decalTexturesCollection"] = decalMakerTexturesArray;

            return node;
        }
        public void LoadJSON(JSONClass dmtcJSON)
        {
            _decalMakerTexturesList = new List<DecalMakerTexture>();

            if (dmtcJSON["decalTexturesCollection"] != null)
            {
                JSONArray decalTexturesArray = dmtcJSON["decalTexturesCollection"].AsArray;
                for (int i = 0; i < decalTexturesArray.Count; i++)
                {
                    JSONClass dmtNode = decalTexturesArray[i].AsObject;
                    DecalMakerTexture newDMT = new DecalMakerTexture();
                    newDMT._decalTexturePath = dmtNode["decalTexturePath"].Value;
                    newDMT._decalTextureAlpha = dmtNode["decalTextureAlpha"].AsFloat;
                    newDMT._decalTextureColor = new HSVColor();
                    newDMT._decalTextureColor.H = dmtNode["decalTextureColorHValue"].AsFloat;
                    newDMT._decalTextureColor.S = dmtNode["decalTextureColorSValue"].AsFloat;
                    newDMT._decalTextureColor.V = dmtNode["decalTextureColorVValue"].AsFloat;
                    _decalMakerTexturesList.Add(newDMT);
                }
            }
        }
        public List<string> GetDecalSelectChoices(string currentSelection)
        {
            List<string> choices = new List<string>();
            int index = 1;
            foreach (DecalMakerTexture dm in _decalMakerTexturesList)
            {
                choices.Add(index.ToString());
                index++;
            }
            choices.Add("Add Decal");
            if (currentSelection != "Select Texture...") choices.Add("Remove Decal "+currentSelection);
            return (choices);
        }
        public void AddDecal()
        {
            DecalMakerTexture newDMT = new DecalMakerTexture();
            _decalMakerTexturesList.Add(newDMT);
        }
        public void RemoveDecal(int index)
        {           
            _decalMakerTexturesList.RemoveAt(index);
        }
    }
    public class DecalMakerTexturesDisplay
    {
        private UIAssist _mvrScript;
        public DecalMakerTextureCollection _selectDecalMakerTextureCollection;

        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _dmTextureSelectorJSON;
        
        public JSONStorableString _dmTextureFileJSON;
        public JSONStorableColor _dmTextureColorJSON;
        public JSONStorableFloat _dmTextureAlphaJSON;

        private UIDynamicPopup _dmTextureSelectorPopup;
        private UIDynamicTextField _dmTextureFileTextField;
        private UIDynamicButton _dmSelectTextureFileSelectButton;
        private UIDynamicColorPicker _dmTextureColorPicker;
        private UIDynamicSlider _dmTextureAlphaSlider;
        

        public DecalMakerTexturesDisplay(UIAssist mvrScript, DecalMakerTextureCollection _dmCollection)
        {
            try
            {
                _selectDecalMakerTextureCollection = _dmCollection;
                _mvrScript = mvrScript;

                HSVColor hsvColor;
                hsvColor.H = 1f;
                hsvColor.S = 1f;
                hsvColor.V = 1f;

                _dmTextureSelectorJSON = new JSONStorableStringChooser("dmTextureSelectorJSON", null, "Select Texture...", "", SwitchDMTexture);
                _dmTextureFileJSON = new JSONStorableString("dmTextureFileJSON", "", SwitchDMTextureFile);
                _dmTextureColorJSON = new JSONStorableColor("dmTextureColorJSON", hsvColor, SwitchDMColorValue);
                _dmTextureAlphaJSON = new JSONStorableFloat("dmTextureAlphaJSON", 1f, SwitchDMAlphaValue, 0f, 1f);
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchDMTexture(string stringVal)
        {
            try
            {
               if (!_isDisplayUpdating)
                {
                    if (stringVal =="Add Decal")
                    {
                        _isDisplayUpdating = true;
                        _selectDecalMakerTextureCollection.AddDecal();
                        _dmTextureSelectorJSON.val = _selectDecalMakerTextureCollection._decalMakerTexturesList.Count.ToString();
                        _isDisplayUpdating = false;
                    }
                    if (stringVal.StartsWith("Remove Decal"))
                    {
                        _isDisplayUpdating = true;
                        _selectDecalMakerTextureCollection.RemoveDecal(int.Parse(stringVal.Substring(13))-1);
                        _dmTextureSelectorJSON.val = "Select Texture...";
                        _isDisplayUpdating = false;
                    }
                    _mvrScript.RefreshRightUIDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDMTextureFile(string stringVal)
        {
            try
            {                
                if (!_isDisplayUpdating)
                {
                    int index = int.Parse(_dmTextureSelectorJSON.val) - 1;
                    _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTexturePath = stringVal;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDMColorValue(JSONStorableColor colorVal)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    int index = int.Parse(_dmTextureSelectorJSON.val) - 1;
                    _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTextureColor = colorVal.val;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDMAlphaValue(float floatVal)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    int index = int.Parse(_dmTextureSelectorJSON.val) - 1;
                    _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTextureAlpha = floatVal;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CloseAllPopup(UIDynamicPopup exception)
        {
            if (_dmTextureSelectorPopup != null && _dmTextureSelectorPopup != exception) { _dmTextureSelectorPopup.popup.visible = false; }
        }
        public void ClearDisplay()
        {
            if (_dmTextureSelectorPopup != null) { _mvrScript.RemovePopup(_dmTextureSelectorPopup); }

            if (_dmTextureFileTextField != null) { _mvrScript.RemoveTextField(_dmTextureFileTextField); }
            if (_dmTextureAlphaSlider != null) { _mvrScript.RemoveSlider(_dmTextureAlphaSlider); }
            if (_dmTextureColorPicker != null) { _mvrScript.RemoveColorPicker(_dmTextureColorPicker); }
            if (_dmSelectTextureFileSelectButton != null) { _dmSelectTextureFileSelectButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_dmSelectTextureFileSelectButton); }

        }
        public void RefreshDisplay(string labelPrefix)
        {
            ClearDisplay();
            _isDisplayUpdating = true;
            _dmTextureSelectorPopup = _mvrScript.CreateSetupPopup(_dmTextureSelectorJSON, true,400, DecalSelectorPopupOpen, labelPrefix+" Decals");
            if (_dmTextureSelectorJSON.val !="Select Texture...")
            {
                int index = int.Parse(_dmTextureSelectorJSON.val)-1;
                _dmTextureFileJSON.val = _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTexturePath;
                _dmTextureFileTextField = _mvrScript.CreateTextField(_dmTextureFileJSON, true);
                _dmTextureFileTextField.height = 1;
                _dmSelectTextureFileSelectButton = _mvrScript.CreateButton("Select "+labelPrefix+" Decal "+(index+1).ToString()+" File",true);
                _dmSelectTextureFileSelectButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", "Custom\\Atom\\Person\\Textures", false, true, false); });
                _dmTextureColorJSON.val = _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTextureColor;
                _dmTextureColorPicker = _mvrScript.CreateColorPicker(_dmTextureColorJSON, true);
                _dmTextureColorPicker.label = labelPrefix + " Decal " + (index + 1).ToString() + " Color";
                _dmTextureAlphaJSON.val = _selectDecalMakerTextureCollection._decalMakerTexturesList[index]._decalTextureAlpha;
                _dmTextureAlphaSlider = _mvrScript.CreateSlider(_dmTextureAlphaJSON, true);
                _dmTextureAlphaSlider.label = labelPrefix + " Decal " + (index + 1).ToString() + " Alpha";
            }
            _isDisplayUpdating = false;
        }

        private void TextureSelected(string filePath)
        {
            try
            {
                if (filePath != "")
                {
                    _dmTextureFileJSON.val = filePath;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void DecalSelectorPopupOpen()
        {
            _mvrScript.CloseAllRightUIPopups(_dmTextureSelectorPopup);

            _dmTextureSelectorJSON.choices = _selectDecalMakerTextureCollection.GetDecalSelectChoices(_dmTextureSelectorJSON.val);
        }
    }

    public class DecalMakerTexture
    {
        public string _decalTexturePath = "";
        public HSVColor _decalTextureColor ;
        public float _decalTextureAlpha = 1f;
        public DecalMakerTexture()
        {           
            _decalTextureColor.H = 0f;
            _decalTextureColor.S = 0f;
            _decalTextureColor.V = 1f;
        }
    }

    public class PluginVariables
    {
        public Dictionary<string, PluginVariable> _pluginVarDict;
        private UIAButton _parentButton;

        public PluginVariables (UIAButton parent)
        {
            try
            {
                _pluginVarDict = new Dictionary<string, PluginVariable>();
                _parentButton = parent;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CopyFrom(PluginVariables pv)
        {
            if (pv != null)
            {
                _pluginVarDict = new Dictionary<string, PluginVariable>();
                foreach (KeyValuePair<string, PluginVariable> kvp in pv._pluginVarDict)
                {
                    PluginVariable newPV = new PluginVariable();
                    newPV._pluginBoolValue = kvp.Value._pluginBoolValue;
                    newPV._pluginFloatValue = kvp.Value._pluginFloatValue;
                    newPV._pluginVarType = kvp.Value._pluginVarType;
                    newPV._pluginStringValue = kvp.Value._pluginStringValue;
                    newPV._pluginColorValue = kvp.Value._pluginColorValue;

                    _pluginVarDict.Add(kvp.Key, newPV);
                }
            }

        }
        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            JSONArray pluginVariablesArray = new JSONArray();

            foreach (KeyValuePair<string, PluginVariable> kvp in _pluginVarDict)
            {
                JSONClass variableNode = new JSONClass();
                variableNode["pluginVarName"] = kvp.Key;
                variableNode["pluginVarType"] = kvp.Value._pluginVarType;

                switch (kvp.Value._pluginVarType)
                {
                    case "BOOL":
                        variableNode["pluginVarBoolValue"].AsBool = kvp.Value._pluginBoolValue;
                        break;
                    case "FLOAT":
                        variableNode["pluginVarFloatValue"].AsFloat = kvp.Value._pluginFloatValue;
                        break;
                    case "STRING":
                    case "STRINGCHOOSER":
                    case "URL":
                        variableNode["pluginVarStringValue"] = kvp.Value._pluginStringValue;
                        break;
                    case "COLOR":
                        variableNode["pluginVarColorHValue"].AsFloat = kvp.Value._pluginColorValue.H ;
                        variableNode["pluginVarColorSValue"].AsFloat = kvp.Value._pluginColorValue.S;
                        variableNode["pluginVarColorVValue"].AsFloat = kvp.Value._pluginColorValue.V;

                        break;
                }

                pluginVariablesArray.Add(variableNode);
            }

            node["pluginVariables"] = pluginVariablesArray;

            return node;
        }

        public void LoadJSON(JSONClass pluginVarJSON)
        {
            _pluginVarDict = new Dictionary<string, PluginVariable>();
            
            if (pluginVarJSON["pluginVariables"] != null)
            {
                JSONArray pluginVariablesArray = pluginVarJSON["pluginVariables"].AsArray;
                for (int i = 0; i < pluginVariablesArray.Count; i++)
                {
                    JSONClass variableNodeJSON = pluginVariablesArray[i].AsObject;
                    PluginVariable pluginVariable = new PluginVariable() ;
                    pluginVariable._pluginVarType = variableNodeJSON["pluginVarType"];

                    switch (pluginVariable._pluginVarType)
                    {
                        case "BOOL":
                            pluginVariable._pluginBoolValue = variableNodeJSON["pluginVarBoolValue"].AsBool;
                            break;
                        case "FLOAT":
                            pluginVariable._pluginFloatValue = variableNodeJSON["pluginVarFloatValue"].AsFloat;
                            break;
                        case "STRING":
                        case "STRINGCHOOSER":
                        case "URL":
                            pluginVariable._pluginStringValue = variableNodeJSON["pluginVarStringValue"];
                            break;
                        case "COLOR":
                            pluginVariable._pluginColorValue = new HSVColor();
                            pluginVariable._pluginColorValue.H = variableNodeJSON["pluginVarColorHValue"].AsFloat;
                            pluginVariable._pluginColorValue.S = variableNodeJSON["pluginVarColorSValue"].AsFloat;
                            pluginVariable._pluginColorValue.V = variableNodeJSON["pluginVarColorVValue"].AsFloat;
                            break;
                    }

                    _pluginVarDict.Add(variableNodeJSON["pluginVarName"], pluginVariable);
                }
            }
        }
        public List<string> GetPluginParamChoices()
        {
            List<string> paramChoices = new List<string>();
            List<string> displayParamChoices = new List<string>();
            paramChoices = _parentButton.GetTargetPluginParamNameChoices("ALL");

            foreach (string choice in paramChoices)
            {
                if (_pluginVarDict.ContainsKey(choice))
                {
                    displayParamChoices.Add("+ " + choice + " +");
                }
                else displayParamChoices.Add("  " + choice + "  ");

            }

            return (displayParamChoices);
        }

        public string GetVarType(string varName)
        {
            string varType = "NONE";

            if (_pluginVarDict.ContainsKey(varName))
            {
                varType = _pluginVarDict[varName]._pluginVarType;
            }

            return (varType);
        }
        public bool GetVarSet(string varName)
        {
            return (_pluginVarDict.ContainsKey(varName));
        }

        public string GetVarTypeFromPlugin (string varName)
        {
            string varType = _parentButton.GetVarTypeFromPlugin(varName);

            return (varType);
        }
        public float GetMinFloatValue(string varName)
        {
            float minFloat = _parentButton.GetMinFloatValue(varName);
            if (_pluginVarDict.ContainsKey(varName))
            {
                float varValue = _pluginVarDict[varName]._pluginFloatValue;
                if (varValue < minFloat) minFloat = varValue;
            }
            return (minFloat);
        }
        public float GetMaxFloatValue(string varName)
        {
            float maxFloat = _parentButton.GetMaxFloatValue(varName);
            if (_pluginVarDict.ContainsKey(varName))
            {
                float varValue = _pluginVarDict[varName]._pluginFloatValue;
                if (varValue > maxFloat) maxFloat = varValue;
            }
            return (maxFloat);
        }
        public List<string> GetStringChooserChoices(string varName)
        {
            List<string> choices = _parentButton.GetStringChooserChoices(varName);
            if (choices.Count < 1) choices.Add("NO VALUES AVAILABLE");
            return (choices);
        }
    }
    public class PluginVariablesDisplay
    {
        private UIAssist _mvrScript;
        public PluginVariables _selectButtonPluginVariables;

        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _pluginVariableSelectorJSON;
        public JSONStorableBool _pluginVariableSetJSON;
        public JSONStorableFloat _pluginVariableFloatValueJSON;
        public JSONStorableString _pluginVariableStringValueJSON;
        public JSONStorableStringChooser _pluginVariableStringChooserValueJSON;
        public JSONStorableBool _pluginVariableBoolValueJSON;
        public JSONStorableColor _pluginVariableColorValueJSON;

        private UIDynamicPopup _pluginVariableSelectorPopup;
        private UIDynamicToggle _pluginVariableSetToggle;
        private UIDynamicToggle _pluginVariableBoolValueToggle;

        private UIDynamicSlider _pluginVariableFloatValueSlider;
        private UIDynamicColorPicker _pluginVariableColorValueColorPicker;
        private UIDynamicTextField _pluginVariableStringValueTextField;
        private UIDynamicPopup _pluginVariableStringValuePopup;

        public PluginVariablesDisplay( UIAssist mvrScript, PluginVariables pluginVariables)
        {
            try
            {
                _selectButtonPluginVariables = pluginVariables;
                _mvrScript = mvrScript;

                HSVColor hsvColor; hsvColor.H = 0f; hsvColor.S = 0f; hsvColor.V = 0f;

                _pluginVariableSelectorJSON = new JSONStorableStringChooser("pluginVariableChooser", null, "Select Parameter...", "", SwitchPluginVariable);
                _pluginVariableSetJSON = new JSONStorableBool("pluginVariableSet",false, SwitchPluginVariableSet);
                _pluginVariableFloatValueJSON = new JSONStorableFloat("pluginVariableFloat", 0f, SwitchFloatValue, 0f, 1f, true);
                _pluginVariableStringValueJSON = new JSONStorableString("pluginVariableString","",SwitchStringValue);
                _pluginVariableStringChooserValueJSON = new JSONStorableStringChooser("pluginVariableStringChooser", null, "", "", SwitchStringValue);
                _pluginVariableBoolValueJSON = new JSONStorableBool("pluginVariableBoolValue", true, SwitchBoolValue);
                _pluginVariableColorValueJSON = new JSONStorableColor("pluginVariableColorValue", hsvColor, SwitchColorValue);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchPluginVariable(string stringVal)
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_pluginVariableSelectorPopup);
                if (!_isDisplayUpdating)
                {
                    _mvrScript.RefreshRightUIDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPluginVariableSet(bool setVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (setVal && !_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable) )
                {
                    PluginVariable pluginVar = new PluginVariable();
                    pluginVar._pluginVarType = _selectButtonPluginVariables.GetVarTypeFromPlugin(pluginVariable);

                    _selectButtonPluginVariables._pluginVarDict.Add(pluginVariable, pluginVar);
                }
                else if (!setVal && _selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                {
                    _selectButtonPluginVariables._pluginVarDict.Remove(pluginVariable);
                }
                if (!_isDisplayUpdating)
                {
                    _mvrScript.RefreshRightUIDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloatValue(float floatVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginFloatValue = floatVal;
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchStringValue(string stringVal)
        {
            try
            {
                if (_pluginVariableStringValuePopup != null) _mvrScript.CloseAllRightUIPopups(_pluginVariableStringValuePopup);
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue = stringVal;
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchBoolValue(bool boolVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginBoolValue = boolVal;
                    }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchColorValue(JSONStorableColor colorVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginColorValue = colorVal.val;
                    }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CloseAllPopup(UIDynamicPopup exception)
        {
            if (_pluginVariableSelectorPopup != null && _pluginVariableSelectorPopup != exception) { _pluginVariableSelectorPopup.popup.visible = false; }
            if (_pluginVariableStringValuePopup != null && _pluginVariableStringValuePopup != exception) { _pluginVariableStringValuePopup.popup.visible = false; }
        }
        public void ClearDisplay()
        {
            if (_pluginVariableSelectorPopup != null) { _mvrScript.RemovePopup(_pluginVariableSelectorPopup); }
            if (_pluginVariableStringValuePopup != null) { _mvrScript.RemovePopup(_pluginVariableStringValuePopup); }
            if (_pluginVariableStringValueTextField != null) { _mvrScript.RemoveTextField(_pluginVariableStringValueTextField); }
            if (_pluginVariableFloatValueSlider != null) { _mvrScript.RemoveSlider(_pluginVariableFloatValueSlider); }
            if (_pluginVariableColorValueColorPicker != null) { _mvrScript.RemoveColorPicker(_pluginVariableColorValueColorPicker); }
            if (_pluginVariableSetToggle != null) { _mvrScript.RemoveToggle(_pluginVariableSetToggle); }

            if (_pluginVariableBoolValueToggle != null) { _mvrScript.RemoveToggle(_pluginVariableBoolValueToggle); }
            
        }
        public void RefreshDisplay(string labelPrefix)
        {
            ClearDisplay();
            _isDisplayUpdating = true;

            _pluginVariableSelectorPopup = _mvrScript.CreateSetupPopup(_pluginVariableSelectorJSON, true, 500f, SkinParamSelectorPopupOpen, labelPrefix+"Plugin Settings");
            _pluginVariableSelectorPopup.labelTextColor = Color.blue;
            if (_pluginVariableSelectorJSON.val!="Select Parameter...")
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                bool isSet = _selectButtonPluginVariables.GetVarSet(pluginVariable);
                _pluginVariableSetJSON.val = isSet;
                _pluginVariableSetToggle = _mvrScript.CreateToggle(_pluginVariableSetJSON, true);
                _pluginVariableSetToggle.label = "Initialise Setting value pre-Action";
                if (isSet)
                {
                    string varType = _selectButtonPluginVariables.GetVarType(pluginVariable);
                    
                    switch (varType)
                    {
                        case "FLOAT":                           
                            _pluginVariableFloatValueJSON.min = _selectButtonPluginVariables.GetMinFloatValue(pluginVariable);
                            _pluginVariableFloatValueJSON.max = _selectButtonPluginVariables.GetMaxFloatValue(pluginVariable);
                            _pluginVariableFloatValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginFloatValue;
                            _pluginVariableFloatValueJSON.constrained = false;
                            _pluginVariableFloatValueSlider = _mvrScript.CreateSlider(_pluginVariableFloatValueJSON, true);
                            _pluginVariableFloatValueSlider.label = pluginVariable;
                            break;
                        case "BOOL":
                            _pluginVariableBoolValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginBoolValue;
                            
                            _pluginVariableBoolValueToggle = _mvrScript.CreateToggle(_pluginVariableBoolValueJSON, true);
                            _pluginVariableBoolValueToggle.label = pluginVariable;
                            break;
                        case "STRINGCHOOSER":
                            _pluginVariableStringChooserValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue;
                            _pluginVariableStringChooserValueJSON.choices = _selectButtonPluginVariables.GetStringChooserChoices(pluginVariable);
                            if (!_pluginVariableStringChooserValueJSON.choices.Contains(_pluginVariableStringChooserValueJSON.val))
                            {
                                _pluginVariableStringChooserValueJSON.val = _pluginVariableStringChooserValueJSON.choices[0];
                                _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue = _pluginVariableStringChooserValueJSON.val;
                            }
                            _pluginVariableStringValuePopup = _mvrScript.CreateSetupPopup(_pluginVariableStringChooserValueJSON,true);
                            _pluginVariableStringValuePopup.label = "Set '"+pluginVariable+"'";
                            break;
                        case "STRING":
                        case "URL":
                            _pluginVariableStringValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue;
                           
                            _pluginVariableStringValueTextField = _mvrScript.CreateTextInput(_pluginVariableStringValueJSON, true);
                            break;
                        case "COLOR":
                            _pluginVariableColorValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginColorValue;
                            _pluginVariableColorValueColorPicker = _mvrScript.CreateColorPicker(_pluginVariableColorValueJSON, true);
                            break;
                    }
                }
            }


            _isDisplayUpdating = false;
        }
        private void SkinParamSelectorPopupOpen()
        {
            _mvrScript.CloseAllRightUIPopups(_pluginVariableSelectorPopup);

            _pluginVariableSelectorJSON.choices = _selectButtonPluginVariables.GetPluginParamChoices();
        }
    }
    public class PluginVariable
    {
        public string _pluginVarType = "NONE";

        public float _pluginFloatValue =0;
        public bool _pluginBoolValue = false;
        public string _pluginStringValue ="";
        public HSVColor _pluginColorValue;
    }
    public class ButtonSkin {
        public float _minFontSize = 4f;
        public float _maxFontSize = 25f;
        public float _minMicroFontSize = 3f;
        public float _maxMicroFontSize = 12f;
        public float _minMiniFontSize = 3f;
        public float _maxMiniFontSize = 15f;
        public float _minSmallFontSize = 4f;
        public float _maxSmallFontSize = 15f;
        public float _minMediumFontSize = 4f;
        public float _maxMediumFontSize = 25f;
        public float _minLargeFontSize = 4f;
        public float _maxLargeFontSize = 35f;
        public bool _fontSizeDefault = true;
        public HSVColor _textOnColor;
        public bool _textOnColorDefault = true;
        public HSVColor _textOffColor;
        public bool _textOffColorDefault = true;
        public string _textAlignment = "MiddleCenter";
        public bool _textAlignmentDefault = true;
        public HSVColor _buttonOnColor;
        public bool _buttonOnColorDefault = true;
        public HSVColor _buttonOffColor;
        public bool _buttonOffColorDefault = true;
        public string _customButtonOnTextureURL = "";
        public bool _customButtonOnTextureDefault = true;
        public string _customButtonOffTextureURL = "";
        public bool _customButtonOffTextureDefault = true;
        public string _buttonFont = "Arial";
        public bool _buttonFontDefault = true;
        public int _maxButtonStates;
        public string _defaultGazeTarget = "Last viewed Female";
        public Texture2D _buttonOnTexture;
        public Texture2D _buttonOffTexture;
        private UIAssist _mvrScript;
        public HSVColor _buttonGazeSelectColor;
        public bool _24HourClock = true;

        public ButtonSkin(int buttonStates, UIAssist mvrScript) {
            try {
                _mvrScript = mvrScript;
                _maxButtonStates = buttonStates;
                _buttonGazeSelectColor = new HSVColor();

                Color gazeButtonColor = new Color(128f / 255f, 142f / 255f, 216f / 255f, 1f);
                Color.RGBToHSV(gazeButtonColor, out _buttonGazeSelectColor.H, out _buttonGazeSelectColor.S, out _buttonGazeSelectColor.V);
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CopyFrom(ButtonSkin bs) {
            _minFontSize = bs._minFontSize;
            _maxFontSize = bs._maxFontSize;
            _minMicroFontSize = bs._minMicroFontSize;
            _maxMicroFontSize = bs._maxMicroFontSize;
            _minMiniFontSize = bs._minMiniFontSize;
            _maxMiniFontSize = bs._maxMiniFontSize;

            _minSmallFontSize = bs._minSmallFontSize;
            _maxSmallFontSize = bs._maxSmallFontSize;
            _minMediumFontSize = bs._minMediumFontSize;
            _maxMediumFontSize = bs._maxMediumFontSize;
            _minLargeFontSize = bs._minLargeFontSize;
            _maxLargeFontSize = bs._maxLargeFontSize;
            _fontSizeDefault = bs._fontSizeDefault;

            _textOnColor.H = bs._textOnColor.H;
            _textOnColor.S = bs._textOnColor.S;
            _textOnColor.V = bs._textOnColor.V;
            _textOnColorDefault = bs._textOnColorDefault;

            _textOffColor.H = bs._textOffColor.H;
            _textOffColor.S = bs._textOffColor.S;
            _textOffColor.V = bs._textOffColor.V;
            _textOffColorDefault = bs._textOffColorDefault;

            _textAlignment = bs._textAlignment;
            _textAlignmentDefault = bs._textAlignmentDefault;

            _buttonOnColor.H = bs._buttonOnColor.H;
            _buttonOnColor.S = bs._buttonOnColor.S;
            _buttonOnColor.V = bs._buttonOnColor.V;
            _buttonOnColorDefault = bs._buttonOnColorDefault;

            _buttonOffColor.H = bs._buttonOffColor.H;
            _buttonOffColor.S = bs._buttonOffColor.S;
            _buttonOffColor.V = bs._buttonOffColor.V;
            _buttonOffColorDefault = bs._buttonOffColorDefault;

            _customButtonOnTextureURL = bs._customButtonOnTextureURL;
            _customButtonOnTextureDefault = bs._customButtonOnTextureDefault;
            _customButtonOffTextureURL = bs._customButtonOffTextureURL;
            _customButtonOffTextureDefault = bs._customButtonOffTextureDefault;
            _buttonFont = bs._buttonFont;
            _buttonFontDefault = bs._buttonFontDefault;
            _maxButtonStates = bs._maxButtonStates;
            _buttonOnTexture = bs._buttonOnTexture;
            _buttonOffTexture = bs._buttonOffTexture;
            _buttonGazeSelectColor = bs._buttonGazeSelectColor;
            _defaultGazeTarget = bs._defaultGazeTarget;
            _24HourClock = bs._24HourClock;

        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["minFontSize"].AsFloat = _minFontSize;
            node["maxFontSize"].AsFloat = _maxFontSize;
            node["minMicroFontSize"].AsFloat = _minMicroFontSize;
            node["maxMicroFontSize"].AsFloat = _maxMicroFontSize;
            node["minMiniFontSize"].AsFloat = _minMiniFontSize;
            node["maxMiniFontSize"].AsFloat = _maxMiniFontSize;

            node["minSmallFontSize"].AsFloat = _minSmallFontSize;
            node["maxSmallFontSize"].AsFloat = _maxSmallFontSize;
            node["minMediumFontSize"].AsFloat = _minMediumFontSize;
            node["maxMediumFontSize"].AsFloat = _maxMediumFontSize;
            node["minLargeFontSize"].AsFloat = _minLargeFontSize;
            node["maxLargeFontSize"].AsFloat = _maxLargeFontSize;

            node["fontSizeDefault"].AsBool = _fontSizeDefault;
            node["textOnColorDefault"].AsBool = _textOnColorDefault;
            node["textOffColorDefault"].AsBool = _textOffColorDefault;
            node["textAlignmentDefault"].AsBool = _textAlignmentDefault;
            node["buttonOnColorDefault"].AsBool = _buttonOnColorDefault;
            node["buttonOffColorDefault"].AsBool = _buttonOffColorDefault;
            node["customButtonOnTextureDefault"].AsBool = _customButtonOnTextureDefault;
            node["customButtonOffTextureDefault"].AsBool = _customButtonOffTextureDefault;
            node["buttonFontDefault"].AsBool = _buttonFontDefault;

            node["textOnColor"]["H"].AsFloat = _textOnColor.H;
            node["textOnColor"]["S"].AsFloat = _textOnColor.S;
            node["textOnColor"]["V"].AsFloat = _textOnColor.V;

            node["textOffColor"]["H"].AsFloat = _textOffColor.H;
            node["textOffColor"]["S"].AsFloat = _textOffColor.S;
            node["textOffColor"]["V"].AsFloat = _textOffColor.V;

            node["buttonOnColor"]["H"].AsFloat = _buttonOnColor.H;
            node["buttonOnColor"]["S"].AsFloat = _buttonOnColor.S;
            node["buttonOnColor"]["V"].AsFloat = _buttonOnColor.V;

            node["buttonOffColor"]["H"].AsFloat = _buttonOffColor.H;
            node["buttonOffColor"]["S"].AsFloat = _buttonOffColor.S;
            node["buttonOffColor"]["V"].AsFloat = _buttonOffColor.V;

            node["textAlignment"] = _textAlignment;
            node["customButtonOnTextureURL"] = _customButtonOnTextureURL;
            node["customButtonOffTextureURL"] = _customButtonOffTextureURL;
            node["buttonFont"] = _buttonFont;
            node["defaultGazeTarget"] = _defaultGazeTarget;

            node["maxButtonStates"].AsInt = _maxButtonStates;

            node["gazeSelectColor"]["H"].AsFloat = _buttonGazeSelectColor.H;
            node["gazeSelectColor"]["S"].AsFloat = _buttonGazeSelectColor.S;
            node["gazeSelectColor"]["V"].AsFloat = _buttonGazeSelectColor.V;

            node["24HourClock"].AsBool = _24HourClock;

            return node;
        }
        public void LoadJSON(JSONClass buttonSkinJSON) {
            if (buttonSkinJSON["minFontSize"] != null) _minFontSize = buttonSkinJSON["minFontSize"].AsFloat;
            if (buttonSkinJSON["maxFontSize"] != null) _maxFontSize = buttonSkinJSON["maxFontSize"].AsFloat;
            if (buttonSkinJSON["minMicroFontSize"] != null) _minMicroFontSize = buttonSkinJSON["minMicroFontSize"].AsFloat;
            if (buttonSkinJSON["maxMicroFontSize"] != null) _maxMicroFontSize = buttonSkinJSON["maxMicroFontSize"].AsFloat;
            if (buttonSkinJSON["minMiniFontSize"] != null) _minMiniFontSize = buttonSkinJSON["minMiniFontSize"].AsFloat;
            if (buttonSkinJSON["maxMiniFontSize"] != null) _maxMiniFontSize = buttonSkinJSON["maxMiniFontSize"].AsFloat;

            if (buttonSkinJSON["minSmallFontSize"] != null) _minSmallFontSize = buttonSkinJSON["minSmallFontSize"].AsFloat;
            if (buttonSkinJSON["maxSmallFontSize"] != null) _maxSmallFontSize = buttonSkinJSON["maxSmallFontSize"].AsFloat;
            if (buttonSkinJSON["minMediumFontSize"] != null) _minMediumFontSize = buttonSkinJSON["minMediumFontSize"].AsFloat;
            if (buttonSkinJSON["maxMediumFontSize"] != null) _maxMediumFontSize = buttonSkinJSON["maxMediumFontSize"].AsFloat;
            if (buttonSkinJSON["minLargeFontSize"] != null) _minLargeFontSize = buttonSkinJSON["minLargeFontSize"].AsFloat;
            if (buttonSkinJSON["maxLargeFontSize"] != null) _maxLargeFontSize = buttonSkinJSON["maxLargeFontSize"].AsFloat;

            if (buttonSkinJSON["fontSizeDefault"] != null) _fontSizeDefault = buttonSkinJSON["fontSizeDefault"].AsBool;
            if (buttonSkinJSON["textOnColorDefault"] != null) _textOnColorDefault = buttonSkinJSON["textOnColorDefault"].AsBool;
            if (buttonSkinJSON["textOffColorDefault"] != null) _textOffColorDefault = buttonSkinJSON["textOffColorDefault"].AsBool;
            if (buttonSkinJSON["textAlignmentDefault"] != null) _textAlignmentDefault = buttonSkinJSON["textAlignmentDefault"].AsBool;
            if (buttonSkinJSON["buttonOnColorDefault"] != null) _buttonOnColorDefault = buttonSkinJSON["buttonOnColorDefault"].AsBool;
            if (buttonSkinJSON["buttonOffColorDefault"] != null) _buttonOffColorDefault = buttonSkinJSON["buttonOffColorDefault"].AsBool;
            if (buttonSkinJSON["customButtonOnTextureDefault"] != null) _customButtonOnTextureDefault = buttonSkinJSON["customButtonOnTextureDefault"].AsBool;
            if (buttonSkinJSON["customButtonOffTextureDefault"] != null) _customButtonOffTextureDefault = buttonSkinJSON["customButtonOffTextureDefault"].AsBool;
            if (buttonSkinJSON["buttonFontDefault"] != null) _buttonFontDefault = buttonSkinJSON["buttonFontDefault"].AsBool;
            if (buttonSkinJSON["fontSizeDefault"] != null) _fontSizeDefault = buttonSkinJSON["fontSizeDefault"].AsBool;

            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.H = buttonSkinJSON["textOnColor"]["H"].AsFloat;
            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.S = buttonSkinJSON["textOnColor"]["S"].AsFloat;
            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.V = buttonSkinJSON["textOnColor"]["V"].AsFloat;

            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.H = buttonSkinJSON["textOffColor"]["H"].AsFloat;
            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.S = buttonSkinJSON["textOffColor"]["S"].AsFloat;
            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.V = buttonSkinJSON["textOffColor"]["V"].AsFloat;

            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.H = buttonSkinJSON["buttonOnColor"]["H"].AsFloat;
            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.S = buttonSkinJSON["buttonOnColor"]["S"].AsFloat;
            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.V = buttonSkinJSON["buttonOnColor"]["V"].AsFloat;

            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.H = buttonSkinJSON["buttonOffColor"]["H"].AsFloat;
            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.S = buttonSkinJSON["buttonOffColor"]["S"].AsFloat;
            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.V = buttonSkinJSON["buttonOffColor"]["V"].AsFloat;

            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.H = buttonSkinJSON["gazeSelectColor"]["H"].AsFloat;
            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.S = buttonSkinJSON["gazeSelectColor"]["S"].AsFloat;
            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.V = buttonSkinJSON["gazeSelectColor"]["V"].AsFloat;

            if (buttonSkinJSON["textAlignment"] != null) _textAlignment = buttonSkinJSON["textAlignment"];
            if (buttonSkinJSON["customButtonOnTextureURL"] != null) _customButtonOnTextureURL = buttonSkinJSON["customButtonOnTextureURL"];
            if (buttonSkinJSON["customButtonOffTextureURL"] != null) _customButtonOffTextureURL = buttonSkinJSON["customButtonOffTextureURL"];
            if (buttonSkinJSON["buttonFont"] != null) _buttonFont = buttonSkinJSON["buttonFont"];
            if (buttonSkinJSON["defaultGazeTarget"] != null) _defaultGazeTarget = buttonSkinJSON["defaultGazeTarget"];
            if (buttonSkinJSON["24HourClock"] != null) _24HourClock = buttonSkinJSON["24HourClock"].AsBool;

            if (buttonSkinJSON["maxButtonStates"] != null) _maxButtonStates = buttonSkinJSON["maxButtonStates"].AsInt;

            if (_customButtonOnTextureURL != "") QueueLoadTexture(_customButtonOnTextureURL, "ON");
            else { _buttonOnTexture = null; }
            if (_customButtonOffTextureURL != "") QueueLoadTexture(_customButtonOffTextureURL, "OFF");
            else { _buttonOffTexture = null; }
        }

        public List<string> GetButtonSkinParamChoices(bool defaultSkinChoices = false) {
            List<string> choices = new List<string>();
            if (defaultSkinChoices) {
                choices.Add("Button(Micro) Font Size");
                choices.Add("Button(Mini) Font Size");
                choices.Add("Button(Small) Font Size");
                choices.Add("Button(Medium) Font Size");
                choices.Add("Button(Large) Font Size");
            }
            else choices.Add("Button Font Size");
            choices.Add("Button Font");
            choices.Add("Button Text Alignment");
            if (_maxButtonStates == 1) { choices.Add("Button Text Color"); choices.Add("Button Color"); choices.Add("Button Thumbnail"); }
            else { choices.Add("Button Text Color (On)"); choices.Add("Button Text Color (Off)"); choices.Add("Button Color (On)"); choices.Add("Button Color (Off)"); choices.Add("Button Thumbnail (On)"); choices.Add("Button Thumbnail (Off)"); }
            if (defaultSkinChoices) choices.Add("Gaze Select Button Color");
            if (defaultSkinChoices) choices.Add("12/24h Wrist Clock");
            _buttonOnColor.V = 0.84f;
            _buttonOffColor.V = 0.4f;
            if (defaultSkinChoices) choices.Add("Default Gaze Target");
            return (choices);
        }
        public void QueueLoadTexture(string url, string loadType)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url)) {
                if (loadType == "ON") _buttonOnTexture = null;
                else _buttonOffTexture = null;
                _mvrScript.RefreshWristUIButtonGrid();
                return;
            }

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = true;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = false;
            queuedImage.linear = false;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.createNormalFromBump = false;
            queuedImage.bumpStrength = 0f;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            switch (loadType) {
                case "ON":
                    queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
                    {
                        Texture2D tex = qi.tex;
                        _buttonOnTexture = tex;
                        _mvrScript.RefreshWristUIButtonGrid();
                    };
                    break;
                case "OFF":
                    queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
                    {
                        Texture2D tex = qi.tex;
                        _buttonOffTexture = tex;
                        _mvrScript.RefreshWristUIButtonGrid();
                    };
                    break;
            }

            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }
    }

    public class ButtonSkinDisplay {
        private UIAssist _mvrScript;
        public ButtonSkin _selectedButtonSkin;

        bool _isDefaultSkin;
        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _buttonSkinParamSelectorJSON;
        public JSONStorableFloat _skinFloatParam1JSON;
        public JSONStorableFloat _skinFloatParam2JSON;
        public JSONStorableColor _skinColorParamJSON;
        public JSONStorableStringChooser _skinStringChooserParamJSON;
        public JSONStorableString _skinStringParamJSON;
        public JSONStorableBool _skinBoolParamJSON;
        public JSONStorableBool _skinSetToCurrentProfileDefaultJSON;

        public UIDynamicPopup _buttonSkinParamSelectorPopup;
        public UIDynamicPopup _skinStringChooserParamPopup;
        private UIDynamicSlider _skinFloatParam1Slider;
        private UIDynamicSlider _skinFloatParam2Slider;
        private UIDynamicColorPicker _skinColorParamColorPicker;
        private UIDynamicToggle _skinBoolParamToggle;
        private UIDynamicToggle _skinSetToCurrentProfileDefaultToggle;

        private UIDynamicTextField _skinStringParamTextField;
        private UIDynamicButton _skinMirrorOnButton;
        private UIDynamicButton _skinSelectTextureButton;
        private UIDynamicButton _skinClearTextureButton;

        private UIDynamic _spacer1;
        private UIDynamic _spacer2;

        public ButtonSkinDisplay(ButtonSkin buttonSkin, UIAssist mvrScript, bool defaultSkin = false) {
            try {
                _selectedButtonSkin = buttonSkin;
                _mvrScript = mvrScript;
                _isDefaultSkin = defaultSkin;
                string initalSkinParamValue = "Button Font Size";
                if (defaultSkin) initalSkinParamValue = "Button(Small) Font Size";
                _buttonSkinParamSelectorJSON = new JSONStorableStringChooser("buttonSkinParamSelect", _selectedButtonSkin.GetButtonSkinParamChoices(), initalSkinParamValue, "", SwitchSkinParam);
                HSVColor hsvColor; hsvColor.H = 0f; hsvColor.S = 0f; hsvColor.V = 0f;
                _skinColorParamJSON = new JSONStorableColor("skinColorParam", hsvColor, SwitchColorParam);
                _skinSetToCurrentProfileDefaultJSON = new JSONStorableBool("skinSetToDefault", true, SwitchSetToDefault);
                _skinFloatParam1JSON = new JSONStorableFloat("skinFloatParam1", 8f, SwitchFloat1, 2f, 40f, true);
                _skinFloatParam2JSON = new JSONStorableFloat("skinFloatParam2", 18f, SwitchFloat2, 2f, 40f, true);
                _skinBoolParamJSON = new JSONStorableBool("skinBoolParam1",true,SwitchBoolParam);
                _skinStringChooserParamJSON = new JSONStorableStringChooser("skinStringChooser", null, "", "", SwitchStringChooser);
                _skinStringParamJSON = new JSONStorableString("skinString", "");

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchBoolParam(bool boolVal)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    if (_buttonSkinParamSelectorJSON.val == "12/24h Wrist Clock")
                    {
                        _selectedButtonSkin._24HourClock = boolVal;
                    }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchStringChooser(string stringVal) {
            try {


                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font") {
                        _selectedButtonSkin._buttonFont = stringVal;
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Button Text Alignment") {
                        _selectedButtonSkin._textAlignment = stringVal;
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Default Gaze Target")
                    {
                        _selectedButtonSkin._defaultGazeTarget = stringVal;
                    }

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloat1(float float1val) {
            try {
                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font Size") _selectedButtonSkin._minFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Micro) Font Size") _selectedButtonSkin._minMicroFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Mini) Font Size") _selectedButtonSkin._minMiniFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") _selectedButtonSkin._minSmallFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") _selectedButtonSkin._minMediumFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") _selectedButtonSkin._minLargeFontSize = float1val;

                    _isDisplayUpdating = true;
                    _skinFloatParam1JSON.val = Mathf.Round(float1val);
                    _isDisplayUpdating = false;

                    if (_skinFloatParam1JSON.val > _skinFloatParam2JSON.val) _skinFloatParam2JSON.val = Mathf.Round(float1val);

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloat2(float float2val) {
            try {
                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font Size") _selectedButtonSkin._maxFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Micro) Font Size") _selectedButtonSkin._maxMicroFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Mini) Font Size") _selectedButtonSkin._maxMiniFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") _selectedButtonSkin._maxSmallFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") _selectedButtonSkin._maxMediumFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") _selectedButtonSkin._maxLargeFontSize = float2val;

                    _isDisplayUpdating = true;
                    _skinFloatParam2JSON.val = Mathf.Round(float2val);
                    _isDisplayUpdating = false;

                    if (_skinFloatParam2JSON.val < _skinFloatParam1JSON.val) _skinFloatParam1JSON.val = Mathf.Round(float2val);

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSetToDefault(bool setToDefault) {
            try {
                if (!_isDisplayUpdating) {
                    switch (_buttonSkinParamSelectorJSON.val) {
                        case "Button Text Color":
                        case "Button Text Color (On)":
                            _selectedButtonSkin._textOnColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Text Color (Off)":
                            _selectedButtonSkin._textOffColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Color":
                        case "Button Color (On)":
                            _selectedButtonSkin._buttonOnColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Color (Off)":
                            _selectedButtonSkin._buttonOffColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Font Size":
                            _selectedButtonSkin._fontSizeDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Font":
                            _selectedButtonSkin._buttonFontDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Text Alignment":
                            _selectedButtonSkin._textAlignmentDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Thumbnail (On)":
                        case "Button Thumbnail":
                            _selectedButtonSkin._customButtonOnTextureDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Thumbnail (Off)":
                            _selectedButtonSkin._customButtonOffTextureDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                    }
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                    _mvrScript.RefreshWristUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchColorParam(JSONStorableColor color) {
            try {
                if (!_isDisplayUpdating) {
                    switch (_buttonSkinParamSelectorJSON.val) {
                        case "Button Text Color":
                        case "Button Text Color (On)":
                            _selectedButtonSkin._textOnColor = color.val;
                            break;
                        case "Button Text Color (Off)":
                            _selectedButtonSkin._textOffColor = color.val;
                            break;
                        case "Button Color":
                        case "Button Color (On)":
                            _selectedButtonSkin._buttonOnColor = color.val;
                            break;
                        case "Button Color (Off)":
                            _selectedButtonSkin._buttonOffColor = color.val;
                            break;
                        case "Gaze Select Button Color":
                            _selectedButtonSkin._buttonGazeSelectColor = color.val;
                            break;
                    }
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSkinParam(string skinParam) {
            try {
                if (!_isDisplayUpdating) {
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void TextureSelected(string filePath) {
            try {
                if (filePath != "") {
                    if (_buttonSkinParamSelectorJSON.val == "Button Thumbnail" || _buttonSkinParamSelectorJSON.val == "Button Thumbnail (On)") {
                        _selectedButtonSkin._customButtonOnTextureURL = filePath;
                        _selectedButtonSkin.QueueLoadTexture(filePath, "ON");
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Button Thumbnail (Off)") {
                        _selectedButtonSkin._customButtonOffTextureURL = filePath;
                        _selectedButtonSkin.QueueLoadTexture(filePath, "OFF");
                    }
                    _mvrScript._lastLoadedThumbnailFolder = FileManagerSecure.GetDirectoryName(filePath);
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        public void ClearDisplay() {
            if (_buttonSkinParamSelectorPopup != null) { _mvrScript.RemovePopup(_buttonSkinParamSelectorPopup); }
            if (_skinStringChooserParamPopup != null) { _mvrScript.RemovePopup(_skinStringChooserParamPopup); }
            if (_skinStringParamTextField != null) { _mvrScript.RemoveTextField(_skinStringParamTextField); }
            if (_skinFloatParam1Slider != null) { _mvrScript.RemoveSlider(_skinFloatParam1Slider); }
            if (_skinFloatParam2Slider != null) { _mvrScript.RemoveSlider(_skinFloatParam2Slider); }
            if (_skinColorParamColorPicker != null) { _mvrScript.RemoveColorPicker(_skinColorParamColorPicker); }
            if (_skinStringParamTextField != null) { _mvrScript.RemoveTextField(_skinStringParamTextField); }
            if (_skinSetToCurrentProfileDefaultToggle != null) { _mvrScript.RemoveToggle(_skinSetToCurrentProfileDefaultToggle); }
            if (_skinMirrorOnButton != null) { _skinMirrorOnButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_skinMirrorOnButton); }
            if (_skinSelectTextureButton != null) { _skinSelectTextureButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_skinSelectTextureButton); }
            if (_skinClearTextureButton != null) { _skinClearTextureButton.button.onClick.RemoveAllListeners(); _mvrScript.RemoveButton(_skinClearTextureButton); }
            if (_skinBoolParamToggle != null) _mvrScript.RemoveToggle(_skinBoolParamToggle);

            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            if (_spacer2 != null) { _mvrScript.RemoveSpacer(_spacer2); }
        }
        public void RefreshDisplay() {
            ClearDisplay();
            _isDisplayUpdating = true;

            if (_isDefaultSkin) {
                _buttonSkinParamSelectorPopup = _mvrScript.CreateSetupPopup(_buttonSkinParamSelectorJSON, !_isDefaultSkin, 500f, SkinParamSelectorPopupOpen, "Button Default Skin Param");
            }
            else {
                _spacer1 = _mvrScript.CreateSpacer(true);
                _spacer1.height = 10;
                _buttonSkinParamSelectorPopup = _mvrScript.CreateSetupPopup(_buttonSkinParamSelectorJSON, !_isDefaultSkin, 500f, SkinParamSelectorPopupOpen, "Button Skin Parameter");
                _skinSetToCurrentProfileDefaultToggle = _mvrScript.CreateToggle(_skinSetToCurrentProfileDefaultJSON, !_isDefaultSkin);
                _skinSetToCurrentProfileDefaultToggle.label = "Set Skin Param to default value";
                switch (_buttonSkinParamSelectorJSON.val) {
                    case "Button Text Color":
                    case "Button Text Color (On)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textOnColorDefault;
                        break;
                    case "Button Text Color (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textOffColorDefault;
                        break;
                    case "Button Color":
                    case "Button Color (On)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonOnColorDefault;
                        break;
                    case "Button Color (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonOffColorDefault;
                        break;

                    case "Button Font Size":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._fontSizeDefault;
                        break;
                    case "Button Font":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonFontDefault;
                        break;
                    case "Button Text Alignment":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textAlignmentDefault;
                        break;
                    case "Button Thumbnail (On)":
                    case "Button Thumbnail":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._customButtonOnTextureDefault;
                        break;
                    case "Button Thumbnail (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._customButtonOffTextureDefault;
                        break;
                }
            }
            _buttonSkinParamSelectorPopup.labelTextColor = Color.blue;
            if (_isDefaultSkin || _skinSetToCurrentProfileDefaultJSON.val == false) {
                string folder;
                switch (_buttonSkinParamSelectorJSON.val) {
                    case "Button Text Color":
                    case "Button Text Color (On)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._textOnColor;
                        break;
                    case "Button Text Color (Off)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._textOffColor;
                        if (_selectedButtonSkin._textOffColor.H != _selectedButtonSkin._textOnColor.H || _selectedButtonSkin._textOffColor.S != _selectedButtonSkin._textOnColor.S || _selectedButtonSkin._textOffColor.V != _selectedButtonSkin._textOnColor.V) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Color", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._textOffColor = _selectedButtonSkin._textOnColor;
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;
                    case "Button Color":
                    case "Button Color (On)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonOnColor;
                        break;
                    case "Button Color (Off)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonOffColor;
                        if (_selectedButtonSkin._buttonOffColor.H != _selectedButtonSkin._buttonOnColor.H || _selectedButtonSkin._buttonOffColor.S != _selectedButtonSkin._buttonOnColor.S || _selectedButtonSkin._buttonOffColor.V != _selectedButtonSkin._buttonOnColor.V) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Color", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._buttonOffColor = _selectedButtonSkin._buttonOnColor;
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;
                    case "Gaze Select Button Color":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        HSVColor buttonGazeSelectDefaultColor = new HSVColor();

                        Color gazeButtonColor = new Color(128f / 255f, 142f / 255f, 216f / 255f, 1f);
                        Color.RGBToHSV(gazeButtonColor, out buttonGazeSelectDefaultColor.H, out buttonGazeSelectDefaultColor.S, out buttonGazeSelectDefaultColor.V);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonGazeSelectColor;
                        _skinColorParamJSON.defaultVal = buttonGazeSelectDefaultColor;
                        break;
                    case "Button Font Size":
                    case "Button(Micro) Font Size":
                    case "Button(Mini) Font Size":
                    case "Button(Small) Font Size":
                    case "Button(Medium) Font Size":
                    case "Button(Large) Font Size":
                        _skinFloatParam1Slider = _mvrScript.CreateSlider(_skinFloatParam1JSON, !_isDefaultSkin);
                        _skinFloatParam2Slider = _mvrScript.CreateSlider(_skinFloatParam2JSON, !_isDefaultSkin);
                        _skinFloatParam1Slider.label = "Min Font Size";
                        _skinFloatParam2Slider.label = "Max Font Size";
                        if (_buttonSkinParamSelectorJSON.val == "Button Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Micro) Font Size")
                        {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minMicroFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxMicroFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Mini) Font Size")
                        {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minMiniFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxMiniFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minSmallFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxSmallFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minMediumFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxMediumFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minLargeFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxLargeFontSize);
                        }
                        break;
                    case "Default Gaze Target":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 250f, SkinStringChooserPopupOpen, "Gaze Target",false);
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._defaultGazeTarget;
                        break;
                    case "Button Font":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 500f, SkinStringChooserPopupOpen, "Select Font");
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._buttonFont;
                        break;
                    case "Button Text Alignment":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 500f, SkinStringChooserPopupOpen, "Select Alignment");
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._textAlignment;
                        break;
                    case "Button Thumbnail (On)":
                    case "Button Thumbnail":
                        _skinStringParamTextField = _mvrScript.CreateTextField(_skinStringParamJSON, !_isDefaultSkin);
                        _skinStringParamTextField.height = 1;
                        _skinStringParamJSON.val = _selectedButtonSkin._customButtonOnTextureURL;
                        _skinSelectTextureButton = _mvrScript.CreateButton("Select Thumbnail Image", !_isDefaultSkin);
                        folder = "Saves";
                        if (_mvrScript._lastLoadedThumbnailFolder != "") folder = _mvrScript._lastLoadedThumbnailFolder;
                        _skinSelectTextureButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", folder, true, true, false); });
                        if (_skinStringParamJSON.val != null && _skinStringParamJSON.val != "") {
                            _skinClearTextureButton = _mvrScript.CreateButton("Clear Thumbnail", !_isDefaultSkin);

                            _skinClearTextureButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._customButtonOnTextureURL = "";
                                    _selectedButtonSkin.QueueLoadTexture("", "ON");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            }
                        ); }
                        break;
                    case "Button Thumbnail (Off)":
                        _skinStringParamTextField = _mvrScript.CreateTextField(_skinStringParamJSON, !_isDefaultSkin);
                        _skinStringParamTextField.height = 1;
                        _skinStringParamJSON.val = _selectedButtonSkin._customButtonOffTextureURL;
                        _skinSelectTextureButton = _mvrScript.CreateButton("Select Thumbnail Image", !_isDefaultSkin);
                        folder = "Saves";
                        if (_mvrScript._lastLoadedThumbnailFolder != "") folder = _mvrScript._lastLoadedThumbnailFolder;
                        _skinSelectTextureButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", folder, true, true, false); });
                        if (_skinStringParamJSON.val != "" && _skinStringParamJSON.val != null) {
                            _skinClearTextureButton = _mvrScript.CreateButton("Clear Thumbnail", !_isDefaultSkin);

                            _skinClearTextureButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._customButtonOffTextureURL = "";
                                    _selectedButtonSkin.QueueLoadTexture("", "OFF");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            }
                        ); }
                        if (_selectedButtonSkin._customButtonOffTextureURL != _selectedButtonSkin._customButtonOnTextureURL) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Thumbnail", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating)
                                { _selectedButtonSkin._customButtonOffTextureURL = _selectedButtonSkin._customButtonOnTextureURL;
                                    _selectedButtonSkin.QueueLoadTexture(_selectedButtonSkin._customButtonOffTextureURL, "OFF");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;
                    case "12/24h Wrist Clock":
                        _skinBoolParamToggle = _mvrScript.CreateToggle(_skinBoolParamJSON,!_isDefaultSkin);
                        _skinBoolParamToggle.label = "Enable 24h clock format";
                        _skinBoolParamJSON.val = _selectedButtonSkin._24HourClock;
                        break;
                }
                if (_skinColorParamColorPicker != null) _skinColorParamColorPicker.label = "Color";
            }
            _isDisplayUpdating = false;
        }

        private void SkinParamSelectorPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_buttonSkinParamSelectorPopup);
            _mvrScript.CloseAllLeftUIPopups(_buttonSkinParamSelectorPopup);

            _buttonSkinParamSelectorJSON.choices = _selectedButtonSkin.GetButtonSkinParamChoices(_isDefaultSkin);
        }
        private void SkinStringChooserPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_skinStringChooserParamPopup);
            _mvrScript.CloseAllLeftUIPopups(_buttonSkinParamSelectorPopup);
            List<string> popupChoices = new List<string>();
            if (_buttonSkinParamSelectorJSON.val == "Button Font") {
                popupChoices.Add("Arial");
                popupChoices.Add("chintzy");
                popupChoices.Add("DroidSansMono");
                popupChoices.Add("Laffayette_Comic_Pro");
            }
            else if (_buttonSkinParamSelectorJSON.val == "Button Text Alignment") {
                popupChoices.Add("MiddleCenter");
                popupChoices.Add("LowerCenter");
                popupChoices.Add("UpperCenter");
                popupChoices.Add("MiddleLeft");
                popupChoices.Add("LowerLeft");
                popupChoices.Add("UpperLeft");
                popupChoices.Add("MiddleRight");
                popupChoices.Add("LowerRight");
                popupChoices.Add("UpperRight");
            }
            else if (_buttonSkinParamSelectorJSON.val == "Default Gaze Target")
            {
                popupChoices.Add("Last viewed Person");
                popupChoices.Add("Last viewed Female");
                popupChoices.Add("Last viewed Male");
            }
            _skinStringChooserParamJSON.choices = popupChoices;
        }

    }

    public class ToolBox
    {

        public ToolBox(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            _heelAdjustTool = new HeelAdjustTool(mvrScript);
            _pluginLoaderTool = new PluginLoaderTool(mvrScript);
            _gazeAssistedSelectTool = new GazeAssistedSelectTool(mvrScript);
        }

        public void RefreshToolboxDisplay(string toolboxSelection)
        {
            if (_toolBoxToggle1 != null) _mvrScript.RemoveToggle(_toolBoxToggle1);
            if (_toolBoxToggle2 != null) _mvrScript.RemoveToggle(_toolBoxToggle2);
            if (_toolBoxPopup1 != null) _mvrScript.RemovePopup(_toolBoxPopup1);
            if (_toolBoxPopup2 != null) _mvrScript.RemovePopup(_toolBoxPopup2);

            switch (toolboxSelection)
            {
                case "Auto Heel Adjust":
                    _toolBoxToggle1 = _mvrScript.CreateToggle(_heelAdjustTool._heelAdjustActiveJSON);
                    _toolBoxToggle1.label = "Auto Load Prestigitis Heel Adjusts";
                    if (_heelAdjustTool._heelAdjustActiveJSON.val)
                    {
                        _toolBoxToggle2 = _mvrScript.CreateToggle(_heelAdjustTool._heelAdjustRaisePeopleJSON);
                        _toolBoxToggle2.label = "Auto Raise Person by Heel Height";
                    }
                    break;
                case "Auto Plugin Loader":
                    _toolBoxPopup1 = _mvrScript.CreateSetupPopup(_pluginLoaderTool._atomTypeJSON);

                    break;
                case "Gaze Assisted Select\n(Edit Mode only)":
                    _toolBoxToggle1 = _mvrScript.CreateToggle(_gazeAssistedSelectTool._gazeAssistActiveJSON);
                    _toolBoxToggle1.label = "Activate Gaze Assisted Select";
                   
                    break;

            }
        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["heelAdjustActive"].AsBool = _heelAdjustTool._heelAdjustActiveJSON.val;
            node["heelAdjustRaisePerson"].AsBool = _heelAdjustTool._heelAdjustRaisePeopleJSON.val;
            _gazeAssistedSelectTool.GetJSON(node);

            return (node);
        }

        public void LoadJSON(JSONClass toolboxJSON)
        {
            if (toolboxJSON["heelAdjustActive"] != null) _heelAdjustTool._heelAdjustActiveJSON.val = toolboxJSON["heelAdjustActive"].AsBool;
            if (toolboxJSON["heelAdjustRaisePerson"] != null) _heelAdjustTool._heelAdjustRaisePeopleJSON.val = toolboxJSON["heelAdjustRaisePerson"].AsBool;
            _gazeAssistedSelectTool.LoadJSON(toolboxJSON);
        }

        private UIAssist _mvrScript;
        private UIDynamicToggle _toolBoxToggle1;
        private UIDynamicToggle _toolBoxToggle2;

        private UIDynamicPopup _toolBoxPopup1;
        private UIDynamicPopup _toolBoxPopup2;

        public HeelAdjustTool _heelAdjustTool;
        public PluginLoaderTool _pluginLoaderTool;
        public GazeAssistedSelectTool _gazeAssistedSelectTool;
    }

    public class GazeAssistedSelectTool
    {
        public GazeAssistedSelectTool(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            _gazeAssistActiveJSON = new JSONStorableBool("gazeAssistedSelectActive", false);

        }

        public JSONClass GetJSON(JSONClass jc)
        {
            jc["gazeAssistedSelectActive"].AsBool = _gazeAssistActiveJSON.val;


            return (jc);
        }

        public void LoadJSON(JSONClass jc)
        {
            if (jc["gazeAssistedSelectActive"] != null) _gazeAssistActiveJSON.val = jc["gazeAssistedSelectActive"].AsBool;
        }

        UIAssist _mvrScript;
        public JSONStorableBool _gazeAssistActiveJSON ;

    }

    public class PluginLoaderTool
    {
        public PluginLoaderTool(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            List<string> atomTypeChoices = new List<string>();
            atomTypeChoices.Add("People");
            atomTypeChoices.Add("Females");
            atomTypeChoices.Add("Males");
            atomTypeChoices.Add("Custom Unity Assets");


            _atomTypeJSON = new JSONStorableStringChooser("atomType", atomTypeChoices, "Females","Atom types",SwitchAtomType);


        }

        public void SwitchAtomType(string atomType)
        {
 //           if (!_mvrScript)
            {
 //               RefreshLeftUIDisplay();
            }
        }

        UIAssist _mvrScript;
        public JSONStorableStringChooser _atomTypeJSON;



    }

    public class HeelAdjustTool
    {
        public HeelAdjustTool(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            _heelAdjustActiveJSON = new JSONStorableBool("heelAdjustActive", true, SwitchHeelAdjustActive);
            _mvrScript.RegisterBool(_heelAdjustActiveJSON);
            _heelAdjustActiveJSON.isRestorable = false;
            _heelAdjustActiveJSON.isStorable = false;
            _heelAdjustRaisePeopleJSON = new JSONStorableBool("heelAdjustRaisePerson", false, SwitchHeelAdjustRaisePerson);
            _mvrScript.RegisterBool(_heelAdjustRaisePeopleJSON);
            _heelAdjustRaisePeopleJSON.isRestorable = false;
            _heelAdjustRaisePeopleJSON.isStorable = false;

        }
        private void SwitchHeelAdjustActive(bool haActive)
        {
            try
            {
                if (haActive)
                {
                    if (_heelAdjustRaisePeopleJSON.val)
                    {
                        // If Raising people is now being switched to active, then disable the temporary suppression.
                        foreach (KeyValuePair<string, HASetting> kvp in _atomHASettings)
                        {
                            if (kvp.Value !=null) kvp.Value._suppressControlNodeHeightReset = false;
                        }
                    }
                    if (!isHAFolderPresent())
                    {
                        if (!_mvrScript._isRestoring) SuperController.LogMessage("UIAssist: Auto heel adjust requires HeelAdjust configs to be created by prestigitis plugin and placed in 'Saves\\HeelAdjust' folder.");
                    }                
                }
                else
                {
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.type == "Person")FootPoseReset(atom);
                    }
                }

                JSONStorable otherUIA = _mvrScript.GetOtherActiveUIAPlugin();
                if (otherUIA != null)
                {
                    JSONStorableBool otherUIAHeelAdjustActiveJSON = otherUIA.GetBoolJSONParam("heelAdjustActive");
                    if (otherUIAHeelAdjustActiveJSON != null) otherUIAHeelAdjustActiveJSON.val = haActive;
                    
                }
                _mvrScript.RefreshLeftUIDisplay();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchHeelAdjustRaisePerson(bool haRaisePerson)
        {
            try
            {
                if (haRaisePerson)
                {
                    foreach (KeyValuePair<string,HASetting> kvp in _atomHASettings)
                    {
                        if (kvp.Value!=null) kvp.Value._suppressControlNodeHeightReset = false;
                    }
                }
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "Person")
                    {
                        if (haRaisePerson) UpdateAtomHeightOffset(atom);
                        else ResetAtomHeightOffset(atom);
                    }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void UpdateExistingHAFiles()
        {
            _existingHAFiles = new List<string>();
            if (FileManagerSecure.DirectoryExists(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName))
            {
                foreach (string file in SuperController.singleton.GetFilesAtPath(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName))
                {
                    string[] pathsplit = file.Split(new string[] { "\\" }, StringSplitOptions.None);
                    _existingHAFiles.Add(pathsplit.Last<String>().Replace("." + Constants._HeelAdjustFileExtension, ""));
                }
            }
        }
        private Transform CreateSingleShoeCollider(Atom atom, string tranformHierarchy, string colliderNamePreFix, Vector3 baseRotation, float colliderSize)
        {
            Transform limbTransform = atom.gameObject.transform.Find(tranformHierarchy);          

            GameObject rescaleGO = new GameObject();

            if (limbTransform == null) return null;
            else
            {
                rescaleGO.transform.position = limbTransform.position;
                rescaleGO.transform.rotation = limbTransform.rotation;
                rescaleGO.transform.parent = limbTransform;
                rescaleGO.transform.localEulerAngles = baseRotation;
                rescaleGO.name = colliderNamePreFix + "ColliderRescaleUIA";
                rescaleGO.transform.localScale = new Vector3(1f, 0.01f, 1f);

                GameObject colliderGO = GameObject.CreatePrimitive(PrimitiveType.Cube);
                colliderGO.transform.position = rescaleGO.transform.position;
                colliderGO.transform.parent = rescaleGO.transform;
                colliderGO.transform.localScale = new Vector3(colliderSize, Constants._ShoeColliderBaseScale, colliderSize);
                colliderGO.name = "ColliderUIA";

                colliderGO.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, 0f);
                colliderGO.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
                colliderGO.GetComponent<Renderer>().enabled = false;
                colliderGO.GetComponent<Renderer>().material.color = Color.green;
            }

            return (rescaleGO.transform);
        }
        private void CreateHAColliders(Atom atom)
        {

           // if there are any colliders active from the HeelAdjust plugin - then remove them (needed because HA plugin doesnt tidy up OnDestroy)
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe", false);

            Transform rShoeToeRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe", _ToeColliderBaseRotationR, 0.05f);
            Transform lShoeToeRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe", _ToeColliderBaseRotationL, 0.05f);
            Transform rShoeHeelRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel", _HeelColliderBaseRotationR, 0.02f);
            Transform lShoeHeelRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel", _HeelColliderBaseRotationL, 0.02f);

        }

        private void ApplySingleColiderHeightRotation(Transform transform, float colliderHeight, Vector3 colliderRotation, Vector3 baseColliderRotation, float colliderUpdateDelay)
        {
            _mvrScript.StartCoroutine(DelayedColliderUpdate(transform,colliderHeight, colliderUpdateDelay));
            transform.transform.localEulerAngles = new Vector3(-colliderRotation.x + baseColliderRotation.x, baseColliderRotation.y, baseColliderRotation.z);
        }
        private IEnumerator DelayedColliderUpdate(Transform transform, float colliderHeight, float colliderUpdateDelay)
        {
            if (colliderUpdateDelay == 0.001f) yield return new WaitForEndOfFrame();
            else
            {
                yield return new WaitForSeconds(colliderUpdateDelay);
                
            }

            Transform colliderT = transform.Find("ColliderUIA");
            
            colliderT.gameObject.GetComponent<Collider>().enabled = true;
            transform.transform.localScale = new Vector3(1f, colliderHeight, 1f);
        }
        private void ApplyHAColliderHeightsRotations(Atom atom, HASetting haSetting,bool vamBuiltInHeel, float colliderUpdateDelay)
        {
            float toeColliderHeight = 0.0001f;
            float heelColliderHeight = 0.0001f;
            if (!vamBuiltInHeel)
            {
                toeColliderHeight = haSetting._toeColliderHeight;
                heelColliderHeight = haSetting._heelColliderHeight;
            }

            Transform rToeColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescaleUIA");
            ApplySingleColiderHeightRotation(rToeColliderRescale, toeColliderHeight, haSetting._toeRotation, _ToeColliderBaseRotationR, colliderUpdateDelay);

            Transform rHeelColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            ApplySingleColiderHeightRotation(rHeelColliderRescale, heelColliderHeight, haSetting._footRotation, _HeelColliderBaseRotationR, colliderUpdateDelay);

            Transform lToeColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe/lShoeToeColliderRescaleUIA");
            ApplySingleColiderHeightRotation(lToeColliderRescale, toeColliderHeight, haSetting._toeRotation, _ToeColliderBaseRotationL, colliderUpdateDelay);

            Transform lHeelColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescaleUIA");
            ApplySingleColiderHeightRotation(lHeelColliderRescale, heelColliderHeight, haSetting._footRotation, _HeelColliderBaseRotationL, colliderUpdateDelay);

//            Transform rHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA/rShoeHeelColliderUIA");
//            Transform lHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescaleUIA/lShoeHeelColliderUIA");
            Transform rHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA/ColliderUIA");
            Transform lHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescaleUIA/ColliderUIA");

            rHeelCollider.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, haSetting._heelColliderBackOffset);
            lHeelCollider.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, haSetting._heelColliderBackOffset);

            //adjust foot pose to the given angles
            Transform rfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot");
            Transform rtoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe");
            Transform lfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot");
            Transform ltoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe");

            Vector3 rf = new Vector3(haSetting._footRotation.x, haSetting._footRotation.y, haSetting._footRotation.z);
            Vector3 rt = new Vector3(haSetting._toeRotation.x - haSetting._footRotation.x, haSetting._toeRotation.y, haSetting._toeRotation.z);
            Vector3 lf = new Vector3(haSetting._footRotation.x, -haSetting._footRotation.y, -haSetting._footRotation.z);
            Vector3 lt = new Vector3(haSetting._toeRotation.x - haSetting._footRotation.x, -haSetting._toeRotation.y, -haSetting._toeRotation.z);

            if (rfoot.GetComponent<DAZBone>().baseJointRotation != rf || rtoe.GetComponent<DAZBone>().baseJointRotation != rt || lfoot.GetComponent<DAZBone>().baseJointRotation != lf || ltoe.GetComponent<DAZBone>().baseJointRotation != lt)
            {
                if (_savedRotationStatesDict.ContainsKey(atom.name)) _savedRotationStatesDict[atom.name]._timeSaved = Time.unscaledTime;
                else
                {
                    ToeFootRotationStates savedRotationStates = GetCurrentRotationStates(atom);
                    _savedRotationStatesDict.Add(atom.name, savedRotationStates);
                }

                ResetRotationStates(atom, null);
                rfoot.GetComponent<DAZBone>().baseJointRotation = rf;
                rtoe.GetComponent<DAZBone>().baseJointRotation = rt;
                lfoot.GetComponent<DAZBone>().baseJointRotation = lf;
                ltoe.GetComponent<DAZBone>().baseJointRotation = lt;
            }
        }
        private ToeFootRotationStates GetCurrentRotationStates(Atom atom)
        {
            
            ToeFootRotationStates rotationStates = new ToeFootRotationStates();

            FreeControllerV3 rf = atom.freeControllers.First(fc => fc.name == "rFootControl");
            FreeControllerV3 rt = atom.freeControllers.First(fc => fc.name == "rToeControl");
            FreeControllerV3 lf = atom.freeControllers.First(fc => fc.name == "lFootControl");
            FreeControllerV3 lt = atom.freeControllers.First(fc => fc.name == "lToeControl");

            rotationStates._rFootRotationState = rf.currentRotationState;
            rotationStates._rToeRotationState = rt.currentRotationState;
            rotationStates._lFootRotationState = lf.currentRotationState;
            rotationStates._lToeRotationState = lt.currentRotationState;
            rotationStates._timeSaved = Time.unscaledTime;

            return (rotationStates);
        }

        public void ResetRotationStates(Atom atom, ToeFootRotationStates rotationStates)
        {
            if (atom != null)
            {
                FreeControllerV3 rf = atom.freeControllers.First(fc => fc.name == "rFootControl");
                FreeControllerV3 rt = atom.freeControllers.First(fc => fc.name == "rToeControl");
                FreeControllerV3 lf = atom.freeControllers.First(fc => fc.name == "lFootControl");
                FreeControllerV3 lt = atom.freeControllers.First(fc => fc.name == "lToeControl");

                if (rotationStates == null)
                {
                    rf.currentRotationState = FreeControllerV3.RotationState.Off;
                    rt.currentRotationState = FreeControllerV3.RotationState.Off;
                    lf.currentRotationState = FreeControllerV3.RotationState.Off;
                    lt.currentRotationState = FreeControllerV3.RotationState.Off;
                }
                else
                {
                    rf.currentRotationState = rotationStates._rFootRotationState;
                    rt.currentRotationState = rotationStates._rToeRotationState;
                    lf.currentRotationState = rotationStates._lFootRotationState;
                    lt.currentRotationState = rotationStates._lToeRotationState;
                }
            }
        }
        public void ResetAllSavedAtomRotationStates(bool setOff = false)
        {
            List<string> atomsToRemoveFromDictionary = new List<string>();
            foreach (KeyValuePair<string, ToeFootRotationStates> kvp in _savedRotationStatesDict)
            {
                if (Time.unscaledTime - kvp.Value._timeSaved > Constants._FCV3ResetDelay)
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(kvp.Key);
                    ResetRotationStates(atom, kvp.Value);
                    atomsToRemoveFromDictionary.Add(kvp.Key);
                }

            }
            foreach (string atomName in atomsToRemoveFromDictionary) _savedRotationStatesDict.Remove(atomName);
        }
        private void AjustLinkedAPHeights(string personUID, float heightAdjust)
        {            
            if (heightAdjust > 0.001f || heightAdjust < -0.001f)
            {
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "AnimationPattern")
                    {
                        JSONStorable storable = atom.GetStorableByID("AnimatedObject");
                        if (storable != null)
                        {
                            JSONClass apAnimatiedObjectJSON = storable.GetJSON();
                            string reciever = apAnimatiedObjectJSON["receiver"];
                            if (reciever != null)
                            {
                                if (reciever.Contains(":") && reciever.Length > reciever.IndexOf(':'))
                                {
                                    string atomName = reciever.Substring(0, reciever.IndexOf(':'));
                                    string nodeName = reciever.Substring(reciever.IndexOf(':') + 1);
                                    if (atomName == personUID && (nodeName == "control" || nodeName == "lToeControl" || nodeName == "rToeControl" || nodeName == "lFootControl" || nodeName == "rFootControl"))
                                    {
                                        FreeControllerV3 fcAP = atom.freeControllers.First(fc => fc.name == "control");
                                        fcAP.transform.position += Vector3.up * heightAdjust;
                                    }
                                }
                            }
                            
                        }
                    }
                }
            }
        }
        private void ResetAtomHeightOffset(Atom atom)
        {
            if (_atomHASettings.ContainsKey(atom.name) && _atomHASettings[atom.name]!=null && !_atomHASettings[atom.name]._suppressControlNodeHeightReset)
            {
               FreeControllerV3 atomControlFCV3 = atom.freeControllers.First(fc => fc.name == "control");
                atomControlFCV3.transform.position -= Vector3.up * _atomHASettings[atom.name]._controlNodeHeightOffset * Constants._ShoeColliderBaseScale;

                AjustLinkedAPHeights(atom.name, -(_atomHASettings[atom.name]._controlNodeHeightOffset * Constants._ShoeColliderBaseScale));
               _atomHASettings[atom.name]._controlNodeHeightOffset = 0f;
            }
        }

        private void UpdateAtomHeightOffset(Atom atom)
        {
            if (_atomHASettings.ContainsKey(atom.name) && _atomHASettings[atom.name]!=null)
            {
                HASetting haSetting = _atomHASettings[atom.name];
                
                float heightAdjustment = 0f;
                heightAdjustment = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale) - haSetting._controlNodeHeightOffset;
                haSetting._controlNodeHeightOffset = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale);

                FreeControllerV3 atomControlFCV3 = atom.freeControllers.First(fc => fc.name == "control");

                atomControlFCV3.transform.position += Vector3.up * heightAdjustment * Constants._ShoeColliderBaseScale;
                AjustLinkedAPHeights(atom.name, heightAdjustment * Constants._ShoeColliderBaseScale);
            }
        }
        private void LoadHASetting (HASetting haSetting, string path)
        {
            string loadedSettings = SuperController.singleton.ReadFileIntoString(path);
            string[] loadedSettingsLines = loadedSettings.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
            string[] footVals = loadedSettingsLines[0].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeVals = loadedSettingsLines[1].Split(new string[] { "," }, StringSplitOptions.None);
            string[] heelColliderVals = loadedSettingsLines[2].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeColliderVals = loadedSettingsLines[3].Split(new string[] { "," }, StringSplitOptions.None);
            haSetting._footRotation.x = float.Parse(footVals[0]);
            haSetting._footRotation.y = float.Parse(footVals[1]);
            haSetting._footRotation.z = float.Parse(footVals[2]);
            haSetting._toeRotation.x = float.Parse(toeVals[0]);
            haSetting._toeRotation.y = float.Parse(toeVals[1]);
            haSetting._toeRotation.z = float.Parse(toeVals[2]);
            haSetting._heelColliderHeight = float.Parse(heelColliderVals[0]);
            haSetting._heelColliderBackOffset = float.Parse(heelColliderVals[1]);
            haSetting._toeColliderHeight = float.Parse(toeColliderVals[0]);
        }
        private void ApplyHA(Atom atom, string path)
        {
            HASetting haSetting =null;
            bool newAtom = !_atomHASettings.ContainsKey(atom.name);
            if (newAtom || _atomHASettings[atom.name] == null) haSetting = new HASetting();
            else haSetting = _atomHASettings[atom.name];

            if (!newAtom && _atomHASettings[atom.name] == null) _atomHASettings[atom.name] = haSetting;

            LoadHASetting(haSetting, path);

            if (newAtom)
            {
                // This is a new Atom (either through scene load or adding a person)
                // Therefore we assume the person is at the correct height
                haSetting._controlNodeHeightOffset = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale);
                // Heel Adjust raising people is not active, then we dont want to lower people - so we temporarily suppress this until HeelAdjust raising people is activated
                if (!_heelAdjustActiveJSON.val || !_heelAdjustRaisePeopleJSON.val) haSetting._suppressControlNodeHeightReset = true;
                _atomHASettings.Add(atom.name, haSetting);
            }

            float colliderUpdateDelay = 0.001f;
            if (_heelAdjustRaisePeopleJSON.val)
            {
                UpdateAtomHeightOffset(atom);
                colliderUpdateDelay = 0.3f;
            }
            

            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            if (rt == null)
            {
                CreateHAColliders(atom);
                rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            }

            if (rt.transform.localScale.y > haSetting._heelColliderHeight)colliderUpdateDelay = 0.001f;
            
            if (path.EndsWith("MeshedVR-HarliHeels.heeladjust")|| path.EndsWith("MeshedVR-CasualDenimShoes.heeladjust")) ApplyHAColliderHeightsRotations(atom, haSetting,true, colliderUpdateDelay);
            else ApplyHAColliderHeightsRotations(atom, haSetting,false, colliderUpdateDelay);
        }
        private void RemoveCollider(Atom atom, string limbHierarchy, string prefix, bool UIA = true)
        {
            string colliderRescaleName = prefix + "ColliderRescale";
  //          string colliderName = prefix + "Collider";
            string colliderName = "Collider";

            if (UIA)
            {
                colliderRescaleName = colliderRescaleName + "UIA";
                colliderName = colliderName + "UIA";
            }

            Transform collider = atom.gameObject.transform.Find(limbHierarchy + "/" + colliderRescaleName + "/" + colliderName);
            if (collider != null) {
                collider.parent = null;
                GameObject.Destroy(collider.gameObject);
            }

            Transform colliderRescale = atom.gameObject.transform.Find(limbHierarchy + "/" + colliderRescaleName);
            if (colliderRescale != null)
            {
                colliderRescale.parent = null;
                GameObject.Destroy(collider.gameObject);
            }
        }

        public void FootPoseReset(Atom atom)
        {
            ResetAtomHeightOffset(atom);
            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescaleUIA");
            if (rt != null)
            {
                if (_savedRotationStatesDict.ContainsKey(atom.name)) _savedRotationStatesDict[atom.name]._timeSaved = Time.unscaledTime;
                else
                {
                    ToeFootRotationStates savedRotationStates = GetCurrentRotationStates(atom);
                    _savedRotationStatesDict.Add(atom.name, savedRotationStates);
                }
                ResetRotationStates(atom, null);
                //reset foot and toe rotations to their defaults
                Vector3 rfootBaseRotation = new Vector3(-15f, 0f, 0f);
                Vector3 rtoeBaseRotation = new Vector3(0f, 0f, 0f);
                Vector3 lfootBaseRotation = new Vector3(-15f, 0f, 0f);
                Vector3 ltoeBaseRotation = new Vector3(0f, 0f, 0f);

                var rfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot");
                var rtoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe");
                var lfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot");
                var ltoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe");

                rfoot.GetComponent<DAZBone>().baseJointRotation = rfootBaseRotation;
                rtoe.GetComponent<DAZBone>().baseJointRotation = rtoeBaseRotation;
                lfoot.GetComponent<DAZBone>().baseJointRotation = lfootBaseRotation;
                ltoe.GetComponent<DAZBone>().baseJointRotation = ltoeBaseRotation;

                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe");
            }
            if (_atomHASettings.ContainsKey(atom.name)) _atomHASettings.Remove(atom.name);
        }
        public bool isPrestigitisHALoaded(Atom atom)
        {
            bool foundHeelAdjustPlugin = false;
            foreach (string storableName in atom.GetStorableIDs())
            {
                if (storableName.StartsWith("plugin#") || storableName.StartsWith("pluginOrig#"))
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType.StartsWith("prestigitis.HeelAdjust"))
                    {
                        foundHeelAdjustPlugin = true;
                        break;
                    }
                }
            }

            return (foundHeelAdjustPlugin);
        }
        public bool isHAFolderPresent()
        {
            bool folderExists = FileManagerSecure.DirectoryExists(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName);
            return (folderExists);
        }
        public void LoadFootPosesForClothing(Atom atom, List<string> activePersonClothing)
        {
            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale");
            if (rt != null && isPrestigitisHALoaded(atom)) return;
            if (!isHAFolderPresent()) return;

            //get intersection of active clothing and files
            List<string> loadableProfiles = activePersonClothing.Intersect(_existingHAFiles).ToList();
            if (loadableProfiles.Count < 1) //intersection list is empty
            {
                FootPoseReset(atom);
                if (_atomHASettings.ContainsKey(atom.name)) _atomHASettings[atom.name] = null;
                else _atomHASettings.Add(atom.name, null);
                
            }
            else
            {
                //a heeladjust file was found for one of the active clothing items, load the first one
                ApplyHA(atom, SuperController.singleton.savesDir + "/" + Constants._HeelAdjustSubfolderName + "/" + loadableProfiles.First().ToString() + "." + Constants._HeelAdjustFileExtension);
            }
        }
        public void AtomNameUpdate(string oldName, string newName)
        {
            if (_savedRotationStatesDict.ContainsKey(oldName))
            {
                _savedRotationStatesDict.Add(newName, _savedRotationStatesDict[oldName]);
                _savedRotationStatesDict.Remove(oldName);
            }
            if (_atomHASettings.ContainsKey(oldName))
            {
                _atomHASettings.Add(newName, _atomHASettings[oldName]);
                _atomHASettings.Remove(oldName);
            }
        }

        public void AtomRemovedUpdate(string oldName)
        {
            if (_savedRotationStatesDict.ContainsKey(oldName))  _savedRotationStatesDict.Remove(oldName);
            if (_atomHASettings.ContainsKey(oldName)) _atomHASettings.Remove(oldName);       
        }

        private UIAssist _mvrScript;
        public JSONStorableBool _heelAdjustActiveJSON;
        public JSONStorableBool _heelAdjustRaisePeopleJSON;
        
        private List<string> _existingHAFiles;
        private Dictionary<string, ToeFootRotationStates> _savedRotationStatesDict = new Dictionary<string, ToeFootRotationStates>();
        private Dictionary<string, HASetting> _atomHASettings = new Dictionary<string, HASetting>();
        private Vector3 _ToeColliderBaseRotationR = new Vector3(-16.7f, 0f, -4.3f);
        private Vector3 _ToeColliderBaseRotationL = new Vector3(-16.7f, 0f, 4.3f);
        private Vector3 _HeelColliderBaseRotationR = new Vector3(-20f, 0f, 0f); //rFoot world rotation: Vector3(13.1f, 9.2f, 1.3f);
        private Vector3 _HeelColliderBaseRotationL = new Vector3(-20f, 0f, 0f); //lFoot world rotation: Vector3(13.1f, 351.4f, 359.4f);
    }
    public class HASetting
    {
        public Vector3 _footRotation;
        public Vector3 _toeRotation;
        public float _heelColliderHeight;
        public float _toeColliderHeight;
        public float _heelColliderBackOffset;
        public float _controlNodeHeightOffset =0f;
        public bool _suppressControlNodeHeightReset = false;

        public JSONClass GetJSON(string atomName)
        {
            JSONClass jc = new JSONClass();

            jc["atomName"] = atomName;
            jc["footRotationX"].AsFloat = _footRotation.x;
            jc["footRotationY"].AsFloat = _footRotation.y;
            jc["footRotationZ"].AsFloat = _footRotation.z;

            jc["toeRotationX"].AsFloat = _toeRotation.x;
            jc["toeRotationY"].AsFloat = _toeRotation.y;
            jc["toeRotationZ"].AsFloat = _toeRotation.z;

            jc["heelColliderHeight"].AsFloat = _heelColliderHeight;
            jc["toeColliderHeight"].AsFloat = _toeColliderHeight;
            jc["heelColliderBackOffset"].AsFloat = _heelColliderBackOffset;
            jc["controlNodeHeightOffset"].AsFloat = _controlNodeHeightOffset;
            jc["suppressControlNodeHeightReset"].AsBool = _suppressControlNodeHeightReset;

            return (jc);
        }
        public void RestoreJSON(JSONClass jc)
        {
            _footRotation.x= jc["footRotationX"].AsFloat;
            _footRotation.y = jc["footRotationY"].AsFloat;
            _footRotation.z = jc["footRotationZ"].AsFloat;

            _toeRotation.x = jc["toeRotationX"].AsFloat ;
            _toeRotation.y = jc["toeRotationY"].AsFloat ;
            _toeRotation.z = jc["toeRotationZ"].AsFloat ;

            _heelColliderHeight =jc["heelColliderHeight"].AsFloat;
            _toeColliderHeight= jc["toeColliderHeight"].AsFloat;
            _heelColliderBackOffset =jc["heelColliderBackOffset"].AsFloat;
            _controlNodeHeightOffset = jc["controlNodeHeightOffset"].AsFloat;
            _suppressControlNodeHeightReset = jc["suppressControlNodeHeightReset"].AsBool;
        }
    }

    public class ToeFootRotationStates
    {
        public FreeControllerV3.RotationState _lToeRotationState;
        public FreeControllerV3.RotationState _lFootRotationState;

        public FreeControllerV3.RotationState _rToeRotationState;
        public FreeControllerV3.RotationState _rFootRotationState;
        public float _timeSaved;
    }
    public class WristUI {
        public WristUI(MVRScript plugin) {
            _mvrUIAssist = plugin;
            GameObject canvasObject = new GameObject();
            _canvas = canvasObject.AddComponent<Canvas>();
            _canvas.pixelPerfect = false;
            _canvas.renderMode = RenderMode.WorldSpace;
            
            
            SuperController.singleton.AddCanvas(_canvas);

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            if (XRSettings.enabled) cs.dynamicPixelsPerUnit = 4f;
            else cs.dynamicPixelsPerUnit = 1.3f;

            cs.scaleFactor = 80f;

            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();

            _canvas.transform.localPosition = new Vector3(0.26f, -0.14f, -1f);
            _canvas.transform.localScale = new Vector3(Constants._localCanvasScale, Constants._localCanvasScale, Constants._localCanvasScale);
            
        }
        public GameObject CreateToggle(Transform parent, float xAnchorOffset, float yAnchorOffset, float width, float height,string label="", int fontSize=12)
        {           

            Transform transform = GameObject.Instantiate<Transform>(_mvrUIAssist.manager.configurableTogglePrefab); 

            GameObject toggleGO = transform.gameObject;
            transform.SetParent(parent, false);
            transform.gameObject.SetActive(true);

            RectTransform rt = transform.GetComponent<RectTransform>();
            rt.pivot = new Vector2(0f, 0f);
            rt.localPosition = new Vector2(xAnchorOffset, yAnchorOffset);

            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            UIDynamicToggle uiToggle = toggleGO.transform.GetComponent<UIDynamicToggle>();
            uiToggle.label = label;
            uiToggle.labelText.fontSize = fontSize;

            GameObject checkBoxGO = transform.Find("Background").gameObject;
            RectTransform checkBoxRT = checkBoxGO.transform.GetComponent<RectTransform>();
            checkBoxRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, height);
            checkBoxRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            checkBoxRT.anchoredPosition = new Vector2(height/2f, 0f);

            GameObject checkMarkGO = transform.Find("Background/Checkmark").gameObject;
            RectTransform checkMarkRT = checkMarkGO.transform.GetComponent<RectTransform>();
            checkMarkRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, height);
            checkMarkRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            GameObject  toggleLabelGO = transform.Find("Label").gameObject;
            RectTransform toggleLabelRT = toggleLabelGO.transform.GetComponent<RectTransform>();
            toggleLabelRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width- height-5f);
            toggleLabelRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height-2f);
            toggleLabelRT.pivot = new Vector2(0f, 0f);
            toggleLabelRT.localPosition = new Vector2(height+2f, 1f);

            return (toggleGO);
        }

        public GameObject CreateScrollableButtonList(float xAnchorOffset, float yAnchorOffset, float width, float height, Color backgroundColor)
        {
            GameObject scrollableButtonListGO = new GameObject();

            AddImageComponent(scrollableButtonListGO, backgroundColor);
            RectTransform imageRT = scrollableButtonListGO.GetComponent<RectTransform>();
            imageRT.anchoredPosition = new Vector2(xAnchorOffset, yAnchorOffset);
            imageRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            imageRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            
            GameObject buttonListViewportGO = CreateButtonListViewport(scrollableButtonListGO,width-20f,height -10f, 5f, 5f);
            buttonListViewportGO.name = "buttonListViewportGO";

             GameObject buttonListScrollbarGO = CreateScrollBar(scrollableButtonListGO, 10f, height-10f, -2.5f, 5f, Scrollbar.Direction.BottomToTop);

            ScrollRect scrollRect = scrollableButtonListGO.AddComponent<ScrollRect>();
            GameObject buttonListContentGO  = buttonListViewportGO.transform.Find("buttonListContentGO").gameObject;
            scrollRect.content = buttonListContentGO.GetComponent<RectTransform>();
            scrollRect.horizontal = false;
            scrollRect.vertical = true;
            scrollRect.movementType = ScrollRect.MovementType.Clamped;
            scrollRect.viewport = buttonListViewportGO.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = buttonListScrollbarGO.GetComponent<Scrollbar>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.Permanent;

            scrollableButtonListGO.transform.SetParent(_canvas.transform, false);
            scrollableButtonListGO.SetActive(true);

            return (scrollableButtonListGO);
        }
        public GameObject CreateButtonListViewport(GameObject parentGO,  float width, float height, float xAnchorOffset, float yAnchorOffset)
        {
            GameObject buttonListViewportGO = new GameObject();

            
            AddImageComponent(buttonListViewportGO, new Color(1f, 1f, 1f));
            buttonListViewportGO.transform.SetParent(parentGO.transform, false);
            buttonListViewportGO.SetActive(true);
            Mask blvpMask = buttonListViewportGO.AddComponent<Mask>();
            blvpMask.showMaskGraphic = false;
            SetSizeAndAnchors(parentGO, buttonListViewportGO, width, height, "topleft", xAnchorOffset, yAnchorOffset);

            GameObject buttonListContentGO = new GameObject();
            AddImageComponent(buttonListContentGO, new Color(0f, 0f, 1f),null,false);
            buttonListContentGO.transform.SetParent(buttonListViewportGO.transform, false);
            buttonListContentGO.SetActive(true);

            SetSizeAndAnchors(buttonListViewportGO, buttonListContentGO, 0.5f, height / 4, "center", xAnchorOffset, yAnchorOffset, 0.5f, 1f);

            buttonListContentGO.name = "buttonListContentGO";

            VerticalLayoutGroup vlg = buttonListContentGO.AddComponent<VerticalLayoutGroup>();
            vlg.childAlignment = TextAnchor.UpperCenter;
            vlg.childForceExpandHeight = true;
            vlg.childForceExpandWidth = true;
            vlg.childControlHeight = false;
            vlg.childControlWidth = false;
            vlg.spacing = 0f;           
   
            ContentSizeFitter csf = buttonListContentGO.AddComponent<ContentSizeFitter>();
            csf.horizontalFit = ContentSizeFitter.FitMode.MinSize;
            csf.verticalFit = ContentSizeFitter.FitMode.MinSize;
           
            return (buttonListViewportGO);
        }

        public GameObject CreateScrollBar(GameObject parentGO, float width, float height, float xAnchorOffset, float yAnchorOffset, Scrollbar.Direction scrollbarDirection)
        {
             Sprite backgroundSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "Background");
            Sprite uiSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "UISprite");

            GameObject scrollbarGO = new GameObject();
            Scrollbar scrollbar = scrollbarGO.AddComponent<Scrollbar>();
            AddImageComponent(scrollbarGO, new Color(1f, 1f, 1f), backgroundSprite);
            scrollbar.SetDirection(scrollbarDirection, false);
            scrollbarGO.transform.SetParent(parentGO.transform, false);
            scrollbarGO.SetActive(true);
            SetSizeAndAnchors(parentGO, scrollbarGO, width, height, "topright", xAnchorOffset, yAnchorOffset);

            GameObject scrollbarSlidingAreaGO = new GameObject();
            scrollbarSlidingAreaGO.AddComponent<RectTransform>();
            scrollbarSlidingAreaGO.transform.SetParent(scrollbarGO.transform, false);
            scrollbarSlidingAreaGO.SetActive(true);
            SetSizeAndAnchors(scrollbarGO, scrollbarSlidingAreaGO, width, height, "center", 0, 0);
         
            GameObject scrollbarHandleGO = new GameObject();
            scrollbarHandleGO.transform.SetParent(scrollbarSlidingAreaGO.transform, false);
            AddImageComponent(scrollbarHandleGO,new Color(1f,1f,1f), uiSprite);
            SetSizeAndAnchors(scrollbarSlidingAreaGO, scrollbarHandleGO, 0f, 0f, "center", 0, 0);

            scrollbar.targetGraphic = scrollbarHandleGO.GetComponent<UnityEngine.UI.Image>();
            scrollbar.handleRect = scrollbarHandleGO.GetComponent<RectTransform>(); ;

            return (scrollbarGO);
        }

        private void AddImageComponent(GameObject go, Color color, Sprite sprite = null, bool enabled = true)
        {
            UnityEngine.UI.Image image = go.AddComponent<UnityEngine.UI.Image>();
            image.color = color;         
            image.sprite = sprite;
            image.type = UnityEngine.UI.Image.Type.Sliced;
            image.enabled = enabled;
        }

        public GameObject CreateUIButtonInParentGO(GameObject parentGO)
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(_mvrUIAssist.manager.configurableButtonPrefab.transform, parentGO.transform,false);
            
            GameObject buttonGO = buttonTransform.gameObject;
            buttonTransform.SetParent(parentGO.transform, false);                     
            return (buttonGO);
            
        }

        public GameObject CreateUIToggleInParentGO(GameObject parentGO, float xAnchorOffset, float yAnchorOffset, float width, float height, string label = "",int fontSize = 12)
        {

            GameObject toggleGO = CreateToggle(parentGO.transform, xAnchorOffset, yAnchorOffset, width, height, label, fontSize);
            return (toggleGO);
        }

        public GameObject CreateUIContainerImageInParentGO(GameObject parentGO, float width, float height, Color panelColor)
        {
            GameObject containerImageGO = new GameObject();

            AddImageComponent(containerImageGO, panelColor);

            containerImageGO.SetActive(true);

            SetSizeAndAnchors(parentGO, containerImageGO, width, height, "center", 0f, 0f);

            containerImageGO.transform.SetParent(parentGO.transform, false);
            return (containerImageGO);
        }

        public UIDynamicButton CreateUIButtonInCanvas(string label, float x, float y, float z = 0, float w = 0, float h = 0,int fontSize = 16) {
            Transform transform = UnityEngine.Object.Instantiate(_mvrUIAssist.manager.configurableButtonPrefab.transform);

            transform.SetParent(_canvas.transform, false);
            transform.gameObject.SetActive(true);

            UIDynamicButton uiButton = transform.GetComponent<UIDynamicButton>();
            uiButton.label = label;
            uiButton.buttonText.fontSize = fontSize;

            if (w > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, w);
            if (h > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);

            transform.localPosition = new Vector3(x, y, z);

            return uiButton;
        }


        public static void SetSizeAndAnchors(GameObject parent, GameObject child, float width, float height, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            RectTransform childRect = child.GetComponent<RectTransform>();
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            SetAnchors(parent, child, placement, offsetX, offsetY,pivotX,pivotY);
            
        }

        public static void SetAnchors(GameObject parent, GameObject child, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            RectTransform parentRect =null;
            if (parent!=null) parentRect = parent.GetComponent<RectTransform>();
            RectTransform childRect = child.GetComponent<RectTransform>();
            SetAnchors(parentRect, childRect, placement, offsetX, offsetY, pivotX, pivotY);
        }


        public static void SetAnchors(RectTransform parent, RectTransform child, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            if (parent == null)
            {
                parent = child.parent.GetComponent<RectTransform>();
            }
            Vector2 pivotPoint = new Vector2(pivotX, pivotY);
            child.pivot = pivotPoint;

            switch (placement)
            {
                case "center":
                    child.transform.localPosition = new Vector3(0, 0, 0);
                    break;
                case "left":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, 0 + offsetY);
                    break;
                case "right":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, 0 - offsetY);
                    break;
                case "top":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMax - child.rect.yMax + offsetX, 0 - offsetY);
                    break;
                case "bottom":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMin - child.rect.yMin + offsetX, 0 + offsetY);
                    break;

                case "topleft":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, parent.rect.yMax - child.rect.yMax - offsetY);
                    break;
                case "topright":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, parent.rect.yMax - child.rect.yMax - offsetY);
                    break;

                case "bottomleft":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, parent.rect.yMin - child.rect.yMin + offsetY);
                    break;
                case "bottomright":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, parent.rect.yMin - child.rect.yMin + offsetY);
                    break;
                //Outside Edges
                case "leftout":
                    child.transform.localPosition = new Vector3(parent.rect.xMin + child.rect.xMin + offsetX, 0 + offsetY);
                    break;
                case "rightout":
                    child.transform.localPosition = new Vector3(parent.rect.xMax + child.rect.xMax + offsetX, 0 - offsetY);
                    break;
                case "topout":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMax + child.rect.yMax + offsetX, 0 - offsetY);
                    break;
                case "bottomout":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMin + child.rect.yMin + offsetX, 0 + offsetY);
                    break;
            }

        }

        public UIDynamicTextField CreateTextField(float x, float y, float z, float width, float height, float margin =5f) {
            Transform textField = GameObject.Instantiate<Transform>(_mvrUIAssist.manager.configurableTextFieldPrefab);

            
            textField.SetParent(_canvas.transform, false);
            RectTransform textFieldRectTransform = textField.GetComponent<RectTransform>();
            
            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);

            Vector2 pivot = new Vector2();
            pivot.x = 0.5f;
            pivot.y = 0f;

            textFieldRectTransform.pivot = pivot;
            textFieldRectTransform.localPosition = new Vector3(x, y, z);
            
            UIDynamicTextField uiTextField = textField.GetComponent<UIDynamicTextField>();

            RectTransform rectT = uiTextField.UItext.GetComponent<RectTransform>();
            //                        rectT.anchoredPosition = new Vector2(5f, -4f);
            //            rectT.pivot = new Vector2(0f, 0.5f);
            //            rectT.localPosition = new Vector2(width / 2f, height / 2f);
            rectT.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, margin, width- (margin*2));
            rectT.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, margin, height - (margin * 2));
            //            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width * 0.8f);
            //            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height * 0.8f);        

            return uiTextField;
        }


        public void OnDestroy() {
            if (SuperController.singleton != null) SuperController.singleton.RemoveCanvas(_canvas);
            _canvas.transform.SetParent(null, false);

            if (_canvas.gameObject != null) GameObject.Destroy(_canvas.gameObject);
        }

        public void Update(Vector3 vrPosition, Quaternion vrRotation, string uiDisplayMode, bool buttonCanvas = false, bool rightHand = false, float verticalHUDOffset = 0f, float horizontalHUDOffset = 0f, float depthHUDOffset = 0f) {
            if (uiDisplayMode == "HUD" || uiDisplayMode == "VRHUD" )
            {
                if (uiDisplayMode == "HUD") depthHUDOffset = 1.5f;
                if (SuperController.singleton.lookCamera != null)
                {
                    Transform cameraT = SuperController.singleton.lookCamera.transform;

                    Vector3 facingPos = cameraT.position + cameraT.forward * 10000000.0f;
                    if (Camera.main != null)
                    {
                        Vector3 endPos = Camera.main.ViewportToWorldPoint(new Vector3(horizontalHUDOffset, verticalHUDOffset, depthHUDOffset));

                        _canvas.transform.LookAt(facingPos, cameraT.up);
                        _canvas.transform.position = endPos;
                    }
                }
            }
            else {
                _canvas.transform.position = vrPosition;
                _canvas.transform.rotation = vrRotation;
                if (uiDisplayMode == "WRIST" && buttonCanvas) _canvas.transform.Translate(0f, -0.05f, 0f, Space.Self);
                float inverseMultiplier = 0f;
                if (rightHand) inverseMultiplier = -1f;
                _canvas.transform.Rotate(180 * inverseMultiplier, 90 + (180 * 0), 188 + (196 * inverseMultiplier));
                if (buttonCanvas) {
                    _canvas.transform.Rotate(-50, 0, 0);
                    _canvas.transform.position = _canvas.transform.position + (_canvas.transform.up * 0.02f);
                }
                float inverseOffetMultiplier = 1f;
                if (rightHand) inverseOffetMultiplier = -1f;
                _canvas.transform.position = _canvas.transform.position + (_canvas.transform.up * 0.02f) + (_canvas.transform.forward * -0.045f) + (_canvas.transform.right * (-0.16f * inverseOffetMultiplier));
            }
        }

        public Canvas _canvas;
        MVRScript _mvrUIAssist;

    }

    static class Constants {
		public const bool _patreonContentEnabled = false;
        public const float _TimeBetweenClothingUpdates = 0.1f;
        public const float _TimeBetweenHAFileUpdates = 1f;
        public const string _HeelAdjustSubfolderName = "HeelAdjust";
        public const string _HeelAdjustFileExtension = "heeladjust";
        public static string _vamFilePath = Application.dataPath.Substring(0, Application.dataPath.Length - 8).Replace('/', '\\');
        public const float _ShoeColliderBaseScale = 0.01f;
        public const float _FCV3ResetDelay = 0.8f;
        public const float _localCanvasScale = 0.001f;
        public const float _gasWidth = 200f;
        public const float _gasMinHeight = 200f;
        public const float _gazeStatusMinWidth = 240f;
        public const float _gazeStatusMaxHeight = 65f;
        public const float _toggleButtonHeight = 24f;
        public static readonly List<string> _gazeRBs = new List <string>(new[] { "hip", "pelvis", "Testes", "rThigh", "lThigh", "rShin", "lShin", "rFoot", "lFoot", "LGlute", "RGlute", "abdomen", "abdomen2", "chest", "rCollar", "lCollar", "rShldr", "lShldr", "lPectoral", "rPectoral", "neck", "head","control" });

    }
	
	static class PatreonFeatures {
		public static void GridLayoutFeature (List<string> buttonGridLayoutChoices){
			buttonGridLayoutChoices =null;
		}
		public static void ClothingPresetActions (UIAssist mvrScript, List<string> mergeClothingGeometryIDs,Atom atom, string mode, string presetFile, bool reverse){
           
		}

		public static void PopulateMergeClothingGeometryIDs (UIAssist mvrScript,List<string> mergeClothingGeometryIDs, string presetFile, Atom atom){
			
		}

        public static void MergePluginStorables (JSONClass atomJSON, JSONArray newPluginStorables, int pluginID)
        {

        }
        public static void MergePluginSettings(Dictionary<string, JSONClass> pluginSettings, JSONArray newPluginStorables, int pluginID)
        {

        }

        public static void RemoveTagClothingItems(Atom atom, Dictionary<DAZClothingItem, List<string>> activeClothingTags, List<String> tagsToRemove)
        {


        }
        public static string AppPresetSuppressLoad(UIAssist mvrScript, Atom atom, string presetFilePath){
			return(null);
		}
		public static string SkinPresetDecalLoad(UIAssist mvrScript, Atom atom, string presetFilePath, bool decalFaceLoad, bool decalBodyLoad, bool decalLimbsLoad, bool decalGenLoad)
        {
			return(null);
		}
	}
	
	public class UIDActiveClothingEditor
    {
        public UIDActiveClothingEditor(UIAssist mvr, WristUI uiButtonCanvas)
        {
        }

        public void CheckForRemovedAtoms()
        {
        }

        public void UpdateAtomNames(string oldName, string newName)
        {
        }
        public void UpadateACE(string atomName, ActiveClothingLists currentActiveClothing)
        {
        }

        public void CreateUI()
        {
        }

        public void RefreshClothing(string atomName)
        {
        }

        public void DestroyUI()
        {
        }
    }
   }
